---@type string
local AddonName = ...
---@class Data
local Data = select(2, ...)

local L = Data.L
local DRList = LibStub("DRList-1.0")

local table_insert = table.insert

local GetClassInfo = GetClassInfo
local GetNumSpecializationsForClassID = C_SpecializationInfo and C_SpecializationInfo.GetNumSpecializationsForClassID or GetNumSpecializationsForClassID
local GetSpecializationInfoForClassID = GetSpecializationInfoForClassID
local GetSpellInfo = GetSpellInfo
local GetSpellName = C_Spell and C_Spell.GetSpellName or GetSpellName
local C_Spell = C_Spell
local GetSpellTexture = C_Spell and C_Spell.GetSpellTexture or GetSpellTexture

local IsRetail = WOW_PROJECT_ID == WOW_PROJECT_MAINLINE
local IsClassic = WOW_PROJECT_ID == WOW_PROJECT_CLASSIC
local IsTBCC = WOW_PROJECT_ID == WOW_PROJECT_BURNING_CRUSADE_CLASSIC
local IsWrath = WOW_PROJECT_ID == WOW_PROJECT_WRATH_CLASSIC


Data.PlayerRoles = {"TANK", "HEALER", "DAMAGER"}

Data.CyrillicToRomanian = { -- source Wikipedia: https://en.wikipedia.org/wiki/Romanization_of_Russian
	["А"] = "a",
	["а"] = "a",
	["Б"] = "b",
	["б"] = "b",
	["В"] = "v",
	["в"] = "v",
	["Г"] = "g",
	["г"] = "g",
	["Д"] = "d",
	["д"] = "d",
	["Е"] = "e",
	["е"] = "e",
	["Ё"] = "e",
	["ё"] = "e",
	["Ж"] = "zh",
	["ж"] = "zh",
	["З"] = "z",
	["з"] = "z",
	["И"] = "i",
	["и"] = "i",
	["Й"] = "i",
	["й"] = "i",
	["К"] = "k",
	["к"] = "k",
	["Л"] = "l",
	["л"] = "l",
	["М"] = "m",
	["м"] = "m",
	["Н"] = "n",
	["н"] = "n",
	["О"] = "o",
	["о"] = "o",
	["П"] = "p",
	["п"] = "p",
	["Р"] = "r",
	["р"] = "r",
	["С"] = "s",
	["с"] = "s",
	["Т"] = "t",
	["т"] = "t",
	["У"] = "u",
	["у"] = "u",
	["Ф"] = "f",
	["ф"] = "f",
	["Х"] = "kh",
	["х"] = "kh",
	["Ц"] = "ts",
	["ц"] = "ts",
	["Ч"] = "ch",
	["ч"] = "ch",
	["Ш"] = "sh",
	["ш"] = "sh",
	["Щ"] = "shch",
	["щ"] = "shch",
	["Ъ"] = "ie",
	["ъ"] = "ie",
	["Ы"] = "y",
	["ы"] = "y",
	["Ь"] = "",
	["ь"] = "",
	["Э"] = "e",
	["э"] = "e",
	["Ю"] = "iu",
	["ю"] = "iu",
	["Я"] = "ia",
	["я"] = "ia"
}

Data.Buttons = {
	Target = TARGET,
	Focus = SET_FOCUS,
	Custom = L.UserDefined
}

Data.ObjectiveAndRespawnPosition = {
	Left = L.LEFT,
	Right = L.RIGHT,
	LeftToTargetCounter = L.LeftToTargetCounter
}

Data.DisplayType = {
	Frame = L.Frame,
	Countdowntext = L.Countdowntext,
}

-- Data.DrCategorys = {
-- 	disorient = L.DR_Disorient,
-- 	incapacitate = L.DR_Incapacitate,
-- 	knockback = L.DR_Knockback,
-- 	silence  = L.DR_Silence,
-- 	stun = L.DR_Stun
-- }

Data.DebuffTypes = {
	HELPFUL = {
		Magic = L.Magic,
	},
	HARMFUL = {
		Magic = L.Magic,
		Disease = L.Disease,
		Poison = L.Poison,
		Curse = L.Curse
	}

}
Data.RandomDebuffType =  {} -- for testmode

do
	local i = 1
	for engName, color in pairs(DebuffTypeColor) do
		Data.RandomDebuffType[i] = engName
		i = i + 1
	end
end

Data.AllPositions = {
	TOPLEFT = L.TOPLEFT,
	TOP = L.TOP,
	TOPRIGHT = L.TOPRIGHT,
	LEFT = L.LEFT,
	CENTER = L.CENTER,
	RIGHT = L.RIGHT,
	BOTTOMLEFT = L.BOTTOMLEFT,
	BOTTOM = L.BOTTOM,
	BOTTOMRIGHT = L.BOTTOMRIGHT
}

Data.BasicPositions = {
	LEFT = L.LEFT,
	RIGHT = L.RIGHT
}

Data.HorizontalDirections = {
	leftwards = L.Leftwards,
	rightwards = L.Rightwards
}
Data.VerticalDirections = {
	upwards = L.Upwards,
	downwards = L.Downwards
}


Data.Interruptdurations = {
	[6552]	= 4,   -- [Warrior] Pummel
	[96231]	= 4,  -- [Paladin] Rebuke
	[231665]= 3, -- [Paladin] Avengers Shield
	[147362]= 3, -- [Hunter] Countershot
	[187707]= 3, -- [Hunter] Muzzle
	[1766]	= 5,   -- [Rogue] Kick
	[183752]= 3, -- [DH] Consume Magic, Disrupt since  8.0.1
	[47528]	= 3,  -- [DK] Mind Freeze
	[91802]	= 2,  -- [DK] Shambling Rush
	[57994]	= 3,  -- [Shaman] Wind Shear
	[115781]= 6, -- [Warlock] Optical Blast
	[19647]	= 6,  -- [Warlock] Spell Lock
	[212619]= 6, -- [Warlock] Call Felhunter
	[132409]= 6, -- [Warlock] Spell Lock
	[171138]= 6, -- [Warlock] Shadow Lock
	[2139]	= 6,   -- [Mage] Counterspell
	[116705]= 4, -- [Monk] Spear Hand Strike
	[106839]= 4, -- [Feral] Skull Bash
	[93985]	= 4,  -- [Feral] Skull Bash
	[351338]= 4 -- [Evoker] Quell
}

-- for spellId, lockoutDuration in pairs(Data.Interruptdurations) do
	-- Data.SpellPriorities[spellId] = 4
-- end


Data.RandomDrCategory = {} --key = number, value = categorieName, used for Testmode
Data.DrCategoryToSpell = {} --key = categorieName, value = table with key = number and value = spellId
Data.SpellPriorities = {}

local i = 1
for categorieName, localizedCategoryName in pairs(DRList:GetCategories()) do
	Data.RandomDrCategory[i] = categorieName
	Data.DrCategoryToSpell[categorieName] = {}
	i = i + 1
end

do
	local drCategoryToPriority = {
		stun = 8,
		disorient = 7,
		incapacitate = 6,
		silence = 5,
		knockback = 2,
		taunt = 1
	}


	-- no more needed since classic supports spellIDs now
	-- if IsClassic then
	-- 	for spellName, spellData in pairs(DRList.spells) do
	-- 		local category = spellData.category
	-- 		table_insert(Data.DrCategoryToSpell[category], spellName)
	-- 		Data.SpellPriorities[spellName] = drCategoryToPriority[category]
	-- 	end
	-- else
	-- 	for spellId, categorieName in pairs(DRList.spells) do
	-- 		table_insert(Data.DrCategoryToSpell[categorieName], spellId)
	-- 		Data.SpellPriorities[spellId] = drCategoryToPriority[categorieName]
	-- 	end
	-- end
	for spellId, categorieName in pairs(DRList.spells) do
		if type(categorieName) == "table" then
			for k, categorie in pairs(categorieName) do
				table_insert(Data.DrCategoryToSpell[categorie], spellId)
			end
		else
			table_insert(Data.DrCategoryToSpell[categorieName], spellId)
		end
		
		Data.SpellPriorities[spellId] = drCategoryToPriority[categorieName]
	end

end


Data.cCduration = {	-- this is basically data from DRList-1 with durations, used for Relentless check
	--[[ INCAPACITATES ]]--
	incapacitate = {
		-- Druid
		[    99] = 3, -- Incapacitating Roar (talent)
		[236025] = 6, -- Main (Honor talent)
		[236026] = 6, -- Main (Honor talent)
		-- Hunter
		[213691] = 4, -- Scatter Shot
		-- Mage
		[   118] = 8, -- Polymorph
		[ 28272] = 8, -- Polymorph (pig)
		[ 28271] = 8, -- Polymorph (turtle)
		[ 61305] = 8, -- Polymorph (black cat)
		[277792] = 8, -- Polymorph (Bumblebee)
		[277787] = 8, -- Polymorph (Direhorn)
		[ 61721] = 8, -- Polymorph (rabbit)
		[ 61780] = 8, -- Polymorph (turkey)
		[126819] = 8, -- Polymorph (procupine)
		[161353] = 8, -- Polymorph (bear cub)
		[161354] = 8, -- Polymorph (monkey)
		[321395] = 8, -- Polymorph (Mawrat)
		[161355] = 8, -- Polymorph (penguin)
		[161372] = 8, -- Polymorph (peacock)
		[ 82691] = 8, -- Ring of Frost
		-- Monk
		[115078] = 4, -- Paralysis
		-- Paladin
		[ 20066] = 8, -- Repentance
		-- Priest
		[ 64044] = 4, -- Psychic Horror (Horror effect)
		-- Rogue
		[  1776] = 4, -- Gouge
		[  6770] = 8, -- Sap
		-- Shaman
		[ 51514] = 8, -- Hex
		[211004] = 8, -- Hex (spider)
		[210873] = 8, -- Hex (raptor)
		[211015] = 8, -- Hex (cockroach)
		[211010] = 8, -- Hex (snake)
		-- Warlock
		[  6789] = 3, -- Mortal Coil
		-- Pandaren
		[107079] = 4 -- Quaking Palm
	},

	--[[ SILENCES ]]--
	silence = {
		-- Death Knight
		[ 47476] = 5, -- Strangulate
		-- Hunter
		[202933] = 4, -- Spider Sting (pvp talent)
		-- Mage
		-- Paladin
		[ 31935] = 3, -- Avenger's Shield
		-- Priest
		[ 15487] = 4, -- Silence
		-- Rogue
		[  1330] = 3, -- Garrote
		-- Blood Elf
		[ 25046] = 2, -- Arcane Torrent (Energy version)
		[ 28730] = 2, -- Arcane Torrent (Priest/Mage/Lock version)
		[ 50613] = 2, -- Arcane Torrent (Runic power version)
		[ 69179] = 2, -- Arcane Torrent (Rage version)
		[ 80483] = 2, -- Arcane Torrent (Focus version)
		[129597] = 2, -- Arcane Torrent (Monk version)
		[155145] = 2, -- Arcane Torrent (Paladin version)
		[202719] = 2  -- Arcane Torrent (DH version)
	},

	--[[ DISORIENTS ]]--
	disorient = {
		-- Druid
		[ 33786] = 6, -- Cyclone
		-- Mage
		[ 31661] = 3, -- Dragon's Breath
		-- Paladin
		[105421] = 6, -- Blinding Light -- FIXME: is this the right category? Its missing from blizzard's list
		-- Priest
		[  8122] = 6, -- Psychic Scream
		-- Rogue
		[  2094] = 8, -- Blind
		-- Warlock
		[  5782] = 6, -- Fear -- probably unused
		[118699] = 6, -- Fear -- new debuff ID since MoP
		-- Warrior
		[  5246] = 5 -- Intimidating Shout (main target)
	},

	--[[ STUNS ]]--
	stun = {
		-- Death Knight
		[108194] = 4, -- Asphyxiate (talent for unholy)
		[221562] = 5, -- Asphyxiate (baseline for blood)
		[207171] = 4, -- Winter is Coming (Remorseless winter stun)
		-- Demon Hunter
		[179057] = 2, -- Chaos Nova
		[200166] = 3, -- Metamorphosis
		[205630] = 3, -- Illidan's Grasp, primary effect
		[211881] = 4, -- Fel Eruption
		-- Druid
		[  5211] = 4, -- Mighty Bash
		[163505] = 4, -- Rake (Stun from Prowl)
		-- Monk
		[120086] = 4, -- Fists of Fury (with Heavy-Handed Strikes, pvp talent)
		[232055] = 3, -- Fists of Fury (new ID in 7.1)
		[119381] = 3, -- Leg Sweep
		-- Paladin
		[   853] = 6, -- Hammer of Justice
		-- Priest
		[200200] = 4, -- Holy word: Chastise
		[226943] = 6, -- Mind Bomb
		-- Rogue
		[  1833] = 4, -- Cheap Shot
	--	[   408] = true, -- Kidney Shot, variable duration
	  --[199804] = true, -- Between the Eyes, variable duration
		-- Shaman
		[118345] = 4, -- Pulverize (Primal Earth Elemental)
		[118905] = 5, -- Static Charge (Capacitor Totem)
		[204399] = 2, -- Earthfury (pvp talent)
		-- Warlock
		[ 89766] = 4, -- Axe Toss (Felguard)
		[ 30283] = 3, -- Shadowfury
		-- Warrior
		[132168] = 2, -- Shockwave
		[132169] = 4, -- Storm Bolt
		-- Tauren
		[ 20549] = 2 -- War Stomp
	}
}

Data.cCdurationBySpellID = {}
for category, spellIDs in pairs(Data.cCduration) do
	Mixin(Data.cCdurationBySpellID, spellIDs)
end


Data.PriorityAuras = {
	HELPFUL = {

		--Death Knight
		[145629] =	5.0,	 	-- Anti-Magic Zone
		[48792] =	5,	 		-- Icebound Fortitude
		[81256] = 	5,	 		-- Dancing Rune Weapon
		[55233] = 	5,	 		-- Vampiric Blood
		[194679] = 	5,	 		-- Rune Tap
		[194844] = 	5.1,		-- Bonestorm
		[219809] = 	5.1,	 	-- Tombstone

		--Demon Hunter
		[187827] = 	5.1,	 	-- Metamorphosis - Vengeance
		[205629] = 	5.1,	 	-- Demonic Trample
		[209426] = 	5.1,	 	-- Darkness
		[212800] = 	5.1,	 	-- Blur
		[196555] = 	5.1,	 	-- Netherwalk
		[206803] = 	5.2, 	 	-- Rain from Above (up)


		--Druid
		[22812] = 	5.1,	 	-- Barkskin
		[22842] = 	5.1,	 	-- Frenzied Regeneration
		[61336] = 	5.1,	 	-- Survival Instincts
		[102342] = 	5.1,	 	-- Ironbark
		[155835] = 	5.1,	 	-- Bristling Fur
		[305497] = 	5.1,	 	-- Thorns (PvP Talent)
		[50334] = 	5.1,	 	-- Berserk (Guardian)
		[247563] = 	5.1,	 	-- Nature's Grasp (Resto Entangling Bark PvP Talent)
		[327037] = 	5.1,	 	-- Kindred Protection (Kyrian Ability)
		[362486] = 	5.2, 	 	-- Keeper of the Grove
		[186265] = 	5.2, 	 	-- Aspect of the Turtle

		--Hunter
		[136] = 	5.1,	 	-- Mend Pet
		[5384] = 	5.1,	 	-- Feign Death
		[53480] = 	5.1,	 	-- Roar of Sacrifice (PvP Talent)
		[54216] = 	5.1,	 	-- Master's Call
		[209997] = 	5.1,	 	-- Play Dead
		[202748] = 	5.1,	 	-- Survival Tactics (PvP Talent)
		[248519] = 	5.2, 	 	-- Interlope (BM PvP Talent)


		--Mage
		[342246] = 	5.1,	 	-- Alter Time (Arcane)
		[110909] = 	5.1,	 	-- Alter Time (Fire/Frost)
		[198111] = 	5.1,	 	-- Temporal Shield (Arcane PvP Talent)
		[198065] = 	5.1,	 	-- Prismatic Cloak (PvP Talent)
		[45438] = 	5.2, 	 	-- Ice Block

		--Monk
		[115176] = 	5.1,	 	-- Zen Meditation
		[120954] = 	5.1,	 	-- Fortifying Brew (Brewmaster)
		[243435] = 	5.1,	 	-- Fortifying Brew (Windwalker/Mistweaver)
		[337294] = 	5.1,	 	-- Roll Out (Legendary)
		[116849] = 	5.1,	 	-- Life Cocoon
		[122278] = 	5.1,	 	-- Dampen Harm
		[125174] = 	5.1,	 	-- Touch of Karma (Buff)
		[122783] = 	5.1,	 	-- Diffuse Magic
		[202162] = 	5.1,	 	-- Avert Harm (Brew PvP Talent)
		[209584] = 	5.1,	 	-- Zen Focus Tea (MW PvP Talent)
		[343249] = 	5.1,	 	-- Escape from Reality (MW Monk Legendary)
		[213664] = 	5.1,	 	-- Nimble Brew (Brew PvP Talent)
		[132578] = 	5.1,	 	-- Invoke Niuzao, the Black Ox
		[202248] = 	5.2, 	 	-- Guided Meditation (Brew PvP Talent)
		[353319] = 	5.2, 	 	-- Peaceweaver

		--Paladin
		[498] = 	5.1,	 	-- Divine Protection
		[1022] = 	5.1,	 	-- Blessing of Protection
		[204018] = 	5.1,	 	-- Blessing of Spellwarding
		[1044] = 	5.1,	 	-- Blessing of Freedom
		[305395] = 	5.1,	 	-- Blessing of Freedom with Unbound Freedom (PvP Talent)
		[6940] = 	5.1,	 	-- Blessing of Sacrifice
		[199448] = 	5.1,	 	-- Blessing of Sacrifice (Ultimate Sacrifice Holy PvP Talent)
		[199450] = 	5.1,	 	-- Ultimate Sacrifice (Holy PvP Talent) - debuff on the paladin
		[31821] = 	5.1,	 	-- Aura Mastery
		[31850] = 	5.1,	 	-- Ardent Defender
		[86659] = 	5.1,	 	-- Guardian of Ancient Kings
		[212641] = 	5.1,	 	-- Guardian of Ancient Kings (Glyphed)
		[184662] = 	5.1,	 	-- Shield of Vengeance
		[199545] = 	5.1,	 	-- Steed of Glory (Protection PvP Talent)
		[205191] = 	5.1,	 	-- Eye for an Eye
		[210256] = 	5.1,	 	-- Blessing of Sanctuary (Ret PvP Talent)
		[210294] = 	5.1,	 	-- Divine Favor (Holy PvP Talent)
		[157128] = 	5.1,	 	-- Saved by the Light
		[642] = 	5.2, 	 	-- Divine Shield
		[228050] = 	5.2,     	-- Guardian of the Forgotten Queen (Protection PvP Talent)

		--Priest
		[337661] = 	5.1,	 	-- Translucent Image (Fade defensive Conduit)
		[33206] = 	5.1,	 	-- Pain Suppression
		[47536] = 	5.1,	 	-- Rapture
		[109964] = 	5.1,	 	-- Spirit Shell
		[47585] = 	5.1,	 	-- Dispersion
		[47788] = 	5.1,	 	-- Guardian Spirit
		[64843] = 	5.1,	 	-- Divine Hymn
		[81782] = 	5.1,	 	-- Power Word: Barrier
		[232707] = 	5.1,	 	-- Ray of Hope (Holy PvP Talent)
		[197862] = 	5.1,	 	-- Archangel (Disc PvP Talent)
		[200183] = 	5.1,	 	-- Apotheosis
		[213610] = 	5.1,	 	-- Holy Ward
		--[27827] = 5.1,	 	-- Spirit of Redemption
		[215769] = 	5.1,	 	-- Spirit of Redemption (Spirit of the Redeemer Holy PvP Talent)
		[211336] = 	5.1,	 	-- Archbishop Benedictus' Restitution (Resurrection Buff)
		[211319] = 	5.1,	 	-- Archbishop Benedictus' Restitution (Debuff)
		[289655] = 	5.1,	 	-- Holy Word: Concentration
		[329543] = 	5.1,	 	-- Divine Ascension (down)
		[15286] = 	5.1,	 	-- Vampiric Embrace
		[19236] = 	5.1,	 	-- Desperate Prayer
		[213602] = 	5.2, 		-- Greater Fade (Holy/Shadow PvP Talent)
		[328530] = 	5.2, 		-- Divine Ascension (up)

		--Rogue
		[1966] = 	5.1,	 	-- Feint
		[5277] = 	5.1,	 	-- Evasion
		[11327] = 	5.1,	 	-- Vanish
		[45182] = 	5.1,	 	-- Cheating Death
		[199027] = 	5.1,	 	-- Veil of Midnight (Subtlety PvP Talent)
		[197003] = 	5.1,	 	-- Maneuverability (PvP Talent)

		--Shaman
		[108281] = 	5.1,	 	-- Ancestral Guidance
		[325174] = 	5.1,	 	-- Spirit Link Totem
		[204293] = 	5.1,	 	-- Spirit Link (PvP Talent)
		[108271] = 	5.1,	 	-- Astral Shift
		[210918] = 	5.1,	 	-- Ethereal Form (Enhancement PvP Talent)
		[114052] = 	5.1,	 	-- Ascendance (Restoration)
		[118337] = 	5.1,	 	-- Harden Skin
		[201633] = 	5.1,	 	-- Earthen Wall Totem
		[260881] = 	5.1,	 	-- Spirit Wolf
		[290641] = 	5.1,	 	-- Ancestral Gift
		[207495] = 	5.1,	 	-- Ancestral Protection (Totem)
		[207498] = 	5.1,	 	-- Ancestral Protection (Player)
		[104773] = 	5.1,	 	-- Unending Resolve
		[108416] = 	5.1,	 	-- Dark Pact
		[221705] = 	5.1,	 	-- Casting Circle (PvP Talent)
		[333889] = 	5.1,	 	-- Fel Domination
		[8178] = 	5.2, 		-- Grounding Totem Effect (PvP Talent)


		--Warrior
		[871] = 	5.1,	 	-- Shield Wall
		[12975] = 	5.1,	 	-- Last Stand
		[97463] = 	5.1,	 	-- Rallying Cry
		[118038] = 	5.1,	 	-- Die by the Sword
		[147833] = 	5.1,	 	-- Intervene
		[184364] = 	5.1,	 	-- Enraged Regeneration
		[197690] = 	5.1,	 	-- Defensive Stance
		[213871] = 	5.1,	 	-- Bodyguard (Prot PvP Talent)
		[236321] = 	5.1,	 	-- War Banner (PvP Talent)
		[227847] = 	5.2, 		-- Bladestorm (Arms)
		[46924] = 	5.2, 		-- Bladestorm (Fury)
		[23920] = 	5.2, 		-- Spell Reflection
		[330279] = 	5.2, 		-- Overwatch (PvP Talent)
		[335255] = 	5.2, 		-- Spell Reflection (Misshapen Mirror Legendary)

		--Death Knight
		[51271] = 	 4.9,	 -- Pillar of Frost
		[63560] = 	 4.9,	 -- Dark Transformation
		[152279] = 	 4.9,	 -- Breath of Sindragosa
		[47568] = 	 4.9,	 -- Empower Rune Weapon
		[207289] = 	 4.9,	 -- Unholy Assault
		[321995] = 	 4.9,	 -- Hypothermic Presence
		[315443] = 	 4.9,	 -- Abomination Limb (Necrolord Ability)
		[311648] = 	 4.9,	 -- Swarming Mist (Venthyr Ability)
		[48707] = 	 5.2, 		-- Anti-Magic Shell
		--Demon Hunter
		[162264] = 	 4.9, -- Metamorphosis - Havoc
		[188501] = 	 4.9, -- Spectral Sight
		[206804] = 	 4.9, -- Rain from Above (down)
		[337567] = 	 4.9, -- Chaotic Blades (Chaos Theory Legendary)
		--Druid		[5217] = 	 4.9 , -- Tiger's Fury
		[29166] = 	 4.9, -- Innervate
		[33891] = 	 4.9, -- Incarnation: Tree of Life (for the menu entry - "Incarnation" tooltip isn't informative)
		[117679] = 	 4.9, -- Incarnation (grants access to Tree of Life form)
		[102543] = 	 4.9, -- Incarnation: King of the Jungle
		[102558] = 	 4.9, -- Incarnation: Guardian of Ursoc
		[102560] = 	 4.9, -- Incarnation: Chosen of Elune
		[106951] = 	 4.9, -- Berserk (Feral)
		[132158] = 	 4.9, -- Nature's Swiftness
		[194223] = 	 4.9, -- Celestial Alignment
		[202425] = 	 4.9, -- Warrior of Elune
		[319454] = 	 4.9, -- Heart of the Wild
		[108291] = 	 4.9, -- with Balance Affinity
		[108292] = 	 4.9, -- with Feral Affinity
		[108293] = 	 4.9, -- with Guardian Affinity
		[108294] = 	 4.9, -- with Resto Affinity
		[323764] = 	 4.9, -- Convoke the Spirits (Night Fae Ability)
		[323546] = 	 4.9, -- Ravenous Frenzy (Venthyr Ability)
		[338142] = 	 4.9, -- Lone Empowerment (Kyrian Ability)
		--Hunter		[19574] = 	 4.9 , -- Bestial Wrath
		[186254] = 	 4.9, -- Bestial Wrath buff on the pet
		[186289] = 	 4.9, -- Aspect of the Eagle
		[193530] = 	 4.9, -- Aspect of the Wild
		[260402] = 	 4.9, -- Double Tap
		[266779] = 	 4.9, -- Coordinated Assault
		[288613] = 	 4.9, -- Trueshot
		--Mage		[66] = 		 4.9 , -- Invisibility (Countdown)
		[32612] = 	 4.9, -- Invisibility
		[113862] = 	 4.9, -- Greater Invisibility
		[12042] = 	 4.9, -- Arcane Power
		[12051] = 	 4.9, -- Evocation
		[12472] = 	 4.9, -- Icy Veins
		[198144] = 	 4.9, -- Ice Form
		[342242] = 	 4.9, -- Time Warp procced by Time Anomality (Arcane Talent)
		[190319] = 	 4.9, -- Combustion
		[198158] = 	 4.9, -- Mass Invisibility (Arcane PvP Talent)
		[205025] = 	 4.9, -- Presence of Mind
		[333100] = 	 4.9, -- Firestorm (Fire Legendary)
		[324220] = 	 4.9, -- Deathborne (Necrolord Ability)
		--Monk		[137639] = 	 4.9 , -- Storm, Earth, and Fire
		[152173] = 	 4.9, -- Serenity
		[310454] = 	 4.9, -- Weapons of Order (Kyrian Ability)
		[202335] = 	 4.9, -- Double Barrel (Brew PvP Talent) - "next cast will..." buff
		--Paladin		[31884] = 	 4.9 }, -- Avenging Wrath
		[216331] = 	 4.9, -- Avenging Crusader (Holy Talent)
		[231895] = 	 4.9, -- Crusade (Retribution Talent)
		[105809] = 	 4.9, -- Holy Avenger
		[152262] = 	 4.9, -- Seraphim
		[215652] = 	 4.9, -- Shield of Virtue (Protection PvP Talent) - "next cast will..." buff
		--Priest		[10060] = 	 4.9 , -- Power Infusion
		[194249] = 	 4.9, -- Voidform
		[197871] = 	 4.9, -- Dark Archangel (Disc PvP Talent) - on the priest
		[197874] = 	 4.9, -- Dark Archangel (Disc PvP Talent) - on others
		[319952] = 	 4.9, -- Surrender to Madness
		[322431] = 	 4.9, -- Thoughtsteal (Buff)
		[325013] = 	 4.9, -- Boon of the Ascended (Kyrian)
				--Rogue		[13750] = 	 4.9 , -- Adrenaline Rush
		[51690] = 	 4.9, -- Killing Spree
		[121471] = 	 4.9, -- Shadow Blades
		[185422] = 	 4.9, -- Shadow Dance
		[207736] = 	 4.9, -- Shadowy Duel
		[212283] = 	 4.9, -- Symbols of Death
		[115192] = 	 4.9, -- Subterfuge
		[256735] = 	 4.9, -- Master Assassin
		[340094] = 	 4.9, -- Master Assassin's Mark (Legendary)
		[345569] = 	 4.9, -- Flagellation (Venthyr Ability)
		[347037] = 	 4.9, -- Sepsis (Night Fae Ability)
		--Shaman		[114049] = 	 4.9 , -- Ascendance
		[114050] = 	 4.9, -- Ascendance (Elemental)
		[114051] = 	 4.9, -- Ascendance (Enhancement)
		[191634] = 	 4.9, -- Stormkeeper (Ele)
		[320137] = 	 4.9, -- Stormkeeper (Enh)
		[204366] = 	 4.9, -- Thundercharge (Enhancement PvP Talent)
		[335903] = 	 4.9, -- Doom Winds
		[320125] = 	 4.9, -- Echoing Shock
		[333957] = 	 4.9, -- Feral Spirit
		[204361] = 	 4.9, -- Bloodlust (Enhancement PvP Talent)
		[204362] = 	 4.9, -- Heroism (Enhancement PvP Talent)
		[327164] = 	 4.9, -- Primordial Wave (Necrolord Ability)
		--Warlock		[113860] = 	 4.9 , -- Dark Soul: Misery (Affliction)
		[113858] = 	 4.9, -- Dark Soul: Instability (Destruction)
		[265273] = 	 4.9, -- Demonic Power (Demonic Tyrant)
		[344566] = 	 4.9, -- Rapid Contagion (Affliction PvP Talent)
		[267171] = 	 4.9, -- Demonic Strength
		[267218] = 	 4.9, -- Nether Portal
		[212295] = 	 5.2,    -- Nether Ward (PvP Talent)

		--Warrior
		[1719] =	 4.9, -- Recklessness
		[107574] = 	 4.9, -- Avatar
		[262228] = 	 4.9, -- Deadly Calm
		[198817] = 	 4.9, -- Sharpen Blade (Arms PvP Talent) - "next cast will..." buff
		[324143] = 	 4.9, -- Conqueror's Banner (Necrolord Ability) - on the warrior
		[325862] = 	 4.9, -- Conqueror's Banner (Necrolord Ability) - on others
	},
	HARMFUL = {
		--Death Knight
		[77606] =	 4.9, -- Dark Simulacrum
		[223929] = 	 4.9, -- Necrotic Wound
		[343294] = 	 4.9, -- Soul Reaper
		[288849] = 	 4.9, -- Crypt Fever (Necromancer's Bargain Unholy PvP Talent)

		--Demon Hunter
		[206649] = 	 4.9, -- Eye of Leotheras
		[203704] = 	 4.9, -- Mana Break

		--Hunter
		[131894] = 	 4.9, -- A Murder of Crows

		--Monk
		[122470] = 	 4.9, -- Touch of Karma (Debuff)
		[344021] = 	 4.9, -- Keefer's Skyreach

		--Paladin
		[343721] = 	 4.9, -- Final Reckoning


		[322459] =	 4.9, -- Thoughtstolen (Shaman)
		[322464] = 	 4.9, -- Thoughtstolen (Mage)
		[322442] = 	 4.9, -- Thoughtstolen (Druid)
		[322462] = 	 4.9, -- Thoughtstolen (Priest - Holy)
		[322457] = 	 4.9, -- Thoughtstolen (Paladin)
		[322463] = 	 4.9, -- Thoughtstolen (Warlock)
		[322461] = 	 4.9, -- Thoughtstolen (Priest - Discipline)
		[322458] = 	 4.9, -- Thoughtstolen (Monk)
		[322460] = 	 4.9, -- Thoughtstolen (Priest - Shadow)

		--Priest
		[323673] =	 4.9, -- Mindgames
		[335467] =	 4.9, -- Devouring Plague
		[34914] =	 4.9, -- Vampiric Touch
		[199845] =	 4.9, -- Psyflay (Psyfiend) debuff

		--Rogue
		[79140] =	 4.9, -- Vendetta
		[328305] =	 4.9, -- Sepsis (Night Fae Ability)

		--Shaman
		[208997] =	 4.9, -- Counterstrike Totem (PvP Talent)

		--Warlock
		[234877] =	 4.9, -- Bane of Shadows (Affliction PvP Talent)
		[316099] =	 4.9, -- Unstable Affliction
		[342938] = 	 4.9, -- Unstable Affliction (Affliction PvP Talent)
		[30213] =	 4.9, -- Legion Strike
		[200587] =	 4.9, -- Fel Fissure (PvP Talent)
		[80240] =	 4.9, -- Havoc
		[200548] = 	 4.9, -- Bane of Havoc (Destro PvP Talent)

		--Warrior
		[208086] =	 4.9, -- Colossus Smash
		[198819] =	 4.9, -- Mortal Strike when applied with Sharpen Blade (50% healing reduc)
	}
}

for spellId, priority in pairs(Data.PriorityAuras) do
	if not Data.SpellPriorities[spellId] then
		Data.SpellPriorities[spellId] = priority
	end
end


Data.BattlegroundspezificBuffs = { --key = mapID, value = table with key = faction(0 for hode, 1 for alliance) value spellId of the flag, minecart
	[1339] = {						-- Warsong Gulch, used to be mapID 443 before BFA
		[0] = 156621, 					-- Alliance Flag
		[1] = 156618 					-- Horde Flag
	},
	[1460] = {						-- Warsong Gulch, used in Classic, TBCC
		[0] = 301091, 					-- Alliance Flag
		[1] = 301089 					-- Horde Flag
	},
	[112] = {						-- Eye of the Storm, used to be mapID 482 before BFA
		[0] = 34976,  					-- Netherstorm Flag
		[1] = 34976						-- Netherstorm Flag
	},
	[1956] = {						-- Eye of the Storm, TBCC
		[0] = 34976,  					-- Netherstorm Flag
		[1] = 34976						-- Netherstorm Flag
	},
	[397] = {						-- Eye of the Storm (mapID RBG only? Not sure why there are two map IDs for Eye of the Storm), used to be mapID 813 before BFA
		[0] = 34976,  					-- Netherstorm Flag
		[1] = 34976						-- Netherstorm Flag
	},
	[206] = {						-- Twin Peaks, used to be mapID 626 before BFA
		[0] = 156621, 					-- Alliance Flag
		[1] = 156618 					-- Horde Flag
	},
	[2345] = {						-- Deephaul Ravine added in Patch 11.0.2
		[0] = 434339, 					-- Deephaul Crystal
		[1] = 434339 					-- Deephaul Crystal
	}
}


Data.BattlegroundspezificDebuffs = { --key = mapID, value = table with key = number and value = debuff name
	[1339] = {						-- Warsong Gulch, used to be mapID 443 before BFA
		46392,						-- Focused Assault
		46393						-- Brutal Assault
	},
	[112] = {						-- Eye of the Storm, used to be mapID 482 before BFA
		46392,						-- Focused Assault
		46393						-- Brutal Assault
	},
	[397] = {						-- Eye of the Storm (mapID RBG only? Not sure why there are two map IDs for Eye of the Storm), used to be mapID 813 before BFA
		46392,						-- Focused Assault
		46393						-- Brutal Assault
	},
	[1956] = {						-- Eye of the Storm, TBCC
		46392,						-- Focused Assault
		46393						-- Brutal Assault
	},
	[206] = {						-- Twin Peaks, used to be mapID 626 before BFA
		46392,						-- Focused Assault
		46393						-- Brutal Assault
	},
	[417] = {						-- Temple of Kotmogu, used to be mapID 856 before BFA
		121164, 					-- Orb of Power, Blue
		121175, 					-- Orb of Power, Purple
		121177, 					-- Orb of Power, Orange
		121176 						-- Orb of Power, Green
	}
}



--[[ 
	Classic Data
	Insignia of the Alliance 5 min CD
	Insignia of the Horde 5 min CD

	Medallions got added in TBC and require level 70

	Medallion of the Alliance 2 min CD
	Medallion of the Horde 2 min CD
 ]]

local trinketCD
if IsRetail or UnitLevel("player") >= 70 then
	trinketCD = 120
else
	trinketCD = 300
end

Data.TrinketData = {
	[195710] = {cd = 180											},		-- 1: Honorable Medallion, 3. min. CD, detected by Combatlog
	[42292]  = {cd = trinketCD, itemID = 37865						},		-- 2: Medallion of the Alliance, Medallion of the Horde used in Classic, TBC, and probably some other Expansions  2 min. CD, detected by Combatlog, should show as Medaillon; used in TBC etc
	[208683] = {cd = 120											}, 		-- 2: Gladiator's Medallion, 2 min. CD, detected by Combatlog
	[336126] = {cd = 120											},		-- 2: Gladiator's Medallion, 2 min. CD, Shadowlands Update
--	[195901] = {cd = 60, fileID = GetSpellTexture(214027)			},		-- 3: Adaptation, 1 min. CD, detected by Aura 195901
--	[214027] = {cd = 60												},		-- 3: Adaptation, 1 min. CD, detected by Aura 195901, for the Arena_cooldownupdate
--	[336135] = {cd = 60												},		-- 3: Adaptation, 1 min. CD, Shadowlands Update
--	[336139] = {cd = 60, fileID = GetSpellTexture(214027)			},		-- 3: Adapted, 1 min. CD, Shadowlands Update
	[196029] = {cd = false											}, 		-- 4: Relentless, passive, no CD
	[336128] = {cd = false											}, 		-- 4: Relentless, passive, no CD, Shadowlands Update
	[363117] = {cd = false											},		-- 5: Gladiator's Fastidious Resolve, Added in Shadowlands Patch 9.2
}


if IsClassic then
	--Classic: there are no spellIDs in the combat log in classic, only spellnames
	Mixin(Data.TrinketData, {
		[GetSpellName(23273)] = {spellId = 23273, cd = 300, itemID = 18856},	-- Immune Charm/Fear/Polymorph	Rogue, Warlock
		[GetSpellName(23274)] = {spellId = 23274, cd = 300, itemID = 18856}, 	-- Immune Fear/Polymorph/Snare	Mage
		[GetSpellName(23276)] = {spellId = 23276, cd = 300, itemID = 18856},	-- Immune Fear/Polymorph/Stun	Paladin, Priest
		[GetSpellName(23227)] = {spellId = 23227, cd = 300, itemID = 18856}		-- Immune Charm/Fear/Stun		Druid
	})
end






--see for covenant colors SharedXML/SharedColorConstants.lua

Data.CovenantIcons = {
	[1] = GetSpellTexture(324739), --Kyrian
	[2] = GetSpellTexture(300728), --Ventyr
	[3] = GetSpellTexture(310143), --Night Fae
	[4] = GetSpellTexture(324631), --Necrolord
}


--C_Covenants.GetCovenantData()
Data.CovenantSpells = {
	--Kyrian
	[324739] = 1,--Summon Steward 				All Classes

	[312202] = 1,--Shackle the Unworthy 		Death Knight
	[306830] = 1,--Elysian Decree 				Demon Hunter
	[326434] = 1,--Kindred Spirits 				Druid
	[308491] = 1,--Resonating Arrow 			Hunter
	[307443] = 1,--Radiant Spark 				Mage
	[310454] = 1,--Weapons of Order 			Monk
	[304971] = 1,--Divine Toll 					Paladin
	[325013] = 1,--Boon of the Ascended 		Priest
	[323547] = 1,--Echoing Reprimand 			Rogue
	[324386] = 1,--Vesper Totem 				Shaman
	[312321] = 1,--Scouring Tithe 				Warrior
	[307865] = 1,--Spear of Bastion 			Warlock


	--Ventyr
	[300728] = 2, --Door of Shadows 			All Classes

	[311648] = 2,--Swarming Mist 				Death Knight
	[317009] = 2,--Sinful Brand  				Demon Hunter
	[323546] = 2,--Ravenous Frenzy				Druid
	[324149] = 2,--Flayed Shot 					Hunter
	[314793] = 2,--Mirrors of Torment 			Mage
	[326860] = 2,--Fallen Order 				Monk
	[316958] = 2,--Ashen Hallow 				Paladin
	[323673] = 2,--Mindgames 					Priest
	[323654] = 2,--Flagellation 				Rogue
	[320674] = 2,--Chain Harvest 				Shaman
	[321792] = 2,--Impending Catastrophe  		Warrior
	[317349] = 2,--Condemn 						Warlock


	--Night Fae
	[310143] = 3, --Soulshape 					All Classes

	[324128] = 3,--Death's Due 					Death Knight
	[323639] = 3,--The Hunt (ability)  			Demon Hunter
	[323764] = 3,--Convoke the Spirits 			Druid
	[328231] = 3,--Wild Spirits 				Hunter
	[314791] = 3,--Shifting Power 				Mage
	[327104] = 3,--Faeline Stomp				Monk
	[328278] = 3,--Blessing of the Seasons 		Paladin
	[327661] = 3,--Fae Guardians 				Priest
	[328305] = 3,--Sepsis 						Rogue
	[328923] = 3,--Fae Transfusion				Shaman
	[325640] = 3,--Soul Rot						Warrior
	[325886] = 3,--Ancient Aftershock 			Warlock


	--Necrolord
	[324631] = 4, --Fleshcraft					All Classes

	[315443] = 4,--Abomination Limb				Death Knight
	[329554] = 4,--Fodder to the Flame  		Demon Hunter
	[325727] = 4,--Adaptive Swarm 				Druid
	[325028] = 4,--Death Chakram 				Hunter
	[324220] = 4,--Deathborne 					Mage
	[325216] = 4,--Bonedust Brew 				Monk
	[328204] = 4,--Vanquisher's Hammer 			Paladin
	[324724] = 4,--Unholy Nova 					Priest
	[328547] = 4,--Serrated Bone Spike 			Rogue
	[326059] = 4,--Primordial Wave 				Shaman
	[325289] = 4,--Decimating Bolt  			Warrior
	[324143] = 4,--Conqueror's Banner 			Warlock
}

Data.RacialSpellIDtoCooldown = {
	 [7744] = {cd = 120, trinketCD = 30						},					--Will of the Forsaken, Undead Racial, 30 sec cooldown trigger on trinket
	[20594] = {cd = 120 									},					--Stoneform, Dwarf Racial
	[58984] = {cd = 120 									},					--Shadowmeld, Night Elf Racial
	[59752] = {cd = IsRetail and 180 or 120, trinketCD = 90 },  				--Every Man for Himself, Human Racial, 90 sec cooldown trigger on trinket, got renamed in Shadowlands to Will to Survive
	[28730] = {cd = 90  									},					--Arcane Torrent, Blood Elf Racial, Mage and Warlock,
	[50613] = {cd = 90  									},					--Arcane Torrent, Blood Elf Racial, Death Knight,
   [202719] = {cd = 90  									},					--Arcane Torrent, Blood Elf Racial, Demon Hunter,
	[80483] = {cd = 90  									},					--Arcane Torrent, Blood Elf Racial, Hunter,
   [129597] = {cd = 90  									},					--Arcane Torrent, Blood Elf Racial, Monk,
   [155145] = {cd = 90  									},					--Arcane Torrent, Blood Elf Racial, Paladin,
   [232633] = {cd = 90  									},					--Arcane Torrent, Blood Elf Racial, Priest,
	[25046] = {cd = 90  									},					--Arcane Torrent, Blood Elf Racial, Rogue,
	[69179] = {cd = 90  									},					--Arcane Torrent, Blood Elf Racial, Warrior,
	[20589] = {cd = 90  									}, 					--Escape Artist, Gnome Racial
	[26297] = {cd = 180 									},					--Berserkering, Troll Racial
	[33702] = {cd = 120 									},					--Blood Fury, Orc Racial, Mage,  Warlock
	[20572]	= {cd = 120 									},					--Blood Fury, Orc Racial, Warrior, Hunter, Rogue, Death Knight
	[33697] = {cd = 120 									},					--Blood Fury, Orc Racial, Shaman, Monk
	[20577] = {cd = 120 									}, 					--Cannibalize, Undead Racial
	[68992]	= {cd = 120 									},					--Darkflight, Worgen Racial
	[59545] = {cd = 180 									},					--Gift of the Naaru, Draenei Racial, Death Knight
	[59543] = {cd = 180 									},					--Gift of the Naaru, Draenei Racial, Hunter
	[59548] = {cd = 180 									},					--Gift of the Naaru, Draenei Racial, Mage
   [121093]	= {cd = 180 									},					--Gift of the Naaru, Draenei Racial, Monk
	[59542] = {cd = 180 									},					--Gift of the Naaru, Draenei Racial, Paladin
	[59544] = {cd = 180 									},					--Gift of the Naaru, Draenei Racial, Priest
	[59547] = {cd = 180 									},					--Gift of the Naaru, Draenei Racial, Shaman
	[28880] = {cd = 180 									},					--Gift of the Naaru, Draenei Racial, Warrior
   [107079] = {cd = 120 									},					--Quaking Palm, Pandaren Racial
	[69041] = {cd = 90  									},					--Rocket Barrage, Goblin Racial
	[69070] = {cd = 90  									},					--Rocket Jump, Goblin Racial
	[20549] = {cd = 90  									},					--War Stomp, Tauren Racial
   [312411] = {cd = 90										},					--Bag of Tricks, Vulpera Racial
   [265221] = {cd = 120										},					--Fireblood, Dark Iron Dwarf Racial
   [255647] = {cd = 150										},					--Light's Judgment, Lightforged Draenei Racial
   [255654] = {cd = 120										},					--Bull Rush, Highmountain Tauren Racial
   [287712] = {cd = 150										}, 					--Haymaker, Kul Tiran Racial
   [260364] = {cd = 180										}					--Arcane Pulse, Nightborne Racial
}

if IsWrath then
	Data.RacialSpellIDtoCooldown = {
		[7744] = {cd = 120 				},					--Will of the Forsaken, Undead Racial, 30 sec cooldown trigger on trinket
	   [20594] = {cd = 120 				},					--Stoneform, Dwarf Racial
	   [58984] = {cd = 120 				},					--Shadowmeld, Night Elf Racial
	   [20549] = {cd = 120 				},					--War Stomp, Tauren Racial
	   [28730] = {cd = 120  			},					--Arcane Torrent, Blood Elf Racial, Mage and Warlock,
	   [50613] = {cd = 120  			},					--Arcane Torrent, Blood Elf Racial, Death Knight,
      [202719] = {cd = 120  			},					--Arcane Torrent, Blood Elf Racial, Demon Hunter,
	   [80483] = {cd = 120  			},					--Arcane Torrent, Blood Elf Racial, Hunter,
      [129597] = {cd = 120  			},					--Arcane Torrent, Blood Elf Racial, Monk,
      [155145] = {cd = 120  			},					--Arcane Torrent, Blood Elf Racial, Paladin,
      [232633] = {cd = 120  			},					--Arcane Torrent, Blood Elf Racial, Priest,
	   [25046] = {cd = 120  			},					--Arcane Torrent, Blood Elf Racial, Rogue,
	   [69179] = {cd = 120  			},					--Arcane Torrent, Blood Elf Racial, Warrior,
	   [33702] = {cd = 120 				},					--Blood Fury, Orc Racial, Mage,  Warlock
	   [20572] = {cd = 120 				},					--Blood Fury, Orc Racial, Warrior, Hunter, Rogue, Death Knight
	   [33697] = {cd = 120 				},					--Blood Fury, Orc Racial, Shaman, Monk
	   [20589] = {cd = 105  			}, 					--Escape Artist, Gnome Racial
	}
end


Data.RacialNameToSpellIDs = {}
Data.Racialnames = {}
for spellId in pairs(Data.RacialSpellIDtoCooldown) do
	local racialName = GetSpellName(spellId)
	if racialName and racialName ~= "" then
		if not Data.RacialNameToSpellIDs[racialName] then
			Data.RacialNameToSpellIDs[racialName] = {}
			Data.Racialnames[racialName] = racialName
		end
		Data.RacialNameToSpellIDs[racialName][spellId] = true
	end
end


-- Data.EnemiesRangeToRange = {}
-- Data.EnemiesItemIDToRange = {}
-- Data.EnemiesRangeToItemID	= {
-- 	[5] = 37727, -- Ruby Acorn --works, also for allies,
-- 	[6] = 63427, -- Worgsaw --works, also for allies,
-- 	[8] = 34368, -- Attuned Crystal Cores --works, also for allies,
-- 	[10] = 32321, -- Sparrowhawk Net
-- 	[15] = 33069, -- Sturdy Rope --doesn't work on allies
-- 	[20] = 10645, -- Gnomish Death Ray --doesn't work on allies
-- 	[25] = 24268, -- Netherweave Net --doesn't work on allies
-- 	[30] = 34191, -- Handful of Snowflakes
-- 	[35] = 18904, -- Zorbin's Ultra-Shrinker
-- 	[40] = 28767, -- The Decapitator --doesn't work on allies
-- 	[45] = 32698, -- Wrangling Rope
-- 	[50] = 116139, -- Haunting Memento
-- 	[60] = 32825, -- Soul Cannon
-- 	[70] = 41265, -- Eyesore Blaster
-- 	[80] = 35278, -- Reinforced Net
-- 	[100] = 33119, -- Malister's Frost Wand --doesn't work on allies
-- }
-- Data.AlliesRangeToRange = {}
-- Data.AlliesItemIDToRange = {}
-- Data.AlliesRangeToItemID	= {
-- 	[5] = 37727, -- Ruby Acorn
-- 	[6] = 63427, -- Worgsaw
-- 	[8] = 34368, -- Attuned Crystal Cores
-- 	[10] = 32321, -- Sparrowhawk Net
-- 	[15] = 6450, -- Silk Bandage
-- 	[20] = 21519, -- Mistletoe
-- 	[25] = 31463, -- Zezzak's Shard
-- 	[30] = 34191, -- Handful of Snowflakes
-- 	[35] = 18904, -- Zorbin's Ultra-Shrinker
-- 	[40] = 34471, -- Vial of the Sunwell
-- 	[45] = 32698, -- Wrangling Rope
-- 	[50] = 116139, -- Haunting Memento
-- 	[60] = 32825, -- Soul Cannon
-- 	[70] = 41265, -- Eyesore Blaster
-- 	[80] = 35278, -- Reinforced Net
-- }

-- for range, itemID in next, Data.EnemiesRangeToItemID do
-- 	Data.EnemiesRangeToRange[range] = range --for testmode
-- 	Data.EnemiesItemIDToRange[itemID] = range --for testmode
-- end

-- for range, itemID in next, Data.AlliesRangeToItemID do
-- 	Data.AlliesRangeToRange[range] = range --for testmode
-- 	Data.AlliesItemIDToRange[itemID] = range --for testmode
-- end




Data.Classes = {}
Data.RolesToSpec = {HEALER = {}, TANK = {}, DAMAGER = {}} --for Testmode only
Data.ClassList = {} -- For TBCC Testmode only

do

	local specIdToRessource = {
		--Death Knight
		[250] = "RUNIC_POWER",	--Blood
		[251] = "RUNIC_POWER",	--Frost
		[252] = "RUNIC_POWER",	--Unholy
		--Demon Hunter
		[577] = "FURY",			--Havoc
		[581] = "PAIN",			--Vengeance
		--Druid
		[102] = "LUNAR_POWER",	--Balance
		[103] = "ENERGY",		--Feral Combat
		[104] = "RAGE",			--Guardian
		[105] = "MANA",			--Restoration
		--Evoker
		[1467] = "Essence", 	--Devastation
		[1468] = "Essence",		--Preservation
		--Hunter
		[253] = "FOCUS",		--Beast Mastery
		[254] = "FOCUS",		--Marksmanship
		[255] = "FOCUS",		--Survival
		--Mage
		[62] = "MANA",			--Arcane
		[63] = "MANA",			--Fire
		[64] = "MANA",			--Frost
		--Monk
		[268] = "ENERGY",		--Brewmaster
		[269] = "ENERGY",		--Windwalker
		[270] = "MANA",			--Mistweaver
		--Paladin
		[65] = "MANA",			--Holy
		[66] = "MANA",			--Protection
		[70] = "MANA",			--Retribution
		--Priest
		[256] = "MANA",			--Discipline
		[257] = "MANA",			--Holy
		[258] = "INSANITY",		--Shadow
		--Rogue
		[259] = "ENERGY",		--Assassination
		[260] = "ENERGY",		--Outlaw
		[261] = "ENERGY",		--Subtlety
		--Shaman,
		[262] = "MAELSTROM",	--Elemental
		[263] = "MAELSTROM",	--Enhancement
		[264] = "MANA",			--Restoration
		--Warlock
		[265] = "MANA",			--Affliction
		[266] = "MANA",			--Demonology
		[267] = "MANA",			--Destruction
		--Warrior
		[71] = "RAGE",			--Arms
		[72] = "RAGE",			--Fury
		[73] = "RAGE"			--Protection
	}

	local ClassRessources = { --used for TBCC
		WARRIOR = "RAGE",
		PALADIN = "MANA",
		HUNTER = "MANA",
		ROGUE = "ENERGY",
		PRIEST = "MANA",
		SHAMAN = "MANA",
		MAGE = "MANA",
		WARLOCK = "MANA",
		DRUID = "MANA",
		MONK = "ENERGY",
		DEATHKNIGHT = "RUNIC_POWER",
		DEMONHUNTER = "FURY",
		EVOKER = "ESSENCE"
	}


	for classID = 1, GetNumClasses() do --example classes[EnglishClass][SpecName].
		local _, classToken = GetClassInfo(classID)
		if classToken then
			Data.Classes[classToken] = {Ressource = ClassRessources[classToken]}-- for Classic, TBCC Wrath, and any other expansions without specs and some brawls

			
			if GetNumSpecializationsForClassID and GetSpecializationInfoByID and GetSpecialization then --HasSpeccs
				for i = 1, GetNumSpecializationsForClassID(classID) do
					local specID,maleSpecName,_,icon,role = GetSpecializationInfoForClassID(classID, i, 2) -- male version				

					Data.Classes[classToken][maleSpecName] = {roleID = role, specID = specID, specIcon = icon, Ressource = specIdToRessource[specID]}
					table.insert(Data.RolesToSpec[role], {classToken = classToken, specName = maleSpecName}) --for testmode

					--if specName == "Танцующий с ветром" then specName = "Танцующая с ветром" end -- fix for russian bug, fix added on 2017.08.27
					local specID,specName,_,icon,role = GetSpecializationInfoForClassID(classID, i, 3) -- female version
					if not Data.Classes[classToken][specName] then --there is a female version of that specName
						Data.Classes[classToken][specName] = Data.Classes[classToken][maleSpecName]
					end
				end
			else
				table.insert(Data.ClassList, classToken)
			end
		end
	end
end

Data["FakeAuras"] = {
	["HARMFUL"] = {
		[117526] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = true,
			["duration"] = 8,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 462650,
			["spellId"] = 117526,
		},
		[207777] = {
			["spellId"] = 207777,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236272,
			["nameplateShowPersonal"] = false,
		},
		[353084] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 6,
			["spellId"] = 353084,
			["sourceUnit"] = "raid6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4067368,
			["dispelName"] = "Magic",
		},
		[131894] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 131894,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 645217,
			["nameplateShowPersonal"] = true,
		},
		[154953] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 154953,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236268,
			["nameplateShowPersonal"] = true,
		},
		[236777] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0.102,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 236777,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135826,
			["nameplateShowPersonal"] = false,
		},
		[352448] = {
			["spellId"] = 352448,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1391773,
			["dispelName"] = "Magic",
		},
		[115994] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 14,
			["spellId"] = 115994,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136132,
			["dispelName"] = "Disease",
		},
		[8122] = {
			["spellId"] = 8122,
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136184,
			["dispelName"] = "Magic",
		},
		[328305] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 9,
			["spellId"] = 328305,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636848,
			["dispelName"] = "Poison",
		},
		[115196] = {
			["spellId"] = 115196,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132274,
			["dispelName"] = "Poison",
		},
		[357305] = {
			["spellId"] = 357305,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136194,
			["nameplateShowPersonal"] = false,
		},
		[344658] = {
			["spellId"] = 344658,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 1.5,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136130,
			["nameplateShowPersonal"] = false,
		},
		[236077] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 236077,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132343,
			["nameplateShowPersonal"] = false,
		},
		[226943] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 4.8,
			["spellId"] = 226943,
			["sourceUnit"] = "raid15",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136173,
			["dispelName"] = "Magic",
		},
		[257284] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 30,
			["spellId"] = 257284,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236188,
			["dispelName"] = "Magic",
		},
		[344021] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 344021,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 606551,
			["nameplateShowPersonal"] = false,
		},
		[264689] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 600,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 264689,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136035,
			["nameplateShowPersonal"] = false,
		},
		[351687] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 351687,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3510985,
			["nameplateShowPersonal"] = false,
		},
		[160067] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 160067,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237430,
			["nameplateShowPersonal"] = false,
		},
		[115804] = {
			["spellId"] = 115804,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "partypet1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132109,
			["nameplateShowPersonal"] = false,
		},
		[79140] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 79140,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 458726,
			["nameplateShowPersonal"] = false,
		},
		[285515] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 6,
			["spellId"] = 285515,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135849,
			["dispelName"] = "Magic",
		},
		[106830] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 106830,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 451161,
			["nameplateShowPersonal"] = true,
		},
		[26013] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 900,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 26013,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132118,
			["nameplateShowPersonal"] = false,
		},
		[5116] = {
			["spellId"] = 5116,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135860,
			["nameplateShowPersonal"] = false,
		},
		[80354] = {
			["spellId"] = 80354,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 600,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 458224,
			["nameplateShowPersonal"] = false,
		},
		[102359] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 6.8,
			["spellId"] = 102359,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 538515,
			["dispelName"] = "Magic",
		},
		[335467] = {
			["spellId"] = 335467,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 252997,
			["dispelName"] = "Disease",
		},
		[17877] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 17877,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136191,
			["nameplateShowPersonal"] = true,
		},
		[157131] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 157131,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1030102,
			["nameplateShowPersonal"] = false,
		},
		[772] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 772,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132155,
			["nameplateShowPersonal"] = true,
		},
		[364851] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 1,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 364851,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 612968,
			["nameplateShowPersonal"] = false,
		},
		[353354] = {
			["spellId"] = 353354,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 4,
			["applications"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3854013,
			["nameplateShowPersonal"] = false,
		},
		[320267] = {
			["spellId"] = 320267,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 1.7,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136090,
			["dispelName"] = "Magic",
		},
		[363830] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 363830,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1396978,
			["nameplateShowPersonal"] = false,
		},
		[186387] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate19",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 186387,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1376038,
			["nameplateShowPersonal"] = false,
		},
		[3600] = {
			["spellId"] = 3600,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136102,
			["dispelName"] = "Magic",
		},
		[118297] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 21,
			["spellId"] = 118297,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135817,
			["dispelName"] = "Magic",
		},
		[196608] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 196608,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236165,
			["nameplateShowPersonal"] = false,
		},
		[228354] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 1,
			["spellId"] = 228354,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1506795,
			["dispelName"] = "Magic",
		},
		[363450] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 180,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3741938,
			["spellId"] = 363450,
		},
		[453] = {
			["spellId"] = 453,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135933,
			["nameplateShowPersonal"] = false,
		},
		[339051] = {
			["spellId"] = 339051,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 9,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 666474,
			["dispelName"] = "Magic",
		},
		[363452] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 180,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3741938,
			["spellId"] = 363452,
		},
		[194310] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["timeMod"] = 1,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["applications"] = 2,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1129420,
			["spellId"] = 194310,
		},
		[47476] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 4,
			["spellId"] = 47476,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136214,
			["dispelName"] = "Magic",
		},
		[118905] = {
			["spellId"] = 118905,
			["nameplateShowAll"] = true,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136013,
			["dispelName"] = "Magic",
		},
		[325640] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 8,
			["spellId"] = 325640,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636850,
			["dispelName"] = "Magic",
		},
		[2094] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = true,
			["duration"] = 8,
			["timeMod"] = 1,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136175,
			["spellId"] = 2094,
		},
		[2096] = {
			["spellId"] = 2096,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135934,
			["nameplateShowPersonal"] = false,
		},
		[288849] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 288849,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136066,
			["nameplateShowPersonal"] = false,
		},
		[354896] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 3.2,
			["spellId"] = 354896,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1398086,
			["dispelName"] = "Poison",
		},
		[204213] = {
			["spellId"] = 204213,
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236216,
			["dispelName"] = "Magic",
		},
		[228358] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 2,
			["isBossAura"] = false,
			["duration"] = 6,
			["spellId"] = 228358,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135850,
			["dispelName"] = "Magic",
		},
		[117405] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 462650,
			["spellId"] = 117405,
		},
		[363459] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 180,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3741938,
			["spellId"] = 363459,
		},
		[363715] = {
			["spellId"] = 363715,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1028995,
			["dispelName"] = "Magic",
		},
		[5246] = {
			["spellId"] = 5246,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132154,
			["nameplateShowPersonal"] = false,
		},
		[2120] = {
			["spellId"] = 2120,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135826,
			["nameplateShowPersonal"] = false,
		},
		[363461] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 180,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3741938,
			["spellId"] = 363461,
		},
		[41425] = {
			["spellId"] = 41425,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135785,
			["nameplateShowPersonal"] = false,
		},
		[1833] = {
			["spellId"] = 1833,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 4,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132092,
			["nameplateShowPersonal"] = false,
		},
		[121175] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = true,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 121175,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 1119888,
			["nameplateShowPersonal"] = false,
		},
		[217694] = {
			["spellId"] = 217694,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = true,
			["isBossAura"] = false,
			["duration"] = 3.076,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236220,
			["dispelName"] = "Magic",
		},
		[82691] = {
			["spellId"] = 82691,
			["nameplateShowAll"] = true,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 464484,
			["dispelName"] = "Magic",
		},
		[332423] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 2.25,
			["spellId"] = 332423,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1990984,
			["dispelName"] = "Magic",
		},
		[1330] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 1330,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132297,
			["nameplateShowPersonal"] = false,
		},
		[121176] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = true,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 121176,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 1119886,
			["nameplateShowPersonal"] = false,
		},
		[209967] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 209967,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 877476,
			["nameplateShowPersonal"] = false,
		},
		[46392] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136004,
			["spellId"] = 46392,
		},
		[1079] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 24,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 1079,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132152,
			["nameplateShowPersonal"] = true,
		},
		[257044] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 1.824,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 461115,
			["spellId"] = 257044,
		},
		[323996] = {
			["spellId"] = 323996,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 0.637,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636838,
			["nameplateShowPersonal"] = false,
		},
		[132169] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 3.6,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 132169,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 613535,
			["nameplateShowPersonal"] = false,
		},
		[135299] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 576309,
			["spellId"] = 135299,
		},
		[269576] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 7.654,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132177,
			["spellId"] = 269576,
		},
		[24394] = {
			["spellId"] = 24394,
			["nameplateShowAll"] = true,
			["duration"] = 5,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132111,
			["nameplateShowPersonal"] = false,
		},
		[205179] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8.064,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 205179,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132886,
			["nameplateShowPersonal"] = true,
		},
		[363473] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3741937,
			["spellId"] = 363473,
		},
		[353254] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353254,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3528274,
			["nameplateShowPersonal"] = false,
		},
		[354788] = {
			["spellId"] = 354788,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 6,
			["applications"] = 1,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4067373,
			["nameplateShowPersonal"] = false,
		},
		[8680] = {
			["spellId"] = 8680,
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 1,
			["timeMod"] = 1,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134197,
			["dispelName"] = "Poison",
		},
		[345335] = {
			["spellId"] = 345335,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3528279,
			["nameplateShowPersonal"] = true,
		},
		[207736] = {
			["spellId"] = 207736,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1020341,
			["nameplateShowPersonal"] = false,
		},
		[345209] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 6.4,
			["spellId"] = 345209,
			["sourceUnit"] = "raid12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236158,
			["dispelName"] = "Disease",
		},
		[336139] = {
			["spellId"] = 336139,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135737,
			["nameplateShowPersonal"] = false,
		},
		[179057] = {
			["spellId"] = 179057,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 4,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135795,
			["dispelName"] = "Magic",
		},
		[324260] = {
			["spellId"] = 324260,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 462651,
			["nameplateShowPersonal"] = false,
		},
		[15407] = {
			["spellId"] = 15407,
			["nameplateShowAll"] = false,
			["duration"] = 3.156,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136208,
			["nameplateShowPersonal"] = false,
		},
		[328605] = {
			["spellId"] = 328605,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 575534,
			["nameplateShowPersonal"] = false,
		},
		[317491] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 317491,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565727,
			["nameplateShowPersonal"] = false,
		},
		[164815] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 13.5,
			["spellId"] = 164815,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236216,
			["dispelName"] = "Magic",
		},
		[192090] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 192090,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 451161,
			["nameplateShowPersonal"] = true,
		},
		[146739] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 18.2,
			["spellId"] = 146739,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136118,
			["dispelName"] = "Magic",
		},
		[217832] = {
			["spellId"] = 217832,
			["nameplateShowAll"] = true,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1380368,
			["dispelName"] = "Magic",
		},
		[15487] = {
			["spellId"] = 15487,
			["nameplateShowAll"] = true,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 458230,
			["dispelName"] = "Magic",
		},
		[321712] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 6.367,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135808,
			["spellId"] = 321712,
		},
		[321329] = {
			["spellId"] = 321329,
			["nameplateShowAll"] = true,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 3.4,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135864,
			["dispelName"] = "Magic",
		},
		[200587] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 200587,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135801,
			["nameplateShowPersonal"] = false,
		},
		[345219] = {
			["spellId"] = 345219,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1035040,
			["nameplateShowPersonal"] = false,
		},
		[316220] = {
			["spellId"] = 316220,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 18,
			["applications"] = 0,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132295,
			["nameplateShowPersonal"] = true,
		},
		[331934] = {
			["spellId"] = 331934,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2015861,
			["nameplateShowPersonal"] = false,
		},
		[50259] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 2.55,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 50259,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135860,
			["nameplateShowPersonal"] = false,
		},
		[354163] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 1,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 354163,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636851,
			["nameplateShowPersonal"] = false,
		},
		[325037] = {
			["spellId"] = 325037,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578207,
			["nameplateShowPersonal"] = false,
		},
		[64695] = {
			["spellId"] = 64695,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 8,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136100,
			["dispelName"] = "Magic",
		},
		[188389] = {
			["spellId"] = 188389,
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 18,
			["applications"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135813,
			["dispelName"] = "Magic",
		},
		[124280] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 6,
			["spellId"] = 124280,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 651728,
			["dispelName"] = "Magic",
		},
		[31661] = {
			["spellId"] = 31661,
			["nameplateShowAll"] = true,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134153,
			["dispelName"] = "Magic",
		},
		[58180] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 10.2,
			["spellId"] = 58180,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236158,
			["dispelName"] = "Disease",
		},
		[207744] = {
			["spellId"] = 207744,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1344647,
			["nameplateShowPersonal"] = true,
		},
		[356723] = {
			["spellId"] = 356723,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132169,
			["dispelName"] = "Poison",
		},
		[124121] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 124121,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132484,
			["nameplateShowPersonal"] = false,
		},
		[81261] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 81261,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 252188,
			["nameplateShowPersonal"] = false,
		},
		[324149] = {
			["spellId"] = 324149,
			["nameplateShowAll"] = false,
			["duration"] = 18,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565719,
			["nameplateShowPersonal"] = false,
		},
		[34914] = {
			["spellId"] = 34914,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = true,
			["isBossAura"] = false,
			["duration"] = 21,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135978,
			["dispelName"] = "Magic",
		},
		[212792] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 5,
			["spellId"] = 212792,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135852,
			["dispelName"] = "Magic",
		},
		[183218] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 2.55,
			["spellId"] = 183218,
			["sourceUnit"] = "nameplate5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1360760,
			["dispelName"] = "Magic",
		},
		[356727] = {
			["spellId"] = 356727,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1412206,
			["dispelName"] = "Poison",
		},
		[335904] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 335904,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1035054,
			["nameplateShowPersonal"] = false,
		},
		[80240] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 12,
			["spellId"] = 80240,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 460695,
			["dispelName"] = "Curse",
		},
		[255909] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236278,
			["spellId"] = 255909,
		},
		[191587] = {
			["spellId"] = 191587,
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132099,
			["dispelName"] = "Disease",
		},
		[51490] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 51490,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237589,
			["nameplateShowPersonal"] = false,
		},
		[356730] = {
			["spellId"] = 356730,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236200,
			["dispelName"] = "Poison",
		},
		[339229] = {
			["spellId"] = 339229,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1309101,
			["nameplateShowPersonal"] = true,
		},
		[132951] = {
			["spellId"] = 132951,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135815,
			["nameplateShowPersonal"] = false,
		},
		[356604] = {
			["spellId"] = 356604,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "nameplate8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4226154,
			["nameplateShowPersonal"] = false,
		},
		[315341] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 21,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 315341,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135610,
			["nameplateShowPersonal"] = true,
		},
		[354050] = {
			["spellId"] = 354050,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 965900,
			["nameplateShowPersonal"] = false,
		},
		[269352] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 9,
			["spellId"] = 269352,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 456563,
			["dispelName"] = "Curse",
		},
		[703] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 18,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 703,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132297,
			["nameplateShowPersonal"] = true,
		},
		[354051] = {
			["spellId"] = 354051,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["sourceUnit"] = "nameplate7",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 965900,
			["dispelName"] = "Magic",
		},
		[208772] = {
			["spellId"] = 208772,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136224,
			["nameplateShowPersonal"] = false,
		},
		[157736] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 18,
			["spellId"] = 157736,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135817,
			["dispelName"] = "Magic",
		},
		[87204] = {
			["spellId"] = 87204,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135945,
			["dispelName"] = "Magic",
		},
		[342938] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 21,
			["spellId"] = 342938,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136228,
			["dispelName"] = "Magic",
		},
		[337956] = {
			["spellId"] = 337956,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1391768,
			["dispelName"] = "Magic",
		},
		[325437] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 2,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 325437,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132344,
			["nameplateShowPersonal"] = false,
		},
		[127797] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 127797,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 571588,
			["nameplateShowPersonal"] = false,
		},
		[2818] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 12,
			["spellId"] = 2818,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132290,
			["dispelName"] = "Poison",
		},
		[205708] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 8,
			["spellId"] = 205708,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236208,
			["dispelName"] = "Magic",
		},
		[31694] = {
			["spellId"] = 31694,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136122,
			["nameplateShowPersonal"] = false,
		},
		[197277] = {
			["spellId"] = 197277,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135959,
			["nameplateShowPersonal"] = true,
		},
		[317009] = {
			["spellId"] = 317009,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565717,
			["dispelName"] = "Magic",
		},
		[367726] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 367726,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1392547,
			["nameplateShowPersonal"] = false,
		},
		[353546] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353546,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 615100,
			["nameplateShowPersonal"] = false,
		},
		[328891] = {
			["spellId"] = 328891,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1769066,
			["nameplateShowPersonal"] = false,
		},
		[197214] = {
			["spellId"] = 197214,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 2,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 524794,
			["nameplateShowPersonal"] = false,
		},
		[323654] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 323654,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565724,
			["nameplateShowPersonal"] = false,
		},
		[137619] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 60,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236364,
			["spellId"] = 137619,
		},
		[353293] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 1.912,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353293,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4067372,
			["nameplateShowPersonal"] = false,
		},
		[50326] = {
			["spellId"] = 50326,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132484,
			["nameplateShowPersonal"] = false,
		},
		[3355] = {
			["dispelName"] = "Magic",
			["isBossAura"] = false,
			["nameplateShowAll"] = true,
			["duration"] = 8,
			["timeMod"] = 1,
			["nameplateShowPersonal"] = false,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135834,
			["spellId"] = 3355,
		},
		[352528] = {
			["spellId"] = 352528,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1500942,
			["nameplateShowPersonal"] = false,
		},
		[319952] = {
			["spellId"] = 319952,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 254090,
			["nameplateShowPersonal"] = false,
		},
		[204242] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10.2,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 204242,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135926,
			["nameplateShowPersonal"] = false,
		},
		[198813] = {
			["spellId"] = 198813,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1348401,
			["nameplateShowPersonal"] = false,
		},
		[353807] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353807,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1117883,
			["nameplateShowPersonal"] = false,
		},
		[271920] = {
			["spellId"] = 271920,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 464162,
			["dispelName"] = "Magic",
		},
		[8326] = {
			["spellId"] = 8326,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132331,
			["nameplateShowPersonal"] = false,
		},
		[71909] = {
			["spellId"] = 71909,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135767,
			["nameplateShowPersonal"] = false,
		},
		[589] = {
			["spellId"] = 589,
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 16,
			["applications"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136207,
			["dispelName"] = "Magic",
		},
		[1943] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 24,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 1943,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132302,
			["nameplateShowPersonal"] = true,
		},
		[364668] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 9,
			["applications"] = 0,
			["sourceUnit"] = "raid6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 364668,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 457636,
			["nameplateShowPersonal"] = false,
		},
		[356620] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4067366,
			["spellId"] = 356620,
		},
		[5211] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 3.6,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 5211,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132114,
			["nameplateShowPersonal"] = false,
		},
		[356621] = {
			["spellId"] = 356621,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4226153,
			["nameplateShowPersonal"] = false,
		},
		[64044] = {
			["spellId"] = 64044,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 1,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237568,
			["dispelName"] = "Magic",
		},
		[271924] = {
			["spellId"] = 271924,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1032474,
			["dispelName"] = "Magic",
		},
		[198688] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 2,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 198688,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 643249,
			["nameplateShowPersonal"] = false,
		},
		[12486] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 2.4,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 12486,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135857,
			["nameplateShowPersonal"] = false,
		},
		[173266] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 173266,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 133032,
			["nameplateShowPersonal"] = false,
		},
		[6770] = {
			["spellId"] = 6770,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 8,
			["applications"] = 1,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132310,
			["nameplateShowPersonal"] = false,
		},
		[118699] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 6,
			["spellId"] = 118699,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136183,
			["dispelName"] = "Magic",
		},
		[5760] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 10,
			["spellId"] = 5760,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136066,
			["dispelName"] = "Poison",
		},
		[346278] = {
			["spellId"] = 346278,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135801,
			["nameplateShowPersonal"] = true,
		},
		[356626] = {
			["spellId"] = 356626,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4226152,
			["nameplateShowPersonal"] = false,
		},
		[6788] = {
			["spellId"] = 6788,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6.304,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135871,
			["nameplateShowPersonal"] = false,
		},
		[980] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 6,
			["isBossAura"] = false,
			["duration"] = 26.4,
			["spellId"] = 980,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136139,
			["dispelName"] = "Curse",
		},
		[853] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 7.5,
			["spellId"] = 853,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135963,
			["dispelName"] = "Magic",
		},
		[3409] = {
			["spellId"] = 3409,
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6.4,
			["applications"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132274,
			["dispelName"] = "Poison",
		},
		[20549] = {
			["spellId"] = 20549,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 2,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132368,
			["nameplateShowPersonal"] = false,
		},
		[363656] = {
			["spellId"] = 363656,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 4.992,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136214,
			["nameplateShowPersonal"] = false,
		},
		[358164] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 120,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 358164,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1030904,
			["nameplateShowPersonal"] = false,
		},
		[334275] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 9.6,
			["spellId"] = 334275,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136162,
			["dispelName"] = "Curse",
		},
		[196840] = {
			["spellId"] = 196840,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 5.100000000000001,
			["sourceUnit"] = "player",
			["canApplyAura"] = true,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135849,
			["dispelName"] = "Magic",
		},
		[356632] = {
			["spellId"] = 356632,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4226150,
			["nameplateShowPersonal"] = false,
		},
		[289308] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 2.4,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 289308,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 629077,
			["nameplateShowPersonal"] = false,
		},
		[323673] = {
			["spellId"] = 323673,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = true,
			["isBossAura"] = true,
			["duration"] = 7,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565723,
			["dispelName"] = "Magic",
		},
		[59638] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 4,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135846,
			["spellId"] = 59638,
		},
		[25771] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 25771,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135952,
			["nameplateShowPersonal"] = false,
		},
		[1715] = {
			["spellId"] = 1715,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 10.2,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132316,
			["nameplateShowPersonal"] = false,
		},
		[12654] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 9.495000000000001,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135818,
			["spellId"] = 12654,
		},
		[323802] = {
			["spellId"] = 323802,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636838,
			["nameplateShowPersonal"] = true,
		},
		[321759] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 17,
			["spellId"] = 321759,
			["sourceUnit"] = "nameplate9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132586,
			["dispelName"] = "Magic",
		},
		[108194] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 108194,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 538558,
			["nameplateShowPersonal"] = false,
		},
		[605] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 4,
			["spellId"] = 605,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136206,
			["dispelName"] = "Magic",
		},
		[328275] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 18,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 328275,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636840,
			["nameplateShowPersonal"] = false,
		},
		[320227] = {
			["spellId"] = 320227,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 600,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 656440,
			["nameplateShowPersonal"] = false,
		},
		[200200] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 200200,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135904,
			["nameplateShowPersonal"] = false,
		},
		[258883] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 258883,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1305149,
			["nameplateShowPersonal"] = true,
		},
		[199845] = {
			["spellId"] = 199845,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136208,
			["nameplateShowPersonal"] = false,
		},
		[211881] = {
			["spellId"] = 211881,
			["nameplateShowAll"] = true,
			["duration"] = 4,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1118739,
			["nameplateShowPersonal"] = false,
		},
		[326868] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 326868,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 615100,
			["nameplateShowPersonal"] = false,
		},
		[323710] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 323710,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578196,
			["nameplateShowPersonal"] = false,
		},
		[121164] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = true,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 121164,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 1119885,
			["nameplateShowPersonal"] = false,
		},
		[198121] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 4,
			["spellId"] = 198121,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135842,
			["dispelName"] = "Magic",
		},
		[121177] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = true,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 121177,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 1119887,
			["nameplateShowPersonal"] = false,
		},
		[201787] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 2,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 201787,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 627606,
			["nameplateShowPersonal"] = false,
		},
		[122470] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 122470,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 651728,
			["nameplateShowPersonal"] = false,
		},
		[208086] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 208086,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 464973,
			["nameplateShowPersonal"] = true,
		},
		[113746] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 113746,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 615342,
			["nameplateShowPersonal"] = false,
		},
		[353248] = {
			["spellId"] = 353248,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3089628,
			["nameplateShowPersonal"] = false,
		},
		[364311] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 600,
			["applications"] = 0,
			["sourceUnit"] = "raid6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 364311,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4038102,
			["nameplateShowPersonal"] = false,
		},
		[324073] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 324073,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578230,
			["nameplateShowPersonal"] = true,
		},
		[50327] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 50327,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132484,
			["nameplateShowPersonal"] = false,
		},
		[42792] = {
			["spellId"] = 42792,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132484,
			["nameplateShowPersonal"] = false,
		},
		[263165] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["sourceUnit"] = "raid15",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 263165,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1386551,
			["nameplateShowPersonal"] = false,
		},
		[113942] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 90,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 113942,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 607512,
			["nameplateShowPersonal"] = false,
		},
		[33786] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 2.4,
			["applications"] = 0,
			["sourceUnit"] = "raid12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 33786,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136022,
			["nameplateShowPersonal"] = false,
		},
		[353472] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353472,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3752258,
			["nameplateShowPersonal"] = false,
		},
		[99] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["sourceUnit"] = "raid12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 99,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132121,
			["nameplateShowPersonal"] = false,
		},
		[212431] = {
			["spellId"] = 212431,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = true,
			["isBossAura"] = false,
			["duration"] = 3,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236178,
			["dispelName"] = "Magic",
		},
		[57723] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 600,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 57723,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136090,
			["nameplateShowPersonal"] = false,
		},
		[22703] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 2.5,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 22703,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135860,
			["nameplateShowPersonal"] = true,
		},
		[61391] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 6,
			["spellId"] = 61391,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135860,
			["dispelName"] = "Magic",
		},
		[116947] = {
			["spellId"] = 116947,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136102,
			["dispelName"] = "Magic",
		},
		[352939] = {
			["spellId"] = 352939,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1041266,
			["nameplateShowPersonal"] = false,
		},
		[209749] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 5,
			["spellId"] = 209749,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 538516,
			["dispelName"] = "Magic",
		},
		[367931] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 367931,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 530250,
			["nameplateShowPersonal"] = false,
		},
		[326939] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 5.771,
			["spellId"] = 326939,
			["sourceUnit"] = "raid12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134427,
			["dispelName"] = "Disease",
		},
		[12323] = {
			["spellId"] = 12323,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136147,
			["nameplateShowPersonal"] = false,
		},
		[356356] = {
			["spellId"] = 356356,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 2,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236319,
			["nameplateShowPersonal"] = true,
		},
		[319504] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 9,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 319504,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135428,
			["nameplateShowPersonal"] = true,
		},
		[70013] = {
			["spellId"] = 70013,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136130,
			["nameplateShowPersonal"] = false,
		},
		[356608] = {
			["spellId"] = 356608,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1305160,
			["nameplateShowPersonal"] = false,
		},
		[225080] = {
			["spellId"] = 225080,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 451167,
			["nameplateShowPersonal"] = false,
		},
		[30283] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 2.4,
			["spellId"] = 30283,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 607865,
			["dispelName"] = "Magic",
		},
		[118] = {
			["spellId"] = 118,
			["nameplateShowAll"] = true,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = true,
			["duration"] = 4,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136071,
			["dispelName"] = "Magic",
		},
		[77505] = {
			["spellId"] = 77505,
			["nameplateShowAll"] = true,
			["duration"] = 1.5,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 451165,
			["nameplateShowPersonal"] = false,
		},
		[163505] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 163505,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132122,
			["nameplateShowPersonal"] = false,
		},
		[122] = {
			["spellId"] = 122,
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135848,
			["dispelName"] = "Magic",
		},
		[214222] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 214222,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135959,
			["nameplateShowPersonal"] = true,
		},
		[1490] = {
			["spellId"] = 1490,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1450142,
			["nameplateShowPersonal"] = false,
		},
		[200196] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 3.2,
			["applications"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 200196,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135886,
			["nameplateShowPersonal"] = false,
		},
		[287280] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 30,
			["spellId"] = 287280,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1360764,
			["dispelName"] = "Magic",
		},
		[325733] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 3,
			["isBossAura"] = false,
			["duration"] = 12,
			["spellId"] = 325733,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578197,
			["dispelName"] = "Magic",
		},
		[333526] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 5.100000000000001,
			["spellId"] = 333526,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 960682,
			["dispelName"] = "Magic",
		},
		[164812] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 16,
			["spellId"] = 164812,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136096,
			["dispelName"] = "Magic",
		},
		[87023] = {
			["spellId"] = 87023,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 252268,
			["nameplateShowPersonal"] = false,
		},
		[339654] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132149,
			["spellId"] = 339654,
		},
		[305485] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 305485,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1385911,
			["nameplateShowPersonal"] = false,
		},
		[352562] = {
			["spellId"] = 352562,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5.499000000000001,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3601538,
			["nameplateShowPersonal"] = true,
		},
		[321519] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 30,
			["spellId"] = 321519,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136242,
			["dispelName"] = "Poison",
		},
		[14914] = {
			["spellId"] = 14914,
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 7,
			["applications"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135972,
			["dispelName"] = "Magic",
		},
		[359078] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 5,
			["spellId"] = 359078,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136066,
			["dispelName"] = "Poison",
		},
		[213691] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3.6,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 213691,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132153,
			["nameplateShowPersonal"] = false,
		},
		[408] = {
			["spellId"] = 408,
			["nameplateShowAll"] = true,
			["duration"] = 6,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132298,
			["nameplateShowPersonal"] = false,
		},
		[234153] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3.905,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 234153,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136169,
			["nameplateShowPersonal"] = false,
		},
		[356616] = {
			["spellId"] = 356616,
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4226151,
			["nameplateShowPersonal"] = false,
		},
		[105771] = {
			["spellId"] = 105771,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 1.7,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132337,
			["nameplateShowPersonal"] = false,
		},
		[360775] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 360775,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 1035041,
			["nameplateShowPersonal"] = false,
		},
		[340045] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 340045,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1380867,
			["nameplateShowPersonal"] = false,
		},
		[185763] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4.8,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 185763,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1373908,
			["nameplateShowPersonal"] = false,
		},
		[199450] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 199450,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135966,
			["nameplateShowPersonal"] = false,
		},
		[87024] = {
			["spellId"] = 87024,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 300,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 252268,
			["nameplateShowPersonal"] = false,
		},
		[214968] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1390942,
			["spellId"] = 214968,
		},
		[1714] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 20,
			["spellId"] = 1714,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136140,
			["dispelName"] = "Curse",
		},
		[262115] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 262115,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132090,
			["nameplateShowPersonal"] = true,
		},
		[247121] = {
			["spellId"] = 247121,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135860,
			["nameplateShowPersonal"] = false,
		},
		[51514] = {
			["spellId"] = 51514,
			["nameplateShowAll"] = true,
			["isBossAura"] = false,
			["duration"] = 4,
			["timeMod"] = 1,
			["canApplyAura"] = true,
			["applications"] = 0,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237579,
			["dispelName"] = "Curse",
		},
		[307871] = {
			["spellId"] = 307871,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565453,
			["nameplateShowPersonal"] = false,
		},
		[232559] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 232559,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136104,
			["nameplateShowPersonal"] = false,
		},
		[340431] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 9,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 340431,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 538040,
			["nameplateShowPersonal"] = false,
		},
		[217200] = {
			["spellId"] = 217200,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2058007,
			["nameplateShowPersonal"] = false,
		},
		[356528] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 4,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 356528,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 366936,
			["nameplateShowPersonal"] = false,
		},
		[340556] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 67,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236169,
			["spellId"] = 340556,
		},
		[323416] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 323416,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3459801,
			["nameplateShowPersonal"] = false,
		},
		[280773] = {
			["spellId"] = 280773,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 294382,
			["nameplateShowPersonal"] = true,
		},
		[255937] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4.25,
			["applications"] = 0,
			["sourceUnit"] = "nameplate5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 255937,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1112939,
			["nameplateShowPersonal"] = false,
		},
		[325203] = {
			["spellId"] = 325203,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = true,
			["isBossAura"] = false,
			["duration"] = 15,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3528292,
			["dispelName"] = "Disease",
		},
		[105421] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 6,
			["spellId"] = 105421,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 571553,
			["dispelName"] = "Magic",
		},
		[212183] = {
			["spellId"] = 212183,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 458733,
			["nameplateShowPersonal"] = false,
		},
		[6789] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = true,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 3,
			["spellId"] = 6789,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 607853,
			["dispelName"] = "Magic",
		},
		[339284] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 339284,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1778229,
			["nameplateShowPersonal"] = false,
		},
		[196364] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 4,
			["spellId"] = 196364,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135975,
			["dispelName"] = "Magic",
		},
		[204263] = {
			["spellId"] = 204263,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 571554,
			["dispelName"] = "Magic",
		},
		[187464] = {
			["spellId"] = 187464,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136202,
			["nameplateShowPersonal"] = false,
		},
		[265931] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 8,
			["spellId"] = 265931,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135807,
			["dispelName"] = "Magic",
		},
		[155722] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 155722,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132122,
			["nameplateShowPersonal"] = true,
		},
		[370037] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4263293,
			["spellId"] = 370037,
		},
		[308498] = {
			["spellId"] = 308498,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565445,
			["nameplateShowPersonal"] = true,
		},
		[341291] = {
			["spellId"] = 341291,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1386547,
			["nameplateShowPersonal"] = false,
		},
	},
	["HELPFUL"] = {
		[199483] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 60,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 461113,
			["spellId"] = 199483,
		},
		[339632] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 339632,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3731082,
			["nameplateShowPersonal"] = false,
		},
		[262652] = {
			["spellId"] = 262652,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 462329,
			["nameplateShowPersonal"] = false,
		},
		[314058] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 314058,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136012,
			["nameplateShowPersonal"] = false,
		},
		[1022] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 10,
			["spellId"] = 1022,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135964,
			["dispelName"] = "Magic",
		},
		[236440] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 236440,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1408835,
			["nameplateShowPersonal"] = false,
		},
		[323524] = {
			["spellId"] = 323524,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3.001,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 840942,
			["nameplateShowPersonal"] = false,
		},
		[20572] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135726,
			["spellId"] = 20572,
		},
		[357028] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 4,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565723,
			["spellId"] = 357028,
		},
		[130] = {
			["spellId"] = 130,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135992,
			["nameplateShowPersonal"] = false,
		},
		[223143] = {
			["spellId"] = 223143,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132872,
			["nameplateShowPersonal"] = false,
		},
		[194627] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 194627,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132206,
			["nameplateShowPersonal"] = false,
		},
		[330943] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 6,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 330943,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135946,
			["nameplateShowPersonal"] = false,
		},
		[352938] = {
			["spellId"] = 352938,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1041266,
			["nameplateShowPersonal"] = false,
		},
		[356263] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565445,
			["spellId"] = 356263,
		},
		[22842] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 22842,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132091,
			["nameplateShowPersonal"] = true,
		},
		[341175] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 900,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 341175,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237274,
			["nameplateShowPersonal"] = false,
		},
		[5215] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 1,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 5215,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 514640,
			["nameplateShowPersonal"] = false,
		},
		[290793] = {
			["spellId"] = 290793,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 537078,
			["nameplateShowPersonal"] = false,
		},
		[328900] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 3,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132206,
			["spellId"] = 328900,
		},
		[185422] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 9,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236279,
			["spellId"] = 185422,
		},
		[48018] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 900,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 48018,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237559,
			["nameplateShowPersonal"] = false,
		},
		[207289] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 207289,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136224,
			["nameplateShowPersonal"] = false,
		},
		[365474] = {
			["spellId"] = 365474,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "mouseover",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4224735,
			["nameplateShowPersonal"] = false,
		},
		[142073] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 142073,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 853211,
			["nameplateShowPersonal"] = false,
		},
		[102558] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "raid12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 102558,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 571586,
			["nameplateShowPersonal"] = false,
		},
		[319952] = {
			["spellId"] = 319952,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 254090,
			["nameplateShowPersonal"] = false,
		},
		[351921] = {
			["spellId"] = 351921,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3622131,
			["nameplateShowPersonal"] = false,
		},
		[127302] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 127302,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 656346,
			["nameplateShowPersonal"] = false,
		},
		[336065] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 336065,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237582,
			["nameplateShowPersonal"] = false,
		},
		[269571] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["timeMod"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["applications"] = 3,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135961,
			["spellId"] = 269571,
		},
		[363943] = {
			["spellId"] = 363943,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["applications"] = 1,
			["sourceUnit"] = "nameplate8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4226157,
			["nameplateShowPersonal"] = false,
		},
		[327371] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 617812,
			["spellId"] = 327371,
		},
		[267270] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 267270,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1624590,
			["nameplateShowPersonal"] = false,
		},
		[23338] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 23338,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132225,
			["nameplateShowPersonal"] = false,
		},
		[353203] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353203,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3528276,
			["nameplateShowPersonal"] = false,
		},
		[58984] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132089,
			["spellId"] = 58984,
		},
		[46740] = {
			["spellId"] = 46740,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134519,
			["nameplateShowPersonal"] = false,
		},
		[332744] = {
			["spellId"] = 332744,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 463858,
			["nameplateShowPersonal"] = false,
		},
		[351414] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 1,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 351414,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3492310,
			["nameplateShowPersonal"] = false,
		},
		[36702] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 36702,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132238,
			["nameplateShowPersonal"] = false,
		},
		[156910] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 156910,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1030095,
			["nameplateShowPersonal"] = false,
		},
		[546] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 600,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 546,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135863,
			["nameplateShowPersonal"] = false,
		},
		[129861] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 129861,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 625906,
			["nameplateShowPersonal"] = false,
		},
		[294384] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 120,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237276,
			["spellId"] = 294384,
		},
		[102560] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 40,
			["applications"] = 0,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 102560,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 571586,
			["nameplateShowPersonal"] = false,
		},
		[53390] = {
			["spellId"] = 53390,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 2,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237590,
			["dispelName"] = "Magic",
		},
		[288503] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 288503,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2475037,
			["nameplateShowPersonal"] = false,
		},
		[261769] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 3,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 261769,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1381295,
			["nameplateShowPersonal"] = false,
		},
		[193356] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1393014,
			["spellId"] = 193356,
		},
		[59752] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0.1,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 59752,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136129,
			["nameplateShowPersonal"] = false,
		},
		[147833] = {
			["spellId"] = 147833,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132365,
			["nameplateShowPersonal"] = false,
		},
		[139] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135953,
			["spellId"] = 139,
		},
		[339400] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236154,
			["spellId"] = 339400,
		},
		[54861] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 54861,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135788,
			["nameplateShowPersonal"] = false,
		},
		[294133] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 294133,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1360759,
			["nameplateShowPersonal"] = false,
		},
		[212283] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 252272,
			["spellId"] = 212283,
		},
		[5487] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 5487,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132276,
			["nameplateShowPersonal"] = false,
		},
		[196555] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["timeMod"] = 1,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 463284,
			["spellId"] = 196555,
		},
		[8936] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 12,
			["spellId"] = 8936,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136085,
			["dispelName"] = "Magic",
		},
		[317920] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 317920,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135933,
			["nameplateShowPersonal"] = false,
		},
		[81782] = {
			["spellId"] = 81782,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 253400,
			["nameplateShowPersonal"] = false,
		},
		[248473] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 120,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136116,
			["spellId"] = 248473,
		},
		[193359] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1269442,
			["spellId"] = 193359,
		},
		[363701] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 363701,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3778581,
			["nameplateShowPersonal"] = false,
		},
		[320224] = {
			["spellId"] = 320224,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 656440,
			["nameplateShowPersonal"] = false,
		},
		[171621] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1045155,
			["spellId"] = 171621,
		},
		[23243] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 23243,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132253,
			["nameplateShowPersonal"] = false,
		},
		[345545] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 345545,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132108,
			["nameplateShowPersonal"] = false,
		},
		[145152] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 2,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 145152,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1033474,
			["nameplateShowPersonal"] = true,
		},
		[194384] = {
			["spellId"] = 194384,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 458720,
			["nameplateShowPersonal"] = false,
		},
		[339664] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 5,
			["spellId"] = 339664,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135969,
			["dispelName"] = "Magic",
		},
		[24378] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 24378,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 136012,
			["nameplateShowPersonal"] = false,
		},
		[307185] = {
			["spellId"] = 307185,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 7200.003,
			["applications"] = 0,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3566840,
			["nameplateShowPersonal"] = false,
		},
		[98727] = {
			["spellId"] = 98727,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 526356,
			["nameplateShowPersonal"] = false,
		},
		[23451] = {
			["spellId"] = 23451,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 132870,
			["nameplateShowPersonal"] = false,
		},
		[332505] = {
			["spellId"] = 332505,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 133863,
			["nameplateShowPersonal"] = false,
		},
		[208963] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 208963,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135829,
			["nameplateShowPersonal"] = false,
		},
		[353477] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1394953,
			["spellId"] = 353477,
		},
		[332506] = {
			["spellId"] = 332506,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 133863,
			["nameplateShowPersonal"] = false,
		},
		[307187] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 7200,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3566841,
			["spellId"] = 307187,
		},
		[212800] = {
			["spellId"] = 212800,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1305150,
			["nameplateShowPersonal"] = true,
		},
		[321510] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 321510,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 960671,
			["nameplateShowPersonal"] = false,
		},
		[306421] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3044237,
			["spellId"] = 306421,
		},
		[346574] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 9,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 346574,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1035055,
			["nameplateShowPersonal"] = false,
		},
		[366779] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 4,
			["sourceUnit"] = "nameplate11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 366779,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237574,
			["nameplateShowPersonal"] = false,
		},
		[264735] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 264735,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136094,
			["nameplateShowPersonal"] = false,
		},
		[323558] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 45,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 2,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565450,
			["spellId"] = 323558,
		},
		[586] = {
			["spellId"] = 586,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135994,
			["nameplateShowPersonal"] = false,
		},
		[323559] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 45,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 3,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565450,
			["spellId"] = 323559,
		},
		[327140] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = true,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 327140,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1030904,
			["nameplateShowPersonal"] = false,
		},
		[323560] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 45,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 4,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565450,
			["spellId"] = 323560,
		},
		[341207] = {
			["spellId"] = 341207,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 462324,
			["nameplateShowPersonal"] = false,
		},
		[118038] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 118038,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132336,
			["nameplateShowPersonal"] = false,
		},
		[339929] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2065620,
			["spellId"] = 339929,
		},
		[292360] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 292360,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2446016,
			["nameplateShowPersonal"] = false,
		},
		[167152] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134062,
			["spellId"] = 167152,
		},
		[332514] = {
			["spellId"] = 332514,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3257748,
			["nameplateShowPersonal"] = false,
		},
		[23036] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 23036,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132484,
			["nameplateShowPersonal"] = false,
		},
		[271646] = {
			["spellId"] = 271646,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1992951,
			["nameplateShowPersonal"] = false,
		},
		[352719] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135127,
			["spellId"] = 352719,
		},
		[185311] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 4,
			["timeMod"] = 1,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1373904,
			["spellId"] = 185311,
		},
		[352720] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 352720,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135127,
			["nameplateShowPersonal"] = false,
		},
		[287504] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 287504,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 606551,
			["nameplateShowPersonal"] = false,
		},
		[292364] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 292364,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2446019,
			["nameplateShowPersonal"] = false,
		},
		[48438] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 7,
			["spellId"] = 48438,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236153,
			["dispelName"] = "Magic",
		},
		[26297] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 26297,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135727,
			["nameplateShowPersonal"] = false,
		},
		[143625] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 143625,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 305703,
			["nameplateShowPersonal"] = false,
		},
		[56814] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 56814,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132319,
			["nameplateShowPersonal"] = false,
		},
		[257946] = {
			["spellId"] = 257946,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 1,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132216,
			["nameplateShowPersonal"] = true,
		},
		[24379] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10.002,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 24379,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 134062,
			["nameplateShowPersonal"] = false,
		},
		[288275] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 900318,
			["spellId"] = 288275,
		},
		[327405] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3546384,
			["spellId"] = 327405,
		},
		[265258] = {
			["spellId"] = 265258,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237566,
			["nameplateShowPersonal"] = false,
		},
		[221886] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1360759,
			["spellId"] = 221886,
		},
		[235313] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 60,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132221,
			["spellId"] = 235313,
		},
		[352981] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 352981,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 515951,
			["nameplateShowPersonal"] = false,
		},
		[121557] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 5,
			["spellId"] = 121557,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 642580,
			["dispelName"] = "Magic",
		},
		[327407] = {
			["spellId"] = 327407,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3536969,
			["nameplateShowPersonal"] = false,
		},
		[122708] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 122708,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 616692,
			["nameplateShowPersonal"] = false,
		},
		[221887] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5.850000000000001,
			["applications"] = 0,
			["sourceUnit"] = "nameplate11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 221887,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1360759,
			["nameplateShowPersonal"] = false,
		},
		[45242] = {
			["spellId"] = 45242,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 2,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 458227,
			["nameplateShowPersonal"] = false,
		},
		[321527] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 120,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 321527,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132142,
			["nameplateShowPersonal"] = false,
		},
		[335082] = {
			["spellId"] = 335082,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 1,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132284,
			["nameplateShowPersonal"] = false,
		},
		[266030] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 8,
			["spellId"] = 266030,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 840406,
			["dispelName"] = "Magic",
		},
		[155777] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 17.001,
			["spellId"] = 155777,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1033478,
			["dispelName"] = "Magic",
		},
		[16595] = {
			["dispelName"] = "Magic",
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 600,
			["nameplateShowPersonal"] = false,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136162,
			["spellId"] = 16595,
		},
		[345569] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 15,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 345569,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565724,
			["nameplateShowPersonal"] = false,
		},
		[339943] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 8,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 339943,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132113,
			["nameplateShowPersonal"] = false,
		},
		[281888] = {
			["spellId"] = 281888,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 901174,
			["nameplateShowPersonal"] = false,
		},
		[320763] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 320763,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135861,
			["nameplateShowPersonal"] = false,
		},
		[16739] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 300,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 16739,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136121,
			["nameplateShowPersonal"] = false,
		},
		[327158] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 327158,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 467095,
			["nameplateShowPersonal"] = false,
		},
		[278310] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 278310,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135844,
			["nameplateShowPersonal"] = false,
		},
		[339946] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 8,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 339946,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132113,
			["nameplateShowPersonal"] = false,
		},
		[327159] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 327159,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 467095,
			["nameplateShowPersonal"] = false,
		},
		[219589] = {
			["spellId"] = 219589,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132221,
			["nameplateShowPersonal"] = false,
		},
		[77762] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 77762,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 451169,
			["nameplateShowPersonal"] = false,
		},
		[280615] = {
			["spellId"] = 280615,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["sourceUnit"] = "nameplate12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135861,
			["nameplateShowPersonal"] = false,
		},
		[334067] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 334067,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236685,
			["nameplateShowPersonal"] = false,
		},
		[206803] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 1,
			["timeMod"] = 1,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1380371,
			["spellId"] = 206803,
		},
		[24252] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132242,
			["spellId"] = 24252,
		},
		[187878] = {
			["spellId"] = 187878,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1370984,
			["nameplateShowPersonal"] = false,
		},
		[327164] = {
			["spellId"] = 327164,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578231,
			["nameplateShowPersonal"] = false,
		},
		[206804] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1380371,
			["spellId"] = 206804,
		},
		[72968] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134411,
			["spellId"] = 72968,
		},
		[32612] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 32612,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132220,
			["nameplateShowPersonal"] = false,
		},
		[209746] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 18,
			["spellId"] = 209746,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236156,
			["dispelName"] = "Magic",
		},
		[336885] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 11.949,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 644385,
			["spellId"] = 336885,
		},
		[48025] = {
			["spellId"] = 48025,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132251,
			["nameplateShowPersonal"] = false,
		},
		[642] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 8,
			["spellId"] = 642,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 524354,
			["dispelName"] = "Magic",
		},
		[77764] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10.52,
			["applications"] = 0,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 77764,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 464343,
			["nameplateShowPersonal"] = false,
		},
		[235450] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 60,
			["spellId"] = 235450,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135991,
			["dispelName"] = "Magic",
		},
		[339956] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 339956,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636569,
			["nameplateShowPersonal"] = false,
		},
		[324867] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 120,
			["timeMod"] = 1,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3586267,
			["spellId"] = 324867,
		},
		[11327] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132331,
			["spellId"] = 11327,
		},
		[52437] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 52437,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132346,
			["nameplateShowPersonal"] = true,
		},
		[2584] = {
			["spellId"] = 2584,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135955,
			["nameplateShowPersonal"] = false,
		},
		[320009] = {
			["spellId"] = 320009,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132123,
			["nameplateShowPersonal"] = false,
		},
		[330752] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 330752,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236869,
			["nameplateShowPersonal"] = false,
		},
		[96499] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132225,
			["spellId"] = 96499,
		},
		[342774] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 306845,
			["spellId"] = 342774,
		},
		[336892] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 1,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 336892,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2065591,
			["nameplateShowPersonal"] = false,
		},
		[93622] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["sourceUnit"] = "raid12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 93622,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132135,
			["nameplateShowPersonal"] = false,
		},
		[327942] = {
			["spellId"] = 327942,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136114,
			["nameplateShowPersonal"] = false,
		},
		[344566] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 344566,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237557,
			["nameplateShowPersonal"] = false,
		},
		[337661] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 337661,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136047,
			["nameplateShowPersonal"] = false,
		},
		[308250] = {
			["spellId"] = 308250,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3061883,
			["nameplateShowPersonal"] = false,
		},
		[116768] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 116768,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 574575,
			["dispelName"] = "Magic",
		},
		[223306] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 223306,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236249,
			["nameplateShowPersonal"] = false,
		},
		[53365] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135882,
			["spellId"] = 53365,
		},
		[287280] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 30,
			["spellId"] = 287280,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1360764,
			["dispelName"] = "Magic",
		},
		[210391] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 5,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 210391,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 461859,
			["nameplateShowPersonal"] = false,
		},
		[279608] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 279608,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2143069,
			["nameplateShowPersonal"] = false,
		},
		[23214] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 23214,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132226,
			["nameplateShowPersonal"] = false,
		},
		[359403] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 359403,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3684810,
			["nameplateShowPersonal"] = false,
		},
		[352498] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 60,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236681,
			["spellId"] = 352498,
		},
		[336642] = {
			["spellId"] = 336642,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132356,
			["nameplateShowPersonal"] = false,
		},
		[192106] = {
			["dispelName"] = "Magic",
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 1800,
			["timeMod"] = 1,
			["nameplateShowPersonal"] = false,
			["canApplyAura"] = true,
			["applications"] = 0,
			["isStealable"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136051,
			["spellId"] = 192106,
		},
		[245686] = {
			["spellId"] = 245686,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 413585,
			["nameplateShowPersonal"] = false,
		},
		[352500] = {
			["spellId"] = 352500,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 40,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3459153,
			["nameplateShowPersonal"] = false,
		},
		[328204] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 2,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 328204,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578228,
			["nameplateShowPersonal"] = true,
		},
		[294701] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 294701,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136106,
			["dispelName"] = "Magic",
		},
		[32645] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 32645,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132287,
			["nameplateShowPersonal"] = true,
		},
		[266058] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 266058,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1908013,
			["nameplateShowPersonal"] = false,
		},
		[327694] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 327694,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636847,
			["nameplateShowPersonal"] = true,
		},
		[319255] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 319255,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237557,
			["nameplateShowPersonal"] = false,
		},
		[364523] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 364523,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1020305,
			["nameplateShowPersonal"] = false,
		},
		[190446] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 190446,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236206,
			["nameplateShowPersonal"] = false,
		},
		[18499] = {
			["spellId"] = 18499,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136009,
			["dispelName"] = "",
		},
		[354550] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 300,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 354550,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 413578,
			["nameplateShowPersonal"] = false,
		},
		[13877] = {
			["spellId"] = 13877,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132350,
			["nameplateShowPersonal"] = true,
		},
		[347901] = {
			["spellId"] = 347901,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134078,
			["nameplateShowPersonal"] = false,
		},
		[190319] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 14,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135824,
			["spellId"] = 190319,
		},
		[330511] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 3,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132171,
			["spellId"] = 330511,
		},
		[253106] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 253106,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1570764,
			["nameplateShowPersonal"] = false,
		},
		[339463] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 339463,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 531009,
			["nameplateShowPersonal"] = false,
		},
		[92155] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 92155,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 458250,
			["nameplateShowPersonal"] = false,
		},
		[358134] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 358134,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1408832,
			["nameplateShowPersonal"] = false,
		},
		[324631] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 3.001,
			["timeMod"] = 1,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3586267,
			["spellId"] = 324631,
		},
		[108843] = {
			["spellId"] = 108843,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135788,
			["nameplateShowPersonal"] = false,
		},
		[90621] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 90621,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 464140,
			["nameplateShowPersonal"] = false,
		},
		[348162] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 348162,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3939983,
			["nameplateShowPersonal"] = false,
		},
		[61295] = {
			["spellId"] = 61295,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 23.4,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 252995,
			["dispelName"] = "Magic",
		},
		[341770] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135945,
			["spellId"] = 341770,
		},
		[362486] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6.68,
			["applications"] = 0,
			["sourceUnit"] = "nameplate17",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 362486,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4067365,
			["nameplateShowPersonal"] = true,
		},
		[97463] = {
			["spellId"] = 97463,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132351,
			["nameplateShowPersonal"] = false,
		},
		[247483] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 247483,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 613399,
			["nameplateShowPersonal"] = false,
		},
		[227151] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 252272,
			["spellId"] = 227151,
		},
		[64844] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 2,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 64844,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237540,
			["dispelName"] = "Magic",
		},
		[23247] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 23247,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132243,
			["nameplateShowPersonal"] = false,
		},
		[353793] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353793,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3528274,
			["nameplateShowPersonal"] = false,
		},
		[204262] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 538568,
			["spellId"] = 204262,
		},
		[337170] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 337170,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236291,
			["nameplateShowPersonal"] = false,
		},
		[365559] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 365559,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3593822,
			["nameplateShowPersonal"] = false,
		},
		[41635] = {
			["spellId"] = 41635,
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 3,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135944,
			["dispelName"] = "Magic",
		},
		[327710] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 327710,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636847,
			["nameplateShowPersonal"] = true,
		},
		[242882] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 242882,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1503309,
			["nameplateShowPersonal"] = false,
		},
		[60305] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 60305,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136146,
			["nameplateShowPersonal"] = false,
		},
		[353286] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353286,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3528274,
			["nameplateShowPersonal"] = false,
		},
		[343312] = {
			["spellId"] = 343312,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 463286,
			["nameplateShowPersonal"] = false,
		},
		[354054] = {
			["spellId"] = 354054,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "arena2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1373912,
			["nameplateShowPersonal"] = false,
		},
		[197871] = {
			["spellId"] = 197871,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1445237,
			["nameplateShowPersonal"] = false,
		},
		[260402] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 537468,
			["spellId"] = 260402,
		},
		[285514] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 285514,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136088,
			["nameplateShowPersonal"] = false,
		},
		[326946] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 326946,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237519,
			["nameplateShowPersonal"] = false,
		},
		[243651] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1581747,
			["spellId"] = 243651,
		},
		[108271] = {
			["spellId"] = 108271,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 538565,
			["nameplateShowPersonal"] = false,
		},
		[190456] = {
			["spellId"] = 190456,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1377132,
			["nameplateShowPersonal"] = true,
		},
		[353036] = {
			["spellId"] = 353036,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3593819,
			["nameplateShowPersonal"] = false,
		},
		[182912] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1114455,
			["spellId"] = 182912,
		},
		[213858] = {
			["spellId"] = 213858,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 18,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 458971,
			["nameplateShowPersonal"] = false,
		},
		[337948] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135920,
			["spellId"] = 337948,
		},
		[197874] = {
			["spellId"] = 197874,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1445237,
			["nameplateShowPersonal"] = false,
		},
		[16870] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 16870,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136170,
			["dispelName"] = "Magic",
		},
		[31224] = {
			["spellId"] = 31224,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136177,
			["nameplateShowPersonal"] = true,
		},
		[61425] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 61425,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236240,
			["nameplateShowPersonal"] = false,
		},
		[341530] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 4,
			["timeMod"] = 1,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1097742,
			["spellId"] = 341530,
		},
		[246851] = {
			["spellId"] = 246851,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2058007,
			["nameplateShowPersonal"] = false,
		},
		[196980] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["timeMod"] = 1,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136129,
			["spellId"] = 196980,
		},
		[268899] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 1,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 268899,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135872,
			["nameplateShowPersonal"] = false,
		},
		[346136] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 346136,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3883761,
			["nameplateShowPersonal"] = false,
		},
		[341533] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 341533,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 965900,
			["nameplateShowPersonal"] = false,
		},
		[1459] = {
			["spellId"] = 1459,
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135932,
			["dispelName"] = "Magic",
		},
		[272481] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "nameplate5",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2027020,
			["spellId"] = 272481,
		},
		[272737] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 272737,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 643246,
			["nameplateShowPersonal"] = false,
		},
		[324143] = {
			["spellId"] = 324143,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578234,
			["nameplateShowPersonal"] = false,
		},
		[97340] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 514262,
			["spellId"] = 97340,
		},
		[116841] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 116841,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 651727,
			["nameplateShowPersonal"] = false,
		},
		[94463] = {
			["spellId"] = 94463,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 255153,
			["nameplateShowPersonal"] = false,
		},
		[36554] = {
			["spellId"] = 36554,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 2,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132303,
			["nameplateShowPersonal"] = false,
		},
		[356369] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 356369,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 828455,
			["nameplateShowPersonal"] = false,
		},
		[356881] = {
			["spellId"] = 356881,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636841,
			["nameplateShowPersonal"] = false,
		},
		[193530] = {
			["spellId"] = 193530,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136074,
			["nameplateShowPersonal"] = false,
		},
		[352790] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 352790,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1769069,
			["nameplateShowPersonal"] = false,
		},
		[332842] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 5,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 332842,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236310,
			["nameplateShowPersonal"] = false,
		},
		[341282] = {
			["spellId"] = 341282,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1386547,
			["nameplateShowPersonal"] = false,
		},
		[268905] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 1,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 268905,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1029583,
			["nameplateShowPersonal"] = false,
		},
		[97341] = {
			["spellId"] = 97341,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 514261,
			["nameplateShowPersonal"] = false,
		},
		[364556] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 364556,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 249176,
			["nameplateShowPersonal"] = true,
		},
		[352536] = {
			["spellId"] = 352536,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1500942,
			["nameplateShowPersonal"] = false,
		},
		[354838] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 45,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 5,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565450,
			["spellId"] = 354838,
		},
		[290134] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2459601,
			["spellId"] = 290134,
		},
		[248646] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 120,
			["applications"] = 1,
			["sourceUnit"] = "nameplate9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 248646,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 623775,
			["nameplateShowPersonal"] = false,
		},
		[116267] = {
			["spellId"] = 116267,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 2,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236219,
			["nameplateShowPersonal"] = false,
		},
		[93825] = {
			["spellId"] = 93825,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 255152,
			["nameplateShowPersonal"] = false,
		},
		[108211] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 108211,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 538440,
			["nameplateShowPersonal"] = false,
		},
		[261434] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1827501,
			["spellId"] = 261434,
		},
		[73685] = {
			["spellId"] = 73685,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["sourceUnit"] = "nameplate12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 462328,
			["nameplateShowPersonal"] = false,
		},
		[325174] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 325174,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237586,
			["nameplateShowPersonal"] = false,
		},
		[335149] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 630787,
			["spellId"] = 335149,
		},
		[23920] = {
			["spellId"] = 23920,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["applications"] = 1,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132361,
			["nameplateShowPersonal"] = false,
		},
		[213610] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 213610,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 458722,
			["dispelName"] = "Magic",
		},
		[333104] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 333104,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135942,
			["nameplateShowPersonal"] = false,
		},
		[364306] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 1,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 364306,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 612968,
			["nameplateShowPersonal"] = false,
		},
		[201846] = {
			["spellId"] = 201846,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 1,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136099,
			["nameplateShowPersonal"] = false,
		},
		[32216] = {
			["spellId"] = 32216,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132342,
			["nameplateShowPersonal"] = false,
		},
		[334896] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 334896,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 460695,
			["nameplateShowPersonal"] = false,
		},
		[330038] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 330038,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3610512,
			["nameplateShowPersonal"] = false,
		},
		[360984] = {
			["spellId"] = 360984,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["applications"] = 1,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135841,
			["nameplateShowPersonal"] = false,
		},
		[354847] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 460692,
			["spellId"] = 354847,
		},
		[53563] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 53563,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236247,
			["nameplateShowPersonal"] = false,
		},
		[323510] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 120,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 323510,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1500933,
			["nameplateShowPersonal"] = false,
		},
		[337984] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 337984,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1016706,
			["nameplateShowPersonal"] = false,
		},
		[47585] = {
			["spellId"] = 47585,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237563,
			["nameplateShowPersonal"] = false,
		},
		[103196] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 103196,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132261,
			["nameplateShowPersonal"] = false,
		},
		[341550] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236274,
			["spellId"] = 341550,
		},
		[107574] = {
			["spellId"] = 107574,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 613534,
			["nameplateShowPersonal"] = false,
		},
		[183435] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 183435,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135889,
			["nameplateShowPersonal"] = false,
		},
		[242897] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 242897,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1608506,
			["nameplateShowPersonal"] = false,
		},
		[365080] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 365080,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 611420,
			["nameplateShowPersonal"] = false,
		},
		[116014] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 609815,
			["spellId"] = 116014,
		},
		[16593] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 16593,
			["sourceUnit"] = "nameplate7",
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135992,
			["dispelName"] = "Magic",
		},
		[267611] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 267611,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 461860,
			["nameplateShowPersonal"] = true,
		},
		[334391] = {
			["spellId"] = 334391,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3061647,
			["nameplateShowPersonal"] = false,
		},
		[337716] = {
			["spellId"] = 337716,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1022947,
			["nameplateShowPersonal"] = false,
		},
		[210324] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 210324,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135957,
			["nameplateShowPersonal"] = false,
		},
		[23505] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 60,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 136012,
			["spellId"] = 23505,
		},
		[210323] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 210323,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135957,
			["nameplateShowPersonal"] = false,
		},
		[288613] = {
			["spellId"] = 288613,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132329,
			["nameplateShowPersonal"] = true,
		},
		[357410] = {
			["spellId"] = 357410,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 300,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134528,
			["nameplateShowPersonal"] = false,
		},
		[354352] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 354352,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636571,
			["nameplateShowPersonal"] = false,
		},
		[202748] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 1.5,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132214,
			["spellId"] = 202748,
		},
		[196608] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10.4,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 196608,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236165,
			["nameplateShowPersonal"] = false,
		},
		[294564] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 294564,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134062,
			["nameplateShowPersonal"] = false,
		},
		[363805] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 363805,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2065634,
			["nameplateShowPersonal"] = false,
		},
		[120043] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 120043,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 603364,
			["nameplateShowPersonal"] = false,
		},
		[774] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 774,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136081,
			["dispelName"] = "Magic",
		},
		[105809] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 105809,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 571555,
			["nameplateShowPersonal"] = false,
		},
		[157504] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 157504,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 971076,
			["nameplateShowPersonal"] = false,
		},
		[157982] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 1,
			["sourceUnit"] = "nameplate17",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 157982,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136107,
			["nameplateShowPersonal"] = false,
		},
		[330366] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 330366,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2000857,
			["nameplateShowPersonal"] = false,
		},
		[323654] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 5,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 323654,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565724,
			["nameplateShowPersonal"] = false,
		},
		[346502] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 2,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 346502,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1450143,
			["nameplateShowPersonal"] = false,
		},
		[68992] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 68992,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 366937,
			["nameplateShowPersonal"] = false,
		},
		[198144] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 25,
			["spellId"] = 198144,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1387355,
			["dispelName"] = "Magic",
		},
		[332862] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 600,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 332862,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136005,
			["nameplateShowPersonal"] = false,
		},
		[305497] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 12,
			["spellId"] = 305497,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136104,
			["dispelName"] = "Magic",
		},
		[338489] = {
			["spellId"] = 338489,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 636335,
			["nameplateShowPersonal"] = false,
		},
		[92231] = {
			["spellId"] = 92231,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 464161,
			["nameplateShowPersonal"] = false,
		},
		[203155] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 203155,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1412205,
			["nameplateShowPersonal"] = true,
		},
		[336874] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 336874,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 646174,
			["nameplateShowPersonal"] = false,
		},
		[12472] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 33,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 12472,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135838,
			["nameplateShowPersonal"] = true,
		},
		[196099] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 196099,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 538443,
			["nameplateShowPersonal"] = false,
		},
		[60118] = {
			["spellId"] = 60118,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236243,
			["nameplateShowPersonal"] = false,
		},
		[783] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 783,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132144,
			["nameplateShowPersonal"] = false,
		},
		[243435] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 243435,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 615341,
			["nameplateShowPersonal"] = false,
		},
		[364322] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 364322,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 615099,
			["nameplateShowPersonal"] = false,
		},
		[188290] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 9.952,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 188290,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136144,
			["nameplateShowPersonal"] = false,
		},
		[353866] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 353866,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3852564,
			["nameplateShowPersonal"] = false,
		},
		[210294] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 0,
			["spellId"] = 210294,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135915,
			["dispelName"] = "Magic",
		},
		[48792] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 48792,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237525,
			["nameplateShowPersonal"] = false,
		},
		[333889] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 333889,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237564,
			["dispelName"] = "Magic",
		},
		[207400] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate17",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 207400,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237574,
			["nameplateShowPersonal"] = false,
		},
		[337470] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 133784,
			["spellId"] = 337470,
		},
		[48265] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 48265,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237561,
			["nameplateShowPersonal"] = false,
		},
		[59542] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "nameplate11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 59542,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135923,
			["nameplateShowPersonal"] = false,
		},
		[5277] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136205,
			["spellId"] = 5277,
		},
		[48707] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 48707,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136120,
			["nameplateShowPersonal"] = false,
		},
		[329543] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 329543,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135992,
			["nameplateShowPersonal"] = false,
		},
		[354351] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 354351,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636570,
			["nameplateShowPersonal"] = false,
		},
		[257622] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2065591,
			["spellId"] = 257622,
		},
		[334403] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 334403,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3061650,
			["nameplateShowPersonal"] = false,
		},
		[119085] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 119085,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 574577,
			["nameplateShowPersonal"] = false,
		},
		[342076] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132208,
			["spellId"] = 342076,
		},
		[367907] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 367907,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3528279,
			["nameplateShowPersonal"] = false,
		},
		[115989] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 115989,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136132,
			["nameplateShowPersonal"] = false,
		},
		[157228] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 10,
			["spellId"] = 157228,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236163,
			["dispelName"] = "Magic",
		},
		[308087] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 308087,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3054095,
			["nameplateShowPersonal"] = false,
		},
		[262022] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 262022,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1851482,
			["nameplateShowPersonal"] = false,
		},
		[236502] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 2,
			["isBossAura"] = false,
			["duration"] = 0,
			["spellId"] = 236502,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136042,
			["dispelName"] = "Magic",
		},
		[93805] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 93805,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 255148,
			["nameplateShowPersonal"] = false,
		},
		[196742] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 7.887000000000001,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 196742,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 988194,
			["nameplateShowPersonal"] = false,
		},
		[356656] = {
			["spellId"] = 356656,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4226154,
			["nameplateShowPersonal"] = true,
		},
		[339298] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 7.535,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 339298,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136168,
			["nameplateShowPersonal"] = false,
		},
		[768] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 768,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132115,
			["nameplateShowPersonal"] = false,
		},
		[352309] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 352309,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3743310,
			["nameplateShowPersonal"] = false,
		},
		[2096] = {
			["spellId"] = 2096,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135934,
			["nameplateShowPersonal"] = false,
		},
		[186257] = {
			["spellId"] = 186257,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132242,
			["nameplateShowPersonal"] = false,
		},
		[333251] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 8,
			["spellId"] = 333251,
			["sourceUnit"] = "raid6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132810,
			["dispelName"] = "Magic",
		},
		[201098] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 201098,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1308528,
			["nameplateShowPersonal"] = false,
		},
		[356658] = {
			["spellId"] = 356658,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4226150,
			["nameplateShowPersonal"] = true,
		},
		[340546] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4.8,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 340546,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 538517,
			["nameplateShowPersonal"] = false,
		},
		[260286] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 3,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 260286,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1117879,
			["nameplateShowPersonal"] = true,
		},
		[353334] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 644380,
			["spellId"] = 353334,
		},
		[356659] = {
			["spellId"] = 356659,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4226152,
			["nameplateShowPersonal"] = true,
		},
		[186258] = {
			["spellId"] = 186258,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 9,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132242,
			["nameplateShowPersonal"] = false,
		},
		[176151] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 7200,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 176151,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 462324,
			["nameplateShowPersonal"] = false,
		},
		[202636] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 202636,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135881,
			["nameplateShowPersonal"] = false,
		},
		[356660] = {
			["spellId"] = 356660,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4226153,
			["nameplateShowPersonal"] = true,
		},
		[345231] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 345231,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132344,
			["nameplateShowPersonal"] = false,
		},
		[19574] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "arena2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132127,
			["spellId"] = 19574,
		},
		[354359] = {
			["spellId"] = 354359,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3897736,
			["nameplateShowPersonal"] = false,
		},
		[356661] = {
			["spellId"] = 356661,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4226151,
			["nameplateShowPersonal"] = true,
		},
		[266779] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 266779,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2065565,
			["nameplateShowPersonal"] = true,
		},
		[72286] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 72286,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 345787,
			["nameplateShowPersonal"] = false,
		},
		[22723] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 22723,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132225,
			["nameplateShowPersonal"] = false,
		},
		[359731] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 604800.057,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 359731,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4217591,
			["nameplateShowPersonal"] = false,
		},
		[192081] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 7,
			["applications"] = 1,
			["sourceUnit"] = "raid12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 192081,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1378702,
			["nameplateShowPersonal"] = true,
		},
		[328530] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 1,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 328530,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 642580,
			["nameplateShowPersonal"] = false,
		},
		[354361] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 354361,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3897746,
			["nameplateShowPersonal"] = false,
		},
		[342680] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 342680,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3747573,
			["nameplateShowPersonal"] = false,
		},
		[337482] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 337482,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 608953,
			["nameplateShowPersonal"] = false,
		},
		[329554] = {
			["spellId"] = 329554,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 25,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3591588,
			["nameplateShowPersonal"] = false,
		},
		[127220] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 127220,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 656545,
			["nameplateShowPersonal"] = false,
		},
		[5217] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 5217,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132242,
			["nameplateShowPersonal"] = false,
		},
		[355897] = {
			["spellId"] = 355897,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135898,
			["nameplateShowPersonal"] = false,
		},
		[327509] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 327509,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135857,
			["nameplateShowPersonal"] = false,
		},
		[363813] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 2,
			["sourceUnit"] = "nameplate14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 363813,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1416160,
			["nameplateShowPersonal"] = false,
		},
		[347458] = {
			["spellId"] = 347458,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 636335,
			["nameplateShowPersonal"] = false,
		},
		[355898] = {
			["spellId"] = 355898,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237570,
			["nameplateShowPersonal"] = false,
		},
		[328210] = {
			["spellId"] = 328210,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1670848,
			["nameplateShowPersonal"] = false,
		},
		[81340] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 81340,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136181,
			["nameplateShowPersonal"] = false,
		},
		[101568] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 101568,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 878217,
			["nameplateShowPersonal"] = false,
		},
		[197003] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 4,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 965900,
			["spellId"] = 197003,
		},
		[74589] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 1800,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 74589,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132288,
			["nameplateShowPersonal"] = false,
		},
		[73325] = {
			["spellId"] = 73325,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 1.5,
			["applications"] = 0,
			["sourceUnit"] = "nameplate8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 463835,
			["nameplateShowPersonal"] = false,
		},
		[357972] = {
			["spellId"] = 357972,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3079681,
			["nameplateShowPersonal"] = false,
		},
		[364686] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 364686,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136233,
			["nameplateShowPersonal"] = false,
		},
		[339928] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 25,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2065620,
			["spellId"] = 339928,
		},
		[315443] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 16,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 315443,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578196,
			["nameplateShowPersonal"] = false,
		},
		[51460] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 2.545,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 51460,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 252272,
			["nameplateShowPersonal"] = false,
		},
		[63637] = {
			["spellId"] = 63637,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132267,
			["nameplateShowPersonal"] = false,
		},
		[352789] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 352789,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 252184,
			["nameplateShowPersonal"] = false,
		},
		[49379] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 49379,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132246,
			["nameplateShowPersonal"] = false,
		},
		[347462] = {
			["spellId"] = 347462,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1392567,
			["nameplateShowPersonal"] = false,
		},
		[353856] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353856,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3811752,
			["nameplateShowPersonal"] = false,
		},
		[341223] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 341223,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1035054,
			["nameplateShowPersonal"] = false,
		},
		[146622] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 146622,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 877361,
			["nameplateShowPersonal"] = false,
		},
		[355315] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3.001,
			["applications"] = 6,
			["sourceUnit"] = "arena2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 355315,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565718,
			["nameplateShowPersonal"] = false,
		},
		[363522] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4224727,
			["spellId"] = 363522,
		},
		[299159] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 299159,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2620864,
			["nameplateShowPersonal"] = false,
		},
		[333961] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 333961,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3257748,
			["nameplateShowPersonal"] = false,
		},
		[237286] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 237286,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1519598,
			["nameplateShowPersonal"] = false,
		},
		[323546] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 1,
			["sourceUnit"] = "arena2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 323546,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565718,
			["nameplateShowPersonal"] = false,
		},
		[46738] = {
			["spellId"] = 46738,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134519,
			["nameplateShowPersonal"] = false,
		},
		[315496] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132306,
			["spellId"] = 315496,
		},
		[60025] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 60025,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236231,
			["nameplateShowPersonal"] = false,
		},
		[15286] = {
			["spellId"] = 15286,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136230,
			["nameplateShowPersonal"] = true,
		},
		[60114] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 60114,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236244,
			["nameplateShowPersonal"] = false,
		},
		[363578] = {
			["spellId"] = 363578,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135752,
			["nameplateShowPersonal"] = false,
		},
		[352858] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 1,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 352858,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236344,
			["nameplateShowPersonal"] = false,
		},
		[115191] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132320,
			["spellId"] = 115191,
		},
		[13750] = {
			["spellId"] = 13750,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136206,
			["nameplateShowPersonal"] = false,
		},
		[356418] = {
			["spellId"] = 356418,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3939187,
			["nameplateShowPersonal"] = false,
		},
		[325216] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 325216,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578227,
			["nameplateShowPersonal"] = true,
		},
		[61336] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "raid12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 61336,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236169,
			["nameplateShowPersonal"] = true,
		},
		[337749] = {
			["spellId"] = 337749,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135954,
			["nameplateShowPersonal"] = false,
		},
		[179872] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 900,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1029736,
			["spellId"] = 179872,
		},
		[342181] = {
			["spellId"] = 342181,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 442743,
			["nameplateShowPersonal"] = false,
		},
		[340563] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 340563,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 538744,
			["nameplateShowPersonal"] = false,
		},
		[364348] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 3,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 364348,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236297,
			["nameplateShowPersonal"] = false,
		},
		[367673] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 367673,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4394634,
			["nameplateShowPersonal"] = false,
		},
		[93816] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 93816,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 409606,
			["nameplateShowPersonal"] = false,
		},
		[296575] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 296575,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134062,
			["nameplateShowPersonal"] = false,
		},
		[364349] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 364349,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 840195,
			["nameplateShowPersonal"] = false,
		},
		[324220] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 39.4,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 324220,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578226,
			["nameplateShowPersonal"] = false,
		},
		[363793] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 29.992,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 363793,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4038102,
			["nameplateShowPersonal"] = false,
		},
		[115192] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 571317,
			["spellId"] = 115192,
		},
		[335903] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 335903,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1035054,
			["nameplateShowPersonal"] = false,
		},
		[353353] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 4,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3854013,
			["spellId"] = 353353,
		},
		[48517] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 39.998,
			["applications"] = 0,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 48517,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236152,
			["nameplateShowPersonal"] = true,
		},
		[164273] = {
			["spellId"] = 164273,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1033494,
			["nameplateShowPersonal"] = false,
		},
		[114051] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 114051,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135791,
			["nameplateShowPersonal"] = false,
		},
		[44521] = {
			["spellId"] = 44521,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136116,
			["nameplateShowPersonal"] = false,
		},
		[58875] = {
			["spellId"] = 58875,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132328,
			["nameplateShowPersonal"] = false,
		},
		[41516] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 41516,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132250,
			["nameplateShowPersonal"] = false,
		},
		[11426] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 60,
			["spellId"] = 11426,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135988,
			["dispelName"] = "Magic",
		},
		[93326] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 93326,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 461140,
			["nameplateShowPersonal"] = false,
		},
		[137639] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 137639,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136038,
			["nameplateShowPersonal"] = false,
		},
		[363911] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 5,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 363911,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 537026,
			["nameplateShowPersonal"] = true,
		},
		[110909] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 609811,
			["spellId"] = 110909,
		},
		[325202] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 325202,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 607849,
			["nameplateShowPersonal"] = false,
		},
		[335198] = {
			["spellId"] = 335198,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236311,
			["nameplateShowPersonal"] = false,
		},
		[291843] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["sourceUnit"] = "arena2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 291843,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2447780,
			["nameplateShowPersonal"] = false,
		},
		[222202] = {
			["spellId"] = 222202,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1393182,
			["nameplateShowPersonal"] = false,
		},
		[34976] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = true,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 34976,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 132487,
			["nameplateShowPersonal"] = false,
		},
		[139595] = {
			["spellId"] = 139595,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 804461,
			["nameplateShowPersonal"] = false,
		},
		[5761] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 5761,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136066,
			["nameplateShowPersonal"] = false,
		},
		[203277] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 514016,
			["spellId"] = 203277,
		},
		[108416] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 108416,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136146,
			["nameplateShowPersonal"] = false,
		},
		[203533] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 463879,
			["spellId"] = 203533,
		},
		[254811] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 254811,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1716281,
			["nameplateShowPersonal"] = false,
		},
		[205766] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 205766,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236208,
			["nameplateShowPersonal"] = false,
		},
		[231442] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 231442,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1530549,
			["nameplateShowPersonal"] = false,
		},
		[3408] = {
			["spellId"] = 3408,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132274,
			["nameplateShowPersonal"] = false,
		},
		[77761] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 77761,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 463283,
			["nameplateShowPersonal"] = false,
		},
		[338825] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 6,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 338825,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135730,
			["nameplateShowPersonal"] = false,
		},
		[8178] = {
			["spellId"] = 8178,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate12",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136039,
			["nameplateShowPersonal"] = false,
		},
		[166646] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 166646,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 611420,
			["nameplateShowPersonal"] = false,
		},
		[359317] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 359317,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4180079,
			["nameplateShowPersonal"] = false,
		},
		[332932] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 332932,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3063268,
			["nameplateShowPersonal"] = false,
		},
		[260242] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 2,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236179,
			["spellId"] = 260242,
		},
		[48518] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 39.998,
			["applications"] = 0,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 48518,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236151,
			["nameplateShowPersonal"] = true,
		},
		[354017] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 14400.004,
			["applications"] = 0,
			["sourceUnit"] = "raid6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 354017,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 651601,
			["nameplateShowPersonal"] = false,
		},
		[296788] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 296788,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2902390,
			["nameplateShowPersonal"] = false,
		},
		[280729] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 280729,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 840193,
			["nameplateShowPersonal"] = false,
		},
		[5384] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 360,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132293,
			["spellId"] = 5384,
		},
		[291944] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 291944,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1850550,
			["nameplateShowPersonal"] = false,
		},
		[1719] = {
			["spellId"] = 1719,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15.48,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 458972,
			["nameplateShowPersonal"] = true,
		},
		[280730] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2153980,
			["spellId"] = 280730,
		},
		[305395] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 305395,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135968,
			["nameplateShowPersonal"] = false,
		},
		[353875] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353875,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3760459,
			["nameplateShowPersonal"] = false,
		},
		[104773] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 104773,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136150,
			["nameplateShowPersonal"] = true,
		},
		[200183] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 200183,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1060983,
			["nameplateShowPersonal"] = true,
		},
		[114108] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 114108,
			["sourceUnit"] = "nameplate17",
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236160,
			["dispelName"] = "Magic",
		},
		[229487] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 229487,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1499689,
			["nameplateShowPersonal"] = false,
		},
		[310143] = {
			["spellId"] = 310143,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3586268,
			["nameplateShowPersonal"] = false,
		},
		[353365] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353365,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 895875,
			["nameplateShowPersonal"] = false,
		},
		[7744] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0.1,
			["applications"] = 0,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 7744,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136187,
			["nameplateShowPersonal"] = false,
		},
		[325268] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 325268,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237400,
			["nameplateShowPersonal"] = false,
		},
		[338788] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 252188,
			["spellId"] = 338788,
		},
		[241260] = {
			["spellId"] = 241260,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 413585,
			["nameplateShowPersonal"] = false,
		},
		[44544] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 44544,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236227,
			["nameplateShowPersonal"] = false,
		},
		[224125] = {
			["spellId"] = 224125,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1032474,
			["nameplateShowPersonal"] = true,
		},
		[347037] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 1,
			["sourceUnit"] = "raid6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 347037,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636848,
			["nameplateShowPersonal"] = false,
		},
		[340068] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 340068,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3614351,
			["nameplateShowPersonal"] = false,
		},
		[6940] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 6940,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135966,
			["nameplateShowPersonal"] = false,
		},
		[127216] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 127216,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 656546,
			["nameplateShowPersonal"] = false,
		},
		[337228] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 337228,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 613533,
			["dispelName"] = "Magic",
		},
		[355575] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 355575,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135943,
			["nameplateShowPersonal"] = false,
		},
		[31884] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 31884,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135875,
			["nameplateShowPersonal"] = true,
		},
		[224126] = {
			["spellId"] = 224126,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 464162,
			["nameplateShowPersonal"] = true,
		},
		[297871] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 297871,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 463570,
			["nameplateShowPersonal"] = false,
		},
		[325748] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 3,
			["isBossAura"] = false,
			["duration"] = 12,
			["spellId"] = 325748,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578197,
			["dispelName"] = "Magic",
		},
		[354648] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 354648,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136028,
			["nameplateShowPersonal"] = false,
		},
		[33206] = {
			["spellId"] = 33206,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135936,
			["nameplateShowPersonal"] = false,
		},
		[227847] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4.908,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 227847,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236303,
			["nameplateShowPersonal"] = false,
		},
		[55001] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 55001,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135992,
			["nameplateShowPersonal"] = false,
		},
		[256735] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 256735,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132109,
			["nameplateShowPersonal"] = false,
		},
		[224127] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 1027879,
			["spellId"] = 224127,
		},
		[355455] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 355455,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565448,
			["nameplateShowPersonal"] = false,
		},
		[2479] = {
			["spellId"] = 2479,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135994,
			["nameplateShowPersonal"] = false,
		},
		[362699] = {
			["spellId"] = 362699,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 1,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3950373,
			["nameplateShowPersonal"] = false,
		},
		[353883] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 353883,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3896756,
			["nameplateShowPersonal"] = false,
		},
		[329038] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 4,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 329038,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132277,
			["nameplateShowPersonal"] = false,
		},
		[152262] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 4,
			["timeMod"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1030103,
			["spellId"] = 152262,
		},
		[199448] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 199448,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135966,
			["nameplateShowPersonal"] = false,
		},
		[320125] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 320125,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1603013,
			["nameplateShowPersonal"] = false,
		},
		[66157] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 66157,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 236346,
			["nameplateShowPersonal"] = false,
		},
		[130092] = {
			["spellId"] = 130092,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 652810,
			["nameplateShowPersonal"] = false,
		},
		[216328] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 5,
			["spellId"] = 216328,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135931,
			["dispelName"] = "Magic",
		},
		[345453] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5.7,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 345453,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636847,
			["nameplateShowPersonal"] = true,
		},
		[203285] = {
			["spellId"] = 203285,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["applications"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 252172,
			["nameplateShowPersonal"] = false,
		},
		[343144] = {
			["spellId"] = 343144,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136204,
			["nameplateShowPersonal"] = false,
		},
		[356968] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 356968,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2101982,
			["nameplateShowPersonal"] = true,
		},
		[290641] = {
			["spellId"] = 290641,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "nameplate12",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 538564,
			["nameplateShowPersonal"] = false,
		},
		[24858] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 24858,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136036,
			["nameplateShowPersonal"] = false,
		},
		[292486] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 292486,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2446019,
			["nameplateShowPersonal"] = false,
		},
		[336832] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 524793,
			["spellId"] = 336832,
		},
		[101643] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 900,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 101643,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 627608,
			["nameplateShowPersonal"] = false,
		},
		[270232] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 270232,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135857,
			["nameplateShowPersonal"] = false,
		},
		[10060] = {
			["spellId"] = 10060,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = true,
			["isBossAura"] = false,
			["duration"] = 20,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135939,
			["dispelName"] = "Magic",
		},
		[367954] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 367954,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136224,
			["nameplateShowPersonal"] = false,
		},
		[352865] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 352865,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132208,
			["nameplateShowPersonal"] = false,
		},
		[322431] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 322431,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3718862,
			["nameplateShowPersonal"] = false,
		},
		[367467] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 367467,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2054278,
			["nameplateShowPersonal"] = false,
		},
		[232698] = {
			["spellId"] = 232698,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136200,
			["nameplateShowPersonal"] = false,
		},
		[23246] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid11",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 23246,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132264,
			["nameplateShowPersonal"] = false,
		},
		[265144] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 265144,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135982,
			["nameplateShowPersonal"] = false,
		},
		[198300] = {
			["spellId"] = 198300,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 971078,
			["nameplateShowPersonal"] = false,
		},
		[171830] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 171830,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1032613,
			["nameplateShowPersonal"] = false,
		},
		[31821] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 31821,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135872,
			["nameplateShowPersonal"] = false,
		},
		[184362] = {
			["spellId"] = 184362,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = true,
			["isBossAura"] = false,
			["duration"] = 5,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136224,
			["dispelName"] = "",
		},
		[1784] = {
			["spellId"] = 1784,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 1,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132320,
			["nameplateShowPersonal"] = false,
		},
		[341741] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 3,
			["isBossAura"] = false,
			["duration"] = 60,
			["spellId"] = 341741,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1445240,
			["dispelName"] = "Magic",
		},
		[290754] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 290754,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134206,
			["dispelName"] = "Magic",
		},
		[209426] = {
			["spellId"] = 209426,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 7.974,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1305154,
			["nameplateShowPersonal"] = false,
		},
		[333943] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 333943,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 613955,
			["nameplateShowPersonal"] = false,
		},
		[80353] = {
			["spellId"] = 80353,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 40,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 458224,
			["dispelName"] = "Magic",
		},
		[193358] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 193358,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1269440,
			["nameplateShowPersonal"] = true,
		},
		[339321] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 255346,
			["spellId"] = 339321,
		},
		[338036] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 338036,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236186,
			["nameplateShowPersonal"] = false,
		},
		[23509] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 23509,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132266,
			["nameplateShowPersonal"] = false,
		},
		[324226] = {
			["spellId"] = 324226,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135905,
			["nameplateShowPersonal"] = false,
		},
		[338804] = {
			["spellId"] = 338804,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 446089,
			["nameplateShowPersonal"] = false,
		},
		[184662] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 184662,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236264,
			["nameplateShowPersonal"] = false,
		},
		[1044] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135968,
			["spellId"] = 1044,
		},
		[183889] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 183889,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1113014,
			["nameplateShowPersonal"] = false,
		},
		[194594] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236185,
			["spellId"] = 194594,
		},
		[339061] = {
			["spellId"] = 339061,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 1,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1604164,
			["nameplateShowPersonal"] = false,
		},
		[290213] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "nameplate17",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 290213,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236153,
			["nameplateShowPersonal"] = false,
		},
		[125174] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 125174,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 651728,
			["nameplateShowPersonal"] = false,
		},
		[364234] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 1,
			["sourceUnit"] = "nameplate20",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 364234,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132383,
			["nameplateShowPersonal"] = true,
		},
		[323764] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["applications"] = 1,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 323764,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636839,
			["nameplateShowPersonal"] = false,
		},
		[126389] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 120,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 126389,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237276,
			["nameplateShowPersonal"] = false,
		},
		[339412] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 339412,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 589546,
			["nameplateShowPersonal"] = false,
		},
		[336372] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 1,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 336372,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134422,
			["nameplateShowPersonal"] = false,
		},
		[363727] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1445240,
			["spellId"] = 363727,
		},
		[33763] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 33763,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134206,
			["dispelName"] = "Magic",
		},
		[157128] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 157128,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1030102,
			["nameplateShowPersonal"] = false,
		},
		[260708] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 260708,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132306,
			["nameplateShowPersonal"] = true,
		},
		[344179] = {
			["spellId"] = 344179,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["applications"] = 1,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237584,
			["nameplateShowPersonal"] = true,
		},
		[232412] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 232412,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1536702,
			["nameplateShowPersonal"] = false,
		},
		[352875] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 366056,
			["spellId"] = 352875,
		},
		[65081] = {
			["spellId"] = 65081,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135982,
			["dispelName"] = "Magic",
		},
		[351433] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 351433,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3190332,
			["nameplateShowPersonal"] = false,
		},
		[113858] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 113858,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 463286,
			["nameplateShowPersonal"] = false,
		},
		[340600] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132292,
			["spellId"] = 340600,
		},
		[33702] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 33702,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135726,
			["nameplateShowPersonal"] = false,
		},
		[114050] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 114050,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135791,
			["nameplateShowPersonal"] = false,
		},
		[207640] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 3,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 207640,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132124,
			["nameplateShowPersonal"] = false,
		},
		[156618] = {
			["spellId"] = 156618,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = true,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 132485,
			["nameplateShowPersonal"] = false,
		},
		[113860] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 113860,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 463286,
			["nameplateShowPersonal"] = true,
		},
		[352086] = {
			["spellId"] = 352086,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 120,
			["applications"] = 3,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132100,
			["nameplateShowPersonal"] = false,
		},
		[334320] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 1,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 334320,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 537517,
			["nameplateShowPersonal"] = false,
		},
		[343594] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "raid14",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 343594,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1029718,
			["nameplateShowPersonal"] = false,
		},
		[325443] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 325443,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134060,
			["nameplateShowPersonal"] = false,
		},
		[307932] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 307932,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3040844,
			["nameplateShowPersonal"] = false,
		},
		[290608] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 290608,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2564351,
			["nameplateShowPersonal"] = false,
		},
		[242551] = {
			["spellId"] = 242551,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 7200,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134924,
			["nameplateShowPersonal"] = false,
		},
		[289555] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 289555,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2530431,
			["nameplateShowPersonal"] = false,
		},
		[85739] = {
			["spellId"] = 85739,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["applications"] = 2,
			["sourceUnit"] = "nameplate5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132369,
			["nameplateShowPersonal"] = false,
		},
		[309658] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 40,
			["spellId"] = 309658,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 133843,
			["dispelName"] = "Magic",
		},
		[294569] = {
			["spellId"] = 294569,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2829669,
			["nameplateShowPersonal"] = false,
		},
		[311193] = {
			["spellId"] = 311193,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3565453,
			["nameplateShowPersonal"] = false,
		},
		[324748] = {
			["spellId"] = 324748,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 636335,
			["nameplateShowPersonal"] = false,
		},
		[124659] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 124659,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 644374,
			["nameplateShowPersonal"] = false,
		},
		[258920] = {
			["spellId"] = 258920,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1344649,
			["nameplateShowPersonal"] = true,
		},
		[23222] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 23222,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132247,
			["nameplateShowPersonal"] = false,
		},
		[340094] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135661,
			["spellId"] = 340094,
		},
		[216338] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 216338,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134062,
			["nameplateShowPersonal"] = false,
		},
		[134359] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 134359,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 894222,
			["nameplateShowPersonal"] = false,
		},
		[193065] = {
			["spellId"] = 193065,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["sourceUnit"] = "nameplate7",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = true,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136176,
			["dispelName"] = "Magic",
		},
		[333957] = {
			["spellId"] = 333957,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237577,
			["nameplateShowPersonal"] = false,
		},
		[347256] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 347256,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3894514,
			["nameplateShowPersonal"] = false,
		},
		[156621] = {
			["spellId"] = 156621,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = true,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 132486,
			["nameplateShowPersonal"] = false,
		},
		[351077] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 351077,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132175,
			["nameplateShowPersonal"] = false,
		},
		[7384] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 7384,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132223,
			["nameplateShowPersonal"] = true,
		},
		[232707] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 5.999000000000001,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 612098,
			["spellId"] = 232707,
		},
		[1850] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 1850,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132120,
			["nameplateShowPersonal"] = false,
		},
		[79206] = {
			["spellId"] = 79206,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["sourceUnit"] = "nameplate12",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = true,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 451170,
			["dispelName"] = "Magic",
		},
		[114052] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 114052,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135791,
			["nameplateShowPersonal"] = false,
		},
		[268898] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 268898,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132178,
			["nameplateShowPersonal"] = false,
		},
		[186289] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 186289,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 612363,
			["nameplateShowPersonal"] = false,
		},
		[46924] = {
			["spellId"] = 46924,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3.95,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236303,
			["nameplateShowPersonal"] = false,
		},
		[212295] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 3,
			["spellId"] = 212295,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135796,
			["dispelName"] = "Magic",
		},
		[232708] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 5.541,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 631503,
			["spellId"] = 232708,
		},
		[216468] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 216468,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 968261,
			["nameplateShowPersonal"] = false,
		},
		[160331] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 300,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 160331,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 133564,
			["nameplateShowPersonal"] = false,
		},
		[48107] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 7.647,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236218,
			["spellId"] = 48107,
		},
		[44535] = {
			["spellId"] = 44535,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135955,
			["nameplateShowPersonal"] = false,
		},
		[339588] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = true,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2143082,
			["spellId"] = 339588,
		},
		[323476] = {
			["spellId"] = 323476,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3459798,
			["nameplateShowPersonal"] = false,
		},
		[229512] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 229512,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1517168,
			["nameplateShowPersonal"] = false,
		},
		[363885] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 2,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 363885,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1060569,
			["nameplateShowPersonal"] = false,
		},
		[198069] = {
			["spellId"] = 198069,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["applications"] = 1,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1394892,
			["nameplateShowPersonal"] = false,
		},
		[367466] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 367466,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 254117,
			["nameplateShowPersonal"] = false,
		},
		[46739] = {
			["spellId"] = 46739,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134519,
			["nameplateShowPersonal"] = false,
		},
		[203554] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 203554,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1408837,
			["nameplateShowPersonal"] = false,
		},
		[340613] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 340613,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 571585,
			["nameplateShowPersonal"] = false,
		},
		[118337] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 118337,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132154,
			["nameplateShowPersonal"] = false,
		},
		[333964] = {
			["spellId"] = 333964,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 3,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1370984,
			["nameplateShowPersonal"] = false,
		},
		[264656] = {
			["spellId"] = 264656,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 348567,
			["nameplateShowPersonal"] = false,
		},
		[352857] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 352857,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1582140,
			["nameplateShowPersonal"] = false,
		},
		[280398] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135921,
			["spellId"] = 280398,
		},
		[205473] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 60,
			["applications"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 205473,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135855,
			["nameplateShowPersonal"] = false,
		},
		[213209] = {
			["spellId"] = 213209,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1387705,
			["nameplateShowPersonal"] = false,
		},
		[216339] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 216339,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132805,
			["nameplateShowPersonal"] = false,
		},
		[195627] = {
			["spellId"] = 195627,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1373908,
			["nameplateShowPersonal"] = false,
		},
		[343173] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136183,
			["spellId"] = 343173,
		},
		[216343] = {
			["spellId"] = 216343,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136000,
			["nameplateShowPersonal"] = false,
		},
		[47788] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237542,
			["spellId"] = 47788,
		},
		[54149] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 54149,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236254,
			["nameplateShowPersonal"] = false,
		},
		[46199] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 46199,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134285,
			["nameplateShowPersonal"] = false,
		},
		[367143] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 131327,
			["timeMod"] = 1,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4238797,
			["spellId"] = 367143,
		},
		[114255] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135981,
			["spellId"] = 114255,
		},
		[17] = {
			["spellId"] = 17,
			["nameplateShowAll"] = false,
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["timeMod"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135940,
			["dispelName"] = "Magic",
		},
		[121471] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 376022,
			["spellId"] = 121471,
		},
		[61447] = {
			["spellId"] = 61447,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236240,
			["nameplateShowPersonal"] = false,
		},
		[228583] = {
			["spellId"] = 228583,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1390942,
			["nameplateShowPersonal"] = false,
		},
		[347600] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 133250,
			["spellId"] = 347600,
		},
		[48108] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1035045,
			["spellId"] = 48108,
		},
		[193538] = {
			["spellId"] = 193538,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["applications"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 571558,
			["nameplateShowPersonal"] = false,
		},
		[138213] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 133666,
			["spellId"] = 138213,
		},
		[347765] = {
			["spellId"] = 347765,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 36,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136214,
			["nameplateShowPersonal"] = false,
		},
		[359545] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 359545,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4216936,
			["nameplateShowPersonal"] = false,
		},
		[264662] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 264662,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236180,
			["nameplateShowPersonal"] = false,
		},
		[8679] = {
			["spellId"] = 8679,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3600.003,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134197,
			["nameplateShowPersonal"] = false,
		},
		[367405] = {
			["spellId"] = 367405,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4224736,
			["nameplateShowPersonal"] = false,
		},
		[194223] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 29,
			["applications"] = 0,
			["sourceUnit"] = "arena2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 194223,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136060,
			["nameplateShowPersonal"] = false,
		},
		[264663] = {
			["spellId"] = 264663,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "partypet1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136231,
			["nameplateShowPersonal"] = false,
		},
		[197548] = {
			["spellId"] = 197548,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135871,
			["nameplateShowPersonal"] = false,
		},
		[231435] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 231435,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1519875,
			["nameplateShowPersonal"] = false,
		},
		[280776] = {
			["spellId"] = 280776,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132346,
			["nameplateShowPersonal"] = true,
		},
		[314791] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 3.06,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3636841,
			["spellId"] = 314791,
		},
		[273104] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 273104,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1786406,
			["nameplateShowPersonal"] = false,
		},
		[212704] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 212704,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132166,
			["nameplateShowPersonal"] = false,
		},
		[157644] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 1,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135810,
			["spellId"] = 157644,
		},
		[281178] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1360757,
			["spellId"] = 281178,
		},
		[247676] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["timeMod"] = 1,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["applications"] = 2,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135882,
			["spellId"] = 247676,
		},
		[117828] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 4,
			["isBossAura"] = false,
			["duration"] = 10,
			["spellId"] = 117828,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236290,
			["dispelName"] = "Magic",
		},
		[65116] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 65116,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136225,
			["nameplateShowPersonal"] = false,
		},
		[337299] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1355357,
			["spellId"] = 337299,
		},
		[200175] = {
			["spellId"] = 200175,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid9",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1672586,
			["nameplateShowPersonal"] = false,
		},
		[121153] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 1,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 121153,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236274,
			["nameplateShowPersonal"] = false,
		},
		[345228] = {
			["spellId"] = 345228,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135884,
			["nameplateShowPersonal"] = false,
		},
		[264667] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 40,
			["spellId"] = 264667,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136224,
			["dispelName"] = "Magic",
		},
		[247677] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 1,
			["sourceUnit"] = "nameplate5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 247677,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135986,
			["nameplateShowPersonal"] = false,
		},
		[321444] = {
			["spellId"] = 321444,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 967521,
			["nameplateShowPersonal"] = false,
		},
		[163024] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 163024,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 986419,
			["nameplateShowPersonal"] = false,
		},
		[342416] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 342416,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 425950,
			["nameplateShowPersonal"] = false,
		},
		[330749] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 330749,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 133018,
			["nameplateShowPersonal"] = false,
		},
		[323491] = {
			["spellId"] = 323491,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 120,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1500963,
			["nameplateShowPersonal"] = false,
		},
		[2823] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 2823,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132290,
			["nameplateShowPersonal"] = false,
		},
		[22812] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 22812,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136097,
			["nameplateShowPersonal"] = true,
		},
		[6673] = {
			["spellId"] = 6673,
			["nameplateShowAll"] = false,
			["duration"] = 3600,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132333,
			["nameplateShowPersonal"] = false,
		},
		[108366] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 108366,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 571320,
			["nameplateShowPersonal"] = false,
		},
		[163025] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 163025,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 986420,
			["nameplateShowPersonal"] = false,
		},
		[135700] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 1,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 135700,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136170,
			["dispelName"] = "Magic",
		},
		[256374] = {
			["spellId"] = 256374,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1723992,
			["nameplateShowPersonal"] = false,
		},
		[337815] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 337815,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 461859,
			["nameplateShowPersonal"] = false,
		},
		[53480] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["timeMod"] = 1,
			["sourceUnit"] = "arena2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 464604,
			["spellId"] = 53480,
		},
		[186265] = {
			["spellId"] = 186265,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132199,
			["nameplateShowPersonal"] = false,
		},
		[345230] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 20,
			["timeMod"] = 1,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134501,
			["spellId"] = 345230,
		},
		[27827] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = true,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132864,
			["spellId"] = 27827,
		},
		[23978] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 23978,
			["isFromPlayerOrPlayerPet"] = false,
			["icon"] = 132870,
			["nameplateShowPersonal"] = false,
		},
		[223819] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 12,
			["applications"] = 1,
			["sourceUnit"] = "nameplate1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 223819,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135897,
			["nameplateShowPersonal"] = true,
		},
		[363136] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 363136,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 4287558,
			["nameplateShowPersonal"] = false,
		},
		[290133] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 290133,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2495963,
			["nameplateShowPersonal"] = false,
		},
		[330927] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135995,
			["spellId"] = 330927,
		},
		[171850] = {
			["spellId"] = 171850,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1338908,
			["nameplateShowPersonal"] = false,
		},
		[193357] = {
			["spellId"] = 193357,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1269441,
			["nameplateShowPersonal"] = true,
		},
		[131347] = {
			["spellId"] = 131347,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1305157,
			["nameplateShowPersonal"] = false,
		},
		[2645] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "player",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136095,
			["spellId"] = 2645,
		},
		[66] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 66,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132220,
			["nameplateShowPersonal"] = false,
		},
		[221883] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5.7,
			["applications"] = 0,
			["sourceUnit"] = "nameplate3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 221883,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1360759,
			["nameplateShowPersonal"] = false,
		},
		[974] = {
			["spellId"] = 974,
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 600,
			["applications"] = 9,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136089,
			["dispelName"] = "Magic",
		},
		[331937] = {
			["spellId"] = 331937,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136088,
			["nameplateShowPersonal"] = false,
		},
		[171851] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 171851,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132224,
			["nameplateShowPersonal"] = false,
		},
		[199600] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 199600,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1269443,
			["nameplateShowPersonal"] = true,
		},
		[323498] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 120,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 323498,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1500948,
			["nameplateShowPersonal"] = false,
		},
		[33660] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 33660,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132232,
			["nameplateShowPersonal"] = false,
		},
		[106951] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 20,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 106951,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236149,
			["nameplateShowPersonal"] = false,
		},
		[32223] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 32223,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135890,
			["nameplateShowPersonal"] = false,
		},
		[54216] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 4,
			["spellId"] = 54216,
			["sourceUnit"] = "raidpet4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236189,
			["dispelName"] = "Magic",
		},
		[331939] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "player",
			["canApplyAura"] = false,
			["applications"] = 1,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236268,
			["spellId"] = 331939,
		},
		[333218] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 8,
			["spellId"] = 333218,
			["sourceUnit"] = "raid6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132810,
			["dispelName"] = "Magic",
		},
		[354016] = {
			["spellId"] = 354016,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 14400,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 651601,
			["nameplateShowPersonal"] = false,
		},
		[324523] = {
			["spellId"] = 324523,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 524486,
			["nameplateShowPersonal"] = false,
		},
		[19236] = {
			["spellId"] = 19236,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237550,
			["nameplateShowPersonal"] = false,
		},
		[312759] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 312759,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3087324,
			["nameplateShowPersonal"] = false,
		},
		[337567] = {
			["spellId"] = 337567,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1309101,
			["nameplateShowPersonal"] = true,
		},
		[191034] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 191034,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236168,
			["nameplateShowPersonal"] = true,
		},
		[106898] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10.52,
			["applications"] = 0,
			["sourceUnit"] = "raid13",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 106898,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 464343,
			["nameplateShowPersonal"] = false,
		},
		[465] = {
			["spellId"] = 465,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135893,
			["nameplateShowPersonal"] = false,
		},
		[1966] = {
			["nameplateShowPersonal"] = true,
			["nameplateShowAll"] = false,
			["duration"] = 6,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132294,
			["spellId"] = 1966,
		},
		[337824] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 18,
			["applications"] = 0,
			["sourceUnit"] = "nameplate11",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 337824,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135909,
			["nameplateShowPersonal"] = false,
		},
		[281195] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 6,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 281195,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136094,
			["nameplateShowPersonal"] = false,
		},
		[312761] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 312761,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3087361,
			["nameplateShowPersonal"] = false,
		},
		[285933] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136185,
			["spellId"] = 285933,
		},
		[367680] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 367680,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1371266,
			["nameplateShowPersonal"] = false,
		},
		[168657] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 5,
			["timeMod"] = 1,
			["sourceUnit"] = "arena2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135467,
			["spellId"] = 168657,
		},
		[339461] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 339461,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136094,
			["nameplateShowPersonal"] = false,
		},
		[199603] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 199603,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1269439,
			["nameplateShowPersonal"] = true,
		},
		[222237] = {
			["spellId"] = 222237,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1450959,
			["nameplateShowPersonal"] = false,
		},
		[330921] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 330921,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135995,
			["nameplateShowPersonal"] = false,
		},
		[46736] = {
			["spellId"] = 46736,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134519,
			["nameplateShowPersonal"] = false,
		},
		[162264] = {
			["spellId"] = 162264,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 7.228,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1247262,
			["nameplateShowPersonal"] = true,
		},
		[191634] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 2,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 191634,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 839977,
			["nameplateShowPersonal"] = false,
		},
		[20707] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 900,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 20707,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136210,
			["nameplateShowPersonal"] = false,
		},
		[77489] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 8.126,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237537,
			["spellId"] = 77489,
		},
		[352917] = {
			["spellId"] = 352917,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132866,
			["nameplateShowPersonal"] = false,
		},
		[222238] = {
			["spellId"] = 222238,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1450960,
			["nameplateShowPersonal"] = false,
		},
		[88741] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 88741,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 461143,
			["nameplateShowPersonal"] = false,
		},
		[338340] = {
			["spellId"] = 338340,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 3,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237549,
			["nameplateShowPersonal"] = false,
		},
		[352881] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 5,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 352881,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3528275,
			["nameplateShowPersonal"] = false,
		},
		[321461] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 321461,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1604164,
			["nameplateShowPersonal"] = false,
		},
		[98204] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 98204,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132117,
			["nameplateShowPersonal"] = false,
		},
		[323500] = {
			["spellId"] = 323500,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 120,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1500911,
			["nameplateShowPersonal"] = false,
		},
		[498] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 498,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 524353,
			["nameplateShowPersonal"] = false,
		},
		[21562] = {
			["spellId"] = 21562,
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135987,
			["dispelName"] = "Magic",
		},
		[111759] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 600,
			["spellId"] = 111759,
			["sourceUnit"] = "raid6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135928,
			["dispelName"] = "Magic",
		},
		[342814] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 8,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 342814,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236160,
			["nameplateShowPersonal"] = false,
		},
		[246152] = {
			["spellId"] = 246152,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2058007,
			["nameplateShowPersonal"] = false,
		},
		[266087] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 30,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 266087,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135804,
			["nameplateShowPersonal"] = false,
		},
		[356501] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 356501,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3931115,
			["nameplateShowPersonal"] = false,
		},
		[184364] = {
			["spellId"] = 184364,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132345,
			["nameplateShowPersonal"] = false,
		},
		[337697] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "raid10",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 337697,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1596903,
			["nameplateShowPersonal"] = false,
		},
		[222240] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 222240,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1450962,
			["nameplateShowPersonal"] = false,
		},
		[16591] = {
			["dispelName"] = "Magic",
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 600,
			["nameplateShowPersonal"] = false,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136187,
			["spellId"] = 16591,
		},
		[69369] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 12,
			["spellId"] = 69369,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132185,
			["dispelName"] = "Magic",
		},
		[47536] = {
			["spellId"] = 47536,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["applications"] = 0,
			["sourceUnit"] = "raid1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 237548,
			["nameplateShowPersonal"] = true,
		},
		[323385] = {
			["spellId"] = 323385,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "raid7",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3459797,
			["nameplateShowPersonal"] = false,
		},
		[2983] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 8,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132307,
			["spellId"] = 2983,
		},
		[324156] = {
			["spellId"] = 324156,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 1,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236174,
			["nameplateShowPersonal"] = false,
		},
		[45438] = {
			["dispelName"] = "Magic",
			["nameplateShowPersonal"] = false,
			["isBossAura"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["timeMod"] = 1,
			["sourceUnit"] = "party1",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["applications"] = 0,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135841,
			["spellId"] = 45438,
		},
		[222241] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 222241,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1450958,
			["nameplateShowPersonal"] = false,
		},
		[315584] = {
			["spellId"] = 315584,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3600,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132273,
			["nameplateShowPersonal"] = false,
		},
		[232405] = {
			["spellId"] = 232405,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1530249,
			["nameplateShowPersonal"] = false,
		},
		[332464] = {
			["spellId"] = 332464,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3849327,
			["nameplateShowPersonal"] = false,
		},
		[65917] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132834,
			["spellId"] = 65917,
		},
		[332462] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 332462,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3849326,
			["nameplateShowPersonal"] = false,
		},
		[213602] = {
			["spellId"] = 213602,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3,
			["applications"] = 1,
			["sourceUnit"] = "nameplate5",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 135994,
			["nameplateShowPersonal"] = true,
		},
		[334511] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 1,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 334511,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134338,
			["nameplateShowPersonal"] = false,
		},
		[258022] = {
			["spellId"] = 258022,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "nameplate6",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1713157,
			["nameplateShowPersonal"] = false,
		},
		[52127] = {
			["spellId"] = 52127,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 3600,
			["applications"] = 3,
			["sourceUnit"] = "nameplate8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132315,
			["nameplateShowPersonal"] = false,
		},
		[324463] = {
			["spellId"] = 324463,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 1,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3163625,
			["nameplateShowPersonal"] = false,
		},
		[73920] = {
			["spellId"] = 73920,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136037,
			["nameplateShowPersonal"] = false,
		},
		[23033] = {
			["spellId"] = 23033,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132483,
			["nameplateShowPersonal"] = false,
		},
		[236181] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "raid8",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 236181,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1408834,
			["nameplateShowPersonal"] = false,
		},
		[363924] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 10,
			["applications"] = 3,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 363924,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 134919,
			["nameplateShowPersonal"] = true,
		},
		[194249] = {
			["spellId"] = 194249,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1386550,
			["nameplateShowPersonal"] = true,
		},
		[329666] = {
			["spellId"] = 329666,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 12,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2115310,
			["nameplateShowPersonal"] = false,
		},
		[118922] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 4,
			["timeMod"] = 1,
			["sourceUnit"] = "raid3",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 461119,
			["spellId"] = 118922,
		},
		[251143] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 60,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 251143,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 627487,
			["nameplateShowPersonal"] = false,
		},
		[156132] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 300,
			["applications"] = 0,
			["canApplyAura"] = false,
			["isStealable"] = false,
			["spellId"] = 156132,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 133867,
			["nameplateShowPersonal"] = false,
		},
		[23161] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["isBossAura"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["canApplyAura"] = true,
			["isStealable"] = false,
			["spellId"] = 23161,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 132238,
			["nameplateShowPersonal"] = false,
		},
		[260734] = {
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["applications"] = 0,
			["isBossAura"] = false,
			["duration"] = 15,
			["spellId"] = 260734,
			["canApplyAura"] = false,
			["isStealable"] = true,
			["nameplateShowPersonal"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136027,
			["dispelName"] = "Magic",
		},
		[354018] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 14400.003,
			["timeMod"] = 1,
			["sourceUnit"] = "raid5",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 651601,
			["spellId"] = 354018,
		},
		[328908] = {
			["spellId"] = 328908,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "raid4",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3528283,
			["nameplateShowPersonal"] = false,
		},
		[93795] = {
			["spellId"] = 93795,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 255150,
			["nameplateShowPersonal"] = false,
		},
		[272770] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "party2",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 1998992,
			["spellId"] = 272770,
		},
		[295386] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "nameplate4",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 2828483,
			["spellId"] = 295386,
		},
		[201633] = {
			["nameplateShowPersonal"] = false,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["timeMod"] = 1,
			["sourceUnit"] = "target",
			["canApplyAura"] = false,
			["applications"] = 0,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 136098,
			["spellId"] = 201633,
		},
		[351913] = {
			["spellId"] = 351913,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 0,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236271,
			["nameplateShowPersonal"] = false,
		},
		[351915] = {
			["spellId"] = 351915,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 10,
			["applications"] = 0,
			["sourceUnit"] = "party3",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 236271,
			["nameplateShowPersonal"] = false,
		},
		[325862] = {
			["spellId"] = 325862,
			["timeMod"] = 1,
			["nameplateShowAll"] = false,
			["duration"] = 15,
			["applications"] = 0,
			["sourceUnit"] = "nameplate2",
			["canApplyAura"] = false,
			["isStealable"] = false,
			["isBossAura"] = false,
			["isFromPlayerOrPlayerPet"] = true,
			["icon"] = 3578234,
			["nameplateShowPersonal"] = false,
		},
	},
}
