---@type string
local AddonName = ...
---@class Data
local Data = select(2, ...)
local defaultLocale = {}


local gameLocale = GetLocale()
if gameLocale == "enGB" then
	gameLocale = "enUS"
end

local errorReported, missingReported = false, false

Data.L = setmetatable({}, { --key set by all non english clients, Table gets accessed to read translations
    __index = function(t, k)  -- t is the normal table (no metatable)
        if defaultLocale[k] then
            if gameLocale ~= "enUS" and not missingReported then
                C_Timer.After(3, function()
                    BattleGroundEnemies:OnetimeInformation("Missing localizations for your ingame language. You can help translating this addon on https://www.curseforge.com/wow/addons/battlegroundenemies/localization")
                end)
                missingReported = true
            end
            --t[k] = defaultLocale[k] --add it to the table so we dont have to invoce the metatable in the future
            return defaultLocale[k]
        else
            C_Timer.After(3, function()
                BattleGroundEnemies:OnetimeDebug("Missing localization entry for['"..k.."']. Please report this to the addon author.")
            end)
            return k
        end
    end
})

local L = defaultLocale --set to L for curseforges system

L["AddPoint"] = "Add Point"
L["Allies"] = "Allies"
L["allies"] = "allies"
L["ally"] = "ally"
L["Ally"] = "Ally"
L["AllyIsTargeted"] = "Ally is targeted"
L["AllyJoined"] = "An ally joined the battleground"
L["AllyLeft"] = "An ally left the battleground"
L["AND"] = "and"
L["AttachToObject"] = "Attach to object"
L["AttachToObject_Desc"] = "Note that not every object might be possible to select, depending on your positioning of the other frames. For example if you anchor the pvp talents at the racials and you are trying to anchor the racials at the pvp talents there will be a error message."
L["Auras_Buffs_Container_Color_Desc"] = "The border color of the container of the buff icons"
L["Auras_CustomFiltering_Conditions_All"] = "All"
L["Auras_CustomFiltering_Conditions_Any"] = "Any"
L["Auras_CustomFiltering_ConditionsMode"] = "Conditions mode"
L["Auras_CustomFiltering_ConditionsMode_Desc"] = [=[Here you can choose how the selected conditions should apply to the auras. 
If you choose "Any" the addon will show the aura as soon as one of your selected conditions is met.
If you choose "All" the addon will only show an aura if it mets all of your selected conditions.]=]
L["Auras_Debuffs_Container_Color_Desc"] = "The border color of the container of the debuff icons"
L["Auras_Enabled"] = "Enable Auras"
L["Auras_Enabled_Desc"] = "When enabled there will be aura icons. There will also be a number for the remaining time and one for the stacks applied"
L["Auras_Filtering_Mode"] = "Filtering mode"
L["Auras_Filtering_Mode_Desc"] = [=[Here you can choose which filter should be applied to the auras.
You can choose between a blizzlike filter which applys the same filter als Blizzard uses on the raid frame and custom filtering conditions]=]
L["Auras_ShowTooltips"] = "Show spell tooltips"
L["AurasCustomConditions"] = "Custom conditions"
L["AurasFiltering_AddSpellID"] = "SpellID"
L["AurasFiltering_AddSpellID_Desc"] = "Enter a spellID or multiple comma-seperated spellIDs. The spellIDs are getting added to your filter list. Only auras whose spellID is on the filter list are getting shown when the spellID filtering is enabled."
L["AurasFiltering_Enabled_Desc"] = "When this option is enabled you will only see %s which you added to the filtering list"
L["AurasFiltering_Filterlist_Desc"] = "Click to remove this %s"
L["AurasFilteringSettings_Desc"] = "Here you can control which aura gets shown"
L["AurasSettings"] = "Auras"
L["AurasSettings_Desc"] = "Settings about auras (buffs and debuffs)"
L["AurasStacktextSettings"] = "Stack text"
L["BarBackground"] = "Background Color"
L["BarHeight_Desc"] = "The height of each row. This setting is not available in combat due to Blizzard's combat protection."
L["BarSettings"] = "Bar settings"
L["BarSettings_Desc"] = "Here you can configure the rows"
L["BarTexture"] = "Bar texture"
L["BarWidth_Desc"] = "The width of each row."
L["BattlegroundSize"] = "Battleground size"
L["BGSize_15"] = "1–15 Players"
L["BGSize_15_Desc"] = "This settings apply for %s in a BG with 1–15 players per side"
L["BGSize_40"] = "16–40 Players"
L["BGSize_40_Desc"] = "This settings apply for %s in a BG with 16–40 players per side"
L["BGSize_5"] = "Arena"
L["BGSize_5_Desc"] = "This settings apply for %s in arenas"
L["BlizzlikeAuraFiltering"] = "Use Blizzard's raidframe filter"
L["BorderSize"] = "Border size"
L["BorderThickness"] = "Border size"
L["BOTTOM"] = "Bottom"
L["BOTTOMLEFT"] = "Bottom left"
L["BOTTOMRIGHT"] = "Bottom right"
L["buff"] = "buff"
L["BuffContainer"] = "Buff container"
L["BuffIcon"] = "Buff icon"
L["Buffs"] = "Buffs"
L["Button"] = "Button"
L["CantAnchorToItself"] = "Can't anchor to itself!"
L["CENTER"] = "Center"
L["Change"] = "Change"
L["Color"] = "Color"
L["Coloring_Enabled"] = "Color by debuff type"
L["Coloring_Enabled_Desc"] = "When this option is enabled you can choose whether the frame's border or the countdown text should be colored by debuff type (Like its know from the default UI, poisons are green, diseases blue etc.)"
L["Columns"] = "Columns"
L["Columns_Desc"] = "In how many colums the players are displayed."
L["Combat"] = "In Combat"
L["CombatIndicator"] = "Combat indicator"
L["CombatLogScanningForEnemiesDisabled"] = "Disabled combat log scanning for enemies"
L["CombatLogScanningForEnemiesEnabled"] = "Combat log scanning to detect enemy players has been enabled."
L["ConfirmDeletePlayerCountProfile"] = "Are you sure you wan't to delete the profile that currently applies to"
L["ConfirmProfileOverride"] = "Are you sure that you want to override the subprofile %s with the subprofile %s"
L["Container_Color"] = "Container border color"
L["ContainerPosition"] = "Container position"
L["ContainerSettings"] = "Container settings"
L["ConvertCyrillic"] = "Convert Cyrillic"
L["ConvertCyrillic_Desc"] = "Converts cyrillic characters which makes it easier to read names if you are playing against russians"
L["Cooldown"] = "Cooldown"
L["CopyFrom"] = "Copy from"
L["CopySettings"] = "Copy settings from"
L["CopySettingsMirrored_Desc"] = "Mirrored meaning that all of the position and expanding direction of modules get replaced by the horizontally opposite point or direction. (For example point \"TOPLEFT\" becomes \"TOPRIGHT\" and container modules like the aura module will grow the opposite direction)"
L["CopySettingsNormal_Desc"] = "Just a simple copy"
L["Countdowntext"] = "Countdown text"
L["Covenant"] = "Covenant"
L["CovenantIcon_Enabled"] = "Enable covenant icons"
L["CovenantIcon_Enabled_Desc"] = "Enable this option if you want to see a covenant icon on the health bar"
L["CovenantIcon_Size_Desc"] = "The size of the covenant icons (width and height)"
L["CovenantIconSettings_Desc"] = "Settings for the covenant icon which is shown on the health bar"
L["CurrentVersion"] = "Current version"
L["Curse"] = "Curse"
L["CustomMacro_Desc"] = [=[This enables you to set a custom macro. A empty field will lead to a empty macro.
Available substitutions: %n gets substituted with the enemy's name

Example: 
/targetexact %n
/cast Polymorph
/say Polymorph on %n
/targetlasttarget

This will target the enemy, cast Polymorph, write a message into the say channel and target your previous target.
Note that macros can only be 255 characters long (including substitutions).]=]
L["Data"] = "Data"
L["debuff"] = "debuff"
L["DebuffContainer"] = "Debuff container"
L["DebuffIcon"] = "Debuff icon"
L["Debuffs"] = "Debuffs"
L["DebuffType_Filtering"] = "Filter by debuff type"
L["DebuffType_Filtering_Desc"] = "When this option is enabled you will only see debuffs of the chose typ. This option is not simulated in the test mode since this would require a huge database with tons of spells and their debuff type."
L["DeletePlayerCountProfile"] = "Delete"
L["DeletePlayerCountProfile_Desc"] = "Click here to delete the profile that currently applies to"
L["DeletePoint"] = "Delete Point %s"
L["DisableArenaFramesInArena"] = "Disable arena frames in arena"
L["DisableArenaFramesInArena_Desc"] = "Disables the Arenaframes in a arena. Also works when sArena is installed"
L["DisableArenaFramesInBattleground"] = "Disable arena frames in battleground"
L["DisableArenaFramesInBattleground_Desc"] = "Disables the Arenaframes in a battleground. Also works when sArena is installed"
L["Disease"] = "Disease"
L["DispellFilter"] = "Filter by dispellability"
L["DisplayType"] = "Display type"
L["Downwards"] = "Downwards"
L["DR_Disorient"] = "Disorient"
L["DR_Incapacitate"] = "Incapacitate"
L["DR_Knockback"] = "Knockback"
L["DR_Root"] = "Root"
L["DR_Silence"] = "Silence"
L["DR_Stun"] = "Stun"
L["DRContainer"] = "DR container"
L["DRTracking"] = "DR Tracking"
L["DrTracking_Container_Color_Desc"] = "The border color of the container of the DR icons"
L["DrTracking_DisplayType_Desc"] = "Here you can choose if the status of the DR should be indicated by a colored frame around the icon or by the textcolor of the countdown text. Note that the Countdown text setting will only work if you aren't using an addon like OmniCC that modifys the textcolor."
L["DrTracking_Enabled"] = "Enable DR Tracking"
L["DrTracking_Enabled_Desc"] = "When enabled there will be icons for DR Tracking next to the enemy row. Green Border: Next CC will be half duration. Yellow Border: Next CC will be quarter duration. Red Border: Next CC will not apply because the player is immune"
L["DrTracking_Spacing"] = "DR Spacing"
L["DrTracking_Spacing_Desc"] = "The spacing between each DR icon"
L["DrTrackingFiltering_Enabled_Desc"] = "When this option is enabled you will only see icons of spells whose category is checked on the filtering list."
L["DrTrackingFiltering_Filterlist_Desc"] = "Click to track/untrack this category"
L["DrTrackingFilteringSettings_Desc"] = "Here you can choose which DR category is getting shown."
L["DrTrackingSettings"] = "DR Tracking"
L["DrTrackingSettings_Desc"] = "Settings regarding DR Tracking"
L["DurationFilter"] = "Filter by duration"
L["DurationFilter_OnlyShowWhenDuration"] = "Only show when duration is smaller or equal than"
L["DurationFilter_OnlyShowWhenDuration_Desc"] = "Any auras whose duration is longer than the selected duration won't make it through the filter"
L["Editmode_Toggle"] = "Toggle edit mode"
L["Editmode_Toggle_Desc"] = "Enables/Disables the edit mode. This mode mode is basically the testmode but it enables you to move all of the frames around. Once you selected a frame the options panel will automatically jump to the corresponding settings. You can fine tune offsets or jump to the general settings for that particular frame there too. "
L["EditmodeDisabled"] = "Editmode disabled"
L[ [=[EditmodeEnabled
]=] ] = "Editmode enabled"
L["EditModeIntroduction"] = "Welcome to edit mode. This function makes it easy for you to arrange the individual modules the way you want. To do this, simply click on one of the blue modules to select it, then you can simply move it to the desired position. If you push it near another module, it anchors itself to it. Once you select a module, it will be highlighted in yellow and the options menu will jump to its settings."
L["Enable_DrawSwipe"] = "Enable swipe texture"
L["Enable_DrawSwipe_Desc"] = "Show a desaturated, darker texture for the remaining part of the cooldown"
L["EnableClique"] = "Use Clique for keybindings"
L["EnableClique_Desc"] = "When this option is enabled Clique will handle all of the keybindings."
L["EnableCustomDRCategoryIcons"] = "Enable custom category icons"
L["EnableCustomDRCategoryIcons_Desc"] = "when enabled you can choose your own icon for each of the diminishing return categories"
L["EnableCustomPlayerCountProfiles"] = "Enable custom profiles for player count"
L["EnableCustomPlayerCountProfiles_Desc"] = "When this option is enabled you will be able to create your own profiles based on player size. The current used profiles for 1–5, 6–15 and 16-40 players won't be lost but disabled and you can use them as a baseline to create your own profiles. You can switch back whenever you want without losing any settings."
L["EnableInArenas"] = "Enable in Arenas"
L["EnableInBattlegrounds"] = "Enable in Battlegrounds"
L["enemies"] = "enemies"
L["Enemies"] = "Enemies"
L["EnemiesTargetingAllies_Enabled_Desc"] = "When you enable this option you will be notified by a sound when a set amount of enemys are targeting an ally"
L["EnemiesTargetingMe_Enabled_Desc"] = "When you enable this option you will be notified by a sound when a set amount of enemys are targeting you "
L["enemy"] = "enemy"
L["Enemy"] = "Enemy"
L["EnemyJoined"] = "An enemy joined the battleground"
L["EnemyLeft"] = "An enemy left the battleground"
L["ErrorTestmodeInCombat"] = "Can't enable testmode in combat"
L["ExportButton"] = "Export profile"
L["ExportButton_Desc"] = "Click here to export your current active profile"
L["FakeEventsDisabled"] = "No longer creating fake events."
L["FakeEventsEnabled"] = "Now creating fake events for testing."
L["Filtering_Enabled"] = "Enable filtering"
L["Filtering_Filterlist"] = "Filter list"
L["FilterSettings"] = "Filter settings"
L["Font"] = "Font"
L["Font_Desc"] = "The main font used by this addon"
L["Font_Outline"] = "Font outline"
L["Font_Outline_Desc"] = "The font's outline"
L["Fontcolor"] = "Font color"
L["Fontcolor_Desc"] = "The color of the text"
L["FontShadow_Enabled"] = "Enable Font shadow"
L["FontShadow_Enabled_Desc"] = "When enabled there will be a shadow around the font. You can choose the color of that shadow"
L["FontShadowColor"] = "Shadow color"
L["FontShadowColor_Desc"] = "The color of the text shadow"
L["FontSize"] = "Font size"
L["FontSize_Desc"] = "The font size of the text"
L["Frame"] = "Frame"
L["Framescale"] = "Scale"
L["Framescale_Desc"] = "The scale of the mainframe. This setting is not available in combat due to Blizzard's combat protection."
L["General"] = "General"
L["GeneralSettings"] = "General settings"
L["GeneralSettings_Desc"] = "Some general settings"
L["GeneralSettingsAllies"] = "This settings apply to allies regardless of the size of the battleground"
L["GeneralSettingsEnemies"] = "This settings apply to enemies regardless of the size of the battleground"
L["HealthBar"] = "Healthbar"
L["HealthBar_Background_Desc"] = "Background Color for the health bar."
L["HealthBar_Texture_Desc"] = "The texture used for the health bars"
L["HealthBarSettings_Desc"] = "Settings for the health bar."
L["HealthTextEnabled"] = "Enable health text"
L["HealthTextSettings"] = "Health text settings"
L["HealthTextType"] = "Health text type"
L["Height"] = "Height"
L["HideArenaframesIn"] = "Hide arena frames in"
L["HideHighestPriority"] = "Hide highest priority"
L["HideHighestPriority_Desc"] = "When enabled, it removes the aura with the highest priority, useful when the highest priority is already shown on the spec or class icon."
L["HideRaidframesIn"] = "hide raid frames in"
L["HideWhenZero"] = "Hide when zero"
L["HighestPriorityAura"] = "Highest priority aura"
L["Highlight_Color"] = "Highlight Color"
L["Highlight_Color_Desc"] = "The color of the player frame your cursor is currently on"
L["HorizontalGrowdirection"] = "Horizontal grow direction"
L["HorizontalGrowdirection_Desc"] = "Lets you choose if the rows grow to the left or to the right side"
L["HorizontalSpacing"] = "Horizontal spacing"
L["IAmTargeted"] = "I am targeted"
L["Icon"] = "Icon"
L["Icons"] = "Icons"
L["IconSize"] = "Icon size"
L["IconsPerRow"] = "Icons per row"
L["Import"] = "Import"
L["ImportButton"] = "Import profile"
L["ImportButton_Desc"] = "Click here to import a profile from a string and overwrite your current active profile"
L["JumpToGeneralOptions"] = "Jump to general options"
L["JumpToGeneralOptions_Desc"] = "Click here to jump to the general settings of this module"
L["JustifyH"] = "Horizontal alignment"
L["JustifyH_Desc"] = "The horizontal alignment of the text"
L["JustifyV"] = "Vertical alignment"
L["JustifyV_Desc"] = "The vertical alignment of the text"
L["KeybindSettings_Desc"] = "Here you can configure what should happen on left- right- and middle click"
L["LEFT"] = "Left"
L["LeftToTargetCounter"] = "Left to target counter"
L["Leftwards"] = "Leftwards"
L["Level"] = "Level"
L["LevelText_Enabled"] = "Show player's level"
L["LevelText_OnlyShowIfNotMaxLevel"] = "Only show if not max level"
L["Locked"] = "Lock"
L["Locked_Desc"] = "Locks the frame in place"
L["Magic"] = "Magic"
L["MainFrameSettings"] = "Main frame settings"
L["MainFrameSettings_Desc"] = "Settings for the main frame which is used for %s"
L["MaxCantBeUndefined"] = "maximun can't be undefined"
L["MaxNeedsToBeEqualOrGreater"] = "max needs to be equal or greater than min"
L["MaxPlayerCount"] = "Maximum player count"
L["MaxPlayers"] = "Maximum Enemys"
L["MaxPlayers_Desc"] = "The addon won't show enemys if there are more enemys than this number"
L["MIDDLE"] = "Middle"
L["MinCantBeUndefined"] = "minimum can't be undefined"
L["MinPlayerCount"] = "Minimum player count"
L["Mirrored"] = "Mirrored"
L["Modules"] = "Modules"
L["ModuleSettings"] = "Module settings"
L["ModuleSpecificSettings"] = "Module specific settings"
L["MoreProfileOptions"] = "More Profile Options"
L["MouseWheelPlayerTargeting"] = "Scroll targeting"
L["MouseWheelPlayerTargeting_Desc"] = "When this option is enabled you will be able to move your mouse over the enemy or ally frames and scroll to target the next or previous player in the list. Please be aware that this can interfer when you are using mouseover casting."
L["MyFocus"] = "My Focus"
L["MyFocus_Color"] = "Focus Color"
L["MyFocus_Color_Desc"] = "The color used for the border indicating your current focus"
L["MyTarget"] = "My target"
L["MyTarget_Color"] = "Target Color"
L["MyTarget_Color_Desc"] = "The color used for the border around the health bar indicating your current target"
L["MyVersion"] = "You are using Version"
L["Name"] = "Name"
L["Name_Desc"] = "Settings for names on the health bar."
L["NewVersionAvailable"] = "A new version is available"
L["None"] = "None"
L["NonPriorityBuffs"] = "Non prioritized buffs"
L["NonPriorityDebuffs"] = "Non prioritized debuffs"
L["Normal"] = "Normal"
L["NormalCastingBar"] = "Normal Casting bar"
L["NotAvailableInCombat"] = [=[
This setting is not available while you are in combat due to Blizzard's combat restrictions.]=]
L["Notifications_Allies_Enabled_Desc"] = "When enabled you will be notified when an ally is leaving or joining the battleground."
L["Notifications_Enabled"] = "Enable Notifications"
L["Notifications_Enemies_Enabled_Desc"] = "When enabled you will be notified when an enemy is leaving or joining the battleground."
L["NoVersion"] = "No addon found"
L["NumericTargetindicator_Enabled_Desc"] = "Shows how many %s are targeting a player."
L["ObjectiveAndRespawn_ObjectiveEnabled"] = "Show Objectives"
L["ObjectiveAndRespawn_ObjectiveEnabled_Desc"] = "Shows in icon for flag, minecart and orb-carriers next to the enemy button"
L["ObjectiveAndRespawn_Position"] = "Position"
L["ObjectiveAndRespawn_Position_Desc"] = "If the icon should be shown left to the spec icon or right to the trinket/racial icons."
L["ObjectiveAndRespawn_RespawnEnabled"] = "Enable Respawn Timer"
L["ObjectiveAndRespawn_RespawnEnabled_Desc"] = "When enabled there will be an icon indicating the remaining time until the enemy is alive."
L["ObjectiveAndRespawn_Width_Desc"] = "Width of the Objective Icon"
L["ObjectiveAndRespawnSettings_Desc"] = "Settings regarding the objective of the battleground"
L["ObjectiveAndRespawnTimer"] = "Objective and respawn timer"
L["OffsetX"] = "Offset X"
L["OffsetY"] = "Offset Y"
L["OldVersion"] = "Old version"
L["OnlyShowForHealers"] = "Only show for healers"
L["OutOfCombat"] = "Out of combat"
L["OverwriteMirroredConfirm"] = "Are you sure you want to overwrite the settings for %s with the mirrored settings of %s"
L["OverwriteNormalConfirm"] = "Are you sure you want to overwrite the settings for %s with the settings of %s?"
L["PlayerCount"] = "Player Count"
L["PlayerCount_Enabled_Desc"] = "If there should a text for the current number of players. Note that this count can be different than the current number of rows shown due to combat protection or leaving allies/enemies in a RBG"
L["players"] = "players"
L["Point"] = "Point"
L["PointAtObject"] = "Point at object"
L["Poison"] = "Poison"
L["Position"] = "Position"
L["PowerBar"] = "Powerbar"
L["PowerBar_Background_Desc"] = "Background Color for the power bar."
L["PowerBar_Enabled"] = "Enable ressource bars"
L["PowerBar_Enabled_Desc"] = "Enable this option if you wan't to see bars for ressources like mana, rage etc."
L["PowerBar_Height_Desc"] = "The heigt of the ressource bar. Making the ressource bar higher will automatically make the health bar shorter"
L["PowerBar_OnlyShowForHealers_Desc"] = "When enabled, it will only show the powerbar for healers."
L["PowerBar_Texture_Desc"] = "The texture used for the power bars"
L["PowerBarSettings_Desc"] = "Settings for the ressource bar"
L["PriorityBuffs"] = "Prioritized buffs"
L["PriorityDebuffs"] = "Prioritized debuffs"
L["Racial"] = "Racial"
L["Racial_Enabled"] = "Enable Racials"
L["Racial_Enabled_Desc"] = "When enabled there will be icons for used racials"
L["Racial_Width_Desc"] = "The width of the racial."
L["RacialFiltering_Enabled_Desc"] = "When this option is enabled you will only see racials which are checked on the filtering list."
L["RacialFiltering_Filterlist_Desc"] = "Click to track/untrack this racial"
L["RacialFilteringSettings_Desc"] = "Here you can select which racial you want to track"
L["RacialSettings"] = "Racials"
L["RacialSettings_Desc"] = "Settings regarding Racials"
L["RangeIndicator_Alpha"] = "Alpha"
L["RangeIndicator_Alpha_Desc"] = "The alpha of the enemy frame when the enemy is out of the chosen range"
L["RangeIndicator_Enabled"] = "Enable range indicator"
L["RangeIndicator_Enabled_Desc"] = "When tis option is enabled the frame will use the chosen alpha when an enemy is out of a chosen range"
L["RangeIndicator_Everything"] = "Modify the alpha of all elements"
L["RangeIndicator_Frames"] = "Modify alpha of"
L["RangeIndicator_Frames_Desc"] = "Set which parts of the enemy row should be shown with a higher transparency when the enemy is out of range."
L["RangeIndicator_Range"] = "Range"
L["RangeIndicator_Range_Desc"] = "When an enemy is more than this distance (in yards) away his frame will use the chosen alpha"
L["RangeIndicator_Settings"] = "Range indicator"
L["RangeIndicator_Settings_Desc"] = "Here you can configure the range indicator."
L["RangeOverlapping"] = "range overlapping with: %s"
L["RBGSpecificSettings"] = "Rated Battleground"
L["RBGSpecificSettings_Desc"] = "This settings only apply in a RBG (Rated Battleground)"
L["ResetGeneralModule_Desc"] = "click here to reset these general settings for this moule"
L["ResetModule"] = "Reset Module"
L["ResetModule_Desc"] = "Click here to reset the settings of this module for %s:%s"
L["ResetModuleSettings"] = "Reset the settings of this module"
L["RestoreDefault"] = "Restore default"
L["RIGHT"] = "Right"
L["Rightwards"] = "Rightwards"
L["RoleIcon_Enabled"] = "Enable role icons"
L["RoleIcon_Enabled_Desc"] = "Enable this option if you want to see a role icon on the health bar"
L["RoleIcon_Size_Desc"] = "The size of the role icons (width and height)"
L["RoleIconSettings"] = "Role icon"
L["RoleIconSettings_Desc"] = "Settings for the role icon which is shown on the health bar"
L["RoleSortingOrder"] = "role sorting order"
L["RoleSortingOrder_Desc"] = "Select in which order the players are placed according to their role"
L["Scale"] = "Scale"
L["ShowDispellable"] = "Show dispellable"
L["ShowDispellable_Desc"] = [=[Show %s that i can remove by decursing, purging, etc.

]=]
L["ShowFramesIn"] = "Show frames in"
L["ShowHighestPriority"] = "Show highest priority"
L["ShowHighestPriority_Desc"] = "When enabled the addon will show an if the player got interruped or is affected by a priority aura"
L["ShowMine"] = "Show mine"
L["ShowMine_Desc"] = "Show %s that i applied"
L["ShowNumbers"] = "Show numbers"
L["ShowNumbers_Desc"] = "Show cooldown numbers, note that this option might not be working when you are using addons like OmniCC"
L["ShowRealmnames"] = "Show Realmnames"
L["ShowRealmnames_Desc"] = "Shows the enemy's realm"
L["ShowSpecIfExists"] = "Show spec if exists"
L["ShowSpecIfExists_Desc"] = "When this option is enabled it will show the spec icon instead of the class icon"
L["ShowTooltips"] = "Show tooltips"
L["ShowTooltips_Desc"] = "Show tooltips when you move the mouse over objects like trinket and aura icons"
L["Side"] = "Side"
L["SideAtObject"] = "Side at object"
L["Size"] = "Size"
L["SmallCastingBar"] = "Small casting bar"
L["SourceFilter"] = "Filter by Source"
L["Spec_AuraDisplay_Enabled"] = "Enable Aura display"
L["Spec_AuraDisplay_Enabled_Desc"] = "When this option is enabled you will see icons for crowd control and interrupts (when active) instead of the spec icon"
L["Spec_Enabled"] = "Enable specialization"
L["Spec_Enabled_Desc"] = "When enabled there will be icons for the players's specialization"
L["Spec_Width_Desc"] = "The Width of the spec icon."
L["SpecClassPriorityOne"] = "SpecClassPriority1"
L["SpecClassPriorityTwo"] = "SpecClassPriority2"
L["SpecSettings"] = "Spec"
L["SpecSettings_Desc"] = "Settings changing the spec icon."
L["SpellID_Filtering"] = "Filter by spellID"
L["Style"] = "Style"
L["SymbolicTargetindicator_Enabled_Desc"] = "Shows a square class-colored icon for each %s that is targeting a player."
L["TargetAmount"] = "target amount"
L["TargetAmount_Ally"] = "The sound is played as soon as a group member is targeted by more than the set number of opponents."
L["TargetAmount_Me"] = "The sound is played as soon as a you are targeted by more than the set number of opponents."
L["TargetCallerUpdated"] = "is now the target caller"
L["TargetCalling"] = "Target Calling"
L["TargetCallingNotificationEnable"] = "Enable Sound Notification"
L["TargetCallingNotificationEnable_Desc"] = "Notifies you when there is a new target (detected via the raid target skull icon)"
L["TargetCallingSetMark"] = "Mark my target with skull icon"
L["TargetCallingSetMark_Desc"] = "Marks your current target automatically with the skull icon when you are the raid leader and there is no raid assistent in your group. It will also mark your curent target if you are not the raid leader but you are a assistant."
L["TargetCallingShowIcon"] = "Show icon on target caller's target"
L["TargetCallingShowIcon_Desc"] = "When enabled, you will see a icon on the frame which shows the target of the target caller"
L["TargetCallingSoundEnable"] = "Notify me on new target"
L["TargetCallingSoundEnable_Desc"] = "When enabled, there will be a sound when the target caller targets a new player"
L["TargetCallingSoundSound"] = "Notification sound"
L["TargetIndicator"] = "Target Indicator"
L["TargetIndicator_Desc"] = "Numeric and symbolic Target Indicators"
L["TargetIndicatorNumeric"] = "Target Indicator Numeric"
L["TargetIndicatorNumeric_HideWhenZero_Desc"] = "When enabled, it will hide the text if no enemy is targeting that player."
L["TargetIndicatorSymbolic"] = "Target Indicator Symbolic"
L["TargetNextAlly"] = "Target next ally"
L["TargetNextEnemy"] = "Target next enemy"
L["TargetPreviousAlly"] = "Target previous ally"
L["TargetPreviousEnemy"] = "Target previous enemy"
L["Testmode_Toggle"] = "Toggle test mode"
L["Testmode_Toggle_Desc"] = [=[Enables/Disables the testing mode. This mode makes it easier to see what impact most of the options have. Its not possible to simulate all options but most of them are covered in this fancy test mode. 
The test mode is disabled if you are currently in a battleground.]=]
L["Testmode_ToggleAnimation"] = "Toggle test mode animation"
L["Testmode_ToggleAnimation_Desc"] = "Enables/Disables the animation of the testmode so you can focus more on a specific setting without getting distracted by the animation."
L["Testmode_UseTeammates"] = "Use group members for testing"
L["Testmode_UseTeammates_Desc"] = "When enabled the testmode will use your current group members instead of fake ones"
L["TestmodeDisabled"] = "Testmode disabled"
L["TestmodeEnabled"] = "Testmode enabled"
L["TestmodeSettings"] = "Test mode"
L["Text"] = "Text"
L["TextSettings"] = "Text settings"
L["Thick"] = "Thick"
L["TOP"] = "Top"
L["TOPLEFT"] = "Top left"
L["TOPRIGHT"] = "Top right"
L["Trinket"] = "Trinket"
L["Trinket_Enabled"] = "Enable Trinkets"
L["Trinket_Enabled_Desc"] = "When enabled there will be icons for Gladiator's Medallion, Adaptation and Relentless."
L["Trinket_Width_Desc"] = "The width of the trinket."
L["TrinketSettings_Desc"] = "Settings regarding Trinkets"
L["UpdatePeriod"] = "Update period"
L["UpdatePeriod_Desc"] = "The time between combat checks in seconds, dont put this too low because this can potentially cause performance issues."
L["Upwards"] = "Upwards"
L["UseBarHeight"] = "Use bar height"
L["UseBigDebuffsPriority"] = "Use BigDebuffs priorities"
L["UseBigDebuffsPriority_Desc"] = "If you enable this option, the addon will use the priorities from the addon BigDebuffs instead. The priorities are used for the modules %s, %s and %s "
L["UseButtonHeight"] = "Use button height"
L["UseButtonHeight_Desc"] = "If you enable this option, the module will use the buttons height. "
L["UserDefined"] = "User-defined"
L["VerticalGrowdirection"] = "Vertical grow direction"
L["VerticalGrowdirection_Desc"] = "Lets you choose if the bars grow upwards or downwards"
L["VerticalPosition"] = "Vertical position"
L["VerticalSpacing"] = "Vertical spacing"
L["Width"] = "Width"


L = Data.L;

if LOCALE_deDE then
L["AddPoint"] = "Punkt hinzufügen"
L["Allies"] = "Verbündete"
L["allies"] = "Verbündete"
L["ally"] = "Verbündeten"
L["Ally"] = "Verbündeter"
L["AllyIsTargeted"] = "Verbündeter wird anvisiert"
L["AllyJoined"] = "Ein Verbündeter hat das Schlachtfeld betreten"
L["AllyLeft"] = "Ein Verbündeter hat das Schlachtfeld verlassen"
L["AND"] = "und"
L["AttachToObject"] = "Anhängen an Objekt"
L["AttachToObject_Desc"] = "Beachte, dass nicht alle Objekte zur Befestigung sinnvoll bzw machbar sind. Wenn du beispielsweise versuchst die PvP-Talente an den Volksfähigkeiten befestigen möchtest und die Volksfähigkeiten an den PvP-Talenten befestigt sind, wird es eine Fehlermeldung geben."
L["Auras_Buffs_Container_Color_Desc"] = "Die Rahmenfarbe des Containers für die Stärkungszaubersymbole"
L["Auras_CustomFiltering_Conditions_All"] = "Alle"
L["Auras_CustomFiltering_Conditions_Any"] = "Irgendeine"
L["Auras_CustomFiltering_ConditionsMode"] = "Bedingungsmodus"
L["Auras_CustomFiltering_ConditionsMode_Desc"] = [=[Hier kannst du auswählen wie die ausgewählten Bedingungen miteinander verknüpft werden sollen. 
Wenn du "Irgendeine" auswählst, wird die Aura angezeigt sobald eine deiner ausgewählten Bedingungen erfüllt ist.
Wenn du "Alle" auswählst, wird die Aura nur angezeigt sobald alle ausgewählten Bedingungen von der Aura erfüllt werden.]=]
L["Auras_Debuffs_Container_Color_Desc"] = "Die Rahmenfarbe des Containers für die Schwächungszaubersymbole"
L["Auras_Enabled"] = "Auren aktivieren"
L["Auras_Enabled_Desc"] = "Falls diese Option aktiviert ist, werden die Auren der Gegner angezeigt. Jede Aura zeigt die verbleibende Zeit und die momentane Anzahl der Stapel an."
L["Auras_Filtering_Mode"] = "Filtermodus"
L["Auras_Filtering_Mode_Desc"] = [=[Hier kannst du auswählen welcher Filter auf die Auren angewandt wird. 
Du kannst zwischen einem Blizzard-Filter (der Filter der für die Schlachtzugsfenster verwendet wird) und benutzerdefinierten Bedingungen wählen.]=]
L["Auras_ShowTooltips"] = "Zaubertooltips anzeigen"
L["AurasCustomConditions"] = "Benutzerdefinierte Bedingungen"
L["AurasFiltering_AddSpellID"] = "Zauber-ID"
L["AurasFiltering_AddSpellID_Desc"] = "Gib eine Zauber oder mehrere per Komma getrennte Zauber-IDs ein. Diese Zauber-IDs werden zu deiner Filterliste hinzugefügt. Nur Debuffs, deren Zauber-ID auf der Filterliste ist, werden angezeigt, wenn das Filtern aktiviert ist."
L["AurasFiltering_Enabled_Desc"] = "Wenn diese Option aktiviert ist, wirst du nur die %s sehen, die auf der Filterliste stehen."
L["AurasFiltering_Filterlist_Desc"] = "Klicken, um diesen %s zu entfernen"
L["AurasFilteringSettings_Desc"] = "Hier kannst du genauer einstellen, welche Auren angezeigt werden."
L["AurasSettings"] = "Auren"
L["AurasSettings_Desc"] = "Einstellungen für Auren (Stärkungs- und Schwächungszauber)"
L["AurasStacktextSettings"] = "Stapeltext"
L["BarBackground"] = "Hintergrundfarbe"
L["BarHeight_Desc"] = "Die Höher jeder Reihe. Diese Einstellung ist aufgrund Einschränkungen seitens Blizzards nicht im Kampf verfügbar."
L["BarSettings"] = "Balkeneinstellungen"
L["BarSettings_Desc"] = "Hier kannst du die einzelnen Reihen konfigurieren"
L["BarTexture"] = "Balkentextur"
L["BarWidth_Desc"] = "Die Breite jeder Reihe."
L["BattlegroundSize"] = "Schlachtfeldgröße"
L["BGSize_15"] = "1–15 Spieler"
L["BGSize_15_Desc"] = "Diese Einstellungen sind nur für %s in einem Schlachtfeld mit 1–15 Spielern pro Seite"
L["BGSize_40"] = "16–40 Spieler"
L["BGSize_40_Desc"] = "Diese Einstellungen sind nur für %s in einem Schlachtfeld mit 16–40 Spielern pro Seite"
L["BGSize_5"] = "Arena"
L["BGSize_5_Desc"] = "Diese Einstellungen sind nur für %s in Arenen aktiv."
L["BlizzlikeAuraFiltering"] = "Blizzards Schlachtfenster-Filter"
L["BorderSize"] = "Rahmengröße"
L["BorderThickness"] = "Rahmengröße"
L["BOTTOM"] = "Unten"
L["BOTTOMLEFT"] = "Unten links"
L["BOTTOMRIGHT"] = "Unten rechts"
L["buff"] = "Stärkungszauber"
L["BuffContainer"] = "Stärkungszauber-Container"
L["BuffIcon"] = "Stärkungszaubersymbol"
L["Buffs"] = "Stärkungszauber"
L["Button"] = "Button"
L["CantAnchorToItself"] = "Kann nicht an sich selbst verankert werden!"
L["CENTER"] = "MItte"
L["Change"] = "Ändern"
L["Color"] = "Farbe"
L["Coloring_Enabled"] = "Nach Schwächungszaubertyp färben"
L["Coloring_Enabled_Desc"] = "Wenn diese Option aktiviert ist, kannst du auswählen ob der Rahmen oder der Countdown-Text nach Schwächungszaubertyp gefärbt werden soll. (So wie es auch das Standard-UI macht; Gifte sind grün, Magie blau etc.)"
L["Columns"] = "Spalten"
L["Columns_Desc"] = "In wieviele Spalten die Spieler unterteilt werden."
L["Combat"] = "Im Kampf"
L["CombatIndicator"] = "Kampfindikator"
L["CombatLogScanningForEnemiesDisabled"] = "Das Durchsuchen des Kampflogs nach Gegnern wurde deaktiviert."
L["CombatLogScanningForEnemiesEnabled"] = "Das Durchsuchen des Kampflogs nach Gegnern wurde aktiviert."
L["ConfirmDeletePlayerCountProfile"] = "Bist du sicher, dass du das Profil löschen möchtest das momentan verwendet wird für"
L["ConfirmProfileOverride"] = "Bist du sicher, dass du das Unterprofil %s mit dem Unterprofil %s überschreiben willst?"
L["Container_Color"] = "Container-Rahmenfarbe"
L["ContainerPosition"] = "Containerposition"
L["ContainerSettings"] = "Container-Einstellungen"
L["ConvertCyrillic"] = "Kyrillisch umwandeln"
L["ConvertCyrillic_Desc"] = "Wandelt kyrillische Zeichen in lateinische Zeichen um. Dies macht es einfacher die Namen russischer Gegner auszusprechen."
L["Cooldown"] = "Abklingzeit"
L["CopyFrom"] = "Kopieren von"
L["CopySettings"] = "Kopiere Einstellungen von"
L["CopySettingsMirrored_Desc"] = "Gespiegelt bedeutet, dass die gesamte Position und Ausdehnungsrichtung der Module durch den horizontal entgegengesetzten Punkt oder die entgegengesetzte Richtung ersetzt wird. (Zum Beispiel wird der Punkt „TOPLEFT“ zu „TOPRIGHT“ und Containermodule wie das Aura-Modul wachsen in die entgegengesetzte Richtung.)"
L["CopySettingsNormal_Desc"] = "Eine einfache Kopie"
L["Countdowntext"] = "Countdown-Text"
L["Covenant"] = "Pakt"
L["CovenantIcon_Enabled"] = "Paktsymbole aktivieren"
L["CovenantIcon_Enabled_Desc"] = "Aktiviere diese Option, falls du ein Paktsymbol auf dem Gesundheitsbalken haben willst."
L["CovenantIcon_Size_Desc"] = "Die Größe der Paktsymbole (Breite und Höhe)"
L["CovenantIconSettings_Desc"] = "Einstellungen für das Paktsymbol, welches auf dem Gesundheitsbalken angezeigt wird."
L["CurrentVersion"] = "Aktuelle Version"
L["Curse"] = "Fluch"
L["CustomMacro_Desc"] = [=[Hier kannst du ein benutzerdefiniertes Makro festlegen. Ein leeres Feld führt zu einem leeren Makro.
Verfügbare Ersetzungen: %n wird mit dem Gegnernamen ersetzt

Beispiel:
/targetexact %n
/cast Verwandlung
/say Verwandlung auf %n
/targetlasttarget

Dies nimmt den Gegner ins Ziel, wirkt Verwandlung auf ihn, schreibt eine Nachricht in den Sagen-Kanal und nimmt dein vorheriges Ziel ins Ziel.
Beachte, dass Makros nur 255 Zeichen lang sein dürfen (inklusive Ersetzungen).]=]
L["Data"] = "Daten"
L["debuff"] = "Schwächungszauber"
L["DebuffContainer"] = "Schwächungszauber-Container"
L["DebuffIcon"] = "Schwächungszaubersymbol"
L["Debuffs"] = "Schwächungszauber"
L["DebuffType_Filtering"] = "Nach Schwächungszaubertyp filtern"
L["DebuffType_Filtering_Desc"] = "Wenn diese Option aktiviert ist werden nur Schwächungszauber des ausgewählten Typs dargestellt. Diese Option wird im Testmodus nicht simuliert, da diese eine große Datenbank mit allen Zaubern und ihren Schwächungszaubertypen benötigen würde."
L["DeletePlayerCountProfile"] = "Löschen"
L["DeletePlayerCountProfile_Desc"] = "Klicke hier um das Profil zu löschen das momentan verwendet wird für"
L["DeletePoint"] = "Punkt %s löschen"
L["DisableArenaFramesInArena"] = "Arenafenster in Arena deaktivieren"
L["DisableArenaFramesInArena_Desc"] = "Deaktiviert die Arenafenster in einer Arena. Dies funktioniert ebenfalls wenn das Addon sArena installiert ist."
L["DisableArenaFramesInBattleground"] = "Arenafenster im Schlachtfeld deaktivieren"
L["DisableArenaFramesInBattleground_Desc"] = "Deaktiviert die Arenafenster in einem Schlachtfeld. Dies funktioniert ebenfalls wenn das Addon sArena installiert ist."
L["Disease"] = "Krankheit"
L["DispellFilter"] = "Nach Bannbarkeit filtern"
L["DisplayType"] = "Anzeigeart"
L["Downwards"] = "Nach unten"
L["DR_Disorient"] = "Desorientierung"
L["DR_Incapacitate"] = "Handlungsunfähigkeit"
L["DR_Knockback"] = "Rückstöße"
L["DR_Root"] = "Bewegungsunfähigkeit"
L["DR_Silence"] = "Stille"
L["DR_Stun"] = "Betäubungen"
L["DRContainer"] = "DR-Container"
L["DRTracking"] = "DR-Tracking"
L["DrTracking_Container_Color_Desc"] = "Die Rahmenfarbe des Containers für die DR-Symbole"
L["DrTracking_DisplayType_Desc"] = "Hier kannst du auswählen wie der DR-Status angezeigt werden soll. Er wird entweder mit einem farbigen Rahmen um das Symbol dargestellt oder mit der Textfarbe des Countdown-Textes. Beachte, dass die Einstellung Countdown-Text nur funktioniert, wenn du kein Addon installiert hast, das die Countdown-Texte verändert, wie z.B. OmniCC."
L["DrTracking_Enabled"] = "DR-Tracking aktivieren"
L["DrTracking_Enabled_Desc"] = "Wenn diese Option aktiviert ist, erscheinen Symbole neben den Balken der Gegner. Erklärung der Rahmenfarben der Symbole: Die Farben stehen für die Dauer des nächsten Kontrolleffekts auf dem Gegner. Grün: halbe Dauer. Gelb: viertel Dauer. Rot: keine Dauer (Gegner ist immun)"
L["DrTracking_Spacing"] = "DR-Abstand"
L["DrTracking_Spacing_Desc"] = "Der Abstand der einzelnen DR-Symbole"
L["DrTrackingFiltering_Enabled_Desc"] = "Wenn diese Option aktiviert ist, wirst du nur die angehakten Kategorien sehen."
L["DrTrackingFiltering_Filterlist_Desc"] = "Klicken, um die Verfolgung dieser Kategorie zu aktivieren/deaktivieren."
L["DrTrackingFilteringSettings_Desc"] = "Hier kannst du einstellen, welche DR-Kategorien angezeigt werden sollen."
L["DrTrackingSettings"] = "DR-Tracking"
L["DrTrackingSettings_Desc"] = "Einstellungen für den DR-Tracker"
L["DurationFilter"] = "Nach Dauer filtern"
L["DurationFilter_OnlyShowWhenDuration"] = "Dauer kleiner als"
L["DurationFilter_OnlyShowWhenDuration_Desc"] = "Jede Aura, deren Dauer länger ist als die ausgewählte Dauer wird nicht angezeigt"
L["Editmode_Toggle"] = "Bearbeitungsmodus ein-/ausschalten"
L["Editmode_Toggle_Desc"] = "Aktiviert/Deaktiviert den Bearbeitungsmodus. Dieser Modus ist fast identisch mit dem Testmodus aber er ermöglicht es einzele Elemente zu verschieben. Sobald du ein Element ausgewählt hast springt das Menü zu den dazugehörigen Einstellungen. Dort kannst du zum Beispiel den Abstand feiner Einstellungen oder zu den generellen Optionen dieses Moduls springen."
L["EditmodeDisabled"] = "Bearbeitungsmodus deaktiviert"
L[ [=[EditmodeEnabled
]=] ] = "Bearbeitungsmodus aktiviert"
L["EditModeIntroduction"] = "Willkommen zum Bearbeitungsmodus. Diese Funktion macht es dir einfach die einzelnen Module so anzuordnen wie du möchtest. Klicke dazu einfach auf eines der blauen Module um es auszuwählen, danach kannst du es einfach an die gewünschte Position verschieben. Wenn du es in die Nähe eines anderen Modules schiebst verankert es sich an diesem. Sobald du ein Modul ausgewählt hast, wird es in gelb hervorgehoben und das Optionsmenü springt zu den dazugehörigen Einstellungen. "
L["Enable_DrawSwipe"] = "Abklingtextur anzeigen"
L["Enable_DrawSwipe_Desc"] = "Zeigt eine dunklere Textur für den Bereich der bereits vergangenen Zeit."
L["EnableClique"] = "Clique für Tastenbelegungen verwenden"
L["EnableClique_Desc"] = "Wenn du diese Option aktivierst verwaltet Clique die Tastenbelegungen."
L["EnableCustomDRCategoryIcons"] = "Eigene Symbole für Kategorien"
L["EnableCustomDRCategoryIcons_Desc"] = "Aktiviere diese Option um deine eigenen Symbole für die einzelnen DR-Kategorien festzulegen"
L["EnableCustomPlayerCountProfiles"] = "Benutzerdefinierte Subprofile für Spielerzahl aktivieren"
L["EnableCustomPlayerCountProfiles_Desc"] = "Wenn du diese Option aktivierst kannst du deine eigenen Profile basierend auf der Spielerzahl erstellen. Die momentan verwendeten Profile für 1–5, 6–15 and 16-40 Spieler gehen nicht verloren, werden aber deaktiviert und du kannst sie als Basis für deine eigenen Profile verwenden. Du kannst diese Option jederzeit deaktivieren ohne Einstellungen zu verlieren."
L["EnableInArenas"] = "In Arenen aktivieren"
L["EnableInBattlegrounds"] = "In Schlachtfeldern aktivieren"
L["enemies"] = "Gegner"
L["Enemies"] = "Gegner"
L["EnemiesTargetingAllies_Enabled_Desc"] = "Wenn diese Option aktiviert ist, wirst du durch einen Ton benachrichtigt, wenn eine bestimmte Anzahl von Feinden einen Verbündeten anvisiert"
L["EnemiesTargetingMe_Enabled_Desc"] = "Wenn diese Option aktiviert ist, wirst du durch einen Ton benachrichtigt, wenn eine bestimmte Anzahl von Feinden dich anvisiert"
L["enemy"] = "Gegner"
L["Enemy"] = "Gegner"
L["EnemyJoined"] = "Ein Gegner hat das Schlachtfeld betreten"
L["EnemyLeft"] = "Ein Gegner hat das Schlachtfeld verlassen"
L["ErrorTestmodeInCombat"] = "Testmodus kann im Kampf nicht aktiviert werden"
L["ExportButton"] = "Profil exportieren"
L["ExportButton_Desc"] = "Klicke hier, um dein momentan aktives Profil zu exportieren"
L["FakeEventsDisabled"] = "Erstellt nicht länger künstliche Ereignisse."
L["FakeEventsEnabled"] = "Erstellt jetzt künstliche Ereignisse zum Testen."
L["Filtering_Enabled"] = "Filtern aktivieren"
L["Filtering_Filterlist"] = "Filterliste"
L["FilterSettings"] = "Filtereinstellungen"
L["Font"] = "Schriftart"
L["Font_Desc"] = "Die Hauptschriftart dieses Addons"
L["Font_Outline"] = "Schriftkontur"
L["Font_Outline_Desc"] = "Der Umriss der Schrift"
L["Fontcolor"] = "Schriftfarbe"
L["Fontcolor_Desc"] = "Die Farbe des Texts"
L["FontShadow_Enabled"] = "Schriftschatten aktivieren"
L["FontShadow_Enabled_Desc"] = "Wenn diese Option aktiviert ist, erscheint ein Schatten um den Text. Du kannst die Farbe dieses Schattens wählen."
L["FontShadowColor"] = "Schattierungsfarbe"
L["FontShadowColor_Desc"] = "Die Farbe des Textschattens"
L["FontSize"] = "Schriftgröße"
L["FontSize_Desc"] = "Die Größe des Texts"
L["Frame"] = "Rahmen"
L["Framescale"] = "Skalierung"
L["Framescale_Desc"] = "Die Skalierung des Hauptrahmens. Diese Einstellung ist im Kampf aufgrund Einschränkungen seitens Blizzards nicht verfügbar."
L["General"] = "Allgemein"
L["GeneralSettings"] = "Allgemeine Einstellungen"
L["GeneralSettings_Desc"] = "Ein paar allgemeine Einstellungen"
L["GeneralSettingsAllies"] = "Diese Einstellungen werden für Verbündete verwendet, unabhängig von der Größe des Schlachtfelds"
L["GeneralSettingsEnemies"] = "Diese Einstellungen werden für Gegner verwendet, unabhängig von der Größe des Schlachtfelds"
L["HealthBar"] = "Gesundheitsbalken"
L["HealthBar_Background_Desc"] = "Die Hintergrundfarbe des Gesundheitsbalkens."
L["HealthBar_Texture_Desc"] = "Die Textur, die für die Gesundheitsbalken verwendet wird"
L["HealthBarSettings_Desc"] = "Einstellungen für den Gesundheitsbalken und daran angebrachte Elemente wie der Namen oder der Zielzähler."
L["HealthTextEnabled"] = "Gesundheitstext aktivieren"
L["HealthTextSettings"] = "Gesundheitstext-Einstellungen"
L["HealthTextType"] = "Art Gesundheitstext"
L["Height"] = "Höhe"
L["HideArenaframesIn"] = "Arenafenster verstecken in"
L["HideHighestPriority"] = "Höchste Priorität verstecken"
L["HideHighestPriority_Desc"] = "Wenn diese Option aktiviert ist, wird die Aura mit der höchsten Priorität entfernt, dies ist nützlich wenn die Aura mit der höchsten Priorität bereits auf dem Klassen- oder Spezialisierungssymbol angezeigt wird. "
L["HideRaidframesIn"] = "Schlachtzugsfenster verstecken in"
L["HideWhenZero"] = "Ausblenden falls Null"
L["HighestPriorityAura"] = "Höchst priorisierte Aura"
L["Highlight_Color"] = "Hervorhebungsfarbe"
L["Highlight_Color_Desc"] = "Die Farbe des Spielerbalkens, auf dem sich der Mauszeiger momentan befindet"
L["HorizontalGrowdirection"] = "Horizontale Ausbreitungsrichtung"
L["HorizontalGrowdirection_Desc"] = "Ob die Spalten sich nach links oder nach rechts erweitern sollen"
L["HorizontalSpacing"] = "Horizontaler Abstand"
L["IAmTargeted"] = "Ich werde anvisiert"
L["Icon"] = "Symbol"
L["Icons"] = "Symbole"
L["IconSize"] = "Symbolgröße"
L["IconsPerRow"] = "Symbole pro Reihe"
L["Import"] = "Import"
L["ImportButton"] = "Profil importieren"
L["ImportButton_Desc"] = "Klicke hier, um ein Profil per Zeichenkette zu importieren und dein momentan aktives Profil zu überschreiben"
L["JumpToGeneralOptions"] = "Zu allgemeinen Einstellungen springen"
L["JumpToGeneralOptions_Desc"] = "Klicke hier, um zu den allgemeinen Einstellungen dieses Moduls zu springen."
L["JustifyH"] = "Horizontale Ausrichtung"
L["JustifyH_Desc"] = "Die horizontale Ausrichtung des Texts"
L["JustifyV"] = "Vertikale Ausrichtung"
L["JustifyV_Desc"] = "Die vertikale Ausrichtung des Texts"
L["KeybindSettings_Desc"] = "Hier kannst du festlegen, was passieren soll, wenn du auf den Gegner mit der linken, rechten und mittleren Maustaste klickst."
L["LEFT"] = "Links"
L["LeftToTargetCounter"] = "Links neben dem Zielzähler"
L["Leftwards"] = "Nach links"
L["Level"] = "Spielerstufe"
L["LevelText_Enabled"] = "Spielerstufe anzeigen"
L["LevelText_OnlyShowIfNotMaxLevel"] = "Nur anzeigen, falls nicht Maximalstufe"
L["Locked"] = "Fixieren"
L["Locked_Desc"] = "Macht den Rahmen unverschiebbar"
L["Magic"] = "Magie"
L["MainFrameSettings"] = "Hauptfenster-Einstellungen"
L["MainFrameSettings_Desc"] = "Einstellungen für das Hauptfenster, das für %s verwendet wird."
L["MaxCantBeUndefined"] = "Maximun kann nicht undefiniert sein"
L["MaxNeedsToBeEqualOrGreater"] = "Maximun muss größer als oder gleich dem Minimum sein"
L["MaxPlayerCount"] = "Maximale Spielerzahl"
L["MaxPlayers"] = "Maximale Gegnerzahl"
L["MaxPlayers_Desc"] = "Das Addon zeigt keine Gegner an, falls die momentane Gegnerzahl größer als diese Zahl ist"
L["MIDDLE"] = "Mittig"
L["MinCantBeUndefined"] = "Minimum kann nicht undefiniert sein"
L["MinPlayerCount"] = "Minimale Spielerzahl"
L["Mirrored"] = "Gespiegelt"
L["Modules"] = "Module"
L["ModuleSettings"] = "Moduleinstellungen"
L["ModuleSpecificSettings"] = "Modulspezifische Einstellungen"
L["MoreProfileOptions"] = "Mehr Profiloptionen"
L["MouseWheelPlayerTargeting"] = "Zielen durch Scrollen"
L["MouseWheelPlayerTargeting_Desc"] = "Wenn diese Option aktiviert ist, kannst du deine Maus über die Rahmen der Gegner oder Verbündeten bewegen und scrollen, um den nächsten oder vorherigen Spieler in der Liste anzuvisieren. Bitte beachte dass diese Option mouesover Tastaturbelegungen beeinträchtigen kann."
L["MyFocus"] = "Mein Fokus"
L["MyFocus_Color"] = "Fokusfarbe"
L["MyFocus_Color_Desc"] = "Die Farbe die für den Rahmen verwendet wird, falls dieser Gegner dein Fokusziel ist."
L["MyTarget"] = "Mein Ziel"
L["MyTarget_Color"] = "Zielfarbe"
L["MyTarget_Color_Desc"] = "Die Farbe die für den Rahmen verwendet wird, falls du diesen Gegner im Ziel hast."
L["MyVersion"] = "Du verwendest Version"
L["Name"] = "Name"
L["Name_Desc"] = "Einstellungen für die Namen auf dem Gesundheitsbalken"
L["NewVersionAvailable"] = "Eine neue Version ist verfügbar"
L["None"] = "Keine"
L["NonPriorityBuffs"] = "Nicht priorisierte Stärkungszauber"
L["NonPriorityDebuffs"] = "Nicht priorisierte Schwächungszauber"
L["Normal"] = "Normal"
L["NormalCastingBar"] = "Normaler Zauberbalken"
L["NotAvailableInCombat"] = [=[
Diese Einstellung ist aufgrund Einschränkungen seitens Blizzards nicht im Kampf verfügbar.]=]
L["Notifications_Allies_Enabled_Desc"] = "Falls diese Option aktiviert ist, wirst du benachrichtigt, sobald ein Verbündeter das Schlachtfeld verlässt oder betritt."
L["Notifications_Enabled"] = "Benachrichtigungen aktivieren"
L["Notifications_Enemies_Enabled_Desc"] = "Falls diese Option aktiviert ist, wirst du benachrichtigt, sobald ein Gegner das Schlachtfeld verlässt oder betritt."
L["NoVersion"] = "Kein Addon gefunden"
L["NumericTargetindicator_Enabled_Desc"] = "Zeigt an wieviele %s diesen Gegner anvisieren"
L["ObjectiveAndRespawn_ObjectiveEnabled"] = "Schlachtfeldziel anzeigen"
L["ObjectiveAndRespawn_ObjectiveEnabled_Desc"] = "Zeigt Symbole für Flaggen-, Minenwagen- oder Kugelträger neben dem Gegner-Button an"
L["ObjectiveAndRespawn_Position"] = "Position"
L["ObjectiveAndRespawn_Position_Desc"] = "Hier kannst du einstellen, ob das Symbol links neben dem Spezialisierungssymbol oder rechts neben den PvP-Talenten/Volksfähigkeiten angezeigt werden soll"
L["ObjectiveAndRespawn_RespawnEnabled"] = "Wiederbelebungs-Timer"
L["ObjectiveAndRespawn_RespawnEnabled_Desc"] = "Falls diese Option aktiviert ist, wird ein Symbol angezeigt, das die verbleibende Zeit bis zur Wiederbelebung des Gegners anzeigt."
L["ObjectiveAndRespawn_Width_Desc"] = "Breite des Schalchtfeldziel-Symbols"
L["ObjectiveAndRespawnSettings_Desc"] = "Einstellungen für das Ziel des Schlachtfelds wie z.B. Flaggen, Kugeln oder Minenwagen"
L["ObjectiveAndRespawnTimer"] = "Ziel und Wiederbelebungstimer"
L["OffsetX"] = "Versatz X"
L["OffsetY"] = "Versatz Y"
L["OldVersion"] = "Alte Version"
L["OnlyShowForHealers"] = "Nur für Heiler anzeigen"
L["OutOfCombat"] = "Außerhalb des Kampfs"
L["OverwriteMirroredConfirm"] = "Bist du sicher, dass du die Einstellungen für %s mit den gespiegelten Einstellungen von %s überschreiben möchtest?"
L["OverwriteNormalConfirm"] = "Bist du sicher, dass du die Einstellungen für %s mit den Einstellungen von %s überschreiben möchtest?"
L["PlayerCount"] = "Spielerzahl"
L["PlayerCount_Enabled_Desc"] = "Falls diese Option aktiviert ist, wird die momentane Spielerzahl angezeigt. Beachte, dass diese Zahl nicht unbedingt der Zahl der momentan gezeigten Spielerreihen entspricht. Falls ein Gegner ein gewertetes Schlachtfeld verlässt, besteht keine Möglichkeit herauszufinden wer gegangen ist, daher bleibt die Reihenzahl gleich. Ebenso kann es passieren, dass Gegner das Schlachtfeld betreten oder verlassen während du im Kampf bist. In diesem Fall besteht auch keine Möglichkeit eine Reihe hinzuzufügen oder zu entfernen. (Dies passiert sobald du den Kampf verlässt.)"
L["players"] = "Spieler"
L["Point"] = "Punkt"
L["PointAtObject"] = "Punkt am Objekt"
L["Poison"] = "Gift"
L["Position"] = "Position"
L["PowerBar"] = "Ressourcenbalken"
L["PowerBar_Background_Desc"] = "Die Hintergrundfarbe des Ressourcenbalkens."
L["PowerBar_Enabled"] = "Ressourcenbalken aktivieren"
L["PowerBar_Enabled_Desc"] = "Falls diese Option aktiviert ist, werden zusätzliche Balken für die Ressourcen der Gegner wie Mana, Wut, Fokus etc. angezeigt."
L["PowerBar_Height_Desc"] = "Die Höhe des Ressourcenbalkens. Wenn du die Höhe des Ressourcenbalkens vergrößerst, verkleinert sich automatisch die Höhe des Gesundheitsbalkens."
L["PowerBar_OnlyShowForHealers_Desc"] = "Falls aktiviert, wird die Ressourcenleiste nur bei Heilern angezeigt."
L["PowerBar_Texture_Desc"] = "Die Textur, die für die Ressourcenbalken verwendet wird"
L["PowerBarSettings_Desc"] = "Einstellungen für den Ressourcenbalken"
L["PriorityBuffs"] = "Priorisierte Stärkungszauber"
L["PriorityDebuffs"] = "Priorisierte Schwächungszauber"
L["Racial"] = "Volksfertigkeit"
L["Racial_Enabled"] = "Volksfähigkeiten aktivieren"
L["Racial_Enabled_Desc"] = "Falls diese Option aktiviert ist, werden Symbole für Volksfähigkeiten angezeigt."
L["Racial_Width_Desc"] = "Die Breite der Volksfähigkeiten."
L["RacialFiltering_Enabled_Desc"] = "Wenn diese Option aktiviert ist, wirst du nur angehakten Volksfähigkeiten sehen."
L["RacialFiltering_Filterlist_Desc"] = "Klicken, um die Verfolgung dieser Volksfähigkeit zu aktivieren/deaktivieren."
L["RacialFilteringSettings_Desc"] = "Hier kannst du festlegen, welche Volksfähigkeiten angezeigt werden"
L["RacialSettings"] = "Volksfähigkeiten"
L["RacialSettings_Desc"] = "Einstellungen bezüglich der Volksfähigkeiten"
L["RangeIndicator_Alpha"] = "Deckkraft"
L["RangeIndicator_Alpha_Desc"] = "Die Deckkraft des Balkens, wenn der Gegner außerhalb der eingestellten Reichweite ist"
L["RangeIndicator_Enabled"] = "Reichweitenindikator aktivieren"
L["RangeIndicator_Enabled_Desc"] = "Wenn diese Option aktiviert ist, wird der Balken des Gegners mit der eingestellten Transparenz angezeigt, falls der Gegner außerhalb der eingestellten Reichweite ist. "
L["RangeIndicator_Everything"] = "Die Deckkraft aller Elemente verändern"
L["RangeIndicator_Frames"] = "Deckkraft verändern von"
L["RangeIndicator_Frames_Desc"] = "Hier kannst du festlegen, welche Elemente des Gegnerbalkens eine höhere Transparenz erhalten sollen, falls der Gegner außer Reichweite ist."
L["RangeIndicator_Range"] = "Entfernung"
L["RangeIndicator_Range_Desc"] = "Falls ein Gegner mehr als diese Entfernung (in Metern) weg ist, wird sein Balken mit der eingestellten Deckkraft angezeigt."
L["RangeIndicator_Settings"] = "Reichweitenindikator"
L["RangeIndicator_Settings_Desc"] = "Einstellungen für den Reichweitenindikator."
L["RangeOverlapping"] = "Bereich überlappt mit: %s"
L["RBGSpecificSettings"] = "Gewertetes Schlachtfeld"
L["RBGSpecificSettings_Desc"] = "Diese Einstellungen/Funktionen sind nur im gewerteten Schlachtfeld wirksam"
L["ResetGeneralModule_Desc"] = "Klicke hier, um diese allgemeinen Einstellungen dieses Moduls zurückzusetzen"
L["ResetModule"] = "Modul zurücksetzen"
L["ResetModule_Desc"] = "Klicke hier, um die Einstellungen dieses Moduls für %s:%s zurückzusetzen"
L["ResetModuleSettings"] = "Die Einstellungen dieses Moduls zurücksetzen"
L["RestoreDefault"] = "Standard wiederherstellen"
L["RIGHT"] = "Rechts"
L["Rightwards"] = "Nach rechts"
L["RoleIcon_Enabled"] = "Rollensymbole aktivieren"
L["RoleIcon_Enabled_Desc"] = "Aktiviere diese Option, falls du ein Rollensymbol auf dem Gesundheitsbalken haben willst."
L["RoleIcon_Size_Desc"] = "Die Größe der Rollensymbole (Breite und Höhe)"
L["RoleIconSettings"] = "Rollensymbol"
L["RoleIconSettings_Desc"] = "Einstellungen für das Rollensymbol, welches auf dem Gesundheitsbalken angezeigt wird."
L["RoleSortingOrder"] = "Rollenreihenfolge"
L["RoleSortingOrder_Desc"] = "Wähle in welcher Reihenfolge die Spieler entsprechend ihrer Rolle platziert werden sollen"
L["Scale"] = "Skalierung"
L["ShowDispellable"] = "Bannbare anzeigen"
L["ShowDispellable_Desc"] = [=[Zeigt %s die ich entfernen kann durch Reinigen, Entgiftung, etc.
]=]
L["ShowFramesIn"] = "Fenster anzeigen in"
L["ShowHighestPriority"] = "Höchste Priorität anzeigen"
L["ShowHighestPriority_Desc"] = "Wenn du diese Option aktivierst wird angezeigt wenn der Spieler unterbrochen oder eine wichtige Aura aktiv ist"
L["ShowMine"] = "Meine anzeigen"
L["ShowMine_Desc"] = "%s anzeigen, die ich aufgebracht habe."
L["ShowNumbers"] = "Zahlen anzeigen"
L["ShowNumbers_Desc"] = "Abklingzeiten anzeigen, beachte, dass diese Option nicht funktionieren könnte, wenn du Addons wie OmniCC benutzt."
L["ShowRealmnames"] = "Realmnamen anzeigen"
L["ShowRealmnames_Desc"] = "Zeigt den Realm des Gegners"
L["ShowSpecIfExists"] = "Spezialisierung anzeigen falls vorhanden"
L["ShowSpecIfExists_Desc"] = "Wenn diese Option aktiviert ist, wird das Spezialisierungssymol anstatt des Klassensymbols angezeigt"
L["ShowTooltips"] = "Tooltips anzeigen"
L["ShowTooltips_Desc"] = "Zeigt Hinweistexte an, wenn du deine Maus über Objekte wie Schmuckstück- oder Aurensymbole bewegst."
L["Side"] = "Seite"
L["SideAtObject"] = "Seite des Objekts"
L["Size"] = "Größe"
L["SmallCastingBar"] = "Kleiner Zauberbalken"
L["SourceFilter"] = "Nach Quelle filtern"
L["Spec_AuraDisplay_Enabled"] = "Auraanzeige aktivieren"
L["Spec_AuraDisplay_Enabled_Desc"] = "Wenn diese Option aktiviert ist, werden Symbole von für Kontrolleffekte und Unterbrechungen (falls aktiv) anstelle des Spezialisierungssymbols angezeigt."
L["Spec_Enabled"] = "Spezialisierung aktivieren"
L["Spec_Enabled_Desc"] = "Falls diese Option aktiviert ist, werden Symbole für Spezialisierung des Spielers angezeigt."
L["Spec_Width_Desc"] = "Die Breite des Spezialisierungssymbols."
L["SpecClassPriorityOne"] = "SpezialisierungKlassePriorität1"
L["SpecClassPriorityTwo"] = "SpezialisierungKlassePriorität2"
L["SpecSettings"] = "Spezialisierung"
L["SpecSettings_Desc"] = "Einstellungen für das Spezialisierungssymbol."
L["SpellID_Filtering"] = "Per Zauber-ID filtern"
L["Style"] = "Stil"
L["SymbolicTargetindicator_Enabled_Desc"] = "Zeigt ein quadratisches klassenfarbenes Symbol für jeden %s, der diesen Gegner im Ziel hat."
L["TargetAmount"] = "Zielanzahl"
L["TargetAmount_Ally"] = "Der Ton wird abgespielt sobald ein Gruppenmitglied von mehr als der eingestellten Zahl an Gegnern anvisiert wird."
L["TargetAmount_Me"] = "Der Ton wird abgespielt sobald Du von mehr als der eingestellten Zahl an Gegnern anvisiert wird."
L["TargetCallerUpdated"] = "ist jetzt der Zielansager"
L["TargetCalling"] = "Zielansage"
L["TargetCallingNotificationEnable"] = "Benachrichtigungston aktivieren"
L["TargetCallingNotificationEnable_Desc"] = "Benachrichtigt dich sobald es ein neues Ziel gibt (ein Schlachtzugssymbol auf einem neuen Gegner)"
L["TargetCallingSetMark"] = "Mein Ziel mit dem Totenkopf markieren"
L["TargetCallingSetMark_Desc"] = "Markiert dein aktuelles Ziel automatisch mit dem Totenkopf wenn du der Schlachtzugsleiter bist und es keinen Assistenten in dem Schlachtzug gibt. Dein aktuelles Ziel wird ebenfalls markiert wenn du ein Assistent."
L["TargetCallingShowIcon"] = "Ein Symbol auf dem Ziel des Zielansagers anzeigen"
L["TargetCallingShowIcon_Desc"] = "Falls aktiviert, wird ein Symbol auf Button des Gegners angezeigt, den der Zielansager im Ziel hat. "
L["TargetCallingSoundEnable"] = "Benachrichtige mich bei neuem Ziel"
L["TargetCallingSoundEnable_Desc"] = "Falls aktiviert, wird ein Ton abgespielt wenn der Zielansager einen anderen Gegner anvisiert"
L["TargetCallingSoundSound"] = "Benachrichtigungston"
L["TargetIndicator"] = "Zielanzeiger"
L["TargetIndicator_Desc"] = "Numerische und Symbolische Zielanzeiger"
L["TargetIndicatorNumeric"] = "Zielindikator Numerisch"
L["TargetIndicatorNumeric_HideWhenZero_Desc"] = "Falls aktviert, wird der Text versteckt wenn kein Gegner diesen Spieler im Ziel hat."
L["TargetIndicatorSymbolic"] = "Zielindikator Symbolisch"
L["TargetNextAlly"] = "Nächsten Verbündeten anvisieren"
L["TargetNextEnemy"] = "Nächsten Gegner anvisieren"
L["TargetPreviousAlly"] = "Vorherigen Verbündeten anvisieren"
L["TargetPreviousEnemy"] = "Vorherigen Gegner anvisieren"
L["Testmode_Toggle"] = "Testmodus ein-/ausschalten"
L["Testmode_Toggle_Desc"] = "Aktiviert/Deaktiviert den Testmodus. Dieser Modus macht es dir einfacher zu sehen, welche Veränderungen die einzelnen Optionen mit sich bringen. Es ist nicht möglich alle Optionen zu simulieren, jedoch sind die meisten Einstellungsmöglichkeiten in diesem Testmodus abgebildet. Der Testmodus ist deaktiviert, falls du dich momentan in einem Schlachtfeld befindest. "
L["Testmode_ToggleAnimation"] = "Testmodus-Animation ein-/ausschalten"
L["Testmode_ToggleAnimation_Desc"] = "Aktiviert/Deaktiviert die Animation des Testmodus, damit du dich besser auf eine einzelne Einstellung konzentrieren kannst und nicht von der Animation abgelenkt wirst."
L["Testmode_UseTeammates"] = "Mit Gruppenmitgliedern testen"
L["Testmode_UseTeammates_Desc"] = "Wenn du diese Option aktivierst, verwendet der Testmodus deine momentanen Gruppenmitglieder anstelle von simulierten Spielern"
L["TestmodeDisabled"] = "Testmodus deaktiviert"
L["TestmodeEnabled"] = "Testmodus aktiviert"
L["TestmodeSettings"] = "Testmodus"
L["Text"] = "Text"
L["TextSettings"] = "Texteinstellungen"
L["Thick"] = "Dick"
L["TOP"] = "Oben"
L["TOPLEFT"] = "Oben links"
L["TOPRIGHT"] = "Oben rechts"
L["Trinket"] = "Anti-CC-Schmuck"
L["Trinket_Enabled"] = "PvP-Talente aktivieren"
L["Trinket_Enabled_Desc"] = "Falls diese Option aktiviert wird, wird angezeigt welches PvP-Talent der ersten Reihe der Gegner benutzt (Medaillon des Gladiators, Anpassung oder Unerbittlich)"
L["Trinket_Width_Desc"] = "Die Breite der Anti-CC-Talente."
L["TrinketSettings_Desc"] = "Einstellungen bezüglich des PvP-Talents in der ersten Reihe"
L["UpdatePeriod"] = "Aktualisierungsperiode"
L["UpdatePeriod_Desc"] = "Die Zeit zwischen den Abfragen in Sekunden ob jemand im Kampf ist oder nicht. Stelle diesen Wert nicht zu niedrig ein denn dies kann zu Leistungsproblemen führen."
L["Upwards"] = "Nach oben"
L["UseBarHeight"] = "Balkenhöhe verwenden"
L["UseBigDebuffsPriority"] = "BigDebuffs-Prioritäten verwenden"
L["UseBigDebuffsPriority_Desc"] = "Wenn du diese Option aktivierst, verwendet das Addon die Prioritäten aus dem Addon BigDebuffs. Diese Prioritäten werden von den Modulen Stärkungszauber, Schwächungszauber und Höchste Priorität verwenden"
L["UseButtonHeight"] = "Buttonhöhe verwenden"
L["UseButtonHeight_Desc"] = "Wenn du diese Option aktivierst, wird die Höhe des Buttons verwendet"
L["UserDefined"] = "Benutzerdefiniert"
L["VerticalGrowdirection"] = "Vertikale Ausbreitungsrichtung"
L["VerticalGrowdirection_Desc"] = "Ob die Reihen nach oben oder nach unten anwachsen sollen"
L["VerticalPosition"] = "Vertikale Position"
L["VerticalSpacing"] = "Vertikaler Abstand"
L["Width"] = "Breite"

elseif LOCALE_esES or LOCALE_esMX then
L["AddPoint"] = "Añadir punto"
L["Allies"] = "Aliados"
L["allies"] = "aliados"
L["ally"] = "aliado"
L["Ally"] = "Aliado"
L["AllyIsTargeted"] = "Aliado es objetivo de"
L["AllyJoined"] = "Un aliado se ha unido al campo de batalla"
L["AllyLeft"] = "Un aliado ha abandonado el campo de batalla"
L["AND"] = "y"
L["AttachToObject"] = "Adjuntar al objeto"
L["AttachToObject_Desc"] = "Tenga en cuenta que no es posible seleccionar todos los objetos, dependiendo de su posición de los otros marcos. Por ejemplo, si ancla los talentos de pvp en las raciales y está tratando de anclarlas en los talentos de pvp, aparecerá un mensaje de error."
L["Auras_Buffs_Container_Color_Desc"] = "El color del borde de los iconos de los buffs"
L["Auras_CustomFiltering_Conditions_All"] = "Todos/as"
L["Auras_CustomFiltering_Conditions_Any"] = "Cualquier/a"
L["Auras_CustomFiltering_ConditionsMode"] = "Modo de condiciones"
L["Auras_CustomFiltering_ConditionsMode_Desc"] = "Aquí se puede elegir cómo se deberían aplicar las condiciones indicadas a las auras. Si eliges \"Cualquiera\", el addon mostrará el aura tan pronto como se cumplan las condiciones indicadas. Si eliges \"Todas\", el addon solo mostrará el aura si se cumplen todas las condiciones indicadas."
L["Auras_Debuffs_Container_Color_Desc"] = "El color del borde de los iconos de los debuff"
L["Auras_Enabled"] = "Habilitar auras"
L["Auras_Enabled_Desc"] = "Cuando esté habilitado habrá iconos de aura. También habrá un número para el tiempo restante y uno para las pilas aplicadas"
L["Auras_Filtering_Mode"] = "Modo de filtrado"
L["Auras_Filtering_Mode_Desc"] = "Aquí puedes elegir qué filtro se aplica a las auras. Puedes elegir entre el filtro Ventisca, el cual aplica el mismo filtro que Blizzard usa en el marco de banda, y las condiciones de filtrado personalizadas"
L["Auras_ShowTooltips"] = "Mostrar información de los hechizos"
L["AurasCustomConditions"] = "Condiciones personalizadas"
L["AurasFiltering_AddSpellID"] = "ID del Hechizo"
L["AurasFiltering_AddSpellID_Desc"] = "Introduce el ID de hechizo o varios ID de hechizos separados por coma. Los ID de hechizos se incluirán a la lista de filtros. Solo las auras que su ID de hechizo estén en la lista de filtros se mostrarán cuando el filtro de ID de hechizo esté activo."
L["AurasFiltering_Enabled_Desc"] = "Cuando esta opción está habilitada, solo verás los buff y debuff que se agregaron a la lista de filtrado"
L["AurasFiltering_Filterlist_Desc"] = "Click para eliminar %s"
L["AurasFilteringSettings_Desc"] = "Aquí puedes controlar qué auras se muestra"
L["AurasSettings"] = "Auras"
L["AurasSettings_Desc"] = "Opciones de auras (buffs y debuffs)"
L["AurasStacktextSettings"] = "Texto apilado"
L["BarBackground"] = "Color de fondo"
L["BarHeight_Desc"] = "La altura de cada fila. Esta configuración no está disponible en combate debido a la protección de combate de Blizzard."
L["BarSettings"] = "Configuración de barra"
L["BarSettings_Desc"] = "Aquí puedes configurar las filas"
L["BarTexture"] = "Textura de barra"
L["BarWidth_Desc"] = "El ancho de cada fila."
L["BattlegroundSize"] = "Tamaño del campo de batalla"
L["BGSize_15"] = "1-15 Jugadores"
L["BGSize_15_Desc"] = "Estos ajustes son aplicados para %s en un Campo de Batalla de 1 a 15 jugadores por equipo"
L["BGSize_40"] = "16-40 Jugadores"
L["BGSize_40_Desc"] = "Estos ajustes son aplicados para %s en un Campo de Batalla de 16 a 40 jugadores por equipo"
L["BGSize_5"] = "Arena"
L["BGSize_5_Desc"] = "Estos ajustes son aplicados para %s en arenas"
L["BlizzlikeAuraFiltering"] = "Filtrado de aura como Blizzard"
L["BorderSize"] = "Tamaño del borde"
L["BorderThickness"] = "Tamaño del borde"
L["BOTTOM"] = "Abajo"
L["BOTTOMLEFT"] = "Inferior izquierda"
L["BOTTOMRIGHT"] = "Inferior derecha"
L["buff"] = "mejora (buff)"
L["BuffContainer"] = "Contenedor de mejoras (buffs)"
L["BuffIcon"] = "Icono de mejora (buff)"
L["Buffs"] = "Mejoras (Buffs)"
L["Button"] = "Botón"
L["CantAnchorToItself"] = "No se puede anclar a sí mismo!"
L["CENTER"] = "Centro"
L["Color"] = "Color"
L["Coloring_Enabled"] = "Color por tipo de debuff"
L["Coloring_Enabled_Desc"] = "Cuando esta opción está habilitada, puede elegir si el borde del marco o el texto de la cuenta atrás, deben ser coloreados por tipo de debuff (como se sabe por la interfaz de usuario predeterminada, los venenos son verdes, las enfermedades son azules, etc.)"
L["Columns"] = "Columnas"
L["Columns_Desc"] = "En cuantas columnas se muestran los jugadores."
L["Combat"] = "En combate"
L["CombatIndicator"] = "Indicador de combate"
L["CombatLogScanningForEnemiesDisabled"] = "Escaneo de registro de combate para enemigos deshabilitado"
L["CombatLogScanningForEnemiesEnabled"] = "El escaneo de registros de combate para detectar jugadores enemigos se ha habilitado."
L["ConfirmProfileOverride"] = "¿Estás seguro de querer sobrescribir el subperfil %s por el subperfil %s?"
L["Container_Color"] = "Color del borde del contenedor"
L["ContainerPosition"] = "Posición del contenedor"
L["ContainerSettings"] = "Configuraciones de contenedores"
L["ConvertCyrillic"] = "Convertir a Cirílico"
L["ConvertCyrillic_Desc"] = "Convierte caracteres cirílicos, lo que facilita la lectura de nombres si juegas contra rusos"
L["CopySettings"] = "Copiar ajustes de"
L["Countdowntext"] = "Texto de cuenta atrás"
L["Covenant"] = "Curia"
L["CovenantIcon_Enabled"] = "Habilitar iconos de Curia"
L["CovenantIcon_Enabled_Desc"] = "Habilite esta opción si desea ver un icono de Curia en la barra de salud"
L["CovenantIcon_Size_Desc"] = "El tamaño de los iconos de Curia (ancho y alto)"
L["CovenantIconSettings_Desc"] = "Configuraciones para el icono de Curia que se muestra en la barra de salud"
L["CurrentVersion"] = "Versión actual"
L["Curse"] = "Maldición"
L["CustomMacro_Desc"] = "Esto le permite configurar una macro personalizada. Un campo vacío conducirá a una macro vacía. Sustituciones disponibles: %n se sustituye por el nombre del enemigo. Ejemplo: /targetexact %n /cast Polymorph /say Polymorph en %n /targetlasttarget Esto apuntará al enemigo, lanzará Polimorfia, escribirá un mensaje en el canal \"dice\" y apuntará a tu objetivo anterior. Tenga en cuenta que las macros solo pueden tener 255 caracteres (incluidas las sustituciones)."
L["Data"] = "Datos"
L["debuff"] = "debilitamiento (debuff)"
L["DebuffContainer"] = "Contenedor de debilitamiento (debuff)"
L["DebuffIcon"] = "Icono de debilitamiento (debuff)"
L["Debuffs"] = "Debilitamientos (debuffs)"
L["DebuffType_Filtering"] = "Filtrar por tipo de debilitamiento (debuff)"
L["DebuffType_Filtering_Desc"] = "Cuando esta opción está habilitada, solo verá debilitamientos (debuff) del tipo elegido. Esta opción no se simula en el modo de prueba, ya que requeriría una gran base de datos con toneladas de hechizos y su tipo de debilitamiento."
L["DeletePoint"] = "Eliminar Punto %s"
L["DisableArenaFramesInArena"] = "Deshabilitar marcos de arena en la arena"
L["DisableArenaFramesInArena_Desc"] = "Desactiva los Marcos de Arena en una arena. También funciona cuando está instalado sArena"
L["DisableArenaFramesInBattleground"] = "Deshabilitar marcos de arena en el campo de batalla"
L["DisableArenaFramesInBattleground_Desc"] = "Desactiva los Marcos de Arena en un campo de batalla. También funciona cuando está instalado \"sArena\""
L["Disease"] = "Enfermedad"
L["DispellFilter"] = "Filtrar por disipabilidad"
L["DisplayType"] = "Tipo de visualización"
L["Downwards"] = "Hacia abajo"
L["DR_Disorient"] = "Desorientar"
L["DR_Incapacitate"] = "Incapacitar"
L["DR_Knockback"] = "Retroceso"
L["DR_Root"] = "Raíz"
L["DR_Silence"] = "Silencio"
L["DR_Stun"] = "Aturdir"
L["DRContainer"] = "Contenedor DR"
L["DRTracking"] = "Seguimiento DR"
L["DrTracking_Container_Color_Desc"] = "El color del borde del contenedor de los iconos DR"
L["DrTracking_DisplayType_Desc"] = "Aquí puede elegir si el estado del DR (Rendimiento Decreciente), debe indicarse con un marco de color alrededor del icono o con el color del texto de la cuenta atrás. Tenga en cuenta que la configuración de texto de cuenta atrás solo funcionará si no está utilizando un complemento como OmniCC que modifica el color del texto."
L["DrTracking_Enabled"] = "Seguimiento DR"
L["DrTracking_Enabled_Desc"] = "Cuando esté habilitado, habrá iconos para el Seguimiento DR (Rendimiento Decreciente) junto a la fila enemiga. Borde verde: el próximo CC será de media duración. Borde amarillo: el próximo CC tendrá una duración de un cuarto. Borde rojo: el siguiente CC no se aplicará porque el jugador es inmune"
L["DrTracking_Spacing"] = "Espaciado DR"
L["DrTracking_Spacing_Desc"] = "El espacio entre cada icono DR"
L["DrTrackingFiltering_Enabled_Desc"] = "Cuando esta opción está habilitada, solo verá iconos de hechizos cuya categoría está marcada en la lista de filtrado."
L["DrTrackingFiltering_Filterlist_Desc"] = "Haga clic para rastrear o dejar de rastrear esta categoría"
L["DrTrackingFilteringSettings_Desc"] = "Aquí puede elegir qué categoría DR se muestra."
L["DrTrackingSettings"] = "Seguimiento DR"
L["DrTrackingSettings_Desc"] = "Configuraciones relacionadas con el seguimiento DR"
L["DurationFilter"] = "Filtrar por duración"
L["DurationFilter_OnlyShowWhenDuration"] = "Mostrar solo cuando la duración sea menor o igual que"
L["DurationFilter_OnlyShowWhenDuration_Desc"] = "Cualquier aura cuya duración sea más larga que la duración seleccionada no pasará por el filtro"
L["Enable_DrawSwipe"] = "Habilitar textura de deslizamiento"
L["Enable_DrawSwipe_Desc"] = "Muestra una textura desaturada y más oscura para la parte restante del tiempo de reutilización."
L["EnableClique"] = "Usar Clique para los atajos de teclado"
L["EnableClique_Desc"] = "Cuando esta opción está habilitada, Clique manejará todos los atajos de teclado."
L["enemies"] = "enemigos"
L["Enemies"] = "Enemigos"
L["EnemiesTargetingAllies_Enabled_Desc"] = "Cuando habilites esta opción, un sonido te notificará cuando una cantidad determinada de enemigos estén apuntando a un aliado."
L["EnemiesTargetingMe_Enabled_Desc"] = "Cuando habilites esta opción, un sonido te notificará cuando una cantidad determinada de enemigos te esté apuntando."
L["enemy"] = "enemigo"
L["Enemy"] = "Enemigo"
L["EnemyJoined"] = "Un enemigo se ha unido al campo de batalla"
L["EnemyLeft"] = "Un enemigo abandono el campo de batalla"
L["ExportButton"] = "Exportar perfil"
L["ExportButton_Desc"] = "Haga clic aquí para exportar su perfil activo actual"
L["FakeEventsDisabled"] = "Ya no se crean eventos falsos."
L["FakeEventsEnabled"] = "Ahora creando eventos falsos para probar."
L["Filtering_Enabled"] = "Habilitar filtrado"
L["Filtering_Filterlist"] = "Lista de filtros"
L["FilterSettings"] = "Configuración de filtros"
L["Font"] = "Fuente"
L["Font_Desc"] = "La fuente principal utilizada por este complemento"
L["Font_Outline"] = "Contorno de la fuente"
L["Font_Outline_Desc"] = "El contorno de la fuente"
L["Fontcolor"] = "Color de fuente"
L["Fontcolor_Desc"] = "El color del texto"
L["FontShadow_Enabled"] = "Habilitar sombra de fuente"
L["FontShadow_Enabled_Desc"] = "Cuando está habilitado, habrá una sombra alrededor de la fuente. Puedes elegir el color de esa sombra"
L["FontShadowColor"] = "Color del sombreado"
L["FontShadowColor_Desc"] = "El color de la sombra del texto"
L["FontSize"] = "Tamaño de fuente"
L["FontSize_Desc"] = "El tamaño de fuente del texto"
L["Frame"] = "Fotograma"
L["Framescale"] = "Escala de fotogramas"
L["Framescale_Desc"] = "La escala del Marco Principal. Esta configuración no está disponible en combate, debido a la protección de combate de Blizzard."
L["General"] = "General"
L["GeneralSettings"] = "Ajustes generales"
L["GeneralSettings_Desc"] = "Algunas configuraciones generales"
L["GeneralSettingsAllies"] = "Esta configuración se aplica a los aliados independientemente del tamaño del campo de batalla."
L["GeneralSettingsEnemies"] = "Esta configuración se aplica a los enemigos independientemente del tamaño del campo de batalla."
L["HealthBar"] = "Barra de salud"
L["HealthBar_Background_Desc"] = "Color de fondo para la barra de salud."
L["HealthBar_Texture_Desc"] = "La textura utilizada para las barras de salud."
L["HealthBarSettings_Desc"] = "Configuración de la barra de salud."
L["HealthTextEnabled"] = "Habilitar texto de salud"
L["HealthTextSettings"] = "Configuración de texto de salud"
L["HealthTextType"] = "Tipo de texto de salud"
L["Height"] = "Altura"
L["HighestPriorityAura"] = "Aura de máxima prioridad"
L["Highlight_Color"] = "Color de resaltado"
L["Highlight_Color_Desc"] = "El color del marco del jugador en el que se encuentra actualmente el cursor"
L["HorizontalGrowdirection"] = "Dirección de crecimiento horizontal"
L["HorizontalGrowdirection_Desc"] = "Te permite elegir si las filas crecen hacia la izquierda o hacia la derecha"
L["HorizontalSpacing"] = "Espaciado horizontal"
L["IAmTargeted"] = "Estoy marcado"
L["Icon"] = "Icono"
L["IconSize"] = "Tamaño del icono"
L["IconsPerRow"] = "Iconos por fila"
L["Import"] = "Importar"
L["ImportButton"] = "Importar perfil"
L["ImportButton_Desc"] = "Haga clic aquí para importar un perfil de una cadena y sobrescribir su perfil activo actual"
L["JustifyH"] = "Alineación horizontal"
L["JustifyH_Desc"] = "La alineación horizontal del texto"
L["JustifyV"] = "Alineación vertical"
L["JustifyV_Desc"] = "La alineación vertical del texto"
L["KeybindSettings_Desc"] = "Aquí puede configurar lo que debería suceder con el clic izquierdo, derecho y central. "
L["LEFT"] = "Izquierda"
L["LeftToTargetCounter"] = "A la izquierda del contador de objetivos"
L["Leftwards"] = "Hacia la izquierda"
L["Level"] = "Nivel"
L["LevelText_Enabled"] = "Mostrar el nivel del jugador"
L["LevelText_OnlyShowIfNotMaxLevel"] = "Mostrar solo si no es el nivel máximo"
L["Locked"] = "Bloquear"
L["Locked_Desc"] = "Bloquea el marco en su lugar"
L["Magic"] = "Magia"
L["MainFrameSettings"] = "Configuración del marco principal"
L["MainFrameSettings_Desc"] = "Configuraciones para el cuadro principal que se usa para %s"
L["MaxPlayers"] = "Enemigos Máximos"
L["MaxPlayers_Desc"] = "El complemento no mostrará enemigos si hay más enemigos que este número"
L["MIDDLE"] = "Medio"
L["ModuleSettings"] = "Configuración del módulo"
L["ModuleSpecificSettings"] = "Configuraciones específicas del módulo"
L["MoreProfileOptions"] = "Más opciones de perfil"
L["MyFocus"] = "Mi enfoque"
L["MyFocus_Color"] = "Color del Foco"
L["MyFocus_Color_Desc"] = "El color utilizado para el borde que indica su Foco actual"
L["MyTarget"] = "Mi objetivo"
L["MyTarget_Color"] = "Color del objetivo"
L["MyTarget_Color_Desc"] = "El color utilizado para el borde alrededor de la barra de salud que indica su objetivo actual"
L["MyVersion"] = "Estás usando la versión"
L["Name"] = "Nombre"
L["Name_Desc"] = "Configuración de nombres en la barra de salud."
L["NewVersionAvailable"] = "Esta disponible una nueva versión"
L["None"] = "Ninguno"
L["NonPriorityBuffs"] = "Mejoras no priorizadas"
L["NonPriorityDebuffs"] = "Debilitamientos no priorizados"
L["Normal"] = "Normal"
L["NormalCastingBar"] = "Barra normal de lanzamiento"
L["NotAvailableInCombat"] = "Esta configuración no está disponible mientras estás en combate debido a las restricciones de combate de Blizzard."
L["Notifications_Allies_Enabled_Desc"] = "Cuando esté habilitado, se le notificará cuando un aliado se vaya o se una al campo de batalla."
L["Notifications_Enabled"] = "Permitir notificaciones"
L["Notifications_Enemies_Enabled_Desc"] = "Cuando esté habilitado, se le notificará cuando un enemigo se vaya o se una al campo de batalla."
L["NoVersion"] = "No se encontró ningún complemento"
L["NumericTargetindicator_Enabled_Desc"] = "Muestra cuántos %s están apuntando a un jugador."
L["ObjectiveAndRespawn_ObjectiveEnabled"] = "Mostrar objetivos"
L["ObjectiveAndRespawn_ObjectiveEnabled_Desc"] = "Se muestra en el icono de la bandera, el vagón de minas y los portadores de orbes junto al botón del enemigo."
L["ObjectiveAndRespawn_Position"] = "Posición"
L["ObjectiveAndRespawn_Position_Desc"] = "Si el icono debe mostrarse a la izquierda del icono de especialización o a la derecha de los iconos del abalorio/racial."
L["ObjectiveAndRespawn_RespawnEnabled"] = "Habilitar temporizador de reaparición"
L["ObjectiveAndRespawn_RespawnEnabled_Desc"] = "Cuando está habilitado, habrá un icono que indica el tiempo restante hasta que el enemigo esté vivo."
L["ObjectiveAndRespawn_Width_Desc"] = "Ancho del icono objetivo"
L["ObjectiveAndRespawnSettings_Desc"] = "Ajustes con respecto al objetivo del campo de batalla."
L["ObjectiveAndRespawnTimer"] = "Objetivo y temporizador de reaparición"
L["OffsetX"] = "Desplazamiento X"
L["OffsetY"] = "Desplazamiento Y"
L["OldVersion"] = "Versión anterior"
L["OutOfCombat"] = "Fuera de combate"
L["PlayerCount"] = "Contador de jugador activado"
L["PlayerCount_Enabled_Desc"] = "Si debe haber un texto para el número actual de jugadores. Tenga en cuenta que este recuento puede ser diferente al número actual de filas que se muestran debido a la protección contra el combate o al dejar aliados/enemigos en un RBG"
L["Point"] = "Punto"
L["PointAtObject"] = "Apuntar al objeto"
L["Poison"] = "Veneno"
L["Position"] = "Posición"
L["PowerBar"] = "Barra de poder"
L["PowerBar_Background_Desc"] = "Color de fondo para la barra de poder."
L["PowerBar_Enabled"] = "Habilitar barras de recursos"
L["PowerBar_Enabled_Desc"] = "Activa esta opción si no quieres ver barras de recursos como maná, ira, etc."
L["PowerBar_Height_Desc"] = "La altura de la barra de recursos. Hacer que la barra de recursos sea más alta automáticamente hará que la barra de salud sea más corta"
L["PowerBar_Texture_Desc"] = "La textura utilizada para las barras de poder"
L["PowerBarSettings_Desc"] = "Configuración de la barra de recursos"
L["PriorityBuffs"] = "Mejoras priorizadas"
L["PriorityDebuffs"] = "Debilitamientos priorizados"
L["Racial"] = "Racial"
L["Racial_Enabled"] = "Habilitar Raciales"
L["Racial_Enabled_Desc"] = "Cuando esté habilitado, habrá iconos para raciales usadas"
L["Racial_Width_Desc"] = "El ancho de la racial."
L["RacialFiltering_Enabled_Desc"] = "Cuando esta opción está habilitada, solo verá las raciales que están marcadas en la lista de filtrado."
L["RacialFiltering_Filterlist_Desc"] = "Haga clic para rastrear o dejar de rastrear esta racial"
L["RacialFilteringSettings_Desc"] = "Aquí puede seleccionar que racial desea rastrear"
L["RacialSettings"] = "Raciales"
L["RacialSettings_Desc"] = "Configuraciones relacionadas con las raciales"
L["RangeIndicator_Alpha"] = "Indicador de racial Alfa"
L["RangeIndicator_Alpha_Desc"] = "El valor alfa del marco enemigo cuando el enemigo está fuera del rango elegido"
L["RangeIndicator_Enabled"] = "Habilitar indicador de rango"
L["RangeIndicator_Enabled_Desc"] = "Cuando esta opción está habilitada, el marco usará el valor alfa elegido cuando un enemigo esté fuera del rango elegido"
L["RangeIndicator_Everything"] = "Modificar el alfa de todos los elementos"
L["RangeIndicator_Frames"] = "Modificar valor alfa de"
L["RangeIndicator_Frames_Desc"] = "Establece qué partes de la fila enemiga deben mostrarse con mayor transparencia cuando el enemigo está fuera de alcance."
L["RangeIndicator_Range"] = "Rango"
L["RangeIndicator_Range_Desc"] = "Cuando un enemigo está a más de esta distancia (en yardas) de distancia, su marco usará el valor alfa elegido"
L["RangeIndicator_Settings"] = "Indicador de rango"
L["RangeIndicator_Settings_Desc"] = "Aquí puede configurar el indicador de rango."
L["RBGSpecificSettings"] = "Campo de batalla puntuado"
L["RBGSpecificSettings_Desc"] = "Esta configuración solo se aplica en un RBG (campo de batalla puntuado)"
L["ResetModule"] = "Reiniciar módulo"
L["ResetModule_Desc"] = "Haga clic aquí para reiniciar los ajustes de este módulo para %s:%s"
L["ResetModuleSettings"] = "Restablecer la configuración de este módulo"
L["RIGHT"] = "Derecha"
L["Rightwards"] = "A la derecha"
L["RoleIcon_Enabled"] = "Habilitar iconos de roles"
L["RoleIcon_Enabled_Desc"] = "Habilite esta opción si desea ver un icono de rol en la barra de salud"
L["RoleIcon_Size_Desc"] = "El tamaño de los iconos de roles (ancho y alto)"
L["RoleIconSettings"] = "Icono de rol"
L["RoleIconSettings_Desc"] = "Configuraciones para el icono de rol que se muestra en la barra de salud"
L["Scale"] = "Escala"
L["ShowDispellable"] = "Mostrar disipable"
L["ShowDispellable_Desc"] = "Muestre %s que puedo eliminar por maldiciones, purgando, etc."
L["ShowMine"] = "Mostrar el mío"
L["ShowMine_Desc"] = "Mostrar %s que apliqué"
L["ShowNumbers"] = "Mostrar números"
L["ShowNumbers_Desc"] = "Mostrar tiempos de reutilización, tenga en cuenta que esta opción puede no funcionar si utiliza complementos como OmniCC"
L["ShowRealmnames"] = "Mostrar nombre de Reinos"
L["ShowRealmnames_Desc"] = "Muestra el Reino del enemigo"
L["ShowTooltips"] = "Mostrar Consejos"
L["ShowTooltips_Desc"] = "Muestre consejos cuando mueva el mouse sobre objetos como iconos de abalorio y auras"
L["Side"] = "Lado"
L["SideAtObject"] = "Lado en el objeto"
L["Size"] = "Tamaño"
L["SmallCastingBar"] = "Barra de lanzamiento pequeña"
L["SourceFilter"] = "Filtrar por fuente"
L["Spec_AuraDisplay_Enabled"] = "Habilitar visualización de aura"
L["Spec_AuraDisplay_Enabled_Desc"] = "Cuando esta opción está habilitada, verá iconos para control de multitudes e interrupciones (cuando está activo) en lugar del icono de especialización"
L["Spec_Enabled"] = "Habilitar especialización"
L["Spec_Enabled_Desc"] = "Cuando esté habilitado, habrá iconos para la especialización de los jugadores"
L["Spec_Width_Desc"] = "El ancho del icono de especialización."
L["SpecSettings"] = "Especialización"
L["SpecSettings_Desc"] = "Ajustes cambiando el icono de especializaciones."
L["SpellID_Filtering"] = "Filtrar por ID de hechizo"
L["Style"] = "Estilo"
L["SymbolicTargetindicator_Enabled_Desc"] = "Muestra un icono cuadrado de color de clase para cada %s que apunta a un jugador."
L["TargetAmount"] = "conteo de objetivos"
L["TargetAmount_Ally"] = "El sonido se reproduce tan pronto como un miembro del grupo es atacado por más del número establecido de oponentes."
L["TargetAmount_Me"] = "El sonido se reproduce tan pronto como eres objetivo de más oponentes que el número establecido."
L["TargetCallerUpdated"] = "es ahora el que marca el objetivo"
L["TargetCalling"] = "Marca los Objetivos"
L["TargetCallingNotificationEnable"] = "Habilitar notificación de sonido"
L["TargetCallingNotificationEnable_Desc"] = "Te notifica cuando hay un nuevo objetivo (detectado a través del icono de la calavera del objetivo del asalto)"
L["TargetCallingSetMark"] = "Marcar mi objetivo con el icono de calavera"
L["TargetCallingSetMark_Desc"] = "Marca tu objetivo actual automáticamente con el icono de la calavera cuando eres el líder del Asalto y no hay ningún asistente del Asalto en tu grupo. También marcará tu objetivo actual si no eres el líder del Asalto pero eres un asistente."
L["TargetCallingShowIcon"] = "Mostrar un icono en el objetivo de la persona que dirige"
L["TargetCallingShowIcon_Desc"] = "Cuando esté habilitado, verá un icono en el marco que muestra el destino de la persona que dirige"
L["TargetCallingSoundEnable"] = "Notificarme sobre un nuevo objetivo"
L["TargetCallingSoundEnable_Desc"] = "Cuando está habilitado, habrá un sonido cuando la persona que dirige, cambie el foco a un nuevo jugador"
L["TargetCallingSoundSound"] = "Sonido de notificación"
L["TargetIndicator"] = "Indicador de objetivo"
L["TargetIndicator_Desc"] = "Indicadores objetivo numéricos y simbólicos"
L["TargetIndicatorNumeric"] = "Indicador numérico de objetivo"
L["TargetIndicatorSymbolic"] = "Indicador de destino simbólico"
L["Testmode_Toggle"] = "Alternar modo de prueba"
L["Testmode_Toggle_Desc"] = "Activa/Desactiva el modo de prueba. Este modo facilita ver qué impacto tienen la mayoría de las opciones. No es posible simular todas las opciones, pero la mayoría de ellas están cubiertas en este sofisticado modo de prueba. El modo de prueba está deshabilitado si actualmente estás en un campo de batalla."
L["Testmode_ToggleAnimation"] = "Alternar animación de modo de prueba"
L["Testmode_ToggleAnimation_Desc"] = "Habilita/deshabilita la animación del modo de prueba para que pueda concentrarse más en una configuración específica sin distraerse con la animación."
L["Testmode_UseTeammates"] = "Usar miembros del grupo para probar"
L["Testmode_UseTeammates_Desc"] = "Cuando está habilitado, el modo de prueba utilizará los miembros actuales de su grupo en lugar de los falsos"
L["TestmodeDisabled"] = "Modo de prueba deshabilitado"
L["TestmodeEnabled"] = "Modo de prueba habilitado"
L["TestmodeSettings"] = "Modo de prueba"
L["Text"] = "Texto"
L["TextSettings"] = "Configuración de texto"
L["Thick"] = "Grueso"
L["TOP"] = "Arriba"
L["TOPLEFT"] = "Arriba Izquierda"
L["TOPRIGHT"] = "Arriba Derecha"
L["Trinket"] = "Abalorio"
L["Trinket_Enabled"] = "Habilitar Abalorios"
L["Trinket_Enabled_Desc"] = "Cuando esté habilitado, habrá iconos para el Medallón del gladiador, Adaptación e Implacable."
L["Trinket_Width_Desc"] = "El ancho del Abalorio."
L["TrinketSettings_Desc"] = "Configuraciones relacionadas con los Abalorios"
L["UpdatePeriod"] = "Periodo de actualización"
L["UpdatePeriod_Desc"] = "El tiempo entre las comprobaciones de combate en segundos, no lo ponga demasiado bajo porque esto puede causar problemas de rendimiento."
L["Upwards"] = "Hacía arriba"
L["UseBarHeight"] = "Usar la altura de la barra"
L["UseBigDebuffsPriority"] = "Usar grandes prioridades de debilitamientos (Debuffs)"
L["UseBigDebuffsPriority_Desc"] = "Si habilita esta opción, el complemento utilizará las prioridades del complemento de grandes debilitamientos en su lugar (BigDebuffs). Las prioridades se utilizan para los módulos %s, %s y %s"
L["UseButtonHeight"] = "Usar la altura del botón"
L["UseButtonHeight_Desc"] = "Si habilita esta opción, el módulo utilizará la altura de los botones."
L["UserDefined"] = "Definido por el usuario"
L["VerticalGrowdirection"] = "Dirección de crecimiento vertical"
L["VerticalGrowdirection_Desc"] = "Te permite elegir si las barras crecen hacia arriba o hacia abajo"
L["VerticalPosition"] = "Posición vertical"
L["VerticalSpacing"] = "Espaciado vertical"
L["Width"] = "Ancho"

elseif LOCALE_frFR then
L["AddPoint"] = "Ajouter un point"
L["Allies"] = "Alliés "
L["allies"] = "alliés "
L["ally"] = "Allié"
L["Ally"] = "Allié"
L["AllyIsTargeted"] = "Un allié est ciblé"
L["AllyJoined"] = "Un allié a rejoint le champ de bataille "
L["AllyLeft"] = "Un allié a quitté le champ de bataille "
L["AND"] = "et "
L["AttachToObject"] = "Attacher à l'objet "
L["AttachToObject_Desc"] = "Notez qu'il est possible que certains objets soient impossibles à sélectionner, ceci dépendant de la position de vos autres cadres. Par exemple, si vous ancrez les talents JcJ aux talents raciaux et que vous essayez d'ancrer les talents raciaux aux talents JcJ, vous rencontrerez un message d'erreur."
L["Auras_Buffs_Container_Color_Desc"] = "La couleur de la bordure du cadre des icônes de buff"
L["Auras_CustomFiltering_Conditions_All"] = "Toutes"
L["Auras_CustomFiltering_Conditions_Any"] = "Quelconque"
L["Auras_CustomFiltering_ConditionsMode"] = "Mode de conditions"
L["Auras_CustomFiltering_ConditionsMode_Desc"] = [=[Ici vous pouvez choisir comment les conditions s'appliquent aux auras.
Si vous choisissez "Quelconque" l'aura apparaîtra dès qu'une condition sélectionnée est remplie.
Si vous choisissez "Toutes" l'aura apparaîtra seulement si elle remplit toutes les conditions sélectionnées.]=]
L["Auras_Debuffs_Container_Color_Desc"] = "La couleur de la bordure du cadre des icônes de débuff"
L["Auras_Enabled"] = "Activer les auras "
L["Auras_Enabled_Desc"] = "Des icônes d'aura apparaîtront une fois activé. Il y aura également un compteur pour le temps restant ainsi que pour le nombre d'applications empilées "
L["Auras_Filtering_Mode"] = "Mode de filtrage"
L["Auras_Filtering_Mode_Desc"] = "Ici, vous pouvez choisir le filtre qui sera appliqué à vos auras. Vous pouvez choisir entre un filtre Blizzlike appliquant le même filtre que celui que Blizzard utilise sur les fenêtres de raid, et des conditions de filtrage personnalisées"
L["Auras_ShowTooltips"] = "Afficher les tooltips de sorts"
L["AurasCustomConditions"] = "Conditions personnalisées"
L["AurasFiltering_AddSpellID"] = "ID de sort"
L["AurasFiltering_AddSpellID_Desc"] = "Entrer un ou plusieurs spellIDs séparés par une virgule. Ces spellIDs seront ajoutées à votre liste de filtres. Seules les auras dont le spellID est inclus dans la liste de filtres seront visibles dès lors que le filtre par spellID est actif."
L["AurasFiltering_Enabled_Desc"] = "Quand cette option est activée, seuls les %s ajoutés à la liste de filtres seront visible"
L["AurasFiltering_Filterlist_Desc"] = "Cliquez pour enlever le %s"
L["AurasFilteringSettings_Desc"] = "Ici vous pouvez gérer quelles auras sont affichées"
L["AurasSettings"] = "Auras"
L["AurasSettings_Desc"] = "Paramètres des auras (buffs et débuffs)"
L["AurasStacktextSettings"] = "Texte des stacks"
L["BarBackground"] = "Couleur du fond "
L["BarHeight_Desc"] = "La hauteur de chaque rangée. Cette option n'est pas disponible en combat."
L["BarSettings"] = "Paramètres des barres"
L["BarSettings_Desc"] = "Ici vous pouvez configurer les rangées "
L["BarTexture"] = "Texture des barres "
L["BarWidth_Desc"] = "La largeur de chaque rangée. Cette option n'est pas accessible en combat. "
L["BattlegroundSize"] = "Taille du champ de bataille "
L["BGSize_15"] = "1–15 Joueurs"
L["BGSize_15_Desc"] = "Ces options s'appliquent aux %s dans un champ de bataille de 1 à 15 joueurs par faction"
L["BGSize_40"] = "16–40 joueurs"
L["BGSize_40_Desc"] = "Ces options s'appliquent aux %s dans un champ de bataille de 16 à 40 joueurs par faction"
L["BGSize_5"] = "Arène"
L["BGSize_5_Desc"] = "Ces options s'appliquent pour %s en arènes"
L["BlizzlikeAuraFiltering"] = "Utiliser le filtre de fenêtres de raid de Blizzard"
L["BorderSize"] = "Taille de la bordure"
L["BorderThickness"] = "Taille de la bordure "
L["BOTTOM"] = "En Bas"
L["BOTTOMLEFT"] = "En Bas à Gauche"
L["BOTTOMRIGHT"] = "En Bas à Droite"
L["buff"] = "buff"
L["BuffContainer"] = "Cadre de buffs"
L["BuffIcon"] = "Icône de buff "
L["Buffs"] = "Buffs"
L["Button"] = "Bouton"
L["CantAnchorToItself"] = "Ne peut pas s'ancrer à soi-même !"
L["CENTER"] = "Centre"
L["Color"] = "Couleur"
L["Coloring_Enabled"] = "Couleur par type de débuff "
L["Coloring_Enabled_Desc"] = "Quand cette option est activée, vous pouvez choisir si le contour du cadre ou le texte de compte à rebours doit être coloré par type de débuff (Comme ils sont connus dans l'interface de base, les poisons sont verts, les maladies bleues etc.) "
L["Columns"] = "Colonnes"
L["Columns_Desc"] = "En combien de colonnes les joueurs sont affichés."
L["Combat"] = "En Combat"
L["CombatIndicator"] = "Indicateur de combat"
L["CombatLogScanningForEnemiesDisabled"] = "Désactiver les journaux de combats pour les ennemis"
L["CombatLogScanningForEnemiesEnabled"] = "Scanner le journal de combat pour détecter les ennemis a été activé"
L["ConfirmProfileOverride"] = "Êtes-vous sûr de vouloir écraser le sous-profil %s avec le sous-profil %s ?"
L["Container_Color"] = "Couleur de la bordure du cadre"
L["ContainerPosition"] = "Position du cadre"
L["ContainerSettings"] = "Page des réglages"
L["ConvertCyrillic"] = "Convertir le cyrillique "
L["ConvertCyrillic_Desc"] = "Convertit les caractères cyrilliques, ce qui rend plus facile la lecture des noms si vous affrontez des joueurs Russes"
L["CopySettings"] = "Copier les paramètres à partir de"
L["Countdowntext"] = "Texte du compte à rebours"
L["Covenant"] = "Icône de congrégation"
L["CovenantIcon_Enabled"] = "Activer les icônes de congrégation"
L["CovenantIcon_Enabled_Desc"] = "Activez cette option si vous voulez voir l'icône de congrégation sur la barre de vie"
L["CovenantIcon_Size_Desc"] = "La taille des icônes de congrégation (longueur et largeur)"
L["CovenantIconSettings_Desc"] = "Paramètres de l'icône de congrégation affichée sur la barre de vie"
L["CurrentVersion"] = "Version actuelle"
L["Curse"] = "Malédiction"
L["CustomMacro_Desc"] = "Cette option permet de configurer une macro personnalisée. Un champ vide sera interprété comme une macro vide. Substitutions disponibles : %n est substitué par le nom de l'ennemi Exemple : \"targetexact %n /cast Polymorph /say Polymorph on %n /targetlasttarget\" Cette macro va cibler l'ennemi, lancer Polymorph, écrire un message dans le canal Dire et cibler la cible précédente. Noter que les macro ne peuvent contenir que 255 caractères (substitutions incluse)."
L["Data"] = "Donnés"
L["debuff"] = "débuff"
L["DebuffContainer"] = "Cadre de débuff"
L["DebuffIcon"] = "Icône de débuff"
L["Debuffs"] = "Débuffs"
L["DebuffType_Filtering"] = "Filtrer par type de débuff"
L["DebuffType_Filtering_Desc"] = "Quand cette option est activée, ne seront visibles que les affaiblissements du type choisi. Cette option n'est pas simulée par le mode test puisque cela requerrait une énorme base de donnée avec une abondante quantité de sorts ainsi que leur type d'affaiblissement."
L["DeletePoint"] = "Supprimer Point %s"
L["DisableArenaFramesInArena"] = "Désactiver le cadre d'arène en arène"
L["DisableArenaFramesInArena_Desc"] = "Désactiver le cadre d'arène en arène. Fonctionne aussi quand sArena est installé"
L["DisableArenaFramesInBattleground"] = "Désactiver le cadre d'arène en champ de bataille"
L["DisableArenaFramesInBattleground_Desc"] = "Désactive les cadres d'arène dans un champ de bataille. Fonctionne également quand sArena est installé."
L["Disease"] = "Maladie"
L["DispellFilter"] = "Filtrer par nature du debuff"
L["DisplayType"] = "Type d'affichage "
L["Downwards"] = "Vers le bas"
L["DR_Disorient"] = "Désorienter"
L["DR_Incapacitate"] = "Incapaciter"
L["DR_Knockback"] = "Repousser"
L["DR_Root"] = "Enraciner"
L["DR_Silence"] = "Mettre au silence"
L["DR_Stun"] = "Étourdir"
L["DRContainer"] = "Cadre des DR"
L["DRTracking"] = "Tracking des rendements décroisants"
L["DrTracking_Container_Color_Desc"] = "La couleur de la bordure du cadre des icônes de DR"
L["DrTracking_DisplayType_Desc"] = "Ici il est possible de choisir si le status du DR devrait être indiqué par un contour coloré autour de l'icône ou par la couleur du texte du compte à rebours. A Noter que le texte du compte à rebours ne fonctionnera que si aucun addon de type OmniCC modifiant la couleur de texte n'est utilisé."
L["DrTracking_Enabled"] = "Activer le tracking des DR"
L["DrTracking_Enabled_Desc"] = "Quand activé, des icônes pour le DR Tracking apparaîtront à côté de la ligne des ennemis. Contour vert : la durée du prochain CC sera de moitié. Contour jaune : la durée du prochain CC sera de quart. Contour rouge :  le prochain CC ne sera pas appliqué car le joueur est immunisé"
L["DrTracking_Spacing"] = "Espacement des DR"
L["DrTracking_Spacing_Desc"] = "L'espacement entre chaque icône de DR"
L["DrTrackingFiltering_Enabled_Desc"] = "Quand cette option est activée, ne seront visibles que les icônes des sorts dont la catégorie est cochée dans la liste de filtre"
L["DrTrackingFiltering_Filterlist_Desc"] = "Cliquer pour traquer ou non cette catégorie"
L["DrTrackingFilteringSettings_Desc"] = "Ici vous pouvez choisir quelle catégorie de DR afficher."
L["DrTrackingSettings"] = "Tracking des DR"
L["DrTrackingSettings_Desc"] = "Paramètres concernant le tracking des DR"
L["DurationFilter"] = "Filtrer par durée"
L["DurationFilter_OnlyShowWhenDuration"] = "Montrer uniquement quand la durée est inférieure ou égale à "
L["DurationFilter_OnlyShowWhenDuration_Desc"] = "N'importe quelle améliorations avec une durée supérieure que la durée sélectionnée ne sera pas affichée après le filtrage"
L["Enable_DrawSwipe"] = "Activer la texture de balayage"
L["Enable_DrawSwipe_Desc"] = "Affiche une texture désaturée et plus sombre pour la partie restante du temps de recharge."
L["EnableClique"] = "Utiliser Clique pour les raccourcis claviers"
L["EnableClique_Desc"] = "Lorsque cette option est activée, Clique gérera tous les raccourcis claviers."
L["enemies"] = "ennemis "
L["Enemies"] = "Ennemis "
L["EnemiesTargetingAllies_Enabled_Desc"] = "Quand cette option est activée, une notification sonore retentira quand la quantité d'ennemis configurée sont en train de cibler un allié"
L["EnemiesTargetingMe_Enabled_Desc"] = "Quand cette option est activée, une notification sonore retentira quand la quantité d'ennemis configurée sont en train de cibler le joueur"
L["enemy"] = "ennemi "
L["Enemy"] = "Ennemi "
L["EnemyJoined"] = "Un ennemi a rejoint le champ de bataille "
L["EnemyLeft"] = "Un ennemi a quitté le champ de bataille "
L["ExportButton"] = "Exporter profile"
L["ExportButton_Desc"] = "Cliquer ici pour exporter votre profil actif"
L["FakeEventsDisabled"] = "Ne plus créer de faux évènements"
L["FakeEventsEnabled"] = "Est en train de créer de faux évènements pour tester"
L["Filtering_Enabled"] = "Activer le filtrage"
L["Filtering_Filterlist"] = "Liste des filtres"
L["FilterSettings"] = "Paramètres des filtres"
L["Font"] = "Police"
L["Font_Desc"] = "La police principale utilisée par cet addon "
L["Font_Outline"] = "Contour de la police"
L["Font_Outline_Desc"] = "Le contour de la police"
L["Fontcolor"] = "Couleur de la police"
L["Fontcolor_Desc"] = "La couleur du texte"
L["FontShadow_Enabled"] = "Activer l'ombre de la police"
L["FontShadow_Enabled_Desc"] = "Quand activée une ombre apparaîtra autour de la police. La couleur de l'ombre est configurable."
L["FontShadowColor"] = "Couleur de l'ombre"
L["FontShadowColor_Desc"] = "La couleur de l'ombre du texte"
L["FontSize"] = "Taille de la police"
L["FontSize_Desc"] = "La taille de la police du texte"
L["Frame"] = "Cadre"
L["Framescale"] = "Échelle"
L["Framescale_Desc"] = "L'échelle du cadre principal. Cette option n'est pas disponible en combat du fait de la protection des combats de Blizzard."
L["General"] = "Général"
L["GeneralSettings"] = "Paramètres généraux"
L["GeneralSettings_Desc"] = "Quelques paramètres généraux"
L["GeneralSettingsAllies"] = "Ces options s'appliquent aux alliées indépendamment de la taille du champ de batailler"
L["GeneralSettingsEnemies"] = "Ces options s'appliquent aux ennemis indépendamment de la taille du champ de batailler"
L["HealthBar"] = "Barre de vie"
L["HealthBar_Background_Desc"] = "Couleur de fond de la barre de vie"
L["HealthBar_Texture_Desc"] = "La texture utilisée pour les barres de vie"
L["HealthBarSettings_Desc"] = "Paramètres pour la barre de vie"
L["HealthTextEnabled"] = "Activer le texte des PV"
L["HealthTextSettings"] = "Réglages du texte des PV"
L["HealthTextType"] = "Type du texte des PV"
L["Height"] = "Hauteur"
L["HighestPriorityAura"] = "Aura avec la plus importante priorité"
L["Highlight_Color"] = "Couleur de surbrillance"
L["Highlight_Color_Desc"] = "La couleur du cadre du joueur sur lequel le curseur se trouve actuellement"
L["HorizontalGrowdirection"] = "Direction de l'extension horizontale"
L["HorizontalGrowdirection_Desc"] = "Laisse la possibilité de choisir si les lignes s'étendent sur le côté gauche ou droit."
L["HorizontalSpacing"] = "Espacement horizontal"
L["IAmTargeted"] = "Je suis ciblé"
L["Icon"] = "Icône"
L["IconSize"] = "Taille de l'icône"
L["IconsPerRow"] = "Icônes par ligne"
L["Import"] = "Importer"
L["ImportButton"] = "Importer profile"
L["ImportButton_Desc"] = "Cliquez ici pour importer un profil à partir d'une chaîne de caractère et écraser votre profile actif actuel"
L["JustifyH"] = "Alignement horizontal"
L["JustifyH_Desc"] = "L'alignement horizontal du texte"
L["JustifyV"] = "Alignement vertical"
L["JustifyV_Desc"] = "L'alignement vertical du texte"
L["KeybindSettings_Desc"] = "Ici il est possible de configurer ce qu'il devrait arriver lors d'un clique-gauche, d'un clique-droit et d'un clique-milieu"
L["LEFT"] = " À Gauche"
L["LeftToTargetCounter"] = "Compteur de cible à gauche"
L["Leftwards"] = "Vers la gauche"
L["Level"] = "Niveau du joueur"
L["LevelText_Enabled"] = "Afficher le niveau du joueur"
L["LevelText_OnlyShowIfNotMaxLevel"] = "Afficher uniquement si pas niveau max."
L["Locked"] = "Verrouillé"
L["Locked_Desc"] = "Verrouille le cadre"
L["Magic"] = "Magie"
L["MainFrameSettings"] = "Options du cadre principal"
L["MainFrameSettings_Desc"] = "Options du cadre principal utilisé par les %s"
L["MaxPlayers"] = "Ennemis maximum"
L["MaxPlayers_Desc"] = "L'addon ne montrera pas les ennemis s'il y a plus d'ennemis que ce nombre"
L["MIDDLE"] = "Milieu"
L["ModuleSettings"] = "Réglages du module"
L["ModuleSpecificSettings"] = "Réglages spécifiques du module"
L["MoreProfileOptions"] = "Plus d'options de profile"
L["MyFocus"] = "Mon Focus"
L["MyFocus_Color"] = "Couleur de la focalisation"
L["MyFocus_Color_Desc"] = "La couleur utilisée pour la bordure indiquant votre focalisation actuelle"
L["MyTarget"] = "Ma Cible"
L["MyTarget_Color"] = "Couleur de la cible"
L["MyTarget_Color_Desc"] = "La couleur utilisée pour la bordure de la barre vie de votre cible actuelle"
L["MyVersion"] = "Vous utilisez la version"
L["Name"] = "Nom"
L["Name_Desc"] = "Paramètres pour les noms sur la barre de vie"
L["NewVersionAvailable"] = "Une nouvelle version est disponible"
L["None"] = "Aucun"
L["NonPriorityBuffs"] = "Améliorations non priorisées"
L["NonPriorityDebuffs"] = "Debuffs non priorisés"
L["Normal"] = "Normal"
L["NormalCastingBar"] = "Barre d'incantation normale"
L["NotAvailableInCombat"] = "Ce paramètre n'est pas disponible en combat"
L["Notifications_Allies_Enabled_Desc"] = "Quand activé vous serez notifié quand un allié entre ou sort d'un champ de bataille."
L["Notifications_Enabled"] = "Activer les notifications"
L["Notifications_Enemies_Enabled_Desc"] = "Quand activé vous serez notifié quand un ennemi quitte ou rejoint le champ de bataille"
L["NoVersion"] = "Aucun add-on trouvé"
L["NumericTargetindicator_Enabled_Desc"] = "Affiche combien d'%s ciblent un joueur"
L["ObjectiveAndRespawn_ObjectiveEnabled"] = "Afficher les objectifs"
L["ObjectiveAndRespawn_ObjectiveEnabled_Desc"] = "Affiche l'icône du drapeau, du char de mine et des transporteurs d'orbes à côté du bouton de l'ennemi."
L["ObjectiveAndRespawn_Position"] = "Position"
L["ObjectiveAndRespawn_Position_Desc"] = "Si l'icône doit être affiché à gauche de l'icône de spécialisation ou à droite des icônes des trinkets/raciaux."
L["ObjectiveAndRespawn_RespawnEnabled"] = "Activer le chronomètre de réapparition"
L["ObjectiveAndRespawn_RespawnEnabled_Desc"] = "Quand actif, affiche une icône indiquant le temps jusqu'à ce que l'ennemie soit en vie."
L["ObjectiveAndRespawn_Width_Desc"] = "Largeur de l'icône d'objectif"
L["ObjectiveAndRespawnSettings_Desc"] = "Options relatives aux objectifs de champ de bataille"
L["ObjectiveAndRespawnTimer"] = "Objectif"
L["OffsetX"] = "Décalage en X"
L["OffsetY"] = "Décalage en Y"
L["OldVersion"] = "Ancienne version"
L["OutOfCombat"] = "Hors combat"
L["PlayerCount"] = "Compteur de joueur"
L["PlayerCount_Enabled_Desc"] = "Doit-il y avoir un texte pour le nombre actuel de joueurs ? A noter que ce nombre peut être différent du nombre actuel de lignes affichées en raison de protection de combat ou d'alliés/ennemies quittant le champs de bataille."
L["Point"] = "Point"
L["PointAtObject"] = "Cibler l'objet"
L["Poison"] = "Poison"
L["Position"] = "Position"
L["PowerBar"] = "Barre de puissance"
L["PowerBar_Background_Desc"] = "Couleur de fond pour la barre de ressource"
L["PowerBar_Enabled"] = "Activer les barres de ressource"
L["PowerBar_Enabled_Desc"] = "Activez cette option si vous voulez voir les barres de ressources comme la mana, la rage etc."
L["PowerBar_Height_Desc"] = "La hauteur de la barre de ressource. Rendre la barre de ressource plus haute réduira automatiquement la hauteur de la barre de vie."
L["PowerBar_Texture_Desc"] = "La texture utilisée pour les barres de puissance"
L["PowerBarSettings_Desc"] = "Paramètres pour la barre de ressource"
L["PriorityBuffs"] = "Améliorations priorisées"
L["PriorityDebuffs"] = "Debuffs priorisés "
L["Racial"] = "Racial"
L["Racial_Enabled"] = "Activer les raciaux"
L["Racial_Enabled_Desc"] = "Quand activé, il y aura des icônes pour les raciaux utilisés"
L["Racial_Width_Desc"] = "La largeur du racial"
L["RacialFiltering_Enabled_Desc"] = "Quand cette option est activée vous ne verrez que les sorts raciaux qui sont cochés dans la liste de filtrage"
L["RacialFiltering_Filterlist_Desc"] = "Cliquer pour suivre / ne pas suivre ce racial"
L["RacialFilteringSettings_Desc"] = "Ici vous pouvez sélectionner quel racial traquer"
L["RacialSettings"] = "Raciaux"
L["RacialSettings_Desc"] = "Paramètres des raciaux"
L["RangeIndicator_Alpha"] = "Alpha"
L["RangeIndicator_Alpha_Desc"] = "L'alpha du cadre ennemi lorsque l'ennemi est hors de la portée choisie"
L["RangeIndicator_Enabled"] = "Activer l'indicateur de portée"
L["RangeIndicator_Enabled_Desc"] = "Lorsque cette option est activée, le cadre utilisera l'alpha choisi lorsqu'un ennemi est hors d'une portée choisie"
L["RangeIndicator_Everything"] = "Modifier l'alpha de tous les éléments"
L["RangeIndicator_Frames"] = "Modifier l'alpha de"
L["RangeIndicator_Frames_Desc"] = "Définissez les parties de la ligne ennemie qui doivent être affichées avec une plus grande transparence lorsque l'ennemi est hors de portée."
L["RangeIndicator_Range"] = "Portée"
L["RangeIndicator_Range_Desc"] = "Lorsqu'un ennemi est à plus de cette distance (en mètres), son cadre utilisera l'alpha choisi"
L["RangeIndicator_Settings"] = "Indicateur de portée"
L["RangeIndicator_Settings_Desc"] = "Ici vous pouvez configurer l'indicateur de portée"
L["RBGSpecificSettings"] = "Champ de bataille coté"
L["RBGSpecificSettings_Desc"] = "Ces paramètres s'appliquent uniquement dans un champ de bataille coté (RBG)"
L["ResetModule"] = "Réinitialiser le module"
L["ResetModule_Desc"] = "Cliquez ici pour réinitialiser les paramètres du module pour %s:%s"
L["ResetModuleSettings"] = "Réinitialiser les paramètres de ce module"
L["RIGHT"] = "À Droite"
L["Rightwards"] = "Vers la droite"
L["RoleIcon_Enabled"] = "Activer les icônes de rôles"
L["RoleIcon_Enabled_Desc"] = "Activez cette option si vous souhaitez voir une icône de rôle sur la barre de santé"
L["RoleIcon_Size_Desc"] = "La taille des icônes de rôles (largeur et longueur)"
L["RoleIconSettings"] = "Icône de rôle"
L["RoleIconSettings_Desc"] = "Paramètres de l'icône de rôle qui s'affiche sur la barre de santé"
L["Scale"] = "Échelle"
L["ShowDispellable"] = "Afficher les auras dissipable"
L["ShowDispellable_Desc"] = "%s est remplacé par \"Buffs\" ou \"Debuffs\""
L["ShowMine"] = "Afficher le mien"
L["ShowMine_Desc"] = "Afficher les %s que j'ai appliqués"
L["ShowNumbers"] = "Afficher les nombres"
L["ShowNumbers_Desc"] = [=[Affiche le temps en nombre pour les rechargements, notez que cette option peut ne pas marcher si vous utilisez des addons tels que OmniCC
]=]
L["ShowRealmnames"] = "Afficher les noms de royaumes"
L["ShowRealmnames_Desc"] = "Afficher le royaume de l'ennemi"
L["ShowTooltips"] = "Afficher les tooltips"
L["ShowTooltips_Desc"] = "Affiche les informations quand vous passez la souris au dessus des objets tels que les bijoux ou les icones d'améliorations/affaiblissements"
L["Side"] = "Le Côté "
L["SideAtObject"] = "Coté de l'objet"
L["Size"] = "Taille"
L["SmallCastingBar"] = "Petite barre d'incantation"
L["SourceFilter"] = "Filtrer par source"
L["Spec_AuraDisplay_Enabled"] = "Activer l'affichage des auras"
L["Spec_AuraDisplay_Enabled_Desc"] = [=[Quand cette option est activée, vous verrez des icônes pour le contrôle des foules et les interruptions (lorsqu'elles sont actives) au lieu de l'icône choisie par défaut

]=]
L["Spec_Enabled"] = "Activer la spécialisation"
L["Spec_Enabled_Desc"] = "Lorsqu'il est activé, les icônes de spécialisation des joueurs sont affichés"
L["Spec_Width_Desc"] = "La largeur de l'icône de spécialisation"
L["SpecSettings"] = "Spécialisation"
L["SpecSettings_Desc"] = "Paramètres modifiant l'icône choisie par défaut."
L["SpellID_Filtering"] = "Filtrer par ID de sort"
L["Style"] = "Style"
L["SymbolicTargetindicator_Enabled_Desc"] = "Affiche un carré avec la couleur de classe de chaque %s qui cible un joueur."
L["TargetAmount"] = "nombre de cibles"
L["TargetAmount_Ally"] = "Le son est joué dès qu'un membre du groupe est ciblé par plus d'adversaires que le nombre défini."
L["TargetAmount_Me"] = "Le son est joué dès que vous êtes ciblé par plus d'adversaires que le nombre défini."
L["TargetCallerUpdated"] = "est maintenant le target caller"
L["TargetCalling"] = "Target Calling"
L["TargetCallingNotificationEnable"] = "Activer les notifications sonores"
L["TargetCallingNotificationEnable_Desc"] = "Vous avertit lorsqu'il y a une nouvelle cible (détectée via l'icône du crâne de cible de raid)"
L["TargetCallingSetMark"] = "Marquer ma cible avec l'icône crâne"
L["TargetCallingSetMark_Desc"] = "Marque automatiquement votre cible actuelle avec l'icône crâne lorsque vous êtes le chef de raid et qu'il n'y a pas d'assistant de raid dans votre groupe. Il marquera également votre cible actuelle si vous n'êtes pas le chef de raid mais que vous êtes un assistant."
L["TargetCallingShowIcon"] = "Montre une icône sur la cible du fésignateur de cible."
L["TargetCallingShowIcon_Desc"] = "Quand activé, vous verrez une icône sur le cadre affichant la cible du target caller."
L["TargetCallingSoundEnable"] = "Me notifier sur une nouvelle cible"
L["TargetCallingSoundEnable_Desc"] = "Quand activé, il y aura un son quand le target caller cible un nouveau joueur"
L["TargetCallingSoundSound"] = "Notification sonore"
L["TargetIndicator"] = "Indicateur de cible "
L["TargetIndicator_Desc"] = "Indicateurs de cible numérique et symbolique"
L["TargetIndicatorNumeric"] = "Compteur de cibles"
L["TargetIndicatorSymbolic"] = "Symboles de cibles"
L["Testmode_Toggle"] = "Basculer en mode test"
L["Testmode_Toggle_Desc"] = "Active/Désactive le mode de test. Ce mode permet facilement de voir l'impact de la plupart des options. Il n'est pas possible de simuler toutes les options mais la majorité sont intégrées dans ce mode de test. Le mode de test est désactivé si vous êtes en champ de bataille."
L["Testmode_ToggleAnimation"] = "Active l'animation du mode de test"
L["Testmode_ToggleAnimation_Desc"] = "Active/Désactive l'animation du mode de test ce qui permet de se concentrer sur un paramètre précis sans être distrait par l'animation."
L["Testmode_UseTeammates"] = "Utiliser les membres du groupe pour les tests"
L["Testmode_UseTeammates_Desc"] = "Lorsqu'il est activé, le mode de test utilisera les membres actuels de votre groupe au lieu de faux joueurs"
L["TestmodeDisabled"] = "Mode test désactivé"
L["TestmodeEnabled"] = "Mode test activé"
L["TestmodeSettings"] = "Mode test"
L["Text"] = "Texte"
L["TextSettings"] = "Réglages texte"
L["Thick"] = "Épais"
L["TOP"] = "En Haut"
L["TOPLEFT"] = "En Haut à Gauche "
L["TOPRIGHT"] = "En Haut à Droite"
L["Trinket"] = "Bijou"
L["Trinket_Enabled"] = "Activer les bijoux"
L["Trinket_Enabled_Desc"] = "Quand actif, il y aura des icônes pour le Médaillon du Gladiateur, Adaptation et Implacable."
L["Trinket_Width_Desc"] = "La largeur du bijou."
L["TrinketSettings_Desc"] = "Paramétrage des bijoux"
L["UpdatePeriod"] = "Intervalle de mise à jour "
L["UpdatePeriod_Desc"] = "Le temps entre les vérifications des combats en secondes. Ne pas mettre trop bas cela peut causer des problèmes de performances"
L["Upwards"] = "Vers le haut"
L["UseBarHeight"] = "Utiliser la hauteur de la barre"
L["UseBigDebuffsPriority"] = "Utiliser les priorités de BigDebuffs"
L["UseBigDebuffsPriority_Desc"] = "Si vous activez cette option, l'addon utilisera les priorités de BigDebuffs à la place. Les priorités sont utilisés pour les modules %s, %s et %s"
L["UseButtonHeight"] = "Utiliser la hauteur des boutons "
L["UseButtonHeight_Desc"] = "Si vous activez cette option, le module utilisera la hauteur des boutons"
L["UserDefined"] = "Défini par l'utilisateur"
L["VerticalGrowdirection"] = "Direction d'agrandissement verticale"
L["VerticalGrowdirection_Desc"] = "Vous permet de choisir si les barres grandissent vers le haut ou vers le bas"
L["VerticalPosition"] = "Position verticale"
L["VerticalSpacing"] = "Espace vertical"
L["Width"] = "Largeur"

elseif LOCALE_itIT then
L["AddPoint"] = "Aggiungi Punto"
L["Allies"] = "Alleati"
L["allies"] = "alleati"
L["ally"] = "alleato"
L["Ally"] = "Alleato"
L["AllyIsTargeted"] = "Alleato Mirato"
L["AllyJoined"] = "un alleato si è unito al campo di battaglia"
L["AllyLeft"] = "un alleato ha lasciato il campo di battaglia"
L["AND"] = "e"
L["AttachToObject"] = "allega all'oggetto"
L["AttachToObject_Desc"] = "Nota che non è possibile selezionare qualunque oggetto, dipende dal posizionamento degli altri riquadri. Ad esempio, ancorando i talenti PvP alle abilità razziali e nel mentre si ancorano le stesse ai talenti PvP, comparirà un messaggio d'errore."
L["Auras_Buffs_Container_Color_Desc"] = "Colore del bordo del contenitore delle icone dei benefici"
L["Auras_CustomFiltering_Conditions_All"] = " filtraggio personalizzato tutte le condizioni"
L["Auras_CustomFiltering_Conditions_Any"] = "Alcuni"
L["Auras_CustomFiltering_ConditionsMode"] = " filtraggio personalizzato modalità condizioni"
L["Auras_CustomFiltering_ConditionsMode_Desc"] = "Qui puoi scegliere come applicare le condizioni sulle aure selezionate. Se scegli \"Qualsiasi\", l'addon mostrerà l'aura non appena viene soddisfatta una delle condizioni selezionate. Se scegli \"Tutto\", l'addon mostrerà un'aura solo se soddisfa tutte le condizioni selezionate."
L["Auras_Debuffs_Container_Color_Desc"] = "Il colore del bordo del contenitore delle icone dei debuff (penalità)"
L["Auras_Enabled"] = "Abilita Aure"
L["Auras_Enabled_Desc"] = "Quando abilitato saranno mostrate le icone delle aure. Ci sarà anche un numero per il tempo rimanente e uno per gli stack applicati"
L["Auras_Filtering_Mode"] = "Modalità di filtraggio"
L["Auras_Filtering_Mode_Desc"] = "Qui puoi scegliere quale filtro applicare alle aure. Puoi scegliere tra un filtro blizzlike che applica lo stesso filtro utilizzato da Blizzard sul frame del raid e condizioni di filtraggio personalizzate"
L["Auras_ShowTooltips"] = "Mostra i suggerimenti degli incantesimi"
L["AurasCustomConditions"] = "Condizioni Personalizzate"
L["AurasFiltering_AddSpellID"] = "ID Incantesimo"
L["AurasFiltering_AddSpellID_Desc"] = "Inserisci uno ID incantesimo o più, separati da virgole. Gli ID incantesimo vengono aggiunti all'elenco dei filtri. Solo le aure il cui ID incantesimo è nell'elenco dei filtri vengono mostrate quando il filtro ID incantesimo è abilitato"
L["AurasFiltering_Enabled_Desc"] = "Quando questa opzione è abilitata vedrai solo %s che hai aggiunto all'elenco dei filtri."
L["AurasFiltering_Filterlist_Desc"] = "Fai clic per rimuovere questo %s"
L["AurasFilteringSettings_Desc"] = "Qui puoi controllare quale aura viene mostrata."
L["AurasSettings"] = "Aure"
L["AurasSettings_Desc"] = "Impostazioni sulle aure (buff e debuff)."
L["AurasStacktextSettings"] = "Impila il testo."
L["BarBackground"] = "Colore di sfondo."
L["BarHeight_Desc"] = "L'altezza di ogni riga. Questa impostazione non è disponibile in combattimento a causa della protezione in combattimento di Blizzard"
L["BarSettings"] = "Impostazioni barra"
L["BarSettings_Desc"] = "Qui puoi modificare le righe di visualizzazione delle barre"
L["BarTexture"] = "Texture della barra"
L["BarWidth_Desc"] = "La larghezza di ogni riga. Questa impostazione non è disponibile in combattimento."
L["BattlegroundSize"] = "Per la modalità di test, puoi selezionare quale tipo del campo di battaglia (1-15 oppure 16-40 giocatore) che vuoi simulare"
L["BGSize_15"] = "1-15 giocatori"
L["BGSize_15_Desc"] = "Queste impostazioni si applicano a %s in un BG con 1–15 giocatori per squadra"
L["BGSize_40"] = "16-40 giocatori"
L["BGSize_40_Desc"] = "Queste impostazioni si applicano a %s in un BG con 16-40 giocatori per squadra"
L["BGSize_5"] = "Arena"
L["BGSize_5_Desc"] = "Queste impostazioni si applicano a %s nelle arene"
L["BlizzlikeAuraFiltering"] = "Usa il filtro raidframe di Blizzard"
L["BorderSize"] = "Dimensione del bordo"
L["BorderThickness"] = "La dimensione dei bordi"
L["BOTTOM"] = "In basso"
L["BOTTOMLEFT"] = "In basso a sinistra"
L["BOTTOMRIGHT"] = "In basso a destra"
L["buff"] = "buff (beneficio)"
L["BuffContainer"] = "Contenitore buff (beneficio)"
L["BuffIcon"] = "Icona buff (beneficio)."
L["Buffs"] = "Buffs (Benefici)"
L["Button"] = "Bottone"
L["CantAnchorToItself"] = "Non può ancorarsi a se stesso!"
L["CENTER"] = "Centro"
L["Color"] = "Colore"
L["Coloring_Enabled"] = "Colore per tipologia di debuff (penalità)"
L["Coloring_Enabled_Desc"] = "Quando questa opzione è attiva si potrà scegliere se il bordo del riquadro, o il testo del conteggio alla rovescia, dovranno colorarsi in base alla tipologia di debuff (come adottate dall'interfaccia predefinita, i veleni in verde, le malattie in blu, ecc.)"
L["Columns"] = "Colonne"
L["Columns_Desc"] = "In quante colonne vengono visualizzati i giocatori."
L["Combat"] = "In combattimento"
L["CombatIndicator"] = "Indicatore combattimento"
L["CombatLogScanningForEnemiesDisabled"] = "Scansione del registro di combattimento disabilitata per i nemici"
L["CombatLogScanningForEnemiesEnabled"] = "È stata abilitata la scansione del registro di combattimento per rilevare i giocatori nemici."
L["ConfirmProfileOverride"] = "Sei sicuro di voler sovrascrivere il sottoprofilo %s con il sottoprofilo %s "
L["Container_Color"] = "Colore del bordo del contenitore"
L["ContainerPosition"] = "Posizione contenitore"
L["ContainerSettings"] = "Impostazioni contenitore"
L["ConvertCyrillic"] = "Converti cirillico"
L["ConvertCyrillic_Desc"] = "Converte i caratteri cirillici che rendono più facile leggere i nomi se stai giocando contro i russi"
L["CopySettings"] = "Copia le impostazioni da"
L["Countdowntext"] = "Testo del conto alla rovescia"
L["Covenant"] = "Congrega"
L["CovenantIcon_Enabled"] = "Abilita le icone della congrega"
L["CovenantIcon_Enabled_Desc"] = "Abilita questa opzione se vuoi vedere un'icona della congrega sulla barra della salute"
L["CovenantIcon_Size_Desc"] = "La dimensione delle icone della congrega (larghezza e altezza)."
L["CovenantIconSettings_Desc"] = "Impostazioni per l'icona della congrega mostrata sulla barra della salute"
L["CurrentVersion"] = "Versione corrente"
L["Curse"] = "Maleficio"
L["CustomMacro_Desc"] = "Ciò consente di impostare una macro personalizzata. Un campo vuoto porterà a una macro vuota. Sostituzioni disponibili: %n viene sostituito con il nome del nemico Esempio: /targetexact %n /cast Metamorfosi /parla Metamorfosi su %n /targetlasttarget Questo bersaglia il nemico, lancia Metamorfosi, scrive un messaggio nel canale parla e prende di mira il tuo obiettivo precedente. Si noti che le macro possono essere lunghe solo 255 caratteri (incluse le sostituzioni)."
L["Data"] = "Dati"
L["debuff"] = "Debuff (penalità)"
L["DebuffContainer"] = "Contenitore Debuff (penalità)"
L["DebuffIcon"] = "Icona Debuff (penalità)"
L["Debuffs"] = "Debuffs (penalità)"
L["DebuffType_Filtering"] = "Filtra per tipo di debuff (penalità)"
L["DebuffType_Filtering_Desc"] = "Quando questa opzione è abilitata vedrai solo i debuff (penalità) del tipo scelto. Questa opzione non è simulata nella modalità test poiché ciò richiederebbe un enorme database con tonnellate di incantesimi e il loro tipo di debuff (penalità)."
L["DeletePoint"] = "Elimina punto %s"
L["DisableArenaFramesInArena"] = "Disabilita gli arenaframes in arena"
L["DisableArenaFramesInArena_Desc"] = "Disabilita gli Arenaframe in un'arena. Funziona anche quando è installato sArena"
L["DisableArenaFramesInBattleground"] = "Disabilita gli arenaframes nel campo di battaglia"
L["DisableArenaFramesInBattleground_Desc"] = "Disabilita gli Arenaframe in un campo di battaglia. Funziona anche quando è installato sArena"
L["Disease"] = "Malattia"
L["DispellFilter"] = "Filtra per dispellabilità"
L["DisplayType"] = "Tipo di visualizzazione"
L["Downwards"] = "Verso il basso"
L["DR_Disorient"] = "Disorientamento"
L["DR_Incapacitate"] = "Incapacitato"
L["DR_Knockback"] = "Contraccolpo"
L["DR_Root"] = "Radicato"
L["DR_Silence"] = "Silenzio"
L["DR_Stun"] = "Stordire"
L["DRContainer"] = "Contenitore DR"
L["DRTracking"] = "Tracciamento DR."
L["DrTracking_Container_Color_Desc"] = "Il colore del bordo del contenitore delle icone DR"
L["DrTracking_DisplayType_Desc"] = "Qui si può scegliere se lo stato del DR deve essere indicato da una cornice colorata attorno all'icona o dal colore del testo del conto alla rovescia. Si noti che l'impostazione del testo del conto alla rovescia funziona solo se non si utilizza un addon come OmniCC che modifica il colore del testo."
L["DrTracking_Enabled"] = "Abilita tracciamento DR"
L["DrTracking_Enabled_Desc"] = "Quando è abilitato, accanto alla riga dei nemici ci saranno le icone per il tracciamento DR. Bordo verde: Il prossimo CC avrà una durata dimezzata. Bordo giallo: Il prossimo CC sarà di un quarto di durata. Bordo rosso: Il prossimo CC non si applica perché il giocatore è immune."
L["DrTracking_Spacing"] = "Spaziatura DR"
L["DrTracking_Spacing_Desc"] = "la spaziatura tra ciascuna icona DR"
L["DrTrackingFiltering_Enabled_Desc"] = "Quando questa opzione è abilitata, vedrai solo le icone degli incantesimi la cui categoria è selezionata nell'elenco dei filtri."
L["DrTrackingFiltering_Filterlist_Desc"] = "Fai clic per tracciare/annullare il tracciamento di questa categoria"
L["DrTrackingFilteringSettings_Desc"] = "Qui puoi scegliere quale categoria DR viene mostrata."
L["DrTrackingSettings"] = "Tracciamento DR"
L["DrTrackingSettings_Desc"] = "Impostazioni relative al tracciamento DR"
L["DurationFilter"] = "Filtra per durata"
L["DurationFilter_OnlyShowWhenDuration"] = "Mostra solo quando la durata è inferiore o uguale a"
L["DurationFilter_OnlyShowWhenDuration_Desc"] = "Tutte le aure che hanno una durata superiore a quella selezionata non passano attraverso il filtro."
L["EnableClique"] = "Usa Clique per l'assegnazione dei tasti"
L["EnableClique_Desc"] = "Quando questa opzione è abilitata, Clique gestirà tutte le assegnazioni di tasti."
L["enemies"] = "nemici"
L["Enemies"] = "Nemici"
L["EnemiesTargetingAllies_Enabled_Desc"] = "Quando abiliti questa opzione verrai avvisato da un suono quando un determinato numero di nemici prende di mira un alleato"
L["EnemiesTargetingMe_Enabled_Desc"] = "Quando abiliti questa opzione, sarai avvisato da un suono quando un determinato numero di nemici ti sta prendendo di mira"
L["enemy"] = "nemico"
L["Enemy"] = "Nemico"
L["EnemyJoined"] = "Un nemico si è unito al campo di battaglia"
L["EnemyLeft"] = "Un nemico ha lasciato il campo di battaglia"
L["ExportButton"] = "Esporta profilo"
L["ExportButton_Desc"] = "Clicca qui per esportare il tuo attuale profilo attivo"
L["FakeEventsDisabled"] = "Non è più possibile creare eventi falsi."
L["FakeEventsEnabled"] = "Ora si stanno creando eventi falsi per i test."
L["Filtering_Enabled"] = "Abilita filtro"
L["Filtering_Filterlist"] = "Lista filtri"
L["FilterSettings"] = "Impostazioni filtro"
L["Font"] = "Tipo di carattere (font)"
L["Font_Desc"] = "Il tipo di carattere principale utilizzato da questo addon"
L["Font_Outline"] = "Contorno del carattere"
L["Font_Outline_Desc"] = "Il contorno del carattere"
L["Fontcolor"] = "Colore del carattere"
L["Fontcolor_Desc"] = "Il colore del testo"
L["FontShadow_Enabled"] = "Abilita l'ombreggiatura dei caratteri"
L["FontShadow_Enabled_Desc"] = "Quando abilitato ci sarà un'ombra attorno al carattere. Puoi scegliere il colore di quell'ombra"
L["FontShadowColor"] = "Colore dell'ombra"
L["FontShadowColor_Desc"] = "Il colore dell'ombreggiatura del testo"
L["FontSize"] = "Dimensione del carattere"
L["FontSize_Desc"] = "La dimensione del carattere del testo"
L["Frame"] = "Frame"
L["Framescale"] = "Scala"
L["Framescale_Desc"] = "La scala del mainframe. Questa impostazione non è disponibile in combattimento a causa della protezione in combattimento di Blizzard."
L["General"] = "Generale"
L["GeneralSettings"] = "Impostazione Generali"
L["GeneralSettings_Desc"] = "Alcune impostazioni generali"
L["GeneralSettingsAllies"] = "Queste impostazioni si applicano agli alleati indipendentemente dalle dimensioni del campo di battaglia"
L["GeneralSettingsEnemies"] = "Queste impostazioni si applicano ai nemici indipendentemente dalle dimensioni del campo di battaglia"
L["HealthBar"] = "Barra della salute"
L["HealthBar_Background_Desc"] = "Colore di sfondo per la barra della salute."
L["HealthBar_Texture_Desc"] = "La texture utilizzata per le barre della salute"
L["HealthBarSettings_Desc"] = "Impostazioni per la barra della salute."
L["HealthTextEnabled"] = "Abilita il testo sulla salute"
L["HealthTextSettings"] = "Impostazioni del testo sulla salute"
L["HealthTextType"] = "Tipo di testo sulla salute"
L["Height"] = "Altezza"
L["HighestPriorityAura"] = "Aura di massima priorità"
L["Highlight_Color"] = "Evidenzia Colore"
L["Highlight_Color_Desc"] = "Il colore della cornice del player su cui si trova attualmente il cursore"
L["HorizontalGrowdirection"] = "Direzione di crescita orizzontale"
L["HorizontalGrowdirection_Desc"] = "Consente di scegliere se le righe crescono a sinistra o a destra"
L["HorizontalSpacing"] = "Spaziatura orizzontale"
L["IAmTargeted"] = "Sono preso di mira"
L["Icon"] = "Icona"
L["IconSize"] = "Dimensione Icona"
L["IconsPerRow"] = "Icone per riga"
L["Import"] = "Importa"
L["ImportButton"] = "Importa profilo"
L["ImportButton_Desc"] = "Fare clic qui per importare un profilo da una stringa e sovrascrivere il profilo attivo corrente"
L["JustifyH"] = "Allineamento orizzontale"
L["JustifyH_Desc"] = "L'allineamento orizzontale del testo"
L["JustifyV"] = "Allineamento verticale"
L["JustifyV_Desc"] = "L'allineamento verticale del testo"
L["KeybindSettings_Desc"] = "Qui puoi configurare cosa dovrebbe accadere al clic sinistro-destro e centrale"
L["LEFT"] = "Sinistra"
L["LeftToTargetCounter"] = "Sinistra al bersaglio del contatore"
L["Leftwards"] = "Verso sinistra"
L["Level"] = "Livello"
L["LevelText_Enabled"] = "Mostra il livello del giocatore"
L["LevelText_OnlyShowIfNotMaxLevel"] = "Mostra solo se il livello non è massimo"
L["Locked"] = "Blocco"
L["Locked_Desc"] = "Blocca il frame in posizione"
L["Magic"] = "Magia"
L["MainFrameSettings"] = "Impostazioni frame principale"
L["MainFrameSettings_Desc"] = "Impostazioni per il frame principale utilizzato per %s"
L["MaxPlayers"] = "Nemici massimi"
L["MaxPlayers_Desc"] = "L'addon non mostrerà i nemici se ci sono più nemici di questo numero"
L["MIDDLE"] = "Centrale"
L["ModuleSettings"] = "Impostazione modulo"
L["ModuleSpecificSettings"] = "Impostazioni specifiche del modulo"
L["MoreProfileOptions"] = "Altre opzioni di profilo"
L["MyFocus"] = "Il mio focus"
L["MyFocus_Color"] = "Colore focus"
L["MyFocus_Color_Desc"] = "Il colore utilizzato per il bordo che indica il focus attuale"
L["MyTarget"] = "Il mio bersaglio"
L["MyTarget_Color"] = "Colore bersaglio"
L["MyTarget_Color_Desc"] = "Il colore utilizzato per il bordo attorno alla barra della salute che indica il tuo bersaglio attuale"
L["MyVersion"] = "Stai utilizzando la versione"
L["Name"] = "Nome"
L["Name_Desc"] = "Impostazioni per i nomi sulla barra della salute."
L["NewVersionAvailable"] = "È disponibile una nuova versione"
L["None"] = "Nessuno"
L["NonPriorityBuffs"] = "Buff non prioritari"
L["NonPriorityDebuffs"] = "Debuff non prioritari"
L["Normal"] = "Normale"
L["NormalCastingBar"] = "Barra di lancio normale"
L["NotAvailableInCombat"] = "Questa impostazione non è disponibile mentre sei in combattimento a causa delle restrizioni al combattimento di Blizzard."
L["Notifications_Allies_Enabled_Desc"] = "Se abilitato, riceverai una notifica quando un alleato lascia o si unisce al campo di battaglia."
L["Notifications_Enabled"] = "Attivare le notifiche"
L["Notifications_Enemies_Enabled_Desc"] = "Se abilitato, riceverai una notifica quando un nemico lascia o si unisce al campo di battaglia."
L["NoVersion"] = "Nessun addon trovato"
L["NumericTargetindicator_Enabled_Desc"] = "Mostra quanti %s stanno prendendo di mira un giocatore."
L["ObjectiveAndRespawn_ObjectiveEnabled"] = "Mostra obiettivi"
L["ObjectiveAndRespawn_ObjectiveEnabled_Desc"] = "Mostra nell'icona per bandiera, carrello da miniera e porta sfere accanto al pulsante nemico"
L["ObjectiveAndRespawn_Position"] = "Posizione"
L["ObjectiveAndRespawn_Position_Desc"] = "Se l'icona deve essere mostrata a sinistra dell'icona delle specifiche o a destra delle icone dei monili/razziali."
L["ObjectiveAndRespawn_RespawnEnabled"] = "Abilita il timer di rigenerazione"
L["ObjectiveAndRespawn_RespawnEnabled_Desc"] = "Quando abilitato ci sarà un'icona che indica il tempo rimanente prima che il nemico sia vivo."
L["ObjectiveAndRespawn_Width_Desc"] = "Larghezza dell'icona dell'obiettivo"
L["ObjectiveAndRespawnSettings_Desc"] = "Impostazioni riguardanti l'obiettivo del campo di battaglia"
L["ObjectiveAndRespawnTimer"] = "Obiettivo e timer di respawn"
L["OffsetX"] = "Spostamento X"
L["OffsetY"] = "Spostamento Y"
L["OldVersion"] = "Vecchia versione"
L["OutOfCombat"] = "Fuori combattimento"
L["PlayerCount"] = "Conteggio giocatori"
L["PlayerCount_Enabled_Desc"] = "Se c'è un testo per il numero attuale di giocatori. Si noti che questo conteggio può essere diverso dal numero attuale di righe mostrate a causa della protezione dal combattimento o della presenza di alleati/nemici in un RBG."
L["Point"] = "Punto"
L["PointAtObject"] = "Punta all'oggetto"
L["Poison"] = "Veleno"
L["Position"] = "Posizione"
L["PowerBar"] = "Barra di potere"
L["PowerBar_Background_Desc"] = "Colore di sfondo per la barra di potere."
L["PowerBar_Enabled"] = "Abilita le barre delle risorse"
L["PowerBar_Enabled_Desc"] = "Abilita questa opzione se non vuoi vedere le barre per risorse come mana, rabbia, ecc."
L["PowerBar_Height_Desc"] = "L'altezza della barra delle risorse. Alzare la barra delle risorse renderà automaticamente più corta la barra della salute"
L["PowerBar_Texture_Desc"] = "La texture utilizzata per le barre di potere"
L["PowerBarSettings_Desc"] = "Impostazioni per la barra delle risorse"
L["PriorityBuffs"] = "Buff prioritari"
L["PriorityDebuffs"] = "Debuff prioritari"
L["Racial"] = "Razziale"
L["Racial_Enabled"] = "Abilita le razziali"
L["Racial_Enabled_Desc"] = "Quando abilitato ci saranno icone per le razziali usate"
L["Racial_Width_Desc"] = "La larghezza della razziale"
L["RacialFiltering_Enabled_Desc"] = "Quando questa opzione è abilitata vedrai solo le razziali selezionate nell'elenco dei filtri."
L["RacialFiltering_Filterlist_Desc"] = "Fai clic per tracciare/annullare il tracciamento di questa razziale"
L["RacialFilteringSettings_Desc"] = "Qui puoi selezionare quale razziale vuoi monitorare"
L["RacialSettings"] = "Razziali"
L["RacialSettings_Desc"] = "Impostazioni riguardanti le razziali"
L["RangeIndicator_Alpha"] = "Alfa"
L["RangeIndicator_Alpha_Desc"] = "L'alfa del frame nemico quando il nemico è fuori dalla portata scelta"
L["RangeIndicator_Enabled"] = "Abilita l'indicatore di portata"
L["RangeIndicator_Enabled_Desc"] = "Quando questa opzione è abilitata, il frame utilizzerà l'alfa scelto quando un nemico è fuori dalla portata prescelta"
L["RangeIndicator_Everything"] = "Modifica l'alfa di tutti gli elementi"
L["RangeIndicator_Frames"] = "Modifica alfa di"
L["RangeIndicator_Frames_Desc"] = "Imposta quali parti della riga nemica devono essere mostrate con maggiore trasparenza quando il nemico è fuori portata."
L["RangeIndicator_Range"] = "Portata"
L["RangeIndicator_Range_Desc"] = "Quando un nemico si trova a una distanza superiore a questa distanza (in yards), il suo frame utilizzerà l'alpha scelto"
L["RangeIndicator_Settings"] = "Indicatore di portata"
L["RangeIndicator_Settings_Desc"] = "Qui è possibile configurare l'indicatore di portata."
L["RBGSpecificSettings"] = "Campo di battaglia classificato"
L["RBGSpecificSettings_Desc"] = "Queste impostazioni si applicano solo in un RBG (Campo di Battaglia Classificato)"
L["ResetModule"] = "Ripristina Modulo"
L["ResetModule_Desc"] = "Fai clic qui per ripristinare le impostazioni di questo modulo per %s:%s"
L["ResetModuleSettings"] = "Ripristina le impostazioni di questo modulo"
L["RIGHT"] = "Destra"
L["Rightwards"] = "Verso destra"
L["RoleIcon_Enabled"] = "Abilita le icone dei ruoli"
L["RoleIcon_Enabled_Desc"] = "Abilita questa opzione se vuoi vedere un'icona del ruolo sulla barra della salute"
L["RoleIcon_Size_Desc"] = "La dimensione delle icone dei ruoli (larghezza e altezza)"
L["RoleIconSettings"] = "Icona del ruolo"
L["RoleIconSettings_Desc"] = "Impostazioni per l'icona del ruolo mostrata sulla barra della salute"
L["Scale"] = "Scala"
L["ShowDispellable"] = "Mostra dispellabile"
L["ShowDispellable_Desc"] = "Mostra %s che posso rimuovere tramite rimuovi maledizione, epurazione, ecc."
L["ShowMine"] = "Mostra il mio"
L["ShowMine_Desc"] = "Mostra %s che ho applicato"
L["ShowNumbers"] = "Mostra numeri"
L["ShowNumbers_Desc"] = "Mostra numeri di recupero, tieni presente che questa opzione potrebbe non funzionare quando utilizzi componenti aggiuntivi come OmniCC"
L["ShowRealmnames"] = "Mostra nomi dei reami"
L["ShowRealmnames_Desc"] = "Mostra il reame del nemico"
L["ShowTooltips"] = "Mostra suggerimenti"
L["ShowTooltips_Desc"] = "Mostra i suggerimenti quando sposti il ​​mouse su oggetti come le icone dei monili e dell'aura"
L["Side"] = "Lato"
L["SideAtObject"] = "Lato all'oggetto"
L["Size"] = "Dimensione"
L["SmallCastingBar"] = "Barra di lancio piccola"
L["SourceFilter"] = "Filtra per Fonte"
L["Spec_AuraDisplay_Enabled"] = "Abilita visualizzazione Aura"
L["Spec_AuraDisplay_Enabled_Desc"] = "Quando questa opzione è abilitata vedrai le icone per la perdita di controllo e le interruzioni (se attive) invece dell'icona delle specifiche"
L["Spec_Enabled"] = "Abilita la specializzazione"
L["Spec_Enabled_Desc"] = "Quando abilitato ci saranno icone per la specializzazione dei giocatori"
L["Spec_Width_Desc"] = "Larghezza dell'icona della specializzazione"
L["SpecSettings"] = "Specializzazione"
L["SpecSettings_Desc"] = "Impostazioni che modificano l'icona della specializzazione."
L["SpellID_Filtering"] = "Filtra per ID incantesimo"
L["Style"] = "Stile"
L["SymbolicTargetindicator_Enabled_Desc"] = "Mostra un'icona quadrata del colore della classe per ogni %s che prende di mira un giocatore."
L["TargetAmount"] = "ammontare bersaglio"
L["TargetAmount_Ally"] = "Il suono viene riprodotto non appena un membro del gruppo viene preso di mira da un numero di avversari superiore a quello stabilito."
L["TargetAmount_Me"] = "Il suono viene riprodotto non appena vieni preso di mira da un numero di avversari superiore a quello stabilito."
L["TargetCallerUpdated"] = "è ora il chiamante del bersaglio"
L["TargetCalling"] = "Chiamata al bersaglio"
L["TargetCallingNotificationEnable"] = "Abilita notifica sonora"
L["TargetCallingNotificationEnable_Desc"] = "Ti avvisa quando c'è un nuovo bersaglio (rilevato tramite l'icona del teschio del bersaglio dell'incursione)"
L["TargetCallingSetMark"] = "Contrassegna il mio obiettivo con l'icona del teschio"
L["TargetCallingSetMark_Desc"] = "Contrassegna automaticamente il tuo bersaglio attuale con l'icona del teschio quando sei il capo del raid e non ci sono assistenti dell'incursione nel tuo gruppo. Segnerà anche il tuo bersaglio attuale se non sei il capo dell'incursione ma sei un assistente."
L["TargetCallingShowIcon"] = "Se abilitato, il target del chiamante riceverà un'icona sul frame"
L["TargetCallingShowIcon_Desc"] = "Quando abilitato, vedrai un'icona sul frame che mostra l'obiettivo del chiamante del bersaglio"
L["TargetCallingSoundEnable"] = "Avvisami sul nuovo bersaglio"
L["TargetCallingSoundEnable_Desc"] = "Se abilitato, verrà emesso un suono quando il chiamante del bersaglio prende di mira un nuovo giocatore"
L["TargetCallingSoundSound"] = "Suono di notifica"
L["TargetIndicator"] = "Indicatore Bersaglio"
L["TargetIndicator_Desc"] = "Indicatori Bersaglio Numerici e Simbolici"
L["TargetIndicatorNumeric"] = "Indicatore di Bersaglio Numerico"
L["TargetIndicatorSymbolic"] = "Indicatore di Bersaglio Simbolico"
L["Testmode_Toggle"] = "Mostra/Nascondi modalità di test"
L["Testmode_Toggle_Desc"] = [=[Attiva/Disattiva la modalità di test. Questa modalità rende più semplice la visualizzazione dell'impatto della maggior parte delle impostazioni disponibili. Non è possibile simulare tutte le opzioni, ma la maggior parte di esse sono presenti in questa sofisticata modalità.
La modalità di test viene disattivata se è in corso un campo di battaglia.]=]
L["Testmode_ToggleAnimation"] = "Mostra/nascondi animazione modalità test"
L["Testmode_ToggleAnimation_Desc"] = "Attiva/Disattiva l'animazione della modalità test così da poter rivolgere maggiormente l'attenzione su un'impostazione specifica senza le distrazioni dovute all'animazione."
L["Testmode_UseTeammates"] = "Usa i membri del gruppo per i test"
L["Testmode_UseTeammates_Desc"] = "Se abilitata, la modalità test utilizzerà i membri del tuo gruppo attuale invece di quelli falsi"
L["TestmodeDisabled"] = "Modalità test disabilitata"
L["TestmodeEnabled"] = "Modalità test abilitata"
L["TestmodeSettings"] = "Modalità test"
L["Text"] = "Testo"
L["TextSettings"] = "Impostazioni testo"
L["Thick"] = "Spessore"
L["TOP"] = "Superiore"
L["TOPLEFT"] = "Superiore sinistro"
L["TOPRIGHT"] = "Superiore destro"
L["Trinket"] = "Monile"
L["Trinket_Enabled"] = "Attiva Monili"
L["Trinket_Enabled_Desc"] = "Quando attivi, compariranno le icone per il Medaglione del Gladiatore, Adattamento e Implacabilità."
L["Trinket_Width_Desc"] = "La larghezza del monile."
L["TrinketSettings_Desc"] = "Impostazioni relative ai Monili"
L["UpdatePeriod"] = "Periodo di aggiornamento"
L["UpdatePeriod_Desc"] = "L'intervallo di tempo tra i controlli di combattimento in secondi, non deve essere troppo basso perché potrebbe causare problemi di prestazioni."
L["Upwards"] = "Verso l'alto"
L["UseBarHeight"] = "Usa altezza barra"
L["UseBigDebuffsPriority"] = "Usa le priorità di BigDebuff"
L["UseBigDebuffsPriority_Desc"] = "Se abiliti questa opzione, l'addon utilizzerà invece le priorità dell'addon BigDebuffs. Le priorità sono utilizzate per i moduli %s, %s e %s"
L["UseButtonHeight"] = "Usa l'altezza del pulsante"
L["UseButtonHeight_Desc"] = "Se abiliti questa opzione, il modulo utilizzerà l'altezza dei pulsanti."
L["UserDefined"] = "Definito dall'utente"
L["VerticalGrowdirection"] = "Direzione di crescita verticale"
L["VerticalGrowdirection_Desc"] = "Permette di scegliere se le barre crescono verso l'alto o verso il basso."
L["VerticalPosition"] = "Posizione verticale"
L["VerticalSpacing"] = "Spaziatura verticale"
L["Width"] = "Larghezza"

elseif LOCALE_koKR then
L["AddPoint"] = "추가 포인트"
L["Allies"] = "아군"
L["allies"] = "아군"
L["ally"] = "아군"
L["Ally"] = "아군"
L["AllyIsTargeted"] = "아군이 공격받고 있습니다"
L["AllyJoined"] = "아군이 전장에 참여했습니다"
L["AllyLeft"] = "아군이 전장을 떠났습니다"
L["AND"] = "그리고"
L["AttachToObject"] = "개체에 붙이기"
L["AttachToObject_Desc"] = "다른 프레임의 위치에 따라 모든 개체를 선택할 수 있진 않습니다. 예를 들어 pvp 특성을 종족 특성에 고정시킨 후 종족 특성을 pvp 특성에 고정시키려고 시도하면 오류 메시지가 표시됩니다."
L["Auras_Buffs_Container_Color_Desc"] = "강화 효과 아이콘 묶음의 테두리 색상입니다"
L["Auras_CustomFiltering_Conditions_All"] = "모두"
L["Auras_CustomFiltering_Conditions_Any"] = "모든"
L["Auras_Debuffs_Container_Color_Desc"] = "약화 효과 아이콘 묶음의 테두리 색상입니다"
L["Auras_Enabled"] = "효과 활성화"
L["Auras_Enabled_Desc"] = "활성화하면 효과 아이콘이 표시됩니다. 남은 시간과 중첩 적용 횟수의 숫자도 표시됩니다."
L["Auras_ShowTooltips"] = "주문 툴팁 표시"
L["AurasFiltering_AddSpellID"] = "주문ID"
L["AurasFiltering_AddSpellID_Desc"] = "주문ID나 콤마로 구분한 여러 개의 주문ID를 입력하세요. 주문ID는 당신의 필터 목록에 추가됩니다. 주문ID 필터가 활성화되었을 때 필터 목록에 있는 주문ID의 효과만 표시됩니다."
L["AurasFiltering_Enabled_Desc"] = "이 옵션을 활성화하면 당신이 필터 목록에 추가한 %s만 확인할 수 있습니다."
L["AurasFiltering_Filterlist_Desc"] = "이 %s|1을;를; 제거하려면 클릭"
L["AurasFilteringSettings_Desc"] = "여기서 표시된 효과를 제어할 수 있습니다"
L["AurasSettings"] = "효과"
L["AurasSettings_Desc"] = "효과를 설정합니다 (강화 효과와 약화 효과)"
L["AurasStacktextSettings"] = "중첩 문자"
L["BarBackground"] = "배경 색상"
L["BarHeight_Desc"] = "각 바의 높이입니다. 이 설정은 블리자드의 전투 보호때문에 전투 중에 설정할 수 없습니다."
L["BarSettings"] = "바 설정"
L["BarSettings_Desc"] = "여기서 바를 구성할 수 있습니다"
L["BarTexture"] = "바 텍스쳐"
L["BarWidth_Desc"] = "각 바의 너비입니다. 이 설정은 블리자드의 전투 보호때문에 전투 중에 설정할 수 없습니다."
L["BattlegroundSize"] = "전장 규모"
L["BGSize_15"] = "1-15인"
L["BGSize_15_Desc"] = "이 설정은 진영 별 1-15인 규모의 전장에서 %s에게 적용됩니다"
L["BGSize_40"] = "16-40인"
L["BGSize_40_Desc"] = "이 설정은 진영 별 16-40인 규모의 전장에서 %s에게 적용됩니다"
L["BorderThickness"] = "테두리 크기"
L["BOTTOM"] = "하단"
L["BOTTOMLEFT"] = "왼쪽 하단"
L["BOTTOMRIGHT"] = "오른쪽 하단"
L["buff"] = "강화 효과"
L["BuffContainer"] = "강화 효과 묶음"
L["BuffIcon"] = "강화 효과 아이콘"
L["Buffs"] = "강화 효과"
L["Button"] = "버튼"
L["CantAnchorToItself"] = "자기 자신에게는 고정할 수 없습니다!"
L["CENTER"] = "중앙"
L["Coloring_Enabled"] = "약화 효과 유형 별 색상"
L["Coloring_Enabled_Desc"] = "이 옵션을 활성화하면 프레임의 테두리 또는 초읽기 문자에 약화 효과 유형에 따라 색상을 입힙니다 (기본 UI에서 알 수 있는 독은 녹색, 질병은 푸른색 등.)"
L["Columns"] = "열"
L["Columns_Desc"] = "플레이어가 표시되는 열의 수입니다."
L["ConfirmProfileOverride"] = "정말 서브프로필 %s|1을;를; %s|1으로;로; 덮어쓸까요?"
L["Container_Color"] = "묶음 테두리 색상"
L["ContainerPosition"] = "묶음 위치"
L["ConvertCyrillic"] = "키릴 문자 전환"
L["ConvertCyrillic_Desc"] = "러시아어로 플레이 중이라면 이름을 읽기 쉽게 만들어 주는 키릴 문자로 전환합니다"
L["CopySettings"] = "의 설정 복사하기"
L["Countdowntext"] = "초읽기 문자"
L["Curse"] = "저주"
L["CustomMacro_Desc"] = [=[활성화하면 사용자 설정 매크로를 설정할 수 있습니다. 빈 영역은 빈 매크로로 이어집니다.
사용할 수 있는 대체어: %n 적의 이름으로 대체됩니다

예제:
/targetexact %n
/cast 변이
/say %n에게 변이
/targetlasttarget

적을 대상 지정하고, 변이를 시전하고, 일반 대화 채널에 메시지를 쓰고 이전 대상을 대상 지정합니다.
매크로는 255자 길이까지만 가능합니다 (대체어 포함).]=]
L["debuff"] = "약화 효과"
L["DebuffContainer"] = "약화 효과 묶음"
L["DebuffIcon"] = "약화 효과 아이콘"
L["Debuffs"] = "약화 효과"
L["DebuffType_Filtering"] = "약화 효과 유형 별 필터"
L["DebuffType_Filtering_Desc"] = "이 옵션을 활성화하면 선택한 유형의 약화 효과만 확인할 수 있습니다. 이 옵션은 많은 주문과 약화 효과 유형의 거대한 데이터베이스가 필요하기 때문에 테스트 모드에서 실현되지 않습니다."
L["DisableArenaFramesInBattleground_Desc"] = "전장에서 투기장 프레임을 비활성합니다. sArena가 설치되어 있을 때도 작동합니다"
L["Disease"] = "질병"
L["DisplayType"] = "표시 형식"
L["Downwards"] = "아래로"
L["DR_Disorient"] = "방향 감각 상실"
L["DR_Incapacitate"] = "행동 불가"
L["DR_Knockback"] = "밀쳐내기"
L["DR_Root"] = "이동 불가"
L["DR_Silence"] = "침묵"
L["DR_Stun"] = "기절"
L["DRContainer"] = "점감 묶음"
L["DrTracking_Container_Color_Desc"] = "점감 아이콘 묶음의 테두리 색상입니다"
L["DrTracking_DisplayType_Desc"] = "여기서 점감 상태를 아이콘 주변의 색상화된 프레임 또는 초읽기 문자의 문자 색상으로 나타낼 지 선택할 수 있습니다. 초읽기 문자 설정은 문자 색상을 변경하는 OmniCC 같은 애드온을 사용하지 않아야 작동합니다."
L["DrTracking_Enabled"] = "점감 추적 활성화"
L["DrTracking_Enabled_Desc"] = "활성화하면 적 바 옆에 점감 추적 아이콘이 표시됩니다. 녹색 테두리: 다음 군중 제어는 절반의 지속시간을 가집니다. 노란색 테두리: 다음 군중 제어는 1/4의 지속시간을 가집니다. 빨간색 테두리: 다음 군중 제어는 플레이어가 면역이므로 적용되지 않습니다"
L["DrTracking_Spacing"] = "점감 간격"
L["DrTracking_Spacing_Desc"] = "각 점감 아이콘의 간격입니다"
L["DrTrackingFiltering_Enabled_Desc"] = "이 옵션을 활성화하면 필터 목록에서 선택한 범주의 주문 아이콘만 표시합니다."
L["DrTrackingFiltering_Filterlist_Desc"] = "이 범주를 추적/추적 해제하려면 클릭"
L["DrTrackingFilteringSettings_Desc"] = "여기서 어떤 점감 범주를 표시할 지 선택할 수 있습니다."
L["DrTrackingSettings"] = "점감 추적"
L["DrTrackingSettings_Desc"] = "점감 추적에 관련된 설정입니다"
L["enemies"] = "적"
L["Enemies"] = "적"
L["enemy"] = "적"
L["Enemy"] = "적"
L["EnemyJoined"] = "적이 전장에 참여했습니다"
L["EnemyLeft"] = "적이 전장을 떠났습니다"
L["Filtering_Enabled"] = "필터 활성화"
L["Filtering_Filterlist"] = "필터 목록"
L["Font"] = "글꼴"
L["Font_Desc"] = "이 애드온에 사용할 주 글꼴입니다"
L["Font_Outline"] = "글꼴 외곽선"
L["Font_Outline_Desc"] = "글꼴의 외곽선입니다"
L["Fontcolor"] = "글꼴 색상"
L["FontShadow_Enabled"] = "글꼴 그림자 활성화"
L["FontShadow_Enabled_Desc"] = "활성화하면 글꼴 주위에 그림자가 생깁니다. 그림자의 색상을 선택할 수 있습니다"
L["FontShadowColor"] = "그림자 색상"
L["FontShadowColor_Desc"] = "문자 그림자의 색상입니다"
L["FontSize"] = "글꼴 크기"
L["Frame"] = "프레임"
L["Framescale"] = "크기 비율"
L["Framescale_Desc"] = "주 프레임의 크기 비율입니다. 이 설정은 블리자드의 전투 보호때문에 전투 중에 설정할 수 없습니다."
L["General"] = "일반"
L["GeneralSettings"] = "일반 설정"
L["GeneralSettings_Desc"] = "일반 설정입니다"
L["GeneralSettingsAllies"] = "이 설정은 전장의 규모에 상관없이 아군에게 적용됩니다"
L["GeneralSettingsEnemies"] = "이 설정은 전장의 규모에 상관없이 적에게 적용됩니다"
L["HealthBar"] = "생명력바"
L["HealthBar_Background_Desc"] = "생명력 바의 배경 색상입니다."
L["HealthBar_Texture_Desc"] = "생명력 바에 사용할 텍스쳐입니다"
L["HealthBarSettings_Desc"] = "생명력 바의 설정입니다."
L["Height"] = "높이"
L["Highlight_Color"] = "강조 색상"
L["Highlight_Color_Desc"] = "당신의 커서가 올려진 플레이어 프레임의 색상"
L["HorizontalGrowdirection"] = "수평 성장 방향"
L["HorizontalGrowdirection_Desc"] = "열이 왼쪽 또는 오른쪽 가장자리로 성장할지 선택합니다"
L["HorizontalSpacing"] = "수평 간격"
L["IconsPerRow"] = "열 당 아이콘"
L["KeybindSettings_Desc"] = "여기서 왼쪽- 오른쪽- 그리고 가운데 클릭 시 어떻게 작동할 지 구성할 수 있습니다"
L["LEFT"] = "왼쪽"
L["LeftToTargetCounter"] = "대상 지정 숫자 왼쪽에"
L["Leftwards"] = "왼쪽"
L["Level"] = "플레이어 레벨"
L["LevelText_Enabled"] = "플레이어의 레벨 표시"
L["LevelText_OnlyShowIfNotMaxLevel"] = "최고 레벨이 아닐 때만 표시"
L["Locked"] = "고정시키기"
L["Locked_Desc"] = "프레임을 현재 위치에 고정시킵니다"
L["Magic"] = "마법"
L["MainFrameSettings"] = "주 프레임 설정"
L["MainFrameSettings_Desc"] = "%s에게 사용할 주 프레임을 설정합니다"
L["MaxPlayers"] = "최대 적"
L["MaxPlayers_Desc"] = "이 숫자보다 적이 더 많으면 애드온은 적을 표시하지 않습니다"
L["MIDDLE"] = "중앙"
L["ModuleSettings"] = "모듈 설정"
L["MyFocus_Color"] = "주시 대상 색상"
L["MyFocus_Color_Desc"] = "당신의 현재 주시 대상을 나타내는 테두리에 사용할 색상입니다"
L["MyTarget_Color"] = "대상 색상"
L["MyTarget_Color_Desc"] = "당신의 현재 대상을 나타내는 생명력 바 주위 테두리에 사용할 색상입니다"
L["Name"] = "이름"
L["Name_Desc"] = "생명력 바 위의 이름을 설정합니다."
L["None"] = "안 함"
L["Normal"] = "보통"
L["NotAvailableInCombat"] = [=[
이 설정은 블리자드의 전투 제한때문에 당신이 전투 중일 때 사용할 수 없습니다.]=]
L["Notifications_Allies_Enabled_Desc"] = "활성화하면 아군이 전장에 참여 또는 탈영 시 당신이 알립니다."
L["Notifications_Enabled"] = "알림 활성화"
L["Notifications_Enemies_Enabled_Desc"] = "활성화하면 적이 전장에 참여 또는 탈영 시 당신이 알립니다."
L["NumericTargetindicator_Enabled_Desc"] = "몇명의 %s|1이;가; 플레이어를 대상 지정 중인 지 표시합니다."
L["ObjectiveAndRespawn_ObjectiveEnabled"] = "목표 표시"
L["ObjectiveAndRespawn_ObjectiveEnabled_Desc"] = "적 버튼 옆에 깃발, 광산 수레와 공-운반자 아이콘을 표시합니다"
L["ObjectiveAndRespawn_Position"] = "위치"
L["ObjectiveAndRespawn_Position_Desc"] = "전문화 아이콘의 왼쪽 또는 장신구/종족 특성 아이콘의 오른쪽에 아이콘을 표시합니다."
L["ObjectiveAndRespawn_RespawnEnabled"] = "재생성 타이머 활성화"
L["ObjectiveAndRespawn_RespawnEnabled_Desc"] = "활성화하면 적이 되살아나기까지 남은 시간을 나타내는 아이콘이 표시됩니다."
L["ObjectiveAndRespawn_Width_Desc"] = "목표 아이콘의 너비입니다"
L["ObjectiveAndRespawnSettings_Desc"] = "전장의 목표와 관련된 설정입니다"
L["ObjectiveAndRespawnTimer"] = "목표"
L["OffsetX"] = "X 위치"
L["OffsetY"] = "Y 위치"
L["PlayerCount"] = "플레이어 인원 수"
L["PlayerCount_Enabled_Desc"] = "현재 플레이어 인원 수를 문자로 표시합니다. 전투 보호 또는 평점제 전장에서 적/아군의 탈영때문에 바에 표시된 현재 숫자와 다를 수 있습니다."
L["Point"] = "지점"
L["PointAtObject"] = "개체의 지점"
L["Poison"] = "독"
L["Position"] = "위치"
L["PowerBar"] = "마력바"
L["PowerBar_Background_Desc"] = "마력 바의 배경 색상입니다."
L["PowerBar_Enabled"] = "마력 바 활성화"
L["PowerBar_Enabled_Desc"] = "마나, 분노 등과 같은 마력 바를 보고 싶다면 이 옵션을 활성화하세요."
L["PowerBar_Height_Desc"] = "마력 바의 높이입니다. 마력 바의 높이를 키우면 자동으로 생명력 바의 높이는 줄어듭니다"
L["PowerBar_Texture_Desc"] = "마력 바에 사용할 텍스쳐입니다"
L["PowerBarSettings_Desc"] = "마력 바 설정"
L["Racial_Enabled"] = "종족 특성 활성화"
L["Racial_Enabled_Desc"] = "활성화하면 종족 특성 아이콘을 표시합니다"
L["Racial_Width_Desc"] = "종족 특성의 너비입니다."
L["RacialFiltering_Enabled_Desc"] = "이 옵션을 활성화하면 당신이 필터 목록에서 선택한 종족 특성만 표시합니다."
L["RacialFiltering_Filterlist_Desc"] = "이 종족 특성을 추적/추적 해제하려면 클릭하세요"
L["RacialFilteringSettings_Desc"] = "여기서 당신이 추적하고 싶은 종족 특성을 선택할 수 있습니다"
L["RacialSettings"] = "종족 특성"
L["RacialSettings_Desc"] = "종족 특성과 관련된 설정입니다"
L["RangeIndicator_Alpha"] = "투명도"
L["RangeIndicator_Alpha_Desc"] = "적이 선택한 사정 거리를 벗어났을 때의 적 프레임 투명도입니다"
L["RangeIndicator_Enabled"] = "사정 거리 지시기 활성화"
L["RangeIndicator_Enabled_Desc"] = "이 옵션을 활성화하면 적이 선택한 사정 거리를 벗어났을 때 프레임에 선택한 투명도를 사용합니다"
L["RangeIndicator_Everything"] = "모든 요소의 투명도를 변경"
L["RangeIndicator_Frames"] = "투명도 변경:"
L["RangeIndicator_Frames_Desc"] = "적이 사정 거리를 벗어났을 때 적 바의 어떤 부분을 가장 높은 투명도로 표시할 지 설정합니다."
L["RangeIndicator_Range"] = "사정 거리"
L["RangeIndicator_Range_Desc"] = "적이 이 거리 (미터 단위)보다 멀어지면 그의 프레임에 선택한 투명도를 사용합니다"
L["RangeIndicator_Settings"] = "사정 거리 지시기"
L["RangeIndicator_Settings_Desc"] = "여기서 사정 거리 지시기를 설정할 수 있습니다."
L["RBGSpecificSettings"] = "평점제 전장"
L["RBGSpecificSettings_Desc"] = "이 설정은 평점제 전장에서만 적용됩니다"
L["RIGHT"] = "오른쪽"
L["Rightwards"] = "오른쪽"
L["RoleIcon_Enabled"] = "역할 아이콘 활성화"
L["RoleIcon_Enabled_Desc"] = "생명력 바 위에 역할 아이콘을 보고 싶다면 이 옵션을 활성화하세요"
L["RoleIcon_Size_Desc"] = "역할 아이콘의 크기 (너비와 높이)"
L["RoleIconSettings"] = "역할 아이콘"
L["RoleIconSettings_Desc"] = "생명력 바 위에 표시되는 역할 아이콘 설정"
L["ShowMine"] = "내 것만 표시"
L["ShowMine_Desc"] = "내가 적용한 %s만 표시합니다"
L["ShowNumbers"] = "숫자 표시"
L["ShowRealmnames"] = "서버 이름 표시"
L["ShowRealmnames_Desc"] = "적의 서버를 표시합니다"
L["Side"] = "가장자리"
L["SideAtObject"] = "개체의 가장자리"
L["Size"] = "크기"
L["Spec_AuraDisplay_Enabled"] = "효과 표시 활성화"
L["Spec_AuraDisplay_Enabled_Desc"] = "이 옵션을 활성화하면 전문화 아이콘 대신 군중 제어와 시전 방해 (활성화 시) 아이콘을 표시합니다"
L["Spec_Enabled"] = "전문화 활성화"
L["Spec_Enabled_Desc"] = "활성화하면 플레이어의 전문화 아이콘이 표시됩니다"
L["Spec_Width_Desc"] = "전문화 아이콘의 너비입니다."
L["SpecSettings"] = "전문화"
L["SpecSettings_Desc"] = "전문화 아이콘을 변경하는 설정입니다."
L["SpellID_Filtering"] = "주문ID 별 필터"
L["SymbolicTargetindicator_Enabled_Desc"] = "플레이어를 대상 지정 중인 각 %s에게 직업 색상 네모 아이콘을 표시합니다."
L["TargetIndicator"] = "대상 지시기"
L["TargetIndicator_Desc"] = "숫자와 상징 대상 지시기"
L["TargetIndicatorNumeric"] = "대상 지정 숫자"
L["TargetIndicatorSymbolic"] = "대상 상징"
L["Testmode_Toggle"] = "테스트 모드 전환"
L["Testmode_Toggle_Desc"] = "테스트 모드를 활성/비활성화 합니다. 이 모드는 대부분의 옵션이 주는 변화를 쉽게 확인할 수 있게 해줍니다. 모든 옵션을 표현해낼 수 없지만 이 테스트 모드에 대부분의 옵션이 포함되어 있습니다. 당신이 전장 내에 있으면 테스트 모드는 비활성화됩니다."
L["Testmode_ToggleAnimation"] = "테스트 모드 애니메이션 전환"
L["Testmode_ToggleAnimation_Desc"] = "애니메이션때문에 산만해지지 않고 특정 설정에 집중할 수 있도록 테스트 모드의 애니메이션을 켜거나/끕니다."
L["TestmodeSettings"] = "테스트 모드"
L["Thick"] = "두꺼운"
L["TOP"] = "상단"
L["TOPLEFT"] = "왼쪽 상단"
L["TOPRIGHT"] = "오른쪽 상단"
L["Trinket"] = "장신구"
L["Trinket_Enabled"] = "장신구 활성화"
L["Trinket_Enabled_Desc"] = "활성화하면 검투사의 메달, 상황 적응과 가혹함의 아이콘을 표시합니다."
L["Trinket_Width_Desc"] = "장신구의 너비입니다."
L["TrinketSettings_Desc"] = "장신구 관련 설정입니다"
L["Upwards"] = "위로"
L["UseBarHeight"] = "바 높이 사용"
L["UserDefined"] = "사용자 정의"
L["VerticalGrowdirection"] = "수직 성장 방향"
L["VerticalGrowdirection_Desc"] = "바가 위나 아래로 성장하도록 선택합니다"
L["VerticalPosition"] = "수직 위치"
L["VerticalSpacing"] = "수직 간격"
L["Width"] = "너비"

elseif LOCALE_ptBR or LOCALE_ptPT then
L["AddPoint"] = "Adicionar Ponto"
L["Allies"] = "Aliados"
L["allies"] = "Aliados"
L["ally"] = "aliado"
L["Ally"] = "Aliado"
L["AllyIsTargeted"] = "Aliado é alvo"
L["AllyJoined"] = "Um aliado entrou no campo de batalha"
L["AllyLeft"] = "Um aliado saiu do campo de batalha"
L["AND"] = "E"
L["AttachToObject"] = "Anexar ao objeto"
L["AttachToObject_Desc"] = "Observe que nem todos os objetos podem ser selecionados, depende do posicionamento dos outros frames. Por exemplo, se você ancorar os talentos pvp nos raciais e estiver tentando ancorar os talentos raciais nos talentos pvp, haverá uma mensagem de erro."
L["Auras_Buffs_Container_Color_Desc"] = "A cor da borda do ícone de buffs"
L["Auras_CustomFiltering_Conditions_All"] = "Todos"
L["Auras_CustomFiltering_Conditions_Any"] = "Qualquer"
L["Auras_CustomFiltering_ConditionsMode"] = "Modo das Condições"
L["Auras_CustomFiltering_ConditionsMode_Desc"] = "Aqui você pode escolher como as condições selecionadas devem ser aplicadas às auras. Se você escolher \"Qualquer\", o complemento mostrará a aura assim que uma das condições selecionadas for atendida. Se você escolher \"Todos\", o complemento só mostrará uma aura se atender a todas as condições selecionadas."
L["Auras_Debuffs_Container_Color_Desc"] = "A cor da borda do container de icones de debuff"
L["Auras_Enabled"] = "Ativar Auras"
L["Auras_Enabled_Desc"] = "Quando ativado, haverá ícones de aura. Haverá também um número para o tempo restante e um para os stacks aplicados"
L["Auras_Filtering_Mode"] = "Modo de filtragem"
L["Auras_Filtering_Mode_Desc"] = "Aqui você pode escolher qual filtro deve ser aplicado às auras. Você pode escolher entre um filtro semelhante a uma nevasca que aplica o mesmo filtro que a Blizzard usa no frame do raid e condições de filtragem personalizadas"
L["Auras_ShowTooltips"] = "Mostrar dicas das magias"
L["AurasCustomConditions"] = "Condições customizadas"
L["AurasFiltering_AddSpellID"] = "IDmagico"
L["AurasFiltering_AddSpellID_Desc"] = "Digite o IDmagico ou múltiplos IDmagicos separados por virgulas. Os IDmagicos serão adicionados a sua lista de filtragem. Apenas auras que o IDmagico esta na lista de filtragem serão mostrados quando a filtragem de IDmagico esta ativada."
L["AurasFiltering_Enabled_Desc"] = "Quando esta opção esta ativa você apenas vera %s que você adicionou na lista de filtragem"
L["AurasFiltering_Filterlist_Desc"] = "Clique para remover this %s"
L["AurasFilteringSettings_Desc"] = "Aqui você pode controlar qual auras é mostrada"
L["AurasSettings"] = "Auras"
L["AurasSettings_Desc"] = "Configurações sobre auras (buffs e debuffs)"
L["AurasStacktextSettings"] = "Juntar textos"
L["BarBackground"] = "Fundo da barra"
L["BarHeight_Desc"] = "A altura de cada linha. Essa configuração não esta disponível em combate devido as proteções de combate da Blizzard."
L["BarSettings"] = "Configurações da barra"
L["BarSettings_Desc"] = "Aqui você pode configurar as linhas"
L["BarTexture"] = "Textura da barra"
L["BarWidth_Desc"] = "A largura de cada linha. Essa configuração não esta disponível em combate devido as proteções de combate da Blizzard."
L["BattlegroundSize"] = "Tamanho do campo de batalha"
L["BGSize_15"] = "1–15 Jogadores"
L["BGSize_15_Desc"] = "Essa configuração se aplica para %s em um CB com 1-15 jogadores por alvo"
L["BGSize_40"] = "16–40 Jogadores"
L["BGSize_40_Desc"] = "Essas configurações se aplicam para %s em um CB com 16-40 pessoas por lado"
L["BGSize_5"] = "Arena"
L["BGSize_5_Desc"] = "Essas configurações se aplicam para %s em arenas"
L["BlizzlikeAuraFiltering"] = "Usar o quadro de raid padrão da Blizzard"
L["BorderSize"] = "Tamanho da Borda"
L["BorderThickness"] = "Tamanho da borda"
L["BOTTOM"] = "Inferior"
L["BOTTOMLEFT"] = "Inferior esquerda"
L["BOTTOMRIGHT"] = "Inferior direita"
L["buff"] = "buff"
L["BuffContainer"] = "Container de buffs"
L["BuffIcon"] = "Ícone de Buff"
L["Buffs"] = "Buffs"
L["Button"] = "Butão"
L["CantAnchorToItself"] = "Não pode ancorar a si mesmo!"
L["CENTER"] = "Centro"
L["Color"] = "Cor"
L["Coloring_Enabled"] = "Cor por tipo de debuff"
L["Coloring_Enabled_Desc"] = "Quando esta opção está habilitada, você pode escolher se a borda do quadro ou o texto da contagem regressiva deve ser colorido pelo tipo de debuff (como é conhecido na UI padrão, os venenos são verdes, as doenças são azuis, etc.)"
L["Columns"] = "Colunas"
L["Columns_Desc"] = "Em quantas colunas os jogadores serão mostrados"
L["Combat"] = "Em Combate"
L["CombatIndicator"] = "Indicador de Combate"
L["CombatLogScanningForEnemiesDisabled"] = "Log de combate para escanear inimigos foi desativado"
L["CombatLogScanningForEnemiesEnabled"] = "Log de combate para escanear inimigos foi ativado"
L["ConfirmProfileOverride"] = "Você tem certeza que deseja sobrepor o subperfil %s com o subperfil %s"
L["Container_Color"] = "Cor da borda do container"
L["ContainerPosition"] = "Posição do container"
L["ContainerSettings"] = "Configurações do Contêiner"
L["ConvertCyrillic"] = "Converter Cyrillico"
L["ConvertCyrillic_Desc"] = "Converte caracteres \"cyrillicos\" os quais deixa mais fácil ler os nomes se você estiver jogando contra jogadores russos."
L["CopySettings"] = "Copia as configurações de"
L["Countdowntext"] = "Texto de contagem regressiva"
L["Covenant"] = "Ícone do pacto"
L["CovenantIcon_Enabled"] = "Ativar ícones de pactos"
L["CovenantIcon_Enabled_Desc"] = "Ative esta opção se você quiser ver um ícone do pacto na barra de vida"
L["CovenantIcon_Size_Desc"] = "O tamanho do ícone dos pactos (largura e altura)"
L["CovenantIconSettings_Desc"] = "Configurações para o ícone dos pactos que é mostrado na barra de vidas"
L["CurrentVersion"] = "Versão atual"
L["Curse"] = "Maldição"
L["CustomMacro_Desc"] = [=[Isto permite você criar um macro customizado. Um campo vazio ocasiona um macro vazio.
Substituições disponíveis: %n é substituído pelo nome do inimigo correspondente.
Exemplo:
/targetexact %n
/cast Polymorph
/say Polymorph on %n
/targetlasttarget
Esse comando vai mirar no inimigo, lançar polimorfia, escrever uma mensagem no canal aberto e colocar como alvo o seu alvo antigo. Note que macros podem ter apenas 255 caracteres de comprimento ( incluindo substituições) ]=]
L["Data"] = "Dados"
L["debuff"] = "debuff"
L["DebuffContainer"] = "Container de debuff"
L["DebuffIcon"] = "Ícone de Debuff"
L["Debuffs"] = "Debuffs"
L["DebuffType_Filtering"] = "Filtrar por tipo de debuff"
L["DebuffType_Filtering_Desc"] = "Quando esta opção esta ativada você apenas vera debuffs do tipo escolhido. Essa opção não é simulada no modo teste já que iria precisar de uma base de dados enorme, com toneladas de magias e seus tipos de debuff."
L["DeletePoint"] = "Deletar Ponto %s"
L["DisableArenaFramesInArena"] = "Desativar quadros de arena dentro da arena"
L["DisableArenaFramesInArena_Desc"] = "Desativa os quadros de arena quando dentro da arena. Também funciona quando sArena esta instalado."
L["DisableArenaFramesInBattleground"] = "Desativa quadro de arenas quando em campos de batalhas"
L["DisableArenaFramesInBattleground_Desc"] = "Desativa os quadros de arena em um campo de batalha. Também funciona se sArena estiver instalado"
L["Disease"] = "Doença"
L["DispellFilter"] = "Filtros por dispel"
L["DisplayType"] = "Tipo de exibição"
L["Downwards"] = "Para baixo"
L["DR_Disorient"] = "Desorientado"
L["DR_Incapacitate"] = "Incapacitado"
L["DR_Knockback"] = "Empurrão"
L["DR_Root"] = "Enraizamento"
L["DR_Silence"] = "Silenciamento"
L["DR_Stun"] = "Atordoamento"
L["DRContainer"] = "Container de DR"
L["DRTracking"] = "Rastreamento de DR"
L["DrTracking_Container_Color_Desc"] = "A cor da borda do container dos ícones de DR"
L["DrTracking_DisplayType_Desc"] = "Aqui você pode escolher se o status do DR deve ser indicado por uma quadro colorido ao redor do ícone ou por um texto colorido com contagem regressiva. Note que as configurações do texto da contagem regressiva vão funcionar apenas se você não tiver usando um addon como OmniCC que modifica a cor do texto "
L["DrTracking_Enabled"] = "Ativar rastreamento de DR"
L["DrTracking_Enabled_Desc"] = [=[Quando ativado vão mostrar ícones para rastreamento de DR próximo a linha do inimigo.
Borda Verde: Próximo CC vai ter metade da duração.
Borda Amarela: Próximo CC vai ter um quarto da duração.
Borda Vermelha: Próximo CC NÂO vai ser aplicado por que o jogador esta imune.]=]
L["DrTracking_Spacing"] = "Espaçamento de DR"
L["DrTracking_Spacing_Desc"] = "O espaçamento entre cada ícone de DR"
L["DrTrackingFiltering_Enabled_Desc"] = "Quando esta opção é ativada você vera apenas ícones de magias as quais a categoria esta marcada na lista de filtragem"
L["DrTrackingFiltering_Filterlist_Desc"] = "Clique aqui para rastrear ou não essa categoria"
L["DrTrackingFilteringSettings_Desc"] = "Aqui você pode escolher quais categorias de DR serão mostradas."
L["DrTrackingSettings"] = "Rastreamento de DR"
L["DrTrackingSettings_Desc"] = "Configurações sobre o rastreamento de DR"
L["DurationFilter"] = "Filtrar por duração"
L["DurationFilter_OnlyShowWhenDuration"] = "Apenas mostrar quando duração é menor ou igual a "
L["DurationFilter_OnlyShowWhenDuration_Desc"] = "Qualquer aura que a duração é maior que a duração selecionada não passaram pelo filtro"
L["Enable_DrawSwipe"] = "Ativar textura de deslize"
L["Enable_DrawSwipe_Desc"] = "Mostra uma textura dessaturada e mais escura para a parte restante do tempo de recarga."
L["EnableClique"] = "Usar Clique para as teclas de atalho"
L["EnableClique_Desc"] = "Quando esta opção é ativada Clique vai cuidar de todas as teclas de atalho."
L["enemies"] = "inimigos"
L["Enemies"] = "Inimigos"
L["EnemiesTargetingAllies_Enabled_Desc"] = "Quando você ativar esta opção você sera notificado por um som quando um determinado numero de inimigos estiverem com um aliado como alvo."
L["EnemiesTargetingMe_Enabled_Desc"] = "Quando você ativar esta opção você sera notificado por um som quando um determinado numero de inimigos estiverem com você como alvo."
L["enemy"] = "Inimigo"
L["Enemy"] = "Inimigo"
L["EnemyJoined"] = "Um inimigo se juntou ao campo de batalha"
L["EnemyLeft"] = "Um inimigo saiu do campo de batalha"
L["ExportButton"] = "Exportar perfil"
L["ExportButton_Desc"] = "Clique aqui para exportar seu perfil ativo"
L["FakeEventsDisabled"] = "Não mais criando eventos falsos"
L["FakeEventsEnabled"] = "Agora criando eventos falsos para teste"
L["Filtering_Enabled"] = "Ativar filtragem"
L["Filtering_Filterlist"] = "Lista de filtros"
L["FilterSettings"] = "Configurações de Filtros"
L["Font"] = "Fonte"
L["Font_Desc"] = "A fonte principal usada por este addOn"
L["Font_Outline"] = "Contorno da fonte"
L["Font_Outline_Desc"] = "O contorno da fonte"
L["Fontcolor"] = "Cor da fonte"
L["Fontcolor_Desc"] = "A cor do texto"
L["FontShadow_Enabled"] = "Ativar sombra da fonte"
L["FontShadow_Enabled_Desc"] = "Quando ativado vai mostrar uma sombra ao redor da fonte. Você pode escolher a cor dessa sombra"
L["FontShadowColor"] = "Cor da sombra"
L["FontShadowColor_Desc"] = "A cor da sombra do texto"
L["FontSize"] = "Tamanho da fonte"
L["FontSize_Desc"] = "O tamanho da fonte do texto"
L["Frame"] = "Quadro"
L["Framescale"] = "Escala"
L["Framescale_Desc"] = "A escala do quadro principal. Essa configuração não esta disponível durante combate devido a restrições de combate da Blizzard."
L["General"] = "Geral"
L["GeneralSettings"] = "Configurações gerais"
L["GeneralSettings_Desc"] = "Algumas configurações gerais"
L["GeneralSettingsAllies"] = "Essas configurações se aplicam a aliados independente do tamanho do campo de batalha"
L["GeneralSettingsEnemies"] = "Essas configurações se aplicam a inimigos independente do tamanho do campo de batalha"
L["HealthBar"] = "Barra de vida"
L["HealthBar_Background_Desc"] = "Cor de fundo para a barras de vida"
L["HealthBar_Texture_Desc"] = "A textura usada para as barras de vida"
L["HealthBarSettings_Desc"] = "Configurações da barra de vida"
L["HealthTextEnabled"] = "Ativar texto de vida"
L["HealthTextSettings"] = "Configurações do texto de vida"
L["HealthTextType"] = "Tipo do texto de vida"
L["Height"] = "Altura"
L["HighestPriorityAura"] = "Aura de maior prioridade"
L["Highlight_Color"] = "Cor de destaque"
L["Highlight_Color_Desc"] = "A cor do quadro do jogado que o cursor esta em cima no momento"
L["HorizontalGrowdirection"] = "Direção do crescimento horizontal"
L["HorizontalGrowdirection_Desc"] = "Permite você escolher se as linha crescem para o lado esquerdo ou direito"
L["HorizontalSpacing"] = "Espaçamento horizontal"
L["IAmTargeted"] = "Eu sou alvo"
L["Icon"] = "Icone"
L["IconSize"] = "Tamanho do Icone"
L["IconsPerRow"] = "Ícones por linha"
L["Import"] = "Importar"
L["ImportButton"] = "Importar perfil"
L["ImportButton_Desc"] = "Clique aqui para importar um perfil de um texto de sobrepor seu perfil atual ativo"
L["JustifyH"] = "Alinhamento Horizontal"
L["JustifyH_Desc"] = "O alinhamento horizontal do texto"
L["JustifyV"] = "Alinhamento vertical"
L["JustifyV_Desc"] = "O alinhamento vertical do texto"
L["KeybindSettings_Desc"] = "Aqui você pode configurar o que deve acontecer quando apertar o botão esquerdo, direito ou do meio do mouse"
L["LEFT"] = "Esquerda"
L["LeftToTargetCounter"] = "A esquerda do contador de alvos"
L["Leftwards"] = "À esquerda"
L["Level"] = "Nível do Jogador"
L["LevelText_Enabled"] = "Mostrar nível do jogador"
L["LevelText_OnlyShowIfNotMaxLevel"] = "Somente mostrar se não no nível máximo"
L["Locked"] = "Travar"
L["Locked_Desc"] = "Travar o frame no lugar"
L["Magic"] = "Magia"
L["MainFrameSettings"] = "Configurações do quadro principal"
L["MainFrameSettings_Desc"] = "Configurações para o quadro principal que é usado por %s"
L["MaxPlayers"] = "Maximo de inimigos"
L["MaxPlayers_Desc"] = "O addon não vai mostrar os inimigos se a quantidade for maior que este numero"
L["MIDDLE"] = "Meio"
L["ModuleSettings"] = "Configurações do modulo"
L["ModuleSpecificSettings"] = "Configurações especificas do Modulo"
L["MoreProfileOptions"] = "Mais opções de perfil"
L["MyFocus"] = "Meu foco"
L["MyFocus_Color"] = "Cor do foco"
L["MyFocus_Color_Desc"] = "A cor usada pela borda indicando seu foco atual "
L["MyTarget"] = "Meu alvo"
L["MyTarget_Color"] = "Cor do alvo"
L["MyTarget_Color_Desc"] = "A cor usada pela borda ao redor da barra de vida indicando o seu alvo atual"
L["MyVersion"] = "Você esta usando a Versão"
L["Name"] = "Nome"
L["Name_Desc"] = "Configurações para nomes na barra de vida."
L["NewVersionAvailable"] = "Uma nova versão esta disponível"
L["None"] = "Nenhum"
L["NonPriorityBuffs"] = "Buffs não priorizados"
L["NonPriorityDebuffs"] = "Debuffs não priorizados"
L["Normal"] = "Normal"
L["NormalCastingBar"] = "Barra de feitiços normal"
L["NotAvailableInCombat"] = "Essa configuração não estará disponível enquanto você esta em combate devido a restrição de combate da Blizzard"
L["Notifications_Allies_Enabled_Desc"] = "Quando ativo, você será notificado quando um aliado está saindo ou entrando no campo de batalha"
L["Notifications_Enabled"] = "Ativa Notificações"
L["Notifications_Enemies_Enabled_Desc"] = "Quando ativo você será notificado quando um inimigo esta saindo ou entrando no campo de batalha"
L["NoVersion"] = "Nenhum addon foi achado"
L["NumericTargetindicator_Enabled_Desc"] = "Mostra quantos %s estão com o jogador como alvo"
L["ObjectiveAndRespawn_ObjectiveEnabled"] = "Mostrar Objetivos"
L["ObjectiveAndRespawn_ObjectiveEnabled_Desc"] = "Mostra o ícone da bandeira, carro da mina ou orbes próximo ao botão do inimigo"
L["ObjectiveAndRespawn_Position"] = "Posição"
L["ObjectiveAndRespawn_Position_Desc"] = "Se o ícone deveria ser mostrado a esquerda do ícone de especialização ou a direita dos ícones de berloques/raciais"
L["ObjectiveAndRespawn_RespawnEnabled"] = "Ativar o tempo de ressuscitação"
L["ObjectiveAndRespawn_RespawnEnabled_Desc"] = "Quando ativado vai mostrar um ícone indicando o tempo restante até o inimigo estar vivo novamente"
L["ObjectiveAndRespawn_Width_Desc"] = "Largura do ícone de objetivo"
L["ObjectiveAndRespawnSettings_Desc"] = "Configurações sobre os objetivos de campos de batalha"
L["ObjectiveAndRespawnTimer"] = "Objetivo"
L["OffsetX"] = "Deslocamento X"
L["OffsetY"] = "Deslocamento Y"
L["OldVersion"] = "Verão antiga"
L["OutOfCombat"] = "Fora de combate"
L["PlayerCount"] = "Contagem de jogadores"
L["PlayerCount_Enabled_Desc"] = "Se deveria ter um texto para o numero atual de jogadores. Note que essa contagem pode ser diferente do numero atualmente sendo mostrado devido a proteção de combate ou aliados/inimigos deixando um campo de batalha."
L["Point"] = "Apontar"
L["PointAtObject"] = "Apontar ao objeto"
L["Poison"] = "Veneno"
L["Position"] = "Posição"
L["PowerBar"] = "Barra de poderes"
L["PowerBar_Background_Desc"] = "Cor de fundo para a barra de poderes"
L["PowerBar_Enabled"] = "Ativar as barras de recursos"
L["PowerBar_Enabled_Desc"] = "Ative essa opção se você quiser ver barra de recursos como mana, furia etc."
L["PowerBar_Height_Desc"] = "A altura da barra de recursos. Fazer a barra de recursos ficar maior vai automaticamente fazer a barra de vida menor"
L["PowerBar_Texture_Desc"] = "A textura usada para a barra de poderes"
L["PowerBarSettings_Desc"] = "Configurações para a barra de recursos"
L["PriorityBuffs"] = "Buffs priorizados"
L["PriorityDebuffs"] = "Debuffs priorizados"
L["Racial"] = "Racial"
L["Racial_Enabled"] = "Ativar Raciais"
L["Racial_Enabled_Desc"] = "Quando ativado, haverá ícones para as raciais usadas"
L["Racial_Width_Desc"] = "A largura das raciais"
L["RacialFiltering_Enabled_Desc"] = "Quando essa opção esta ativa você vera apenas raciais que estão selecionadas na lista de filtros"
L["RacialFiltering_Filterlist_Desc"] = "Clique para rastrear/desrastrear essa racial"
L["RacialFilteringSettings_Desc"] = "Aqui você pode selecionar quais raciais você quer rastrear"
L["RacialSettings"] = "Raciais"
L["RacialSettings_Desc"] = "Configurações sobre raciais"
L["RangeIndicator_Alpha"] = "Transparência"
L["RangeIndicator_Alpha_Desc"] = "A transparência do quadro do inimigo quando ele esta fora do alcance escolhido"
L["RangeIndicator_Enabled"] = "Ativa o indicador de alcance"
L["RangeIndicator_Enabled_Desc"] = "Quando esta opção é ativada o quadro vai usar a transparência escolhida quando o inimigo esta fora do alcance selecionado"
L["RangeIndicator_Everything"] = "Modifica a transparência de todos os elementos"
L["RangeIndicator_Frames"] = "Modificar a transparência de"
L["RangeIndicator_Frames_Desc"] = "Seleciona quais partes da coluna de inimigos deve ser mostrada com uma transparência maior quando o inimigo está fora do alcance"
L["RangeIndicator_Range"] = "Alcance"
L["RangeIndicator_Range_Desc"] = "Quando um inimigo está mais longe que essa distancia ( em metros) o quadro dele vai usar a transparência escolhida."
L["RangeIndicator_Settings"] = "Indicador de alcance"
L["RangeIndicator_Settings_Desc"] = "Aqui você pode configurar o indicador de alcance"
L["RBGSpecificSettings"] = "Campos de batalha ranqueados"
L["RBGSpecificSettings_Desc"] = "Essa configuração é aplicável somente em RBG (Campos de batalha rankeados)"
L["ResetModule"] = "Resetar modulo"
L["ResetModule_Desc"] = "Clique aqui para resetar as configurações desse modulo para %s:%s"
L["ResetModuleSettings"] = "Resetar as configurações deste modulo"
L["RIGHT"] = "Direita"
L["Rightwards"] = "Para a direita"
L["RoleIcon_Enabled"] = "Ativar ícones de função"
L["RoleIcon_Enabled_Desc"] = "Ative essa opção se quiser ver um ícone com a função ao lado da barra de vida"
L["RoleIcon_Size_Desc"] = "O tamanho dos ícones de função ( largura e altura)"
L["RoleIconSettings"] = "Ícone da função"
L["RoleIconSettings_Desc"] = "Configurações para o ícone de função que é mostrado na barra de vida"
L["Scale"] = "Escala"
L["ShowDispellable"] = "Mostrar dispelaveis"
L["ShowDispellable_Desc"] = "Mostrar %s que podem ser removidos por remover maldições, purgar etc"
L["ShowMine"] = "Mostrar os meus"
L["ShowMine_Desc"] = "Mostrar %s que eu aplicar"
L["ShowNumbers"] = "Mostrar números"
L["ShowNumbers_Desc"] = "Mostra os números do tempo de recarga, note que esta opção pode não funcionar se você estiver usando addons como o OmniCC"
L["ShowRealmnames"] = "Mostrar nomes verdadeiros"
L["ShowRealmnames_Desc"] = "Mostra o reino do inimigo"
L["ShowTooltips"] = "Mostrar dicas"
L["ShowTooltips_Desc"] = "Mostra dicas quando você move o mouse por cima de objetos como berloques e ícones de auras"
L["Side"] = "lado"
L["SideAtObject"] = "ao lado do objeto"
L["Size"] = "Tamanho"
L["SmallCastingBar"] = "Pequena barra de lançamento"
L["SourceFilter"] = "Fonte do filtro"
L["Spec_AuraDisplay_Enabled"] = "Ativar display de auras"
L["Spec_AuraDisplay_Enabled_Desc"] = "Quando esta opção esta ativa você vera ícones de controle de unidades e interrupções (quando ativa) em vez do ícone de especializações"
L["Spec_Enabled"] = "Ativa especialização"
L["Spec_Enabled_Desc"] = "Quando ativado vão ter ícones para as especializações dos players"
L["Spec_Width_Desc"] = "A largura do ícone de especialização"
L["SpecSettings"] = "Especialização"
L["SpecSettings_Desc"] = "Configurações para o ícone de mudança de spec."
L["SpellID_Filtering"] = "Filtrar por ID de habilidade"
L["Style"] = "Estilo"
L["SymbolicTargetindicator_Enabled_Desc"] = "Mostra um quadrado colorido por classe para cada %s que tem o jogador como alvo"
L["TargetAmount"] = "quantidade de alvos"
L["TargetAmount_Ally"] = "O som é tocado assim que um membro do grupo é alvo pela quantidade de oponentes selecionada"
L["TargetAmount_Me"] = "O som é tocado assim que você é alvo pela quantidade de oponentes selecionada"
L["TargetCallerUpdated"] = "é agora o chamador de alvo"
L["TargetCalling"] = "Chamada de Alvo"
L["TargetCallingNotificationEnable"] = "Ativa notificações por som"
L["TargetCallingNotificationEnable_Desc"] = "Notifica você quando tem um novo alvo ( detectado pela alvo com caveira da raid)"
L["TargetCallingSetMark"] = "Marca meu alvo com o ícone da caveira"
L["TargetCallingSetMark_Desc"] = "Marca o seu alvo atual automaticamente com o ícone da caveira quando você é o líder da raid e não tem assistentes no seu grupo. Também marcará o seu alvo atual se você não é o raid líder mas é um assistente."
L["TargetCallingShowIcon"] = "se ativado o alvo do chamador de alvo vai receber um ícone em seu quadro"
L["TargetCallingShowIcon_Desc"] = "Quando ativado, você verá um ícone no quadro que mostra o alvo do chamador de alvos"
L["TargetCallingSoundEnable"] = "Notifique-me em um novo alvo"
L["TargetCallingSoundEnable_Desc"] = "Quando ativado, tocará um som quando o chamador de alvo mudar o alvo para um novo jogador"
L["TargetCallingSoundSound"] = "Som de notificação"
L["TargetIndicator"] = "Indicador de Alvo"
L["TargetIndicator_Desc"] = "Indicadores de alvos numéricos e simbólicos"
L["TargetIndicatorNumeric"] = "Contador de Alvos"
L["TargetIndicatorSymbolic"] = "Símbolos de alvos"
L["Testmode_Toggle"] = "Ativa/Desativa o modo de Testes"
L["Testmode_Toggle_Desc"] = "Ativa/Desativa o modo de testes. Esse modo deixa fácil perceber qual impacto a maioria da opções tem. Não é possível simular todas as opções mas a maior parte delas são cobertas pelo modo de testes. O modo de testes é desativado se você está dentro de um campo de batalha."
L["Testmode_ToggleAnimation"] = "Ativa/Desativa animação no modo teste"
L["Testmode_ToggleAnimation_Desc"] = "Ativa/Desativa a animação do modo teste para que você foque mais em uma configuração especifica sem ser distraído pela animação."
L["Testmode_UseTeammates"] = "Usar membros do grupo para testar"
L["Testmode_UseTeammates_Desc"] = "Quando ativado o modo de teste vai usar os membros do seu grupo em vez de membros falsos"
L["TestmodeDisabled"] = "Modo teste desativado"
L["TestmodeEnabled"] = "Modo teste ativado"
L["TestmodeSettings"] = "Modo Teste"
L["Text"] = "Texto"
L["TextSettings"] = "Opções de texto"
L["Thick"] = "Espessa"
L["TOP"] = "Topo"
L["TOPLEFT"] = "Topo esquerda"
L["TOPRIGHT"] = "Topo direita"
L["Trinket"] = "Berloque"
L["Trinket_Enabled"] = "Ativar Berloques"
L["Trinket_Enabled_Desc"] = "Quando ativado mostrará os ícones para Medalhão do Gladiador, Adaptação e Incansável"
L["Trinket_Width_Desc"] = "A largura do Berloque"
L["TrinketSettings_Desc"] = "Configurações sobre Berloques"
L["UpdatePeriod"] = "Período de atualização"
L["UpdatePeriod_Desc"] = "O tempo entre a checagem de combate em segundos, não coloque isto muito baixo por que pode potencialmente ocasionar problemas de performance"
L["Upwards"] = "Para cima"
L["UseBarHeight"] = "User altura da barra"
L["UseBigDebuffsPriority"] = "Usar a prioridade do BigDebuffs "
L["UseBigDebuffsPriority_Desc"] = "Se você ativar esta opção, o addon vai usar as prioridades do addon BigDebuffs em vez do padrao. As prioridades são usadas pelos modulos %s, %s e %s"
L["UseButtonHeight"] = "Usar altura do botão"
L["UseButtonHeight_Desc"] = "Se você ativar esta opção, o modulo vai usar a altura do botão"
L["UserDefined"] = "Definida pelo usuário"
L["VerticalGrowdirection"] = "Direção de crescimento vertical"
L["VerticalGrowdirection_Desc"] = "Permite você escolher se as barras crescem para cima ou para baixo."
L["VerticalPosition"] = "Posição vertical"
L["VerticalSpacing"] = "Espaçamento vertical"
L["Width"] = "Largura"

elseif LOCALE_ruRU then
L["AddPoint"] = "Добавить точку"
L["Allies"] = "Союзники"
L["allies"] = "союзники"
L["ally"] = "союзник"
L["Ally"] = "Союзник"
L["AllyIsTargeted"] = "Союзник взят в цель"
L["AllyJoined"] = "Союзник присоединился к полю боя"
L["AllyLeft"] = "Союзник покинул поле боя"
L["AND"] = "и"
L["AttachToObject"] = "Прикрепить к объекту"
L["AttachToObject_Desc"] = "Обратите внимание, что не каждый объект может быть выбран, в зависимости от расположения других фреймов. Например, если вы привязали pvp-таланты к расовым способностям, а потом пытаетесь привязать расовые способности к pvp-талантам, то появится сообщение об ошибке."
L["Auras_Buffs_Container_Color_Desc"] = "Цвет границы окна баффов"
L["Auras_CustomFiltering_Conditions_All"] = "Все"
L["Auras_CustomFiltering_Conditions_Any"] = "Любые"
L["Auras_CustomFiltering_ConditionsMode"] = "Режим условий"
L["Auras_CustomFiltering_ConditionsMode_Desc"] = "Здесь вы можете выбрать, как выбранные условия должны применяться к аурам. Если вы выберете «Любой», аддон отобразит ауру, как только будет выполнено одно из выбранных вами условий. Если вы выберете «Все», аддон будет отображать ауру только в том случае, если он соответствует всем выбранным вами условиям."
L["Auras_Debuffs_Container_Color_Desc"] = "Цвет границы окна дебаффов"
L["Auras_Enabled"] = "Включить ауры"
L["Auras_Enabled_Desc"] = "Когда включено, то будут значки ауры. А также будет количество времени перезарядки, а другой - для примененных стаков"
L["Auras_Filtering_Mode"] = "Режим фильтрации"
L["Auras_Filtering_Mode_Desc"] = "Здесь вы можете выбрать, какой фильтр применять к аурам. Вы можете выбрать между фильтром, подобным blizzlike, который применяет тот же фильтр, что и Blizzard, применяемый к рамке рейда, и пользовательскими условиями фильтрации."
L["Auras_ShowTooltips"] = "Показать подсказки заклинаний"
L["AurasCustomConditions"] = "Пользовательские условия"
L["AurasFiltering_AddSpellID"] = "ID Зaклинания"
L["AurasFiltering_AddSpellID_Desc"] = "Введите ID или несколько ID, разделенными запятыми. ID добавляются в список фильтров. При включенной фильтрации ID отображается, только ауры, у которых указаны в списке фильтров."
L["AurasFiltering_Enabled_Desc"] = "Если этот параметр включен, вы увидите только %s, которые были добавлены в список фильтрации"
L["AurasFiltering_Filterlist_Desc"] = "Нажмите, чтобы удалить этот %s"
L["AurasFilteringSettings_Desc"] = "Здесь Вы можете контролировать, какая аура отображается"
L["AurasSettings"] = "Ауры"
L["AurasSettings_Desc"] = "Настройки аур (баффы и дебаффы)"
L["AurasStacktextSettings"] = "Текст стаков"
L["BarBackground"] = "Цвет Фона"
L["BarHeight_Desc"] = "Высота каждой полосы. Изменение этих настроек недоступно во время боя из-за ограничений интерфейса."
L["BarSettings"] = "Настройки полосы"
L["BarSettings_Desc"] = "Здесь вы можете настроить полосы"
L["BarTexture"] = "Текстура полосы"
L["BarWidth_Desc"] = "Ширина каждой полосы. Изменение этих настроек недоступно во время боя из-за ограничений интерфейса."
L["BattlegroundSize"] = "Размер поля боя"
L["BGSize_15"] = "1-15 игроков"
L["BGSize_15_Desc"] = "Эта настройка применяется для %s в ПБ с 1-15 игроками на каждой стороне"
L["BGSize_40"] = "16-40 игроков"
L["BGSize_40_Desc"] = "Эта настройка применяется для %s в ПБ с 16-40 игроками на каждой стороне"
L["BGSize_5"] = "Арена"
L["BGSize_5_Desc"] = "Эта настройка применяется для %s на аренах."
L["BlizzlikeAuraFiltering"] = "Использовать фильтр рейд фреймов от Blizzard"
L["BorderSize"] = "Размер границы"
L["BorderThickness"] = "Толщина границы"
L["BOTTOM"] = "Нижний"
L["BOTTOMLEFT"] = "Нижний левый"
L["BOTTOMRIGHT"] = "Нижний правый"
L["buff"] = "бафф"
L["BuffContainer"] = "Контейнер баффов"
L["BuffIcon"] = "Иконка баффа"
L["Buffs"] = "Баффы"
L["Button"] = "Кнопка"
L["CantAnchorToItself"] = "Невозможно закрепить на самого себя!"
L["CENTER"] = "Центр "
L["Change"] = "Изменить"
L["Color"] = "Цвет"
L["Coloring_Enabled"] = "Цвет по типу дебаффа"
L["Coloring_Enabled_Desc"] = "Включение этой опции дает возможность изменять цвет границы фрейма или текста отсчета по типу дебаффа (как в интерфейсе по умолчанию, яды - зеленые, болезни - синие и т. д.)"
L["Columns"] = "Столбцы"
L["Columns_Desc"] = "Сколько столбцов на игроках отображаются."
L["Combat"] = "В бою"
L["CombatIndicator"] = "Индикатор боя"
L["CombatLogScanningForEnemiesDisabled"] = "Сканирование журнала боя отключено."
L["CombatLogScanningForEnemiesEnabled"] = "Сканирование журнала боя для определения вражеских игроков включено."
L["ConfirmDeletePlayerCountProfile"] = "Вы уверены что хотите удалить профиль примененный к"
L["ConfirmProfileOverride"] = "Вы уверены, что хотите переопределить подпрофиль %s с подпрофилем %s"
L["Container_Color"] = "Цвет границы контейнера"
L["ContainerPosition"] = "Положение контейнера"
L["ContainerSettings"] = "Настройки контейнера"
L["ConvertCyrillic"] = "Преобразовывать кириллицу"
L["ConvertCyrillic_Desc"] = "Преобразовывать кириллические символы для более легкого чтения имен, если вы играете против русских"
L["CopyFrom"] = "Копировать из"
L["CopySettings"] = "Копировать настройки из"
L["Countdowntext"] = "Текст отсчёта"
L["Covenant"] = "Ковенант"
L["CovenantIcon_Enabled"] = "Включить иконки ковенантов"
L["CovenantIcon_Enabled_Desc"] = "Включите это, если вы хотите отобразить иконку ковенанта на полоске здоровья"
L["CovenantIcon_Size_Desc"] = "Размер иконок ковенанта (ширина и высота)"
L["CovenantIconSettings_Desc"] = "Настройки для иконки ковенанта, которая отображается на полоске здоровья"
L["CurrentVersion"] = "Текущая версия"
L["Curse"] = "Curse"
L["CustomMacro_Desc"] = [=[При включении позволяет Вам задать пользовательский макрос. Пустое поле приведет к пустому макросу.
Доступные замены: %n получает замененного именем противника

Пример:
/cast Polymorph
/say Polymorph on %n
/targetlasttarget

Если цель противник, применяет Превращение, и пишет сообщение в канал say и далее выделяет вашу предыдущую цель.
Обратите внимание, что макросы можно писать только до 255 символов (включая замены).]=]
L["Data"] = "данные"
L["debuff"] = "дебафф"
L["DebuffContainer"] = "Контейнер дебаффов"
L["DebuffIcon"] = "Иконка дебаффа"
L["Debuffs"] = "Дебаффы"
L["DebuffType_Filtering"] = "Фильтр по типу дебаффа"
L["DebuffType_Filtering_Desc"] = "Когда этот параметр включен, то вы увидите только дебаффы выбранного типа. Этот параметр не моделируется в тестовом режиме, так как для этого потребуется огромная база данных с тоннами заклинаний и их типом баффа."
L["DeletePlayerCountProfile"] = "Удалить"
L["DeletePlayerCountProfile_Desc"] = "Нажать здесь для удаления профиля примененного к"
L["DeletePoint"] = "Удалить точку %s"
L["DisableArenaFramesInArena"] = "Отключить арена фреймы на арене"
L["DisableArenaFramesInArena_Desc"] = "Отключает аренафреймы на арене. Также работает при установленном sArena"
L["DisableArenaFramesInBattleground"] = "Отключить арена фреймы на полях боя"
L["DisableArenaFramesInBattleground_Desc"] = "Отключает рамки арены на поле боя. Также работает при установке sArena"
L["Disease"] = "Болезнь"
L["DispellFilter"] = "Фильтровать по рассеиваемости"
L["DisplayType"] = "Тип отображения"
L["Downwards"] = "Вниз"
L["DR_Disorient"] = "Дезориентация"
L["DR_Incapacitate"] = "Паралич"
L["DR_Knockback"] = "Отбрасывание"
L["DR_Root"] = "Сковывание"
L["DR_Silence"] = "Немота"
L["DR_Stun"] = "Оглушение"
L["DRContainer"] = "Контейнер DR"
L["DRTracking"] = "Отслеживание DR"
L["DrTracking_Container_Color_Desc"] = "Цвет границы контейнера иконок DR"
L["DrTracking_DisplayType_Desc"] = "Здесь Вы можете выбрать, должен ли статус DR быть с окрашенной рамкой, вокруг значка или быть, с цветовой перезарядкой в тексте. Обратите внимание, что настройки перезарядки текста будут работать - только в том случае, если Вы не используете аддон типа OmniCC, который изменяет цвет текста."
L["DrTracking_Enabled"] = "Включить слежение за DR"
L["DrTracking_Enabled_Desc"] = "Когда включено, слева от вражеских полос будут иконки слежения за диминишингом (DR). Зелёная рамка: Следующий контроль будет действовать половину времени. Жёлтая рамка: Следующий контроль будет действовать четверть времени. Красная рамка: Следующий контроль не будет действовать, потому что цель имунна."
L["DrTracking_Spacing"] = "Интервал DR"
L["DrTracking_Spacing_Desc"] = "Расстояние между иконками DR"
L["DrTrackingFiltering_Enabled_Desc"] = "При включении будут отображаться только иконки заклинаний из выбранных категорий."
L["DrTrackingFiltering_Filterlist_Desc"] = "Щёлкните чтобы отслеживать/не отслеживать эту категорию"
L["DrTrackingFilteringSettings_Desc"] = "Здесь можно выбрать какие категории DR будут отобржаться."
L["DrTrackingSettings"] = "Отслеживание DR"
L["DrTrackingSettings_Desc"] = "Настройки касающиеся отслеживания DR"
L["DurationFilter"] = "Фильтровать по длительности"
L["DurationFilter_OnlyShowWhenDuration"] = "Показывать только если длительность меньше или равна"
L["DurationFilter_OnlyShowWhenDuration_Desc"] = "Любая аура, которая длится дольше выбранного значения, не будет пропущена фильтром"
L["Enable_DrawSwipe"] = "Включить текстуру смахивания"
L["Enable_DrawSwipe_Desc"] = "Показать ненасыщенную, более тёмную текстуру для оставшейся части восстановления"
L["EnableClique"] = "Используйте Clique для привязки клавиш"
L["EnableClique_Desc"] = "Когда этот параметр включен, Clique будет обрабатывать все сочетания клавиш."
L["EnableCustomDRCategoryIcons"] = "Включить категорию своих иконок"
L["EnableCustomDRCategoryIcons_Desc"] = "Когда включено, вы можете выбрать свои иконки для каждой категории наказания"
L["EnableCustomPlayerCountProfiles"] = "Включить свои профили для посчета игроков"
L["EnableCustomPlayerCountProfiles_Desc"] = "Когда настройка включена, она дает вам создавать свои профили под конкретное количество игроков. Текущие профили расчитанные на 1-5, 6-16 и 16-40 ироков утерены не будут, но отключаться и сможете использовать свои профили, вместо стандартных. В можете переключать их когда вам захочеться без потери настроек."
L["EnableInArenas"] = "Включить на Аренах"
L["EnableInBattlegrounds"] = "Включить на Полях боя"
L["enemies"] = "враги"
L["Enemies"] = "Враги"
L["EnemiesTargetingAllies_Enabled_Desc"] = "Когда вы включите эту функцию, вы будете оповещены звуком, когда определённое количество врагов целят союзника"
L["EnemiesTargetingMe_Enabled_Desc"] = "Когда вы включите эту функцию, вы будете оповещены звуком, когда определённое количество врагов целят вас "
L["enemy"] = "враг"
L["Enemy"] = "Враг"
L["EnemyJoined"] = "Враг присоединился к полю боя"
L["EnemyLeft"] = "Враг покинул поле боя"
L["ErrorTestmodeInCombat"] = "Нельзя включить тестовый режим в бою"
L["ExportButton"] = "Экспортировать профиль"
L["ExportButton_Desc"] = "Нажмите сюда для экспорта вашего текущего активного профиля"
L["FakeEventsDisabled"] = "Тестовое создание фейковых событий отключено."
L["FakeEventsEnabled"] = "Тестовое создание фейковых событий включено."
L["Filtering_Enabled"] = "Включить фильтрацию"
L["Filtering_Filterlist"] = "Список фильтров"
L["FilterSettings"] = "Настройки фильтра"
L["Font"] = "Шрифт"
L["Font_Desc"] = "Основной шрифт, используемый этим аддоном"
L["Font_Outline"] = "Контур шрифта"
L["Font_Outline_Desc"] = "Контур для шрифта"
L["Fontcolor"] = "Цвет шрифта"
L["Fontcolor_Desc"] = "Цвет текста"
L["FontShadow_Enabled"] = "Включить тень шрифта"
L["FontShadow_Enabled_Desc"] = "Когда включено будет тень вокруг шрифта. Вы можете выбрать цвет для этой тени"
L["FontShadowColor"] = "Цвет тени"
L["FontShadowColor_Desc"] = "Цвет тени у текста"
L["FontSize"] = "Размер шрифта"
L["FontSize_Desc"] = "Размер шрифта текста"
L["Frame"] = "Рамка"
L["Framescale"] = "Масштаб"
L["Framescale_Desc"] = "Масштаб главной рамки. Эта настройка недоступна в бою из-за боевой защиты Blizzard."
L["General"] = "Основное"
L["GeneralSettings"] = "Основные настройки"
L["GeneralSettings_Desc"] = "Некоторые основные настройки"
L["GeneralSettingsAllies"] = "Эти настройки применяются к союзникам, независимо от размера поля боя"
L["GeneralSettingsEnemies"] = "Эти настройки применяются к врагам, независимо от размера поля боя"
L["HealthBar"] = "Полоса здоровья"
L["HealthBar_Background_Desc"] = "Цвет фона для полос здоровья."
L["HealthBar_Texture_Desc"] = "Текстура, используемая для полосы здоровья"
L["HealthBarSettings_Desc"] = "Настройки для полосы здоровья."
L["HealthTextEnabled"] = "Включить текст здоровья"
L["HealthTextSettings"] = "Настройки текста здоровья"
L["HealthTextType"] = "Тип текста здоровья"
L["Height"] = "Высота"
L["HideArenaframesIn"] = "Скрыть рамки арены в"
L["HideRaidframesIn"] = "Скрыть рамки рейда в"
L["HighestPriorityAura"] = "Аура с максимальным приоритетом"
L["Highlight_Color"] = "Цвет выделения"
L["Highlight_Color_Desc"] = "Цвет рамки игрока, на которым находится Ваш курсор"
L["HorizontalGrowdirection"] = "Горизонтальное направление роста"
L["HorizontalGrowdirection_Desc"] = "Позволяет выбрать, куда растут строки влево или вправо"
L["HorizontalSpacing"] = "Горизонтальный интервал"
L["IAmTargeted"] = "Меня целят "
L["Icon"] = "Иконка"
L["IconSize"] = "Размер иконки"
L["IconsPerRow"] = "Иконка в строке"
L["Import"] = "Импорт"
L["ImportButton"] = "Импортировать профиль"
L["ImportButton_Desc"] = "Нажмите сюда для импортирования профиля из строки и перезаписи вашего текущего активного профиля"
L["JustifyH"] = "Горизонтальное выравнивание"
L["JustifyH_Desc"] = "Горизонтальное выравнивание текста"
L["JustifyV"] = "Вертикальное выравнивание"
L["JustifyV_Desc"] = "Вертикальное выравнивание текста"
L["KeybindSettings_Desc"] = "Здесь Вы можете настроить то, что должно произойти при нажатии левой-правой и средней кнопки"
L["LEFT"] = "Слева"
L["LeftToTargetCounter"] = "Слева от счетчика целей"
L["Leftwards"] = "Влево"
L["Level"] = "Уровень игрока"
L["LevelText_Enabled"] = "Показать уровень игрока"
L["LevelText_OnlyShowIfNotMaxLevel"] = "Показывать только до максимального уровня"
L["Locked"] = "Закрепить"
L["Locked_Desc"] = "Закрепляет рамку на месте"
L["Magic"] = "Магия"
L["MainFrameSettings"] = "Основные настройки рамки"
L["MainFrameSettings_Desc"] = "Настройки основной рамки, который используется для %s"
L["MaxPlayerCount"] = "Максимально игроков"
L["MaxPlayers"] = "Максимально врагов"
L["MaxPlayers_Desc"] = "Аддон не показывает врагов если есть больше врагов, чем заданное число"
L["MIDDLE"] = "Посредине"
L["MinPlayerCount"] = "Минимально игроков"
L["Modules"] = "Модули"
L["ModuleSettings"] = "Настройки модуля"
L["ModuleSpecificSettings"] = "Настройки конкретного модуля"
L["MoreProfileOptions"] = "Больше опций профиля"
L["MyFocus"] = "Моя запомненная цель"
L["MyFocus_Color"] = "Окрашивать запомненную цель"
L["MyFocus_Color_Desc"] = "Цвет, используемый для границ, указывающую, на Вашу текущую запомненную цель"
L["MyTarget"] = "Моя цель"
L["MyTarget_Color"] = "Окрашивать цель"
L["MyTarget_Color_Desc"] = "Цвет, используемый для границ вокруг полос здоровья, указывающую текущую цель"
L["MyVersion"] = "Вы используете версию "
L["Name"] = "Имена"
L["Name_Desc"] = "Настройки имён для полос здоровья."
L["NewVersionAvailable"] = "Доступна новая версия"
L["None"] = "Нет"
L["NonPriorityBuffs"] = "Неприоритетные баффы"
L["NonPriorityDebuffs"] = "Неприоритетные дебаффы"
L["Normal"] = "Обычный"
L["NormalCastingBar"] = "Обычная полоса применения"
L["NotAvailableInCombat"] = "Эта настройка недоступна, когда Вы в бою из-за боевых ограничений Blizzard."
L["Notifications_Allies_Enabled_Desc"] = "Когда включено, Вы будете уведомлены, когда союзник покидает поле боя или присоединяется к нему."
L["Notifications_Enabled"] = "Включить уведомления"
L["Notifications_Enemies_Enabled_Desc"] = "Когда включено, Вы будете уведомлены, когда враг покидает поле боя или присоединяется к нему."
L["NoVersion"] = "Аддон не найден"
L["NumericTargetindicator_Enabled_Desc"] = "Показывает, сколько %s нацелились на игрока."
L["ObjectiveAndRespawn_ObjectiveEnabled"] = "Показывать Ключевые Задачи"
L["ObjectiveAndRespawn_ObjectiveEnabled_Desc"] = "Показывать иконку флага, вагонетки или сферы могущества рядом с полоской"
L["ObjectiveAndRespawn_Position"] = "Позиция"
L["ObjectiveAndRespawn_Position_Desc"] = "Где должна быть иконка Ключевых задач, слева от иконки специализаций или справа от иконок Медальон/рассовых способностей."
L["ObjectiveAndRespawn_RespawnEnabled"] = "Включить Таймер Возрождения"
L["ObjectiveAndRespawn_RespawnEnabled_Desc"] = "При включении, появляется иконка отображающая время до возрождения врага."
L["ObjectiveAndRespawn_Width_Desc"] = "Ширина иконки Ключевых Задач"
L["ObjectiveAndRespawnSettings_Desc"] = "Настройки отображения Ключевых Задач на поле боя"
L["ObjectiveAndRespawnTimer"] = "Ключевые Задачи"
L["OffsetX"] = "Смещение X"
L["OffsetY"] = "Смещение Y"
L["OldVersion"] = "Устаревшая версия"
L["OutOfCombat"] = "Выход из боя"
L["PlayerCount"] = "Подсчет игроков"
L["PlayerCount_Enabled_Desc"] = "Если требуется текст для текущего числа игроков. Обратите внимание, что этот счетчик может отличаться от текущего количества показанных строк, связи с боевой защитой или покиданием союзников/врагов из РПБ"
L["players"] = "игроков"
L["Point"] = "Точка "
L["PointAtObject"] = "Точка на объекте"
L["Poison"] = "Яд"
L["Position"] = "Позиция "
L["PowerBar"] = "Особый ресурс"
L["PowerBar_Background_Desc"] = "Цвет фона для полос ресурсов."
L["PowerBar_Enabled"] = "Включить полосу ресурсов"
L["PowerBar_Enabled_Desc"] = "Включите этот параметр, если Вы не видите полос ресурса, чтобы увидеть такие, как мана, ярость и т. д."
L["PowerBar_Height_Desc"] = "Высота полос ресурсов. Если сделать панель ресурсов выше, она автоматически сократится"
L["PowerBar_Texture_Desc"] = "Текстура используемая для полосы ресурсов"
L["PowerBarSettings_Desc"] = "Настройки для полос особого ресурса"
L["PriorityBuffs"] = "Приоритетные баффы"
L["PriorityDebuffs"] = "Приоритетные дебаффы"
L["Racial"] = "Расовая"
L["Racial_Enabled"] = "Включить расовые способности"
L["Racial_Enabled_Desc"] = "Когда включено будут значки для использованных расовых способностей"
L["Racial_Width_Desc"] = "Ширина значка расовых способностей."
L["RacialFiltering_Enabled_Desc"] = "Когда этот параметр включен, Вы увидите только расы, которые проверяются в списке фильтрации."
L["RacialFiltering_Filterlist_Desc"] = "Нажмите, чтобы отслеживать / не отслеживать эту расовую способность"
L["RacialFilteringSettings_Desc"] = "Здесь вы можете выбрать расовую способность, которую Вы хотите отследить"
L["RacialSettings"] = "Расовые способности"
L["RacialSettings_Desc"] = "Настройки, касающиеся расовых способностей"
L["RangeIndicator_Alpha"] = "Прозрачность"
L["RangeIndicator_Alpha_Desc"] = "Прозрачность полоски, когда враг вне заданной зоны досягаемости"
L["RangeIndicator_Enabled"] = "Включить индикатор расстояния"
L["RangeIndicator_Enabled_Desc"] = "Когда этот параметр включен, рамка будет использовать выбранную прозрачность, когда враг вне заданной зоны досягаемости"
L["RangeIndicator_Everything"] = "Всё"
L["RangeIndicator_Frames"] = "Изменить прозрачность на"
L["RangeIndicator_Frames_Desc"] = "Задайте, какие части врага ряда, должны быть показаны с большей прозрачностью, когда противник находится вне досягаемости."
L["RangeIndicator_Range"] = "Расстояние"
L["RangeIndicator_Range_Desc"] = "Когда враг больше, чем это расстояние (в метрах), то на его рамку будет использоваться выбранная прозрачность"
L["RangeIndicator_Settings"] = "Индикатор расстояния"
L["RangeIndicator_Settings_Desc"] = "Здесь Вы можете настроить индикатор расстояния."
L["RBGSpecificSettings"] = "Рейтинговые ПБ"
L["RBGSpecificSettings_Desc"] = "Эти настройки применяются только в РПБ (Рейтинговое поле боя)"
L["ResetGeneralModule_Desc"] = "нажать здесь, чтобы сбросить основные настройки для этого модуля"
L["ResetModule"] = "Сбросить модуль"
L["ResetModule_Desc"] = "Нажмите сюда для сброса настроек модуля %s:%s"
L["ResetModuleSettings"] = "Сбросить настройки текущего модуля"
L["RestoreDefault"] = "Вернуть по-умолчанию"
L["RIGHT"] = "Справа "
L["Rightwards"] = "Направо"
L["RoleIcon_Enabled"] = "Включить значки ролей"
L["RoleIcon_Enabled_Desc"] = "Включите этот параметр, если Вы хотите увидеть значок роли на полосе здоровья"
L["RoleIcon_Size_Desc"] = "Размер значков для роли (ширина и высота)"
L["RoleIconSettings"] = "Значок ролей"
L["RoleIconSettings_Desc"] = "Настройки значка роли, отображаемого на полосе здоровья"
L["Scale"] = "Масштаб"
L["ShowDispellable"] = "Показать рассеиваемые"
L["ShowDispellable_Desc"] = "Показать %s, которые я могу рассеять или снять."
L["ShowFramesIn"] = "Показ рамок в"
L["ShowMine"] = "Показывать только мой"
L["ShowMine_Desc"] = "Показывать только %s что я применил(а)"
L["ShowNumbers"] = "Отображать время"
L["ShowNumbers_Desc"] = "Показывать кулдаун, эта функция может не работать, если вы используете такие аддоны как OmniCC"
L["ShowRealmnames"] = "Отображать имена игровых миров"
L["ShowRealmnames_Desc"] = "Отображает имя игрового мира врагов"
L["ShowTooltips"] = "Показать подсказки"
L["ShowTooltips_Desc"] = "Показывать подсказки, когда указатель мыши наводится на иконки типа ауры или медальона"
L["Side"] = "Сторона "
L["SideAtObject"] = "Сторона объекта"
L["Size"] = "Размер"
L["SmallCastingBar"] = "Маленькая полоса применения"
L["SourceFilter"] = "Фильтровать по источнику"
L["Spec_AuraDisplay_Enabled"] = "Включить отображение аур"
L["Spec_AuraDisplay_Enabled_Desc"] = "Когда этот параметр включен, вместо значка специализаций отображаются значки для управления толпой и прерываний (когда активна)"
L["Spec_Enabled"] = "Включить специализацию"
L["Spec_Enabled_Desc"] = "При включении будут значки для специализации игроков"
L["Spec_Width_Desc"] = "Ширина значка специализации."
L["SpecSettings"] = "Специализации"
L["SpecSettings_Desc"] = "Настройки, изменяющие значок специализации."
L["SpellID_Filtering"] = "Фильтр по spellID"
L["Style"] = "Стиль"
L["SymbolicTargetindicator_Enabled_Desc"] = "Показывает квадрат, на классовых окрашенных значках для каждого %s, который нацелился на игрока."
L["TargetAmount"] = "количество целей"
L["TargetAmount_Ally"] = "Звук проигрывается, как только участник вашей группы взят в цель, превышающим определенное число, количеством врагов"
L["TargetAmount_Me"] = "Звук проигрывается, как только вы взяты в цель, превышающим определенное число, количеством врагов"
L["TargetCallerUpdated"] = "теперь наводчик "
L["TargetCalling"] = "Наводка"
L["TargetCallingNotificationEnable"] = "Включить звуковое оповещение"
L["TargetCallingNotificationEnable_Desc"] = "Уведомить о появлении новой цели (обнаруживается через метку черепа)"
L["TargetCallingSetMark"] = "Пометить мою цель меткой черепа"
L["TargetCallingSetMark_Desc"] = "Автоматически отмечает текущую цель черепом, когда вы лидер рейда и в вашей группе нет помощников. Также отмечает вашу цель, если вы не лидер рейда, но являетесь помощником."
L["TargetCallingShowIcon"] = "Показать иконку на цели наводчика "
L["TargetCallingShowIcon_Desc"] = "Когда функция включена, то вы увидите иконку, показывающую цель наводчика на его фрейме"
L["TargetCallingSoundEnable"] = "Уведомить о новой цели"
L["TargetCallingSoundEnable_Desc"] = "Когда функция включена, воспроизводится звук при смены цели наводчика"
L["TargetCallingSoundSound"] = "Звук уведомления"
L["TargetIndicator"] = "Индикатор цели"
L["TargetIndicator_Desc"] = "Цифровое и визуальное отображение индикаторов цели"
L["TargetIndicatorNumeric"] = "Счетчик целей"
L["TargetIndicatorSymbolic"] = "Метка Нацеливания"
L["Testmode_Toggle"] = "Тестовый режим"
L["Testmode_Toggle_Desc"] = [=[Включает/выключает режим тестирования. Этот режим позволяет вам легче увидеть на то, что влияет большинства параметров. Не возможно смоделировать все варианты, но большинство из них показаны в этом причудливом тестовом режиме.
Тестовый режим отключен, если Вы находитесь на поле боя.]=]
L["Testmode_ToggleAnimation"] = "Вкл/Выкл анимацию"
L["Testmode_ToggleAnimation_Desc"] = "Включает/выключает анимацию тестового режима, так что Вы можете сосредоточиться на конкретной настройке, не отвлекаясь на анимации."
L["Testmode_UseTeammates"] = "Использовать участников группы для тестового режима"
L["Testmode_UseTeammates_Desc"] = "Когда включено, тестовый режим будет использовать ваших текущих членов группы вместо искусственно созданных"
L["TestmodeDisabled"] = "Тестовый режим выключен"
L["TestmodeEnabled"] = "Тестовый режим включен"
L["TestmodeSettings"] = "Тестовый режим"
L["Text"] = "Текст"
L["TextSettings"] = "Настройки текста"
L["Thick"] = "Толстый"
L["TOP"] = "Верх"
L["TOPLEFT"] = "Верхний левый"
L["TOPRIGHT"] = "В правом верхнем углу"
L["Trinket"] = "Медальон"
L["Trinket_Enabled"] = "Включить Медальоны"
L["Trinket_Enabled_Desc"] = "Когда включено, появится значки для Медальона гладиатора, Адаптации и Непреклонности."
L["Trinket_Width_Desc"] = "Ширина медальона."
L["TrinketSettings_Desc"] = "Настройки, касающиеся медальонов"
L["UpdatePeriod"] = "Период обновления"
L["UpdatePeriod_Desc"] = "Время между проверками в бою, не используйте слишком низкое значение, т.к. это может вызвать проблемы с производительностью."
L["Upwards"] = "Выше"
L["UseBarHeight"] = "Использовать высоту полос"
L["UseBigDebuffsPriority"] = "Использовать приоритеты из BigDebuffs"
L["UseBigDebuffsPriority_Desc"] = "Если вы включите эту опцию, аддон будет использовать приоритеты из аддона BigDebuffs вместо собственных. Приоритеты используются для модулей \"%s\", \"%s\" и \"%s\""
L["UseButtonHeight"] = "Использовать высоту кнопок"
L["UseButtonHeight_Desc"] = "Если вы включите эту опцию, модуль будет использовать высоту кнопок."
L["UserDefined"] = "Пользовательский"
L["VerticalGrowdirection"] = "Вертикальное направление роста"
L["VerticalGrowdirection_Desc"] = "Позволяет выбрать, будут ли полосы расти вверх или вниз"
L["VerticalPosition"] = "Вертикальная позиция"
L["VerticalSpacing"] = "Вертикальный интервал"
L["Width"] = "Ширина"

elseif LOCALE_zhCN then
L["AddPoint"] = "新增对齐点"
L["Allies"] = "友方"
L["allies"] = "友方"
L["ally"] = "友方"
L["Ally"] = "友方"
L["AllyIsTargeted"] = "队友被集火"
L["AllyJoined"] = "一名友方玩家加入了战场"
L["AllyLeft"] = "一名友方玩家离开了战场"
L["AND"] = "和"
L["AttachToObject"] = "目标锚点"
L["AttachToObject_Desc"] = "请注意，并不是每个目标都可以选择，这取决于其他框架设置的定位。例如，你降锚点选择在种族技能或者天赋上时，将会有错误信息提示你。"
L["Auras_Buffs_Container_Color_Desc"] = "增益效果图标区域的外框颜色"
L["Auras_CustomFiltering_Conditions_All"] = "全部符合"
L["Auras_CustomFiltering_Conditions_Any"] = "任意符合"
L["Auras_CustomFiltering_ConditionsMode"] = "条件模式"
L["Auras_CustomFiltering_ConditionsMode_Desc"] = "你可以在此处设置所选条件应如何应用于光环。如果设为“任意符合”，插件会显示满足任何一个条件的光环。如果设为“全部符合”，插件仅显示满足所有条件的光环。"
L["Auras_Debuffs_Container_Color_Desc"] = "减益效果图标区域的外框颜色"
L["Auras_Enabled"] = "开启光环"
L["Auras_Enabled_Desc"] = "启用时，会显示你在敌方身上的所有光环图标，同时会显示剩余时间和堆叠层数"
L["Auras_Filtering_Mode"] = "过滤模式"
L["Auras_Filtering_Mode_Desc"] = "您可以在此处选择应用于光环的过滤器。您可以在 blizzlike 过滤器和自定义过滤条件之间进行选择，blizzlike 过滤器与暴雪在团队框架上使用相同的过滤器"
L["Auras_ShowTooltips"] = "鼠标悬停显示法术信息"
L["AurasCustomConditions"] = "自定义条件"
L["AurasFiltering_AddSpellID"] = "法术ID"
L["AurasFiltering_AddSpellID_Desc"] = "输入一个或多个想要加入到过滤列表的法术ID，使用逗号來分离。启用过滤时，只会显示在列表中法术ID的光环图标。"
L["AurasFiltering_Enabled_Desc"] = "启用时，只能看到已经加入到过滤列表中的%s。"
L["AurasFiltering_Filterlist_Desc"] = "点击删除这个%s"
L["AurasFilteringSettings_Desc"] = "你可以控制显示哪些光环"
L["AurasSettings"] = "光环"
L["AurasSettings_Desc"] = "光环设置 (增益和减益效果)。"
L["AurasStacktextSettings"] = "层数文字"
L["BarBackground"] = "背景颜色"
L["BarHeight_Desc"] = "每行的高度。由于暴雪的功能保护，此设置无法在战斗中使用。"
L["BarSettings"] = "血条设置"
L["BarSettings_Desc"] = "这里可以设置横列。"
L["BarTexture"] = "血条材质"
L["BarWidth_Desc"] = "每行的宽度 ，由于暴雪的战斗保护，此设置无法在战斗中使用。"
L["BattlegroundSize"] = "战场大小"
L["BGSize_15"] = "1-15 名玩家"
L["BGSize_15_Desc"] = "此设置适用于%s在每方1-15人的战场"
L["BGSize_40"] = "16-40 名玩家"
L["BGSize_40_Desc"] = "此设置适用于%s在每方16-40人的战场"
L["BGSize_5"] = "竞技场"
L["BGSize_5_Desc"] = "此设置适用于竞技场中的%s"
L["BlizzlikeAuraFiltering"] = "使用暴雪自带团队框架过滤器"
L["BorderSize"] = "边框尺寸"
L["BorderThickness"] = "边框大小"
L["BOTTOM"] = "下"
L["BOTTOMLEFT"] = "左下"
L["BOTTOMRIGHT"] = "右下"
L["buff"] = "增益效果"
L["BuffContainer"] = "增益效果区域"
L["BuffIcon"] = "增益效果图标"
L["Buffs"] = "增益效果"
L["Button"] = "按钮"
L["CantAnchorToItself"] = "不能将锚点设置在此"
L["CENTER"] = "中"
L["Change"] = "更改"
L["Color"] = "颜色"
L["Coloring_Enabled"] = "通过减益效果显示颜色"
L["Coloring_Enabled_Desc"] = "启用时，可以选择框架边框或倒计时文本是否要通过减益效果的类型显示颜色（和默认界面一致，中毒是绿色，疾病是蓝色等等）"
L["Columns"] = "纵列"
L["Columns_Desc"] = "玩家需要显示成多少纵列"
L["Combat"] = "战斗中"
L["CombatIndicator"] = "战斗指示器"
L["CombatLogScanningForEnemiesDisabled"] = "禁用对敌方的战斗日志扫描"
L["CombatLogScanningForEnemiesEnabled"] = "已启用检测敌方玩家的战斗日志扫描"
L["ConfirmDeletePlayerCountProfile"] = "你确定要删除这个配置文件吗？ 该配置文件当前用于"
L["ConfirmProfileOverride"] = "你确定要将%s设置从%s设置覆盖吗"
L["Container_Color"] = "区域边框颜色"
L["ContainerPosition"] = "区域位置"
L["ContainerSettings"] = "从%s复制设定"
L["ConvertCyrillic"] = "转换俄文"
L["ConvertCyrillic_Desc"] = "如果使用俄文版游戏时转换俄文，使名字更容易阅读。"
L["Cooldown"] = "冷却"
L["CopyFrom"] = "复制自"
L["CopySettings"] = "复制设置"
L["CopySettingsMirrored_Desc"] = "镜像表示这个模块的位置和生长方向都被替换为水平反转的位置和方向。（例如，“左上”会变为“右上”；类似光环模块的容器类模块会以反方向生长。）"
L["CopySettingsNormal_Desc"] = "只是一个简单的拷贝"
L["Countdowntext"] = "倒计时文本"
L["Covenant"] = "盟约图标"
L["CovenantIcon_Enabled"] = "启用盟约图标"
L["CovenantIcon_Enabled_Desc"] = "启用时，会在血条上显示盟约图标"
L["CovenantIcon_Size_Desc"] = "盟约图标大小 (宽和高)"
L["CovenantIconSettings_Desc"] = "血条上盟约图标设置"
L["CurrentVersion"] = "当前版本"
L["Curse"] = "诅咒"
L["CustomMacro_Desc"] = [=[这里您可以设置自定义宏。空字段将导致空宏。
可用的替换符号: %n 被替换为敌人的名字。

范例:
/targetexact %n
/施放 变形术
/说 已將 %n 变羊
/targetlasttarget

这个宏会将敌人选为目标施、放变形术、发送一条信息到“说”频道，然后选择您的前一个目标为当前目标
注意宏最长只能使用 255 个字符(包阔替换的名字)。]=]
L["Data"] = "数据"
L["debuff"] = "减益效果"
L["DebuffContainer"] = "减益效果区域"
L["DebuffIcon"] = "减益效果图标"
L["Debuffs"] = "减益效果"
L["DebuffType_Filtering"] = "通过减益效果过滤"
L["DebuffType_Filtering_Desc"] = "如果启用此选项，将只看到所选择类型的减益效果。测试模式中无法模拟这个选项，因为这需要一个庞大的法术和减益效果数据库。"
L["DeletePlayerCountProfile"] = "删除"
L["DeletePlayerCountProfile_Desc"] = "点击此处删除配置文件，该配置文件当前用于"
L["DeletePoint"] = "删除对齐点%s"
L["DisableArenaFramesInArena"] = "在竞技场内禁用竞技场框架"
L["DisableArenaFramesInArena_Desc"] = "在竞技场内禁用竞技场框架。也适用于sArena插件。"
L["DisableArenaFramesInBattleground"] = "在战场中禁用竞技场框架"
L["DisableArenaFramesInBattleground_Desc"] = "在战场中禁用竞技场框架，也适用于Sarena插件。"
L["Disease"] = "疾病"
L["DispellFilter"] = "按是否可驱散过滤"
L["DisplayType"] = "显示类型"
L["Downwards"] = "向下"
L["DR_Disorient"] = "迷惑"
L["DR_Incapacitate"] = "瘫痪"
L["DR_Knockback"] = "击退"
L["DR_Root"] = "定身"
L["DR_Silence"] = "沉默"
L["DR_Stun"] = "昏迷"
L["DRContainer"] = "控制技能区域"
L["DRTracking"] = "控制递减监控"
L["DrTracking_Container_Color_Desc"] = "控制技能图标区域的外框颜色"
L["DrTracking_DisplayType_Desc"] = "这里你可以选择控制递减状态是由图标边框颜色显示或者倒计时文本显示，请注意，倒计时文本计时只能使用一个插件如果你使用omnicc请关闭计时数字."
L["DrTracking_Enabled"] = "开启控制递减监视"
L["DrTracking_Enabled_Desc"] = "当此选项启用时，会在敌方横列边上显示控制技能递减的图标。绿色边框：下一个控制技能持续一半时间。黄色边框：下一个控制技能持续四分之一时间。红色边框：下一个控制技能无效，因为玩家已经免疫。"
L["DrTracking_Spacing"] = "控制递减间距"
L["DrTracking_Spacing_Desc"] = "控制递减图标之间的距离。"
L["DrTrackingFiltering_Enabled_Desc"] = "当此选项启用时，你只会看到在过滤列表中选中的法术图标."
L["DrTrackingFiltering_Filterlist_Desc"] = "点击监视/不监视类型"
L["DrTrackingFilteringSettings_Desc"] = "这里可以选择显示哪种控制递减类型"
L["DrTrackingSettings"] = "控制递减监视"
L["DrTrackingSettings_Desc"] = "控制时间和递减效果的监视。"
L["DurationFilter"] = "按持续时间过滤"
L["DurationFilter_OnlyShowWhenDuration"] = "仅当持续时间较短或相同时显示"
L["DurationFilter_OnlyShowWhenDuration_Desc"] = "任何持续时间超过所选持续时间的光环都会被过滤"
L["Editmode_Toggle"] = "切换编辑模式"
L["Editmode_Toggle_Desc"] = "开启/关闭编辑模式。基本上编辑模式就是测试模式，但是它可以让你移动所有的框体。选择一个框体时插件会自动跳转到对应的设置页面。你可以微调偏移量，或者到综合设置中为特定框体进行设置。"
L["EditmodeDisabled"] = "编辑模式关闭"
L[ [=[EditmodeEnabled
]=] ] = "编辑模式开启"
L["EditModeIntroduction"] = "欢迎进入编辑模式。这个功能可以让你按照自己的想法轻松地排布各种模块。点击一个蓝色模块以选中，然后你可以拖动到你想摆放的位置。如果你把它放到另一个模块的附近，它会自动锚定到临近的模块。当你选中一个模块时，它会显示为黄色高亮，同时选项菜单跳转到这个模块的设置页面。"
L["Enable_DrawSwipe"] = "开启冷却计时效果"
L["Enable_DrawSwipe_Desc"] = "在冷却计时时显示效果"
L["EnableClique"] = "使用Clique设置按键"
L["EnableClique_Desc"] = "当启用此项时，Clique插件会接管所有按键设置。"
L["EnableCustomDRCategoryIcons"] = "开启自定义的控制递减分类图标"
L["EnableCustomDRCategoryIcons_Desc"] = "当开启时，你可以为控制递减分类选择自定义的图标"
L["EnableCustomPlayerCountProfiles"] = "为玩家数量开启自定义配置文件"
L["EnableCustomPlayerCountProfiles_Desc"] = "当此选项开启时，你可以自定义基于玩家数量的配置文件。当前使用的1-5，6-15和16-40的配置文件会被禁用而不会丢失。你可以使用它们作为你自定义的配置文件的基准。你可以随时切换回来，而不会丢失任何设置。"
L["EnableInArenas"] = "在竞技场中开启"
L["EnableInBattlegrounds"] = "在战场中开启"
L["enemies"] = "敌方"
L["Enemies"] = "敌方"
L["EnemiesTargetingAllies_Enabled_Desc"] = "启用此选项时，当设定数量的敌人选中一个队友时，你会收到声音通知"
L["EnemiesTargetingMe_Enabled_Desc"] = "启用此选项时，当设定数量的敌人选中你时，你会收到声音通知"
L["enemy"] = "敌方"
L["Enemy"] = "敌方"
L["EnemyJoined"] = "一个敌对玩家加入了战场"
L["EnemyLeft"] = "一个敌对玩家离开了战场"
L["ErrorTestmodeInCombat"] = "无法在战斗中开启测试模式"
L["ExportButton"] = "导出配置文件"
L["ExportButton_Desc"] = "点击此处导出你的配置文件"
L["FakeEventsDisabled"] = "不再创建假事件。"
L["FakeEventsEnabled"] = "现在为测试创建假事件。"
L["Filtering_Enabled"] = "开启过滤"
L["Filtering_Filterlist"] = "过滤列表"
L["FilterSettings"] = "过滤器设置选项"
L["Font"] = "字体"
L["Font_Desc"] = "插件所使用的主要字体。"
L["Font_Outline"] = "文字描边"
L["Font_Outline_Desc"] = "文字描边"
L["Fontcolor"] = "文字颜色"
L["Fontcolor_Desc"] = "文本字体颜色"
L["FontShadow_Enabled"] = "开启文字阴影"
L["FontShadow_Enabled_Desc"] = "启用时，字体周围会出现阴影。你可以选择阴影的颜色。"
L["FontShadowColor"] = "阴影颜色"
L["FontShadowColor_Desc"] = "文字阴影颜色"
L["FontSize"] = "字体大小"
L["FontSize_Desc"] = "文本字体大小"
L["Frame"] = "框架"
L["Framescale"] = "缩放大小"
L["Framescale_Desc"] = "主要框架的縮放大小。由于暴雪的战斗保护，此设置无法在战斗中使用。"
L["General"] = "常规"
L["GeneralSettings"] = "常规设置"
L["GeneralSettings_Desc"] = "一些基本的设置。"
L["GeneralSettingsAllies"] = "这种设置适用于友方，不管战场大小。"
L["GeneralSettingsEnemies"] = "这种设置适用于敌方，不管战场大小。"
L["HealthBar"] = "血条"
L["HealthBar_Background_Desc"] = "血条背景颜色"
L["HealthBar_Texture_Desc"] = "血条材质"
L["HealthBarSettings_Desc"] = "血条设置。"
L["HealthTextEnabled"] = "打开血条文字"
L["HealthTextSettings"] = "血条文字设置"
L["HealthTextType"] = "血条文字类型"
L["Height"] = "高度"
L["HideArenaframesIn"] = "隐藏竞技场框架于"
L["HideHighestPriority"] = "隐藏最高优先级"
L["HideHighestPriority_Desc"] = "开启时，最高优先级的光环会被移除。当职业或专精图标已经显示最高优先级光环时，这个功能有作用。"
L["HideRaidframesIn"] = "隐藏团队框架于"
L["HideWhenZero"] = "为0时隐藏"
L["HighestPriorityAura"] = "最高优先级光环"
L["Highlight_Color"] = "高亮颜色"
L["Highlight_Color_Desc"] = "当前鼠标指向的玩家框架颜色"
L["HorizontalGrowdirection"] = "水平增长方向"
L["HorizontalGrowdirection_Desc"] = "选择水平向左或者向右方向增长"
L["HorizontalSpacing"] = "水平间距"
L["IAmTargeted"] = "我被集火"
L["Icon"] = "图标"
L["Icons"] = "图标"
L["IconSize"] = "图标大小"
L["IconsPerRow"] = "每行图标"
L["Import"] = "导入"
L["ImportButton"] = "导入配置"
L["ImportButton_Desc"] = "点击此处从字符串导入并覆盖当前设定的配置文件"
L["JumpToGeneralOptions"] = "跳转到综合设置"
L["JumpToGeneralOptions_Desc"] = "点击以跳转到这个模块的综合设置"
L["JustifyH"] = "水平对齐"
L["JustifyH_Desc"] = "文本水平对齐"
L["JustifyV"] = "垂直对齐"
L["JustifyV_Desc"] = "文本垂直对齐"
L["KeybindSettings_Desc"] = "这里你可以设置鼠标左键、中键和右键点击配置。"
L["LEFT"] = "左侧"
L["LeftToTargetCounter"] = "目标数量的左侧"
L["Leftwards"] = "向左"
L["Level"] = "玩家等级"
L["LevelText_Enabled"] = "显示玩家等级"
L["LevelText_OnlyShowIfNotMaxLevel"] = "只显示非满级玩家的等级"
L["Locked"] = "锁定位置"
L["Locked_Desc"] = "锁定框架位置。"
L["Magic"] = "魔法"
L["MainFrameSettings"] = "主框架设置"
L["MainFrameSettings_Desc"] = "%s所使用的主框架设置"
L["MaxCantBeUndefined"] = "最大值不可以是未定义"
L["MaxNeedsToBeEqualOrGreater"] = "最大值必须等于或大于最小值"
L["MaxPlayerCount"] = "最大玩家数量"
L["MaxPlayers"] = "最多敌方玩家数量"
L["MaxPlayers_Desc"] = "超过这个数量的敌方玩家插件将不会显示。"
L["MIDDLE"] = "中间"
L["MinCantBeUndefined"] = "最小值不可以是未定义"
L["MinPlayerCount"] = "最小玩家数量"
L["Mirrored"] = "镜像"
L["Modules"] = "模块"
L["ModuleSettings"] = "模块设置"
L["ModuleSpecificSettings"] = "特定模块设置"
L["MoreProfileOptions"] = "更多配置文件选项"
L["MouseWheelPlayerTargeting"] = "鼠标滚轮选择目标"
L["MouseWheelPlayerTargeting_Desc"] = "开启这个选项时，你可以移动你的鼠标指针到敌方或者友方框体上，然后滚动滚轮以选择列表上的下一个或者上一个玩家。"
L["MyFocus"] = "我的焦点"
L["MyFocus_Color"] = "焦点颜色"
L["MyFocus_Color_Desc"] = "用于显示当前焦点的边框的颜色。"
L["MyTarget"] = "我的目标"
L["MyTarget_Color"] = "目标颜色"
L["MyTarget_Color_Desc"] = "用于显示当前目标的边框的颜色。"
L["MyVersion"] = "你正在使用版本"
L["Name"] = "名字"
L["Name_Desc"] = "血条上名字的设置。"
L["NewVersionAvailable"] = "有新的版本可用"
L["None"] = "无"
L["NonPriorityBuffs"] = "非优先级buffs"
L["NonPriorityDebuffs"] = "非优先级debuffs"
L["Normal"] = "一般"
L["NormalCastingBar"] = "一般施法条"
L["NotAvailableInCombat"] = "由于暴雪的战斗保护，此设置无法在战斗中使用。"
L["Notifications_Allies_Enabled_Desc"] = "启用时，您将在友方离开或加入战场时得到通知。"
L["Notifications_Enabled"] = "开启通知"
L["Notifications_Enemies_Enabled_Desc"] = " 开启时，敌方玩家加入或离开时会显示通知"
L["NoVersion"] = "插件未找到"
L["NumericTargetindicator_Enabled_Desc"] = "显示有多少%s的当前目标是这个玩家"
L["ObjectiveAndRespawn_ObjectiveEnabled"] = "显示目标"
L["ObjectiveAndRespawn_ObjectiveEnabled_Desc"] = "在敌方按钮旁显示旗帜、矿车、或携带宝珠图标。"
L["ObjectiveAndRespawn_Position"] = "位置"
L["ObjectiveAndRespawn_Position_Desc"] = "图标显示在专精图标的左侧，或饰品/种族技能图标的右侧。"
L["ObjectiveAndRespawn_RespawnEnabled"] = "开启复活计时"
L["ObjectiveAndRespawn_RespawnEnabled_Desc"] = "启用时，会显示敌方再次复活剩余时间的图标。"
L["ObjectiveAndRespawn_Width_Desc"] = "目标图标的宽度。"
L["ObjectiveAndRespawnSettings_Desc"] = "关于战场目标的设置。"
L["ObjectiveAndRespawnTimer"] = "目标和复活计时器"
L["OffsetX"] = "水平位置"
L["OffsetY"] = "垂直位置"
L["OldVersion"] = "旧版本"
L["OnlyShowForHealers"] = "仅为治疗职责显示"
L["OutOfCombat"] = "退出战斗"
L["OverwriteMirroredConfirm"] = "你确定要把%s的设置覆盖为%s的镜像设置吗？"
L["OverwriteNormalConfirm"] = "你确定要把%s的设置覆盖为%s的设置吗？"
L["PlayerCount"] = "玩家数量"
L["PlayerCount_Enabled_Desc"] = "是否要显示当前玩家数量。请注意，这个数量可能会和当前列表显示不同，因为插件在战斗中保护功能及有人离开随机战场。"
L["players"] = "玩家"
L["Point"] = "点"
L["PointAtObject"] = "目标点"
L["Poison"] = "中毒"
L["Position"] = "位置"
L["PowerBar"] = "能量条"
L["PowerBar_Background_Desc"] = "能量条背景颜色。"
L["PowerBar_Enabled"] = "开启能量条"
L["PowerBar_Enabled_Desc"] = "启用时，会显示法力、怒气等资源。"
L["PowerBar_Height_Desc"] = "能量条的高度。能量条越高生命条就会越窄。"
L["PowerBar_OnlyShowForHealers_Desc"] = "开启时，仅显示治疗职责玩家的能量条"
L["PowerBar_Texture_Desc"] = "能量条材质。"
L["PowerBarSettings_Desc"] = "能量条设置。"
L["PriorityBuffs"] = "优先级buffs"
L["PriorityDebuffs"] = "优先级debuffs"
L["Racial"] = "种族"
L["Racial_Enabled"] = "开启种族技能"
L["Racial_Enabled_Desc"] = "启用时，显示已使用的种族技能图标。"
L["Racial_Width_Desc"] = "种族技能宽度"
L["RacialFiltering_Enabled_Desc"] = "启用时，只会看到过滤列表中的种族技能。"
L["RacialFiltering_Filterlist_Desc"] = "点击监视/取消监视这个种族技能。"
L["RacialFilteringSettings_Desc"] = "在这里可以选择你想要追踪的种族技能"
L["RacialSettings"] = "种族技能"
L["RacialSettings_Desc"] = "种族技能设置。"
L["RangeIndicator_Alpha"] = "透明度"
L["RangeIndicator_Alpha_Desc"] = "敌人在超出指定范围内，敌方框架的透明度。"
L["RangeIndicator_Enabled"] = "开启距离提示"
L["RangeIndicator_Enabled_Desc"] = "启用时，敌人超出指定距离时，敌方框架会显示成指定的透明度。"
L["RangeIndicator_Everything"] = "更改所有透明度"
L["RangeIndicator_Frames"] = "更改透明度"
L["RangeIndicator_Frames_Desc"] = "当敌人超出距离范围时，敌方框架应该具有更高的透明度。"
L["RangeIndicator_Range"] = "距离范围"
L["RangeIndicator_Range_Desc"] = "敌人在这个距离范围以外时，敌方框架会显示成指定的透明度。"
L["RangeIndicator_Settings"] = "距离提示"
L["RangeIndicator_Settings_Desc"] = "这里可以调整距离提示"
L["RangeOverlapping"] = "范围重合：%s"
L["RBGSpecificSettings"] = "评级战场"
L["RBGSpecificSettings_Desc"] = "这些设置只会生效于评级战场。"
L["ResetGeneralModule_Desc"] = "点击此处重置这个模块的常规设置"
L["ResetModule"] = "重置模块"
L["ResetModule_Desc"] = "单击此处为%s重置此模块的设置：%s"
L["ResetModuleSettings"] = "重置此模块的设置"
L["RestoreDefault"] = "恢复默认"
L["RIGHT"] = "右侧"
L["Rightwards"] = "向右"
L["RoleIcon_Enabled"] = "开启职责图标"
L["RoleIcon_Enabled_Desc"] = "启用时，会在血条上显示职责图标。"
L["RoleIcon_Size_Desc"] = "职责图标的大小 (宽和高)"
L["RoleIconSettings"] = "职责图标"
L["RoleIconSettings_Desc"] = "血条上职责图标的设置"
L["RoleSortingOrder"] = "职责排序顺序"
L["RoleSortingOrder_Desc"] = "基于玩家的职责，选择以什么顺序进行排序"
L["Scale"] = "比例"
L["ShowDispellable"] = "显示可驱散的法术"
L["ShowDispellable_Desc"] = "显示可以通过解除诅咒、净化术等等移除的%s"
L["ShowFramesIn"] = "显示框架于"
L["ShowHighestPriority"] = "显示最高优先级"
L["ShowHighestPriority_Desc"] = "开启时，插件会在玩家被打断或者被一个优先级光环影响时显示其图标"
L["ShowMine"] = "只显示自己的"
L["ShowMine_Desc"] = "只显示自己施放的%s"
L["ShowNumbers"] = "显示数字"
L["ShowNumbers_Desc"] = "显示冷却计时，请注意，当你使用 OmniCC 等插件时，此选项可能不起作用"
L["ShowRealmnames"] = "显示服务器"
L["ShowRealmnames_Desc"] = "显示敌方的服务器名"
L["ShowSpecIfExists"] = "如果存在时，显示专精"
L["ShowSpecIfExists_Desc"] = "当该选项启用时，将会用专精图标替代职业图标"
L["ShowTooltips"] = "显示鼠标提示"
L["ShowTooltips_Desc"] = "当您将鼠标移动到饰品和光环图标等目标时显示鼠标提示"
L["Side"] = "方向"
L["SideAtObject"] = "目标方向"
L["Size"] = "大小"
L["SmallCastingBar"] = "小型施法条"
L["SourceFilter"] = "按照来源过滤"
L["Spec_AuraDisplay_Enabled"] = "开启光环显示"
L["Spec_AuraDisplay_Enabled_Desc"] = "启用此选项后，将会在专精图标上显示控制技能和打断技能图标（当控制技能在目标身上时）"
L["Spec_Enabled"] = "开启专精"
L["Spec_Enabled_Desc"] = "启用时，显示玩家的专精图标"
L["Spec_Width_Desc"] = "专精图标宽度。"
L["SpecClassPriorityOne"] = "职业专精和优先级模块1"
L["SpecClassPriorityTwo"] = "职业专精和优先级模块2"
L["SpecSettings"] = "专精"
L["SpecSettings_Desc"] = "专精图标设置。"
L["SpellID_Filtering"] = "法术id过滤"
L["Style"] = "风格"
L["SymbolicTargetindicator_Enabled_Desc"] = "在被敌方选中的目标%s 上显示方形的职业颜色图标"
L["TargetAmount"] = "选中数量"
L["TargetAmount_Ally"] = "当队友被超过设定数量的敌方选为目标，就会播放声音"
L["TargetAmount_Me"] = "当你被超过设定数量的敌方选为目标，就会播放声音。"
L["TargetCallerUpdated"] = "是当前喊火指挥"
L["TargetCalling"] = "喊火指挥"
L["TargetCallingNotificationEnable"] = "启用音效通知"
L["TargetCallingNotificationEnable_Desc"] = "通知你有新的目标（通过目标骷髅图标检测）"
L["TargetCallingSetMark"] = "用骷髅图标标记我的目标"
L["TargetCallingSetMark_Desc"] = "当你是团队领袖且没有团队助理时，会自动使用骷髅图标来标记当前目标。如果你不是团队领袖而是助理，也会标记你当前的目标。"
L["TargetCallingShowIcon"] = "喊火指挥的目标上显示图标"
L["TargetCallingShowIcon_Desc"] = "启用后，将在框架上看到一个图标，显示喊火指挥的目标"
L["TargetCallingSoundEnable"] = "有新目标时通知我"
L["TargetCallingSoundEnable_Desc"] = "启用后，当喊火指挥选中新目标时会发出声音"
L["TargetCallingSoundSound"] = "通知音效"
L["TargetIndicator"] = "目标指示"
L["TargetIndicator_Desc"] = "数字和符号的目标指示器。"
L["TargetIndicatorNumeric"] = "目标指示器数字"
L["TargetIndicatorNumeric_HideWhenZero_Desc"] = "开启时，如果没有敌方的目标是该玩家，文字会隐藏"
L["TargetIndicatorSymbolic"] = "目标指示器符号"
L["TargetNextAlly"] = "选中下一个队友"
L["TargetNextEnemy"] = "选中下一个敌人"
L["TargetPreviousAlly"] = "选中上一个被选中的队友"
L["TargetPreviousEnemy"] = "选中上一个被选中的敌人"
L["Testmode_Toggle"] = "测试模式"
L["Testmode_Toggle_Desc"] = [=[启用/禁用测试模式，这种模式使你更容易看到大多数选项功能。模拟所有选项是不可能的，但大多数设置选项均已在这个测试模式中。
当你在战场中时，测试模式将会被禁用。]=]
L["Testmode_ToggleAnimation"] = "测试模式动画"
L["Testmode_ToggleAnimation_Desc"] = "开启/禁用测试模式的动画，使你可以更专注于特定的设置而不受动画影响."
L["Testmode_UseTeammates"] = "测试时使用队友"
L["Testmode_UseTeammates_Desc"] = "启用测试模式时显示当前队友，而非假的玩家"
L["TestmodeDisabled"] = "测试模式已禁用"
L["TestmodeEnabled"] = "测试模式已启用"
L["TestmodeSettings"] = "测试模式"
L["Text"] = "文字"
L["TextSettings"] = "文字设置"
L["Thick"] = "饰品"
L["TOP"] = "上"
L["TOPLEFT"] = "左上"
L["TOPRIGHT"] = "右上"
L["Trinket"] = "饰品"
L["Trinket_Enabled"] = "开启饰品"
L["Trinket_Enabled_Desc"] = "启用时会显示徽章等饰品冷却时间"
L["Trinket_Width_Desc"] = "饰品宽度"
L["TrinketSettings_Desc"] = "饰品设置"
L["UpdatePeriod"] = "更新间隔"
L["UpdatePeriod_Desc"] = "检查进出战斗时间，不要设置太低，因为这可能会导致游戏内性能问题。"
L["Upwards"] = "向上"
L["UseBarHeight"] = "使用横列高度"
L["UseBigDebuffsPriority"] = "使用 BigDebuffs的 优先级"
L["UseBigDebuffsPriority_Desc"] = "如果启用此选项，插件将使用BigDebuff插件优先级。优先级用于%s、%s和%s模块"
L["UseButtonHeight"] = "使用按钮高度"
L["UseButtonHeight_Desc"] = "如果启用此选项，模块将使用按钮高度。"
L["UserDefined"] = "玩家自定义"
L["VerticalGrowdirection"] = "垂直增长方向"
L["VerticalGrowdirection_Desc"] = "选择横列向上或者向下增长"
L["VerticalPosition"] = "垂直位置"
L["VerticalSpacing"] = "垂直间距"
L["Width"] = "宽度"

elseif LOCALE_zhTW then
L["AddPoint"] = "新增對齊點"
L["Allies"] = "友方"
L["allies"] = "友方"
L["ally"] = "友方"
L["Ally"] = "友方"
L["AllyIsTargeted"] = "攻擊目標是我方"
L["AllyJoined"] = "一位盟友加入了戰場"
L["AllyLeft"] = "一位盟友離開了戰場"
L["AND"] = "和"
L["AttachToObject"] = "對齊物件"
L["AttachToObject_Desc"] = "請注意，並不是每個物件都可以選擇，會依據和其他框架的相對位置而定。例如，如果已經將 PVP 天賦對齊到種族技能，而又嘗試將種族技能對齊到 PVP 天賦時，會出現錯誤訊息。"
L["Auras_Buffs_Container_Color_Desc"] = "增益效果圖示區域的外框顏色。"
L["Auras_CustomFiltering_Conditions_All"] = "全部"
L["Auras_CustomFiltering_Conditions_Any"] = "任何"
L["Auras_CustomFiltering_ConditionsMode"] = "條件模式"
L["Auras_CustomFiltering_ConditionsMode_Desc"] = "在這裡可以選擇，所選的條件要如何套用到光環。選擇 \"任何\" 時，只要符合任一條件就會顯示光環。選擇 \"全部\" 時，只有符合全部的條件才會顯示光環。"
L["Auras_Debuffs_Container_Color_Desc"] = "減益效果圖示區域的外框顏色。"
L["Auras_Enabled"] = "啟用光環"
L["Auras_Enabled_Desc"] = "啟用時，會顯示光環圖示，同時也會顯示剩餘時間和堆疊層數的數字。"
L["Auras_Filtering_Mode"] = "過濾模式"
L["Auras_Filtering_Mode_Desc"] = "在這裡可以選擇，哪種過濾方式要套用到光環。選擇暴雪過濾方式時，會和遊戲內建框架顯示相同的光環，或是選擇自訂過濾條件。"
L["Auras_ShowTooltips"] = "顯示法術的滑鼠提示"
L["AurasCustomConditions"] = "自訂條件"
L["AurasFiltering_AddSpellID"] = "法術 ID"
L["AurasFiltering_AddSpellID_Desc"] = "輸入一個或多個要加入到過濾清單的法術 ID，使用逗號來分隔。啟用過濾時，只會顯示法術 ID 在清單內的光環。"
L["AurasFiltering_Enabled_Desc"] = "啟用時，只會看到已經加入到過濾清單內的%s。"
L["AurasFiltering_Filterlist_Desc"] = "點一下移除這個%s。"
L["AurasFilteringSettings_Desc"] = "控制要顯示哪些光環。"
L["AurasSettings"] = "光環"
L["AurasSettings_Desc"] = "光環的設定 (增益和減益效果)。"
L["AurasStacktextSettings"] = "層數文字"
L["BarBackground"] = "背景顏色"
L["BarHeight_Desc"] = "每個橫列的高度。因為暴雪的功能保護，戰鬥中無法調整設定。"
L["BarSettings"] = "條列設定"
L["BarSettings_Desc"] = "這裡可以調整橫列。"
L["BarTexture"] = "血量條材質"
L["BarWidth_Desc"] = "每個橫列的寬度。"
L["BattlegroundSize"] = "戰場大小"
L["BGSize_15"] = "1–15 名玩家"
L["BGSize_15_Desc"] = "在每方 1-15 名玩家的戰場，設定會套用到%s。"
L["BGSize_40"] = "16-40 名玩家"
L["BGSize_40_Desc"] = "在每方 16-40 名玩家的戰場，設定會套用到%s。"
L["BGSize_5"] = "競技場"
L["BGSize_5_Desc"] = "在競技場，設定會套用到%s。"
L["BlizzlikeAuraFiltering"] = "使用暴雪團隊框架過濾方式"
L["BorderSize"] = "邊框大小"
L["BorderThickness"] = "邊框粗細"
L["BOTTOM"] = "下"
L["BOTTOMLEFT"] = "左下"
L["BOTTOMRIGHT"] = "右下"
L["buff"] = "增益效果"
L["BuffContainer"] = "增益效果區域"
L["BuffIcon"] = "增益效果圖示"
L["Buffs"] = "增益效果"
L["Button"] = "按鈕"
L["CantAnchorToItself"] = "不能對齊到自己!"
L["CENTER"] = "中"
L["Change"] = "更改"
L["Color"] = "顏色"
L["Coloring_Enabled"] = "依據減益效果類型顯示顏色"
L["Coloring_Enabled_Desc"] = "啟用時，可以選擇框架邊框或倒數文字是否要依據減益效果的類型來顯示顏色 (和預設介面相同，中毒是綠色、疾病是藍色...等等)。"
L["Columns"] = "直行"
L["Columns_Desc"] = "玩家要顯示成多少個直行。"
L["Combat"] = "戰鬥中"
L["CombatIndicator"] = "戰鬥指示器"
L["CombatLogScanningForEnemiesDisabled"] = "停用掃描敵方的戰鬥記錄"
L["CombatLogScanningForEnemiesEnabled"] = "已啟用掃描戰鬥記錄來偵測敵方玩家。"
L["ConfirmDeletePlayerCountProfile"] = "是否確定要刪除目前正在使用的設定檔"
L["ConfirmProfileOverride"] = "是否確定要覆蓋子設定檔  %s，使用子設定檔 %s 取代"
L["Container_Color"] = "區域外框顏色"
L["ContainerPosition"] = "區域位置"
L["ContainerSettings"] = "區域設定"
L["ConvertCyrillic"] = "轉換西里爾字母"
L["ConvertCyrillic_Desc"] = "不是使用俄文版遊戲時轉換西里爾字母，讓名字更容易閱讀。"
L["Cooldown"] = "冷卻"
L["CopyFrom"] = "複製自"
L["CopySettings"] = "複製設定，從"
L["CopySettingsMirrored_Desc"] = "鏡像意味著模組的所有位置和延伸方向都被水平相反的點或方向取代。 (例如 \"TOPLEFT\" 點變為 \"TOPRIGHT\"，並且像光環模組這樣的容器模組將向相反方向延伸）"
L["CopySettingsNormal_Desc"] = "只是簡單的複製"
L["Countdowntext"] = "倒數時間文字"
L["Covenant"] = "誓盟"
L["CovenantIcon_Enabled"] = "啟用誓盟圖示"
L["CovenantIcon_Enabled_Desc"] = "啟用時，會在血條上顯示誓盟圖示。"
L["CovenantIcon_Size_Desc"] = "誓盟圖示的大小 (寬和高)"
L["CovenantIconSettings_Desc"] = "設定血條上所顯示的誓盟圖示"
L["CurrentVersion"] = "目前版本"
L["Curse"] = "詛咒"
L["CustomMacro_Desc"] = [=[這裡可以設定自訂巨集，空白的欄位會產生空巨集。
可用的替換符號: %n 會替換成敵人的名字。

範例:
/targetexact %n
/施放 變形術
/說 已將 %n 變羊
/targetlasttarget

這個巨集會將敵人選為目標、施放變形術、輸出一段訊息到說話頻道，然後選取前一個目標為當前目標。

請注意，巨集最長只能使用 255 個字元(包含替換後的敵人名字)。]=]
L["Data"] = "資料"
L["debuff"] = "減益效果"
L["DebuffContainer"] = "減益效果區域"
L["DebuffIcon"] = "減益效果圖示"
L["Debuffs"] = "減益效果"
L["DebuffType_Filtering"] = "依據減益效果類型過濾"
L["DebuffType_Filtering_Desc"] = "啟用時，只會看所選類型的減益效果。測試模式中不會模擬這個選項，因為需要龐大的資料庫才能模擬出魔獸海量的法術和減益效果類型。"
L["DeletePlayerCountProfile"] = "刪除"
L["DeletePlayerCountProfile_Desc"] = "點這裡刪除目前正在使用的設定檔"
L["DeletePoint"] = "刪除對齊點"
L["DisableArenaFramesInArena"] = "競技場中停用競技場框架"
L["DisableArenaFramesInArena_Desc"] = "在競技場中停用競技場框架，有安裝 sArena 時也適用。"
L["DisableArenaFramesInBattleground"] = "戰場中停用競技場框架"
L["DisableArenaFramesInBattleground_Desc"] = "在戰場中停用停用競技場框架，同樣適用於 sArena 插件。"
L["Disease"] = "疾病"
L["DispellFilter"] = "依據是否可驅散過濾"
L["DisplayType"] = "顯示類型"
L["Downwards"] = "向下"
L["DR_Disorient"] = "暈眩"
L["DR_Incapacitate"] = "癱瘓"
L["DR_Knockback"] = "擊退"
L["DR_Root"] = "定身"
L["DR_Silence"] = "沉默"
L["DR_Stun"] = "昏迷"
L["DRContainer"] = "控場區域"
L["DRTracking"] = "控場遞減監控"
L["DrTracking_Container_Color_Desc"] = "控場圖示區域的外框顏色。"
L["DrTracking_DisplayType_Desc"] = "選擇控場追蹤的指示要使用圖示周圍的彩色外框，還是彩色的倒數文字。請注意，沒有使用像是 OmniCC 這類型會更改文字顏色的插件時，彩色的倒數文字才能正常運作。"
L["DrTracking_Enabled"] = "啟用控場追蹤"
L["DrTracking_Enabled_Desc"] = "啟用時會在敵方的橫列旁顯示控場效果遞減的監控圖示。綠色邊框：下一個控場技能時間減半，黃色邊框：下一個控場技能時間只有四分之一，紅色邊框：下一個控場技能無效，因為玩家已經免疫。"
L["DrTracking_Spacing"] = "控場追蹤間距"
L["DrTracking_Spacing_Desc"] = "控場追蹤圖示之間的距離。"
L["DrTrackingFiltering_Enabled_Desc"] = "啟用時，只會看到過濾清單中所勾選的類別中的法術圖示。"
L["DrTrackingFiltering_Filterlist_Desc"] = "點一下以追蹤/取消追蹤這個類別。"
L["DrTrackingFilteringSettings_Desc"] = "選擇要顯示的控場追蹤類別。"
L["DrTrackingSettings"] = "控場追蹤"
L["DrTrackingSettings_Desc"] = "控場時間和遞減效果追蹤的設定。"
L["DurationFilter"] = "依持續時間過濾"
L["DurationFilter_OnlyShowWhenDuration"] = "只有持續時間較短或相同時才顯示"
L["DurationFilter_OnlyShowWhenDuration_Desc"] = "任何比選擇的持續時間長的光環都會被過濾掉"
L["Editmode_Toggle"] = "切換編輯模式"
L["Editmode_Toggle_Desc"] = "啟用/停用編輯模式。此模式基本上是測試模式，但能移動所有框架。選擇框架後，選項面板將自動跳到對應的設定。您也可以微調偏移量或跳到該特定框架的一般設定。"
L["EditmodeDisabled"] = "編輯模式已停用"
L[ [=[EditmodeEnabled
]=] ] = "編輯模式已啟用"
L["EditModeIntroduction"] = "歡迎使用編輯模式。此功能可讓您輕鬆地按照想要的方式排列各個模組。為此，只需點一下其中一個藍色模組來選取它，然後只要將它移動到所需位置即可。如果將它移動到另一個模組附近，它就會將自己固定在該模組上。一旦選擇了一個模組，將以黃色顯著標示，並且選項選單將跳到其設定。"
L["Enable_DrawSwipe"] = "啟用轉圈倒數材質"
L["Enable_DrawSwipe_Desc"] = "冷卻時間的剩餘部分顯示去色、較暗的材質"
L["EnableClique"] = "使用 Clique 的按鍵設定"
L["EnableClique_Desc"] = "啟用時，Clique 插件會接管所有按鍵設定。"
L["EnableCustomDRCategoryIcons"] = "啟用自訂類別圖示"
L["EnableCustomDRCategoryIcons_Desc"] = "啟用時，可以為每個遞減效果類別選擇自己的圖示。"
L["EnableCustomPlayerCountProfiles"] = "啟用玩家數量設定檔"
L["EnableCustomPlayerCountProfiles_Desc"] = "啟用此選項時，能夠根據玩家數量建立自訂設定檔。目前使用的1-5人、6-15人和16-40人的設定檔仍會保留，但會被停用，可以將它們作為建立自訂設定檔的基礎。可以隨時切換回去而不會遺失任何設定。"
L["EnableInArenas"] = "競技場中啟用"
L["EnableInBattlegrounds"] = "戰場中啟用"
L["enemies"] = "敵方"
L["Enemies"] = "敵方"
L["EnemiesTargetingAllies_Enabled_Desc"] = "啟用時，當一票敵人的目標是我方時，會播放通知音效。"
L["EnemiesTargetingMe_Enabled_Desc"] = "啟用時，當一票敵人的目標是你時，會播放通知音效。"
L["enemy"] = "敵方"
L["Enemy"] = "敵方"
L["EnemyJoined"] = "一個敵方玩家加入了戰場"
L["EnemyLeft"] = "一個敵方玩家離開了戰場"
L["ErrorTestmodeInCombat"] = "戰鬥中無法啟用測試模式"
L["ExportButton"] = "匯出設定檔"
L["ExportButton_Desc"] = "點一下這裡匯出目前使用的設定檔"
L["FakeEventsDisabled"] = "不會再建立假的事件。"
L["FakeEventsEnabled"] = "現在會建立假的事件以供測試。"
L["Filtering_Enabled"] = "啟用過濾"
L["Filtering_Filterlist"] = "過濾清單"
L["FilterSettings"] = "過濾方式設定"
L["Font"] = "字體"
L["Font_Desc"] = "插件所使用的主要字體。"
L["Font_Outline"] = "文字外框"
L["Font_Outline_Desc"] = "文字的外框。"
L["Fontcolor"] = "顏色"
L["Fontcolor_Desc"] = "文字的顏色"
L["FontShadow_Enabled"] = "啟用文字陰影"
L["FontShadow_Enabled_Desc"] = "啟用時，文字周圍會顯示陰影，可以自行選擇陰影的顏色。"
L["FontShadowColor"] = "陰影顏色"
L["FontShadowColor_Desc"] = "文字陰影的顏色。"
L["FontSize"] = "文字大小"
L["FontSize_Desc"] = "文字的大小"
L["Frame"] = "外框"
L["Framescale"] = "縮放大小"
L["Framescale_Desc"] = "主要框架的縮放大小。因為暴雪的功能保護，戰鬥中無法調整設定。"
L["General"] = "一般"
L["GeneralSettings"] = "一般設定"
L["GeneralSettings_Desc"] = "一些基本的設定。"
L["GeneralSettingsAllies"] = "設定會套用到友方，不論戰場的大小。"
L["GeneralSettingsEnemies"] = "設定會套用到敵方，不論戰場的大小。"
L["HealthBar"] = "血量條"
L["HealthBar_Background_Desc"] = "血量條的背景顏色。"
L["HealthBar_Texture_Desc"] = "血量條的材質。"
L["HealthBarSettings_Desc"] = "血量條的設定。"
L["HealthTextEnabled"] = "啟用血量文字"
L["HealthTextSettings"] = "血量文字設定"
L["HealthTextType"] = "血量文字類型"
L["Height"] = "高度"
L["HideArenaframesIn"] = "隱藏競技場框架，在"
L["HideHighestPriority"] = "隱藏最高優先順序"
L["HideHighestPriority_Desc"] = "啟用時，會移除最高優先順序的光環。當高優先順序的光環已經顯示在專精或職業圖示上時很有用。"
L["HideRaidframesIn"] = "隱藏團隊框架，在"
L["HideWhenZero"] = "數值是零時隱藏"
L["HighestPriorityAura"] = "最高優先順序光環"
L["Highlight_Color"] = "顯著標示顏色"
L["Highlight_Color_Desc"] = "滑鼠目前指向的玩家框架顏色。"
L["HorizontalGrowdirection"] = "水平增長方向"
L["HorizontalGrowdirection_Desc"] = "選擇直行要向左或向右增長。"
L["HorizontalSpacing"] = "水平間距"
L["IAmTargeted"] = "攻擊目標是我"
L["Icon"] = "圖示"
L["Icons"] = "圖示"
L["IconSize"] = "圖示大小"
L["IconsPerRow"] = "每個橫列的圖示數量"
L["Import"] = "匯入"
L["ImportButton"] = "匯入設定檔"
L["ImportButton_Desc"] = "點一下這裡從文字字串匯入設定檔，並且取代目前使用的設定檔。"
L["JumpToGeneralOptions"] = "跳到一般選項"
L["JumpToGeneralOptions_Desc"] = "點這裡跳到此模組的一般設定"
L["JustifyH"] = "水平對齊"
L["JustifyH_Desc"] = "文字水平對齊"
L["JustifyV"] = "垂直對齊"
L["JustifyV_Desc"] = "文字垂直對齊"
L["KeybindSettings_Desc"] = "這裡可以設定滑鼠左鍵、右鍵和中鍵點擊時的功能。"
L["LEFT"] = "左側"
L["LeftToTargetCounter"] = "目標數量的左側"
L["Leftwards"] = "向左"
L["Level"] = "等級"
L["LevelText_Enabled"] = "顯示玩家等級"
L["LevelText_OnlyShowIfNotMaxLevel"] = "只有不是滿等時才顯示等級"
L["Locked"] = "鎖定位置"
L["Locked_Desc"] = "鎖定框架的位置。"
L["Magic"] = "魔法"
L["MainFrameSettings"] = "主要框架設定"
L["MainFrameSettings_Desc"] = "%s所使用的主要框架設定。"
L["MaxCantBeUndefined"] = "必須設定最大值"
L["MaxNeedsToBeEqualOrGreater"] = "最大值必須大於或等於最小值"
L["MaxPlayerCount"] = "最多玩家人數"
L["MaxPlayers"] = "最多敵方數量"
L["MaxPlayers_Desc"] = "超過這個數字的敵方不會顯示出來。"
L["MIDDLE"] = "中間"
L["MinCantBeUndefined"] = "必須設定最小值"
L["MinPlayerCount"] = "最少玩家人數"
L["Mirrored"] = "鏡像"
L["Modules"] = "模組"
L["ModuleSettings"] = "模組設定"
L["ModuleSpecificSettings"] = "模組專屬設定"
L["MoreProfileOptions"] = "更多設定檔選項"
L["MouseWheelPlayerTargeting"] = "滾動選取目標"
L["MouseWheelPlayerTargeting_Desc"] = "啟用時，可以將滑鼠移動到敵方或友方框架上，並滾動滾輪來選取清單中的下一個或上一個玩家。請注意，當您使用滑鼠指向施法時，可能會產生干擾。"
L["MyFocus"] = "我的專注目標"
L["MyFocus_Color"] = "專注目標顏色"
L["MyFocus_Color_Desc"] = "用來顯著標示當前專注目標的血條外框顏色。"
L["MyTarget"] = "我的當前目標"
L["MyTarget_Color"] = "目標顏色"
L["MyTarget_Color_Desc"] = "用來顯著標示當前目標的血條外框顏色。"
L["MyVersion"] = "你的版本是"
L["Name"] = "名字"
L["Name_Desc"] = "血條上名字的設定。"
L["NewVersionAvailable"] = "已有新版本可以使用"
L["None"] = "無"
L["NonPriorityBuffs"] = "非優先的增益"
L["NonPriorityDebuffs"] = "非優先的減益"
L["Normal"] = "一般"
L["NormalCastingBar"] = "一般施法條"
L["NotAvailableInCombat"] = "因為暴雪的功能保護，戰鬥中無法調整設定。"
L["Notifications_Allies_Enabled_Desc"] = "啟用時， 友方玩家加入或離開戰場時會顯示通知。"
L["Notifications_Enabled"] = "啟用通知"
L["Notifications_Enemies_Enabled_Desc"] = "啟用時， 敵方玩家加入或離開戰場時會顯示通知。"
L["NoVersion"] = "沒有找到插件"
L["NumericTargetindicator_Enabled_Desc"] = "顯示有多少%s的當前目標是這個玩家。"
L["ObjectiveAndRespawn_ObjectiveEnabled"] = "顯示任務目標"
L["ObjectiveAndRespawn_ObjectiveEnabled_Desc"] = "在敵方按鈕旁顯示旗幟、礦車、或攜帶異能球圖示。"
L["ObjectiveAndRespawn_Position"] = "位置"
L["ObjectiveAndRespawn_Position_Desc"] = "圖示要顯示在專精圖示的左側，或是飾品/種族技能圖示的右側。"
L["ObjectiveAndRespawn_RespawnEnabled"] = "顯示重生時間"
L["ObjectiveAndRespawn_RespawnEnabled_Desc"] = "啟用時，會顯示敵人再次復活的剩餘時間圖示。"
L["ObjectiveAndRespawn_Width_Desc"] = "任務目標圖示的寬度。"
L["ObjectiveAndRespawnSettings_Desc"] = "戰場任務目標的設定。"
L["ObjectiveAndRespawnTimer"] = "任務目標和重生計時"
L["OffsetX"] = "水平位置"
L["OffsetY"] = "垂直位置"
L["OldVersion"] = "舊版本"
L["OnlyShowForHealers"] = "只顯示補師"
L["OutOfCombat"] = "非戰鬥中"
L["OverwriteMirroredConfirm"] = "是否確定要覆蓋 %​​s 的設定 (使用 %s 的鏡像設定)？"
L["OverwriteNormalConfirm"] = "是否確定要覆蓋 %​​s 的設定 (使用 %s 的設定)？"
L["PlayerCount"] = "玩家數量"
L["PlayerCount_Enabled_Desc"] = "是否要顯示目前玩家數量的文字。請注意，這個數量可能會和目前的橫列數量不同，因為插件的戰鬥中保護功能或離開隨機戰場的人數。"
L["players"] = "玩家"
L["Point"] = "對齊點"
L["PointAtObject"] = "物件的對齊點"
L["Poison"] = "中毒"
L["Position"] = "位置"
L["PowerBar"] = "能量條"
L["PowerBar_Background_Desc"] = "能量條的背景顏色。"
L["PowerBar_Enabled"] = "啟用資源條"
L["PowerBar_Enabled_Desc"] = "啟用時，會顯示法力、怒氣等資源。"
L["PowerBar_Height_Desc"] = "資源條的高度。資源條愈高生命條就會愈窄。"
L["PowerBar_OnlyShowForHealers_Desc"] = "啟用時，只有補師會顯示能量條。"
L["PowerBar_Texture_Desc"] = "能量條的材質。"
L["PowerBarSettings_Desc"] = "資源條的設定。"
L["PriorityBuffs"] = "優先的增益"
L["PriorityDebuffs"] = "優先的減益"
L["Racial"] = "種族"
L["Racial_Enabled"] = "啟用種族特長"
L["Racial_Enabled_Desc"] = "啟用時，會顯示已使用的種族特長圖示。"
L["Racial_Width_Desc"] = "種族技能的寬度。"
L["RacialFiltering_Enabled_Desc"] = "啟用時，只會看到過濾清單中所勾選的種族技能。"
L["RacialFiltering_Filterlist_Desc"] = "點一下以監控/取消監控這個種族技能。"
L["RacialFilteringSettings_Desc"] = "在這裡選擇要追蹤的種族技能。"
L["RacialSettings"] = "種族特長"
L["RacialSettings_Desc"] = "種族特長的設定。"
L["RangeIndicator_Alpha"] = "透明度"
L["RangeIndicator_Alpha_Desc"] = "敵人超出指定距離時，敵方框架的透明度。"
L["RangeIndicator_Enabled"] = "啟用距離指示"
L["RangeIndicator_Enabled_Desc"] = "啟用時，敵人超出指定距離時，敵方框架會顯示成指定的透明度。"
L["RangeIndicator_Everything"] = "更改全部的透明度"
L["RangeIndicator_Frames"] = "更改透明度"
L["RangeIndicator_Frames_Desc"] = "當敵人超出距離範圍時，橫列的哪個部分要變得比較透明。"
L["RangeIndicator_Range"] = "距離範圍"
L["RangeIndicator_Range_Desc"] = "敵人在這個距離 (碼) 以外時，敵方框架會顯示成指定的透明度。"
L["RangeIndicator_Settings"] = "距離指示"
L["RangeIndicator_Settings_Desc"] = "這裡可以調整距離指示。"
L["RangeOverlapping"] = "範圍重疊: %s"
L["RBGSpecificSettings"] = "積分戰場"
L["RBGSpecificSettings_Desc"] = "這些設定只會套用到積分戰場。"
L["ResetGeneralModule_Desc"] = "點擊此處重置該模組的一般設定"
L["ResetModule"] = "重置模組"
L["ResetModule_Desc"] = "點一下這裡重置%s模組的設定:%s"
L["ResetModuleSettings"] = "重置此模組的設定"
L["RestoreDefault"] = "恢復成預設值"
L["RIGHT"] = "右側"
L["Rightwards"] = "向右"
L["RoleIcon_Enabled"] = "啟用角色圖示"
L["RoleIcon_Enabled_Desc"] = "啟用時，會在血量條上顯示角色圖示。"
L["RoleIcon_Size_Desc"] = "角色圖示的大小 (寬和高)。"
L["RoleIconSettings"] = "角色圖示"
L["RoleIconSettings_Desc"] = "血量條上面的角色圖示設定。"
L["RoleSortingOrder"] = "角色職責排列順序"
L["RoleSortingOrder_Desc"] = "根據角色職責選擇玩家的排列順序"
L["Scale"] = "縮放大小"
L["ShowDispellable"] = "顯示可驅散的"
L["ShowDispellable_Desc"] = "顯示我可以使用驅散、淨化...等移除的%s。"
L["ShowFramesIn"] = "顯示框架，在"
L["ShowHighestPriority"] = "顯示最高優先順序"
L["ShowHighestPriority_Desc"] = "啟用時，插件將顯示被打斷或受到優先光環的影響的玩家"
L["ShowMine"] = "顯示我的"
L["ShowMine_Desc"] = "顯示我施放的%s。"
L["ShowNumbers"] = "顯示數字"
L["ShowNumbers_Desc"] = "顯示冷卻時間數字，請注意，使用 OmniCC 插件時，此選項會沒有作用。"
L["ShowRealmnames"] = "顯示伺服器"
L["ShowRealmnames_Desc"] = "顯示敵方的伺服器名稱。"
L["ShowSpecIfExists"] = "顯示專精 (如果有的話)"
L["ShowSpecIfExists_Desc"] = "啟用時，會顯示專精圖示而不是職業圖示"
L["ShowTooltips"] = "顯示滑鼠提示"
L["ShowTooltips_Desc"] = "滑鼠指向飾品或光環圖示等物件時顯示滑鼠提示。"
L["Side"] = "方向"
L["SideAtObject"] = "物件的方向"
L["Size"] = "大小"
L["SmallCastingBar"] = "小型施法條"
L["SourceFilter"] = "依據來源過濾"
L["Spec_AuraDisplay_Enabled"] = "顯示光環"
L["Spec_AuraDisplay_Enabled_Desc"] = "啟用時，會顯示群體控場和斷法技能的圖示 (有啟用時)，取代專精圖示。"
L["Spec_Enabled"] = "啟用專精"
L["Spec_Enabled_Desc"] = "啟用時，會顯示玩家專精的圖示。"
L["Spec_Width_Desc"] = "專精圖示的寬度。"
L["SpecClassPriorityOne"] = "SpecClassPriority1"
L["SpecClassPriorityTwo"] = "SpecClassPriority2"
L["SpecSettings"] = "專精"
L["SpecSettings_Desc"] = "專精圖示的設定。"
L["SpellID_Filtering"] = "依據法術 ID 過濾"
L["Style"] = "樣式"
L["SymbolicTargetindicator_Enabled_Desc"] = "在當前目標是玩家的%s上顯示方形的職業顏色圖示。"
L["TargetAmount"] = "目標數量"
L["TargetAmount_Ally"] = "當隊友被超過指定數量的敵方選為目標時，就會立刻播放聲音。 "
L["TargetAmount_Me"] = "當你被超過指定數量的敵方選為目標時，就會立刻播放聲音。 "
L["TargetCallerUpdated"] = "是目前的發號施令者"
L["TargetCalling"] = "發號施令"
L["TargetCallingNotificationEnable"] = "啟用音效通知"
L["TargetCallingNotificationEnable_Desc"] = "通知你有新的目標 (偵測骷髏頭標)"
L["TargetCallingSetMark"] = "幫我的目標上骷髏頭標"
L["TargetCallingSetMark_Desc"] = "當你是團隊隊長，而且隊中沒有團隊助理時，自動將你的當前目標標記骷髏圖示。當你不是隊長，但是助理時，也會上標記。"
L["TargetCallingShowIcon"] = "發號施令者的目標要顯示圖示"
L["TargetCallingShowIcon_Desc"] = "啟用時，發號施令者的當前目標的框架上面會顯示圖示。"
L["TargetCallingSoundEnable"] = "有新目標時通知我"
L["TargetCallingSoundEnable_Desc"] = "啟用時，發號施令者選取了新的目標時，會播放音效。"
L["TargetCallingSoundSound"] = "通知音效"
L["TargetIndicator"] = "目標指示"
L["TargetIndicator_Desc"] = "數字和符號的目標指示器。"
L["TargetIndicatorNumeric"] = "目標指示器數字"
L["TargetIndicatorNumeric_HideWhenZero_Desc"] = "啟用時，如果玩家沒有被敵方選取，則會隱藏文字"
L["TargetIndicatorSymbolic"] = "目標指示器符號"
L["TargetNextAlly"] = "選取下一個盟友"
L["TargetNextEnemy"] = "選取下一個敵人"
L["TargetPreviousAlly"] = "選取前一個盟友"
L["TargetPreviousEnemy"] = "選取前一個敵人"
L["Testmode_Toggle"] = "切換測試模式"
L["Testmode_Toggle_Desc"] = [=[啟用/停用測試模式，更容易觀察調整設定選項時的變化。無法模擬全部的選項效果，但在這個華麗的測試模式中，已經包含了絕大部分。

已經在戰場中時，會停用測試模式。]=]
L["Testmode_ToggleAnimation"] = "切換測試動畫"
L["Testmode_ToggleAnimation_Desc"] = "啟用/停用測試模式的動畫效果，以便能夠更專注於設定某個選項，而不會被動畫干擾。"
L["Testmode_UseTeammates"] = "測試時使用隊友"
L["Testmode_UseTeammates_Desc"] = "啟用測試模式時會顯示目前的隊友，而不是假的玩家。"
L["TestmodeDisabled"] = "測試模式已停用"
L["TestmodeEnabled"] = "測試模式已啟用"
L["TestmodeSettings"] = "測試模式"
L["Text"] = "文字"
L["TextSettings"] = "文字設定"
L["Thick"] = "飾品"
L["TOP"] = "上"
L["TOPLEFT"] = "左上"
L["TOPRIGHT"] = "右上"
L["Trinket"] = "飾品"
L["Trinket_Enabled"] = "啟用飾品"
L["Trinket_Enabled_Desc"] = "啟用時會顯示鬥士徽章等飾品圖示。"
L["Trinket_Width_Desc"] = "飾品的寬度。"
L["TrinketSettings_Desc"] = "飾品的設定。"
L["UpdatePeriod"] = "更新間隔"
L["UpdatePeriod_Desc"] = "多久檢查一次是否在戰鬥中，請不要輸入太低的數值，可能會導致影響遊戲效能問題。"
L["Upwards"] = "向上"
L["UseBarHeight"] = "使用橫列高度"
L["UseBigDebuffsPriority"] = "使用 BigDebuffs 的優先順序"
L["UseBigDebuffsPriority_Desc"] = "啟用此選項時，會改用來自大型控場圖示插件 BigDebuffs 的優先順序。優先順序會用於%s、%s 和%s模組。"
L["UseButtonHeight"] = "使用按鈕高度"
L["UseButtonHeight_Desc"] = "啟用此選項時，模組會使用按鈕的高度。"
L["UserDefined"] = "玩家自訂的"
L["VerticalGrowdirection"] = "垂直增長方向"
L["VerticalGrowdirection_Desc"] = "選擇橫列要向上或向下增長。"
L["VerticalPosition"] = "垂直位置"
L["VerticalSpacing"] = "垂直間距"
L["Width"] = "寬度"

end